/*
 *  Implementation of COSS Naming Service for MICO
 *  Copyright (C) 1997 Kai-Uwe Sattler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <iostream.h>
#include <fstream.h>
#include <Naming_impl.h>

void NamingContext_impl::_shutdown ()
{
  CORBA::ORB_var orb = CORBA::ORB_instance ("mico-local-orb");
  CORBA::BOA_var boa = orb->BOA_instance ("mico-local-boa");
  cerr << "shutdown name service" <<  endl;
  boa->deactivate_impl (CORBA::ImplementationDef::_nil ());
  orb->shutdown (TRUE);
}

CosNaming::NamingContext_ptr
NamingContext_impl::_create_context (const char *s)
{
  CORBA::ORB::ObjectTag_var tag = CORBA::ORB::string_to_tag (s);
  NamingContext_impl* impl = new NamingContext_impl (tag);
  CosNaming::NamingContext_ptr nctx = impl;
  return CosNaming::NamingContext::_duplicate (nctx);
}

NamingContext_impl::NamingContext_impl (const CORBA::BOA::ReferenceData &tag)
    : CosNaming::NamingContext_skel (tag)
{
}

NamingContext_impl::NamingContext_impl (CORBA::Object_ptr obj) :
  CosNaming::NamingContext_skel (obj)
{
  ifstream in (obj->_ident ());
    cout << "restore persistent state: " << obj->_ident () << endl;
  while (! in.eof ()) {
    int flag;
    string id, kind;
    string objref;
    string tmp;

    in >> flag >> id >> kind >> objref;
    /*
    cout << "type = " << flag << "\nid = " << id
	 << "\nkind = " << kind << "\nobjref = "
         << objref << endl;
	 */
    if (objref.size () == 0)
      continue;
    NmCtxBinding binding;

    tmp = id.substr (1, kind.size () - 2);
    binding.name.kind = CORBA::string_dup (tmp.c_str ());
    tmp = id.substr (1, id.size () - 2);
    binding.name.id = CORBA::string_dup (tmp.c_str ());
    binding.btype = (flag == 1 ? CosNaming::ncontext : CosNaming::nobject);
    binding.obj = _orbnc()->string_to_object (objref.c_str ());
    table[tmp.c_str ()] = binding;
    /*
    cout << "-- bound name = " << tmp
         << " to object = " << objref << endl;
	 */
  }

}

void NamingContext_impl::_do_bind (const CosNaming::Name& n, 
				   CORBA::Object_ptr obj,
				   CosNaming::BindingType btype,
				   CORBA::Boolean rebind)
{
  if (n.length () == 0) 
    mico_throw (CosNaming::NamingContext::InvalidName ());

  BindingTable::iterator i = table.find (n[0].id);
  if (n.length () == 1) {
    if (i != table.end () && !rebind)
      mico_throw (CosNaming::NamingContext::AlreadyBound ());

    else {
      NmCtxBinding binding;

      binding.name = n[0];
      binding.btype = btype;
      binding.obj = CORBA::Object::_duplicate (obj);
      table[n[0].id] = binding;
      /*
      cout << "bound name = " << n[0].id 
           << " to object = " << _orbnc()->object_to_string (obj) << endl;
       */
      _save_object ();
    }
  }
  else {
    if (i == table.end ()) {
      CosNaming::NamingContext::NotFound exc;
      exc.why = CosNaming::NamingContext::missing_node;
      exc.rest_of_name = n;
      mico_throw (exc);
    }

    NmCtxBinding binding = (*i).second;
    if (binding.btype != CosNaming::ncontext)
      mico_throw (CosNaming::NamingContext::CannotProceed ());

    CosNaming::NamingContext_var nc = 
      CosNaming::NamingContext::_narrow (binding.obj);

    CosNaming::Name rest;
    rest.length (n.length () - 1);

    for (CORBA::ULong k = 0; k < rest.length (); k++)
      rest[k] = n[k + 1];

    switch (btype) {
    case CosNaming::nobject:
      if (rebind)
	nc->rebind (rest, obj);
      else
	nc->bind (rest, obj);
      break;

    case CosNaming::ncontext:
      CosNaming::NamingContext_var nc2 =
	CosNaming::NamingContext::_narrow (obj);

      if (rebind)
	nc->rebind_context (rest, nc2);
      else
	nc->bind_context (rest, nc2);
      break;
    }
  }
}

void NamingContext_impl::bind (const CosNaming::Name& n, 
			       CORBA::Object_ptr obj)
{
  _do_bind (n, obj, CosNaming::nobject, FALSE);
}

void NamingContext_impl::rebind (const CosNaming::Name& n, 
				 CORBA::Object_ptr obj)
{
  _do_bind (n, obj, CosNaming::nobject, TRUE);
}

void NamingContext_impl::bind_context (const CosNaming::Name& n, 
				       CosNaming::NamingContext_ptr nc)
{
  _do_bind (n, nc, CosNaming::ncontext, FALSE);
}

void NamingContext_impl::rebind_context (const CosNaming::Name& n, 
					 CosNaming::NamingContext_ptr nc)
{
  _do_bind (n, nc, CosNaming::ncontext, TRUE);
}

CORBA::Object_ptr NamingContext_impl::resolve (const CosNaming::Name& n)
{
  if (n.length () == 0)
    mico_throw (CosNaming::NamingContext::InvalidName ());

  BindingTable::iterator i = table.find (n[0].id);
  if (i == table.end ()) {
    CosNaming::NamingContext::NotFound exc;
    exc.why = CosNaming::NamingContext::missing_node;
    exc.rest_of_name = n;
    mico_throw (exc);
  }

  NmCtxBinding binding = (*i).second;

  if (n.length () == 1) {
    /*
    cout << "resolved objref = " 
	 << _orbnc()->object_to_string (binding.obj) << endl;
    */
    assert (! CORBA::is_nil (binding.obj));
    return CORBA::Object::_duplicate (binding.obj);
  }
  else {
    if (binding.btype != CosNaming::ncontext)
      mico_throw (CosNaming::NamingContext::CannotProceed ());

    CosNaming::Name rest;
    rest.length (n.length () - 1);
    for (CORBA::ULong k = 0; k < rest.length (); k++)
      rest[k] = n[k + 1];

    CosNaming::NamingContext_var nc = 
      CosNaming::NamingContext::_narrow (binding.obj);
    return nc->resolve (rest);
  }
}

void NamingContext_impl::unbind (const CosNaming::Name& n)
{
  if (n.length () == 0)
    mico_throw (CosNaming::NamingContext::InvalidName ());

  BindingTable::iterator i = table.find (n[0].id);
  if (i == table.end ()) {
    CosNaming::NamingContext::NotFound exc;
    exc.why = CosNaming::NamingContext::missing_node;
    exc.rest_of_name = n;
    mico_throw (exc);
  }

  NmCtxBinding binding = (*i).second;

  if (n.length () == 1) {
    table.erase (i);
  }
  else {
    if (binding.btype != CosNaming::ncontext) 
      mico_throw (CosNaming::NamingContext::CannotProceed ());

    CosNaming::Name rest;
    rest.length (n.length () - 1);
    for (CORBA::ULong k = 0; k < rest.length (); k++)
      rest[k] = n[k + 1];

    CosNaming::NamingContext_var nc = 
      CosNaming::NamingContext::_narrow (binding.obj);
    nc->unbind (rest);
  }
}

CosNaming::NamingContext_ptr NamingContext_impl::new_context ()
{
  return NamingContext_impl::_create_context ();
}

CosNaming::NamingContext_ptr 
NamingContext_impl::bind_new_context (const CosNaming::Name& n)
{
  CosNaming::NamingContext_ptr nc = new_context ();
  bind_context (n, nc);
  return nc;
}

void NamingContext_impl::destroy ()
{
  if (table.size () > 0)
    mico_throw (CosNaming::NamingContext::NotEmpty ());

  CORBA::release (_this ());
}

void NamingContext_impl::list (CORBA::ULong how_many, 
			       CosNaming::BindingList*& bl, 
			       CosNaming::BindingIterator_ptr& bi)
{
  CORBA::ULong num = table.size () < how_many ? table.size () : how_many;
  bl = new CosNaming::BindingList (num);
  bl->length (num);
  bi = CosNaming::BindingIterator::_nil ();
  CORBA::ULong l = 0;
  BindingTable::iterator i = table.begin ();

  for (; l < num; l++, i++) {
    CosNaming::Binding b;
    (*bl)[l].binding_name.length (1);
    (*bl)[l].binding_name[0] = (*i).second.name;
    (*bl)[l].binding_type = (*i).second.btype;
  }
  if (how_many < table.size ()) {
    CORBA::ULong rn = table.size () - how_many;
    CosNaming::BindingList rest (rn);
    rest.length (rn);
    for (l = rn; l < table.size (); l++, i++) {
      rest[l - rn].binding_name.length (1);
      rest[l - rn].binding_name[0] = (*i).second.name;
      rest[l - rn].binding_type = (*i).second.btype;
    }
    bi = new BindingIterator_impl (rest);
    CosNaming::BindingIterator::_duplicate (bi);
  }
}


CORBA::Boolean NamingContext_impl::_save_object ()
{
  ofstream out (_ident ());
  BindingTable::iterator i = table.begin ();
  for (; i != table.end (); i++) {
    NmCtxBinding b = (*i).second;
    out << b.btype << " \"" << b.name.id.in() << "\" \""
	<< b.name.kind.in() << "\" "
	<< _orbnc()->object_to_string (b.obj) 
	<< endl;
  }
  return TRUE;
}

BindingIterator_impl::BindingIterator_impl (const CosNaming::BindingList& bl) 
  : idx (0), blist (bl)
{
}

CORBA::Boolean BindingIterator_impl::next_one (CosNaming::Binding*& b)
{
  if (idx < blist.length ()) {
    b = new CosNaming::Binding (blist[idx++]);
    return TRUE;
  }
  else {
    b = new CosNaming::Binding;
    return FALSE;
  }
}

CORBA::Boolean 
BindingIterator_impl::next_n (CORBA::ULong how_many, 
			      CosNaming::BindingList*& bl)
{
  CORBA::ULong rlen = blist.length () - idx;
  if (rlen > 0) {
    CORBA::ULong num = how_many < rlen ? how_many : rlen;
    bl = new CosNaming::BindingList (num);
    bl->length (num);
    for (CORBA::ULong i = 0; i < num; i++, idx++)
      (*bl)[i] = blist[idx];
    return TRUE;
  }
  else {
    bl = new CosNaming::BindingList (0);
    bl->length (0);
    return FALSE;
  }
}

void BindingIterator_impl::destroy ()
{
  CORBA::release (this);
}

CORBA::Boolean NamingContextLoader::restore (CORBA::Object_ptr obj)
{
  if (::strcmp (obj->_repoid (), "IDL:CosNaming/NamingContext:1.0") == 0) {
    (void)new NamingContext_impl (obj);
    return TRUE;
  }
  return FALSE;
}


