/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#include <fstream.h>
#include <ctype.h>
#include <stdio.h>
#include "codegen-c++-common.h"
#include <mico/template_impl.h>
#include <mico/util.h>


#define TVAR( tc ) ( is_variable( tc ) ? "TVarVar" : "TFixVar" )



CodeGenCPPCommon::CodeGenCPPCommon( DB &db, IDLParam& params )
  : CodeGen( db.get_idl_container() ),
    CodeGenCPPUtil( db, params )
{
}


void
CodeGenCPPCommon::emit_common()
{
  _current_scope = CORBA::Container::_duplicate( _container );
  
  emitForwardDcl( _container );
  
  for( CORBA::ULong i = 0; i < _idl_objs.length(); i++ ) {
    emit_idl_obj( _idl_objs[ i ] );
  }

  // Close open modules
  enter_scope( _container, "" );

  // emit POA Skeleton definitions
  if (_params.poa && !_params.pseudo) {
    for (CORBA::ULong i2=0; i2<_idl_objs.length(); i2++) {
      emit_poa_obj (_idl_objs[i2]);
    }
    enter_scope (_container, "POA_");
  }
}

void CodeGenCPPCommon::emit_idl_obj( CORBA::IRObject_ptr obj )
{
  if( check_for_included_defn( obj ) )
    return;
    
  CORBA::Contained_var c = CORBA::Contained::_narrow( obj );
  if( !CORBA::is_nil( c ) ) {
    CORBA::Container_var new_scope = c->defined_in();
    enter_scope( new_scope, "" );
    set_name_prefix( new_scope );
  }
  
  switch( obj->def_kind() ) {
  case CORBA::dk_Interface:
    {
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( obj );
      emit_Interface( in );
      break;
    }
  case CORBA::dk_Struct:
    {
      CORBA::StructDef_var s = CORBA::StructDef::_narrow( obj );
      emit_Struct( s );
      break;
    }
  case CORBA::dk_Union:
    {
      CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( obj );
      emit_Union( u );
      break;
    }
  case CORBA::dk_Constant:
    {
      CORBA::ConstantDef_var co = CORBA::ConstantDef::_narrow( obj );
      emit_Constant( co );
      break;
    }
  case CORBA::dk_Exception:
    {
      CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( obj );
      emit_Exception( e );
      break;
    }
  case CORBA::dk_Enum:
    {
      CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( obj );
      emit_Enum( e );
      break;
    }
  case CORBA::dk_Alias:
    {
      CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( obj );
      emit_Alias( a );
      break;
    }
  case CORBA::dk_String:
    {
      CORBA::StringDef_var s = CORBA::StringDef::_narrow( obj );
      emit_StringDef( s );
      break;
    }
  case CORBA::dk_Wstring:
    {
      CORBA::WstringDef_var ws = CORBA::WstringDef::_narrow( obj );
      emit_WstringDef( ws );
      break;
    }
  case CORBA::dk_Sequence:
    {
      CORBA::SequenceDef_var s = CORBA::SequenceDef::_narrow( obj );
      emit_SequenceDef( s );
      break;
    }
  case CORBA::dk_Fixed:
    {
      CORBA::FixedDef_var f = CORBA::FixedDef::_narrow( obj );
      emit_FixedDef( f );
      break;
    }
  case CORBA::dk_Array:
    {
      CORBA::ArrayDef_var a = CORBA::ArrayDef::_narrow( obj );
      emit_ArrayDef( a );
      break;
    }
  default:
    assert( 0 );
  }
}

void
CodeGenCPPCommon::emit_marshal_prototype( CORBA::IRObject_ptr obj )
{
  if( !_params.sii )
    return;

  /*
   * Here we declare a variable that references an instance
   * of a marshaller. The variable is initialized in the
   * implemantation.
   */
  o << "MICO_EXPORT_VAR_DECL CORBA::StaticTypeInfo *";
  emit_marshaller_ref( obj );
  o << ";" << endl;

  /*
   * When using includes, a typedef might refer to an included
   * IDL definition. The following function allocates a
   * marshaller instance that is used in another file to
   * initialize the typedef marshaller.
   */
  o << "CORBA::StaticTypeInfo* _get";
  emit_marshaller_ref( obj );
  o << "();" << endl << endl;
}

void CodeGenCPPCommon::emitPrototypes( CORBA::InterfaceDef_ptr in,
				       bool as_pure_virtual )
{
  CORBA::InterfaceDef::FullInterfaceDescription_var desc;
  desc = in->describe_interface();

  // Emit attributes
  for( CORBA::ULong j = 0; j < desc->attributes.length(); j++ ) {
    CORBA::AttributeDescription attr;
    attr = desc->attributes[ j ];
    if( as_pure_virtual )
      o << "virtual ";
    CORBA::IDLType_var type_id = lookup_attribute_by_id( attr.id );
    emit_type_for_result( type_id );
    o << " " << ID( attr.name ) << "()";
    if( as_pure_virtual )
      o << " = 0";
    o << ";" << endl;
    if( attr.mode == CORBA::ATTR_NORMAL ) {
      if( as_pure_virtual )
	o << "virtual ";
      o << "void " << ID(attr.name) << "( ";
      emit_type_for_param( type_id, CORBA::PARAM_IN );
      o << " value )";
      if( as_pure_virtual )
	o << " = 0";
      o << ";" << endl;
    }
  }

  // Emit operations
  for( CORBA::ULong j0 = 0; j0 < desc->operations.length(); j0++ ) {
    CORBA::OperationDescription op;
    op = desc->operations[ j0 ];
    if( as_pure_virtual )
      o << "virtual ";
    CORBA::IDLType_var type_id = lookup_result_by_id( op.id );
    emit_type_for_result( type_id );
    o << " " << ID(op.name) << "(";
    CORBA::ParDescriptionSeq p = op.parameters;
    CORBA::ULong k;
    for( k = 0; k < p.length(); k++ ) {
      if( k != 0 )
	o << ",";
      o << " ";
      emit_type_for_param( p[ k ].type_def, p[ k ].mode );
      o << " " << ID(p[ k ].name);
    }

    // Do we have a context?
    CORBA::ContextIdSeq& ctx = op.contexts;
    
    if( ctx.length() > 0 ) {
      if( p.length() > 0 )
	o << ", ";
      o << "CORBA::Context_ptr _ctx";
    }
    
    if( k != 0 || ctx.length() > 0 )
      o << " ";
    o << ")";
    if( as_pure_virtual )
      o << " = 0";
    o << ";" << endl;
  }
}



void CodeGenCPPCommon::emitLocalDecls( CORBA::InterfaceDef_ptr in )
{
  CORBA::ContainedSeq_var c;
  c = in->contents( CORBA::dk_all, 1 );
  for( CORBA::ULong i = 0; i < c->length(); i++ ) {
    switch( c[ i ]->def_kind() ) {
    case CORBA::dk_Operation:
    case CORBA::dk_Attribute:
      // Operation & attribute declarations are generated in emitPrototypes()
      break;
    case CORBA::dk_Struct:
      {
	CORBA::StructDef_var s = CORBA::StructDef::_narrow( c[ i ] );
	emit_Struct( s );
	break;
      }
    case CORBA::dk_Union:
      {
	CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( c[ i ] );
	emit_Union( u );
	break;
      }
    case CORBA::dk_Constant:
      {
	CORBA::ConstantDef_var co = CORBA::ConstantDef::_narrow( c[ i ] );
	emit_Constant( co );
	break;
      }
    case CORBA::dk_Exception:
      {
	CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( c[ i ] );
	emit_Exception( e );
	break;
      }
    case CORBA::dk_Enum:
      {
	CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( c[ i ] );
	emit_Enum( e );
	break;
      }
    case CORBA::dk_Alias:
      {
	CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( c[ i ] );
	emit_Alias( a );
	break;
      }
    default:
      assert( 0 );
    }
  }
}

void
CodeGenCPPCommon::emit_StringDef( CORBA::StringDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_String );

  if( _params.pseudo )
    return;

  save_name_prefix();
  set_name_prefix( CORBA::Container::_nil() );

  o.switchStream( 1 );
  
  CORBA::TypeCode_var type_tc = s->type();
	  
  // Must be a bounded string
  assert( type_tc->length() > 0 );
  o << "extern CORBA::TypeCodeConst _tc_string_";
  o << type_tc->length();
  o << ";" << endl << endl;

  o.switchStream( 0 );

  restore_name_prefix();
}

void
CodeGenCPPCommon::emit_WstringDef( CORBA::WstringDef_ptr ws )
{
  assert( ws->def_kind() == CORBA::dk_Wstring );

  if( _params.pseudo )
    return;
  
  save_name_prefix();
  set_name_prefix( CORBA::Container::_nil() );
  
  o.switchStream( 1 );
  
  CORBA::TypeCode_var type_tc = ws->type();
	  
  // Must be a bounded wstring
  assert( type_tc->length() > 0 );
  o << "extern CORBA::TypeCodeConst _tc_wstring_";
  o << type_tc->length();
  o << ";" << endl << endl;

  o.switchStream( 0 );

  restore_name_prefix();
}

void
CodeGenCPPCommon::emit_SequenceDef( CORBA::SequenceDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_Sequence );

  if( _params.pseudo )
    return;
  
  save_name_prefix();
  set_name_prefix( CORBA::Container::_nil() );

  o.switchStream( 1 );
  
  CORBA::TypeCode_var type_tc = s->type();
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  emit_sequence( s );
  o << " &_s );" << endl;
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  emit_sequence( s );
  o << " &_s );" << endl << endl;

  if( !is_base_sequence( s ) )
    emit_marshal_prototype( s );
  
  o.switchStream( 0 );

  restore_name_prefix();
}

void
CodeGenCPPCommon::emit_FixedDef( CORBA::FixedDef_ptr f )
{
  assert( f->def_kind() == CORBA::dk_Fixed );

  if( _params.pseudo )
    // XXX should still emit _var and _out type for Fixed
    return;
  
  save_name_prefix();
  set_name_prefix( CORBA::Container::_nil() );

  o.switchStream( 1 );
  
  CORBA::TypeCode_var type_tc = f->type();
  string fixed_name = "Fixed_";
  fixed_name += xdec( type_tc->fixed_digits() );
  fixed_name += "_";
  if( type_tc->fixed_scale() < 0 )
    fixed_name += string( "n" ) +
      xdec( -( type_tc->fixed_scale() ) );
  else
    fixed_name += xdec( type_tc->fixed_scale() );
  
  // We have to generate _var and _out type for a Fixed,
  // because unlike anonymous types, it can be passed as a
  // parameter type [20.11.1]
  
  // Emit _var type for Fixed
  o << "typedef TFixVar<";
  emit_fixed( f );
  o << " > " << fixed_name << "_var;" << endl;
  
  // Emit _out for Fixed
  o << "typedef ";
  emit_fixed( f );
  o << "& " << fixed_name << "_out;" << endl;
  
  // Emit TypeCode for Fixed
  o << "extern CORBA::TypeCodeConst ";
  emit_type_code_name( f );
  o << ";" << endl;
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, ";
  emit_fixed( f );
  o << " &_f );" << endl;
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  emit_fixed( f );
  o << " &_f );" << endl << endl;

  emit_marshal_prototype( f );

  o.switchStream( 0 );

  restore_name_prefix();
}

void
CodeGenCPPCommon::emit_ArrayDef( CORBA::ArrayDef_ptr a )
{
  assert( a->def_kind() == CORBA::dk_Array );

  if( _params.pseudo )
    return;
  
  save_name_prefix();
  set_name_prefix( CORBA::Container::_nil() );

  o.switchStream( 1 );
  
  CORBA::TypeCode_var type_tc = a->type();
  if( type_tc->kind() != CORBA::tk_alias ) {
    //XXXarray, second argument should be const
    o << "CORBA::Boolean operator<<=( CORBA::Any &_a, ";
    emit_array( a );
    o << " (&_s)";
    emit_array_suffix( a );
    o << " );" << endl;
    o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
    emit_array( a );
    o << " (&_s)";
    emit_array_suffix( a );
    o << " );" << endl << endl;
  }

  // SII marshaller
  emit_marshal_prototype( a );
  
  o.switchStream( 0 );

  restore_name_prefix();
}

void CodeGenCPPCommon::emit_Interface( CORBA::InterfaceDef_ptr in )
{
  assert( in->def_kind() == CORBA::dk_Interface );

  set_name_prefix( in );
  
  CORBA::String_var n = in->name();
  string className = (const char *) ID(n);
  
  o << endl;

  //-----COMMON------------------------------------------------------
  o << "// Common definitions for interface " << className << endl;
  
  o << "class " << className << " : ";
 
  CORBA::InterfaceDefSeq_var bases = in->base_interfaces();
  if( bases->length() == 0 ) {
    // This interface has no base interfaces
    o << "virtual public CORBA::Object" << endl;
  } else {
    // This interface *has* base interfaces
    o << indent << endl;
    for( CORBA::ULong i = 0; i < bases->length(); i++ ) {
      o << "virtual public ";
      CORBA::String_var base_name = bases[ i ]->absolute_name();
      o << ID(base_name);
      if( i + 1 != bases->length() )
	o << ",";
      o << endl;
    }
    o << exdent;
  }
  
  o << BL_OPEN;
  
  o << "public:" << indent << endl;
  o << "virtual ~" << className << "();" << endl;

  o << "static " << className << "_ptr _duplicate( ";
  o << className << "_ptr obj );" << endl;
  o << "static " << className << "_ptr _narrow( CORBA::Object_ptr obj );"
    << endl;
  o << "static " << className << "_ptr _nil();" << endl << endl;
  
  o << "virtual void *_narrow_helper( const char *repoid );" << endl;

  if (!_params.pseudo) {
    o << "static vector<CORBA::Narrow_proto> *_narrow_helpers;" << endl;
    o << "static bool _narrow_helper2( CORBA::Object_ptr obj );" << endl;
  }

  o << endl;

  // Emit local declarations
  emitLocalDecls( in );
  
  // Generate prototypes
  emitPrototypes( in, true );
  
  // Generate epilogue
  o << exdent << "protected:" << indent << endl;
  o << className << "() {};" << exdent << endl;
  o << "private:" << indent << endl;
  o << className << "( const " << className << "& );" << endl;
  o << "void operator=( const " << className << "& );" << exdent << endl;
  o << BL_CLOSE_SEMI << endl;

  if (!_params.pseudo) {
    // Emit TypeCode for interface
    CORBA::Container_var def_in = in->defined_in();
    if( def_in->def_kind() == CORBA::dk_Repository )
      // We are on the toplevel
      o << "extern";
    else if( def_in->def_kind() == CORBA::dk_Module )
      // The current scope is not top level (inside a module)
      o << "MICO_EXPORT_VAR_DECL";
    else
      // The current scope is not top level (not inside a module)
      o << "static";
    o << " CORBA::TypeCodeConst _tc_" << className << ";" << endl << endl;

    // operator<<= for interface
    o.switchStream( 1 );
    save_name_prefix();
    set_name_prefix( CORBA::Container::_nil() );
    o << "CORBA::Boolean operator<<=( CORBA::Any &a, const ";
    emit_idl_type_name( in );
    o << " obj );" << endl;

    // operator>>= for interface
    o << "CORBA::Boolean operator>>=( const CORBA::Any &a, ";
    emit_idl_type_name( in );
    o << " &obj );" << endl << endl;

    emit_marshal_prototype( in );
    
    o.switchStream( 0 );
    restore_name_prefix();
  }

  if (_params.pseudo) {
    // No stubs or skeletons for PIDL
    o << endl;
    return;
  }

  //-----STUB--------------------------------------------------------
  o << "// Stub for interface " << className << endl;
  
  o << "class " << className << _STUB << 
    " : virtual public " << className;

  if( bases->length() == 0 ) {
    // This interface has no base interfaces
    o << endl;
  } else {
    // This interface *has* base interfaces
    o << "," << indent << endl;
    for( CORBA::ULong i = 0; i < bases->length(); i++ ) {
      o << "virtual public ";
      CORBA::String_var base_name = bases[ i ]->absolute_name();
      o << ID(base_name);
      o << _STUB;
      if( i + 1 != bases->length() )
	o << ",";
      o << endl;
    }
    o << exdent;
  }
  
  o << BL_OPEN;
  
  o << "public:" << indent << endl;
  o << "virtual ~" << className << _STUB << "();" << endl;
  
  // Generate prototypes
  emitPrototypes( in, false );
  
  // Generate epilogue
  o << exdent << "private:" << indent << endl;
  o << "void operator=( const " << className << _STUB << "& );"
    << exdent << endl;
  o << BL_CLOSE_SEMI << endl;

  //-----SKELETON----------------------------------------------------
  // Generate skeleton class for interface [18.3.1(6)], [18.4.5]
  if (_params.boa) {
    o << "class " << className << _SKEL << " :" << indent << endl;
    o << "virtual public MethodDispatcher," << endl;
    o << "virtual public " << className << exdent << endl;
    o << BL_OPEN;
    o << "public:" << indent << endl;
    o << className << _SKEL;
    o << "( const CORBA::BOA::ReferenceData & = CORBA::BOA::ReferenceData() );";
    o << endl;
    o << "virtual ~" << className << _SKEL << "();" << endl;
    o << className << _SKEL << "( CORBA::Object_ptr obj );" << endl;
    o << "virtual bool dispatch( CORBA::ServerRequest_ptr _req, "
      "CORBA::Environment &_env );" << endl;
    o << className << "_ptr _this();" << exdent << endl << endl;
    o << BL_CLOSE_SEMI << endl;
  }
}

/*
 * Generates default constructor, copy constructor and assignment
 * operator for structs and unions. These have to be generated due
 * to a g++ bug (g++ doesn't expand templates properly for the implicit
 * generated versions of these methods). Note that according to
 * the CORBA2 specs these shouldn't be defined (see [16.9(1)])
 */
void
CodeGenCPPCommon::emit_DefaultMembers( char *name )
{
  o << "#ifdef HAVE_EXPLICIT_STRUCT_OPS" << endl;

  // Default constructor
  o << name << "();" << endl;

  // Destructor
  o << "~" << name << "();" << endl;

  // Copy constructor
  o << name << "( const " << name << "& s );" << endl;
  
  // Assignment operator
  o << name << "& operator=( const " << name << "& s );" << endl;

  o << "#endif //HAVE_EXPLICIT_STRUCT_OPS" << endl;
}
  


void CodeGenCPPCommon::emit_Struct( CORBA::StructDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_Struct );

  CORBA::String_var tmp = s->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::StructMemberSeq_var m = s->members();
  CORBA::TypeCode_var tc = s->type();

  if( !tc->is_freestanding() ) {
    // This struct contains somewhere nested a recursive sequence
    // which refers not to this struct. We therefore need to emit
    // forward declarations of those outer level structs because
    // the SequenceIndTmpl will make reference to them.
    for( CORBA::ULong i = 0; i < m->length(); i++ ) {
      if( m[ i ].type->is_recursive_seq() ) {
	o << "struct ";
	CORBA::SequenceDef_var seq =
	  CORBA::SequenceDef::_narrow( m[ i ].type_def );
	assert( !CORBA::is_nil( seq ) );
	CORBA::IDLType_var elem = seq->element_type_def();
	CORBA::Contained_var con = CORBA::Contained::_narrow( elem );
	assert( !CORBA::is_nil( con ) );
	tmp = con->name();
	o << ID(tmp) << ";" << endl;
      }
    }
  }
      
  o << "struct " << name << " " << BL_OPEN;
  emit_DefaultMembers( name );
  for( CORBA::ULong i = 0; i < m->length(); i++ ) {
    emit_type_for_variable( m[ i ].type_def, automatic );
    o << " " << ID(m[ i ].name);
    emit_array_suffix( m[ i ].type_def );
    o << ";" << endl;
  }
  o << BL_CLOSE_SEMI << endl;

  // Emit _var type
  o << "typedef " << TVAR( tc ) << "<" << name << "> "
    << name << "_var;" << endl;
  

  // Emit _out type
  o << "typedef " << name << "_var " << name << "_out;";
  o << endl << endl;

  if (_params.pseudo) {
    return;
  }

  // Emit TypeCode for struct
  if( tc->is_freestanding() ) {
    CORBA::Container_var def_in = s->defined_in();
    if( def_in->def_kind() == CORBA::dk_Repository )
      // We are on the toplevel
      o << "extern";
    else if( def_in->def_kind() == CORBA::dk_Module )
      // The current scope is not top level (inside a module)
      o << "MICO_EXPORT_VAR_DECL";
    else
      // The current scope is not top level (not inside a module)
      o << "static";
    o << " CORBA::TypeCodeConst _tc_" << name << ";" << endl << endl;
  }
  
  // operator<<=
  o.switchStream( 1 );
  tmp = s->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs_name << " &_s );" << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs_name << " &_s );" << endl << endl;

  emit_marshal_prototype( s );
  
  o.switchStream( 0 );
}



void CodeGenCPPCommon::emit_Union( CORBA::UnionDef_ptr u )
{
  assert( u->def_kind() == CORBA::dk_Union );

  CORBA::String_var tmp = u->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::UnionMemberSeq_var members = u->members();
  CORBA::TypeCode_var union_tc = u->type();
  
  if( !union_tc->is_freestanding() ) {
    // This union contains somewhere nested a recursive sequence
    // which refers not to this union. We therefore need to emit
    // forward declarations of those outer level unions because
    // the SequenceIndTmpl will make reference to them.
    for( CORBA::ULong i = 0; i < members->length(); i++ ) {
      if( members[ i ].type->is_recursive_seq() ) {
	o << "class ";
	CORBA::SequenceDef_var seq =
	  CORBA::SequenceDef::_narrow( members[ i ].type_def );
	assert( !CORBA::is_nil( seq ) );
	CORBA::IDLType_var elem = seq->element_type_def();
	CORBA::Contained_var con = CORBA::Contained::_narrow( elem );
	assert( !CORBA::is_nil( con ) );
	tmp = con->name();
	o << ID(tmp) << ";" << endl;
      }
    }
  }

  o << "class " << name << " " << BL_OPEN;
  o << "public:" << indent << endl;

  CORBA::IDLType_var discr = u->discriminator_type_def();

  // Emit declaration for the discriminator
  emit_type_for_variable( discr, automatic );
  o << " _discriminator;" << endl << endl;

  // Generate member variables. Note: these can't be put in a C++
  // union as one would probably expect. The reason is that MICO
  // struct's are mapped to C++ struct's with explicit default
  // constructors and destructors (this is necessary due to a
  // g++ bug). But struct's with constructors are not allowed
  // in an union.
  CORBA::String_var last_name;
  last_name = (const char *) "";
  o << "struct __m " << BL_OPEN;

  bool has_explicit_default = false;
  
  for( CORBA::ULong i = 0; i < members->length(); i++ ) {
    CORBA::UnionMember& m = members[ i ];
    CORBA::TypeCode_var tc = members[ i ].label.type();
    if( is_octet( tc ) )
      has_explicit_default = true;

    if( strcmp( last_name, m.name ) == 0 )
      // We already generated a declaration for this name
      continue;
    emit_type_for_variable( m.type_def, automatic );
    o << " " << ID(m.name);
    emit_array_suffix( m.type_def );
    o << ";" << endl;
    last_name = m.name;
  }
  o << exdent << "} _m;" << endl << endl;
  
  o << exdent << "public:" << indent << endl;
  
  emit_DefaultMembers( name );
  o << endl;
  
  // Emit discriminator functions
  o << "void _d( ";
  emit_type_for_param( discr, CORBA::PARAM_IN );
  o << " _p );" << endl;
  emit_type_for_result( discr );
  o << " _d() const;" << endl << endl;

  // Generate member accessor functions
  last_name = (const char *) "";
  for( CORBA::ULong i0 = 0; i0 < members->length(); i0++ ) {
    CORBA::UnionMember &m = members[ i0 ];

    if( strcmp( last_name, m.name ) == 0 )
      continue;
    last_name = m.name;
    
    CORBA::IDLType_var m_type = CORBA::IDLType::_duplicate( m.type_def );
    CORBA::String_var m_name = (const char *) ID(m.name);
    CORBA::TypeCode_var tc = m_type->type();
    if( is_string( tc ) ) {
      // Accessor functions for string
      o << "void " << m_name << "( char *_p );" << endl;
      o << "void " << m_name << "( const char *_p );" << endl;
      o << "void " << m_name << "( const CORBA::String_var &_p );" << endl;
      o << "const char *" << m_name << "() const;" << endl << endl;
    } else if( is_wstring( tc ) ) {
      // Accessor functions for wstring
      o << "void " << m_name << "( wchar_t *_p );" << endl;
      o << "void " << m_name << "( const wchar_t *_p );" << endl;
      o << "void " << m_name << "( const CORBA::WString_var &_p );" << endl;
      o << "const wchar_t *" << m_name << "() const;" << endl << endl;
    } else if( is_array( tc ) && !is_alias( tc ) ) {
      // We have an anonymous array. See [16.10(7)]
      o << "typedef ";
      emit_array( m_type );
      o << " " << "_" << m_name << "_slice";
      emit_array_slice_suffix( m_type );
      o << ";" << endl;
      o << "void " << m_name << "( ";
      emit_array( m_type );
      o << " _p";
      emit_array_suffix( m_type );
      o << " );" << endl;
      o << "_" << m_name << "_slice* " << m_name << "();" << endl << endl;
    } else if( is_array( tc ) || is_objref( tc ) ||
	       is_type_code( tc ) || is_base_type( tc ) ) {
      // Accessor functions for array or object reference
      o << "void " << m_name << "( ";
      if( is_array( tc ) )
	// Mapping for arrays differs from standard PARAM_IN mapping!
	emit_idl_type_name( m_type );
      else
	emit_type_for_param( m_type, CORBA::PARAM_IN );
      o << " _p );" << endl;
      emit_type_for_result( m_type );
      o << " " << m_name << "() const;" << endl << endl;
    } else if( is_sequence( tc ) && !is_alias( tc ) ) {
      // An array, which is not referred to by an alias. Note
      // That this is a special case, as emit_type_for_param
      // does not handle this situation
      o << "void " << m_name << "( const ";
      emit_sequence( m_type );
      o << "& _p );" << endl;
      o << "const ";
      emit_sequence( m_type );
      o << "& " << m_name << "() const;" << endl;
      emit_sequence( m_type );
      o << "& " << m_name << "();" << endl;
    } else {
      // Gotta be struct, union or sequence
      o << "void " << m_name << "( ";
      emit_type_for_param( m_type, CORBA::PARAM_IN );
      o << " _p );" << endl;
      o << "const ";
      emit_idl_type_name( m_type );
      o << "& " << m_name << "() const;" << endl;
      emit_idl_type_name( m_type );
      o << "& " << m_name << "();" << endl << endl;
    }
  }
  if( !has_explicit_default ) {
    // This union has no explicit default case label
    string default_label =
      generate_union_default_case_label( discr, members );
    if( default_label != "" ) {
      // We have a case label left for an implicit default member
      o << "void _default();" << endl << endl;
    }
  }

  o << exdent << BL_CLOSE_SEMI << endl;
      
  // Emit _var type
  o << "typedef " << TVAR( union_tc ) << "<"
    << name << "> " << name << "_var;" << endl;

  // Emit _out type
  o << "typedef " << name << "_var " << name << "_out;";
  o << endl << endl;

  if (_params.pseudo) {
    return;
  }

  // Emit TypeCode for union
  if( union_tc->is_freestanding() ) {
    CORBA::Container_var def_in = u->defined_in();
    if( def_in->def_kind() == CORBA::dk_Repository )
      // We are on the toplevel
      o << "extern";
    else if( def_in->def_kind() == CORBA::dk_Module )
      // The current scope is not top level (inside a module)
      o << "MICO_EXPORT_VAR_DECL";
    else
      // The current scope is not top level (not inside a module)
      o << "static";
    o << " CORBA::TypeCodeConst _tc_" << name << ";" << endl << endl;
  }
  
  // operator<<=
  o.switchStream( 1 );
  tmp = u->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs_name << " &_u );" << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs_name << " &_u );" << endl << endl;

  // SII marshaller
  emit_marshal_prototype( u );
  
  o.switchStream( 0 );
}



void CodeGenCPPCommon::emit_Constant( CORBA::ConstantDef_ptr c )
{
  assert( c->def_kind() == CORBA::dk_Constant );
  CORBA::String_var tmp = c->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::Container_var def_in = c->defined_in();

  if( def_in->def_kind() == CORBA::dk_Repository ||
      def_in->def_kind() == CORBA::dk_Interface ) {
    o << "static const ";
  } else {
    o << "MICO_EXPORT_VAR_DECL const ";
  }
  CORBA::IDLType_var type = c->type_def();
  emit_base_type( type );
  o << " " << name;

  // If this const is declared on a nested scope, we need to
  // defer its initialization because some C++ compilers
  // choke on this
  if( def_in->def_kind() == CORBA::dk_Repository ) {
    // Only initialize if we are on global scope
    o << " = ";
    if( type->def_kind() == CORBA::dk_Fixed )
      o << "(CORBA::LongDouble) ";
    CORBA::Any* a = c->value();
    CORBA::TypeCode_var tc = a->type();
    if( tc->kind() == CORBA::tk_wstring || tc->kind() == CORBA::tk_wchar )
      o << "L";
    o << a;
    delete a;
  }
  o << ";" << endl;
}



void CodeGenCPPCommon::emit_Exception( CORBA::ExceptionDef_ptr e )
{
  assert( e->def_kind() == CORBA::dk_Exception );

  CORBA::String_var tmp = e->name();
  CORBA::String_var name = (const char *) ID(tmp);
  
  o << "struct " << name << " : public CORBA::UserException " << BL_OPEN;
  CORBA::StructMemberSeq_var m = e->members();
  emit_DefaultMembers( name );

  o << "void _throwit() const;" << endl;
  o << "const char *_repoid() const;" << endl;
  if (!_params.pseudo) {
    o << "void _encode( CORBA::DataEncoder &en ) const;" << endl;
  }
  o << "CORBA::Exception *_clone() const;" << endl;
  o << "static " << name << " *_narrow( CORBA::Exception *ex );" << endl;
  
  for( CORBA::ULong i = 0; i < m->length(); i++ ) {
    emit_type_for_variable( m[ i ].type_def, automatic );
    o << " " << ID(m[ i ].name);
    emit_array_suffix( m[ i ].type_def );
    o << ";" << endl;
  }
  o << BL_CLOSE_SEMI << endl;

  // Emit _var type
  o << "typedef ExceptVar<" << name << "> " << name << "_var;" << endl;

  // Emit _out type
  o << "typedef " << name << "_var " << name << "_out;";
  o << endl << endl;

  if (_params.pseudo) {
    return;
  }

  // Emit TypeCode for exception
  CORBA::Container_var def_in = e->defined_in();
  if( def_in->def_kind() == CORBA::dk_Repository )
    // We are on the toplevel
    o << "extern";
  else if( def_in->def_kind() == CORBA::dk_Module )
    // The current scope is not top level (inside a module)
    o << "MICO_EXPORT_VAR_DECL";
  else
    // The current scope is not top level (not inside a module)
    o << "static";
  o << " CORBA::TypeCodeConst _tc_" << name << ";" << endl << endl;
  
  // operator<<=
  o.switchStream( 1 );
  tmp = e->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "CORBA::Boolean operator<<=( CORBA::Any &a, const ";
  o << abs_name << " &e );" << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &a, ";
  o << abs_name << " &e );" << endl << endl;

  // SII marshaller
  emit_marshal_prototype( e );
  
  o.switchStream( 0 );
}



void CodeGenCPPCommon::emit_Enum( CORBA::EnumDef_ptr e )
{
  assert( e->def_kind() == CORBA::dk_Enum );

  CORBA::String_var tmp = e->name();
  CORBA::String_var name = (const char *) ID(tmp);
  o << "enum " << name << " " << BL_OPEN;

  CORBA::EnumMemberSeq_var m = e->members();
  
  for( CORBA::ULong i = 0; i < m->length(); i++ ) {
    if( i != 0 )
      o << "," << endl;
    o << ID(m[ i ]);
    if( i == 0 )
      o << " = 0";
  }
  o << endl;
  o << BL_CLOSE_SEMI << endl;

  // Emit _out for enum
  o << "typedef " << name << "& " << name << "_out;";
  o << endl << endl;
  
  if (_params.pseudo) {
    return;
  }

  // Emit TypeCode for enum
  CORBA::Container_var def_in = e->defined_in();
  if( def_in->def_kind() == CORBA::dk_Repository )
    // We are on the toplevel
    o << "extern";
  else if( def_in->def_kind() == CORBA::dk_Module )
    // The current scope is not top level (inside a module)
    o << "MICO_EXPORT_VAR_DECL";
  else
    // The current scope is not top level (not inside a module)
    o << "static";
  o << " CORBA::TypeCodeConst _tc_" << name << ";" << endl << endl;
  
  // operator<<=
  o.switchStream( 1 );
  tmp = e->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "CORBA::Boolean operator<<=( CORBA::Any &a, const ";
  o << abs_name << " &e );" << endl << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &a, ";
  o << abs_name << " &e );" << endl << endl;

  emit_marshal_prototype( e );
  
  o.switchStream( 0 );
}



void CodeGenCPPCommon::emit_Alias( CORBA::AliasDef_ptr a )
{
  assert( a->def_kind() == CORBA::dk_Alias );
  
  CORBA::IDLType_var orig = a->original_type_def();
  assert( !CORBA::is_nil( orig ) );
  
  CORBA::TypeCode_var orig_tc = orig->type();

  CORBA::String_var tmp = a->name();
  CORBA::String_var name = (const char *) ID(tmp);
  
  if( is_string( orig_tc ) ) {
    o << "typedef char* " << name << ";" << endl;
    o << "typedef CORBA::String_var ";
    o << name << "_var;" << endl;
    o << "typedef CORBA::String_out ";
    o << name << "_out;" << endl << endl;
  } else if( is_wstring( orig_tc ) ) {
    o << "typedef CORBA::WChar* " << name << ";" << endl;
    o << "typedef CORBA::WString_var ";
    o << name << "_var;" << endl;
    o << "typedef CORBA::WString_out ";
    o << name << "_out;" << endl << endl;
  } else if( is_objref( orig_tc ) ) {
    CORBA::String_var abs_name;
    
    if( orig_tc->kind() == CORBA::tk_alias ) {
      // we have an alias of an alias
      CORBA::AliasDef_var alias = CORBA::AliasDef::_narrow( orig );
      assert( !CORBA::is_nil( alias ) );
      abs_name = alias->absolute_name();
    } else {
      // gotta be an interface
      assert( orig_tc->kind() == CORBA::tk_objref );
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( orig );
      if( !CORBA::is_nil( in ) ) {
	// Must be a regular object reference
	abs_name = in->absolute_name();
      } else {
	// We have an "Object" object reference
	abs_name = (const char *) "CORBA::Object";
      }
    }
    abs_name = (const char *) ID(abs_name);

    o << "typedef " << abs_name << " " << name << ";" << endl;
    o << "typedef " << abs_name << "_ptr " << name << "_ptr;" << endl;
    o << "typedef " << abs_name << "Ref " << name << "Ref;" << endl;
    o << "typedef " << abs_name << "_var " << name << "_var;" << endl;
    o << "typedef " << name << "_var " << name << "_out;" << endl;
    o << endl;
  } else {
    o << "typedef ";
    emit_type_for_variable( orig, automatic );
    o << " " << name;
    emit_array_suffix( orig );
    o << ";" << endl;
    
    if( is_array( orig_tc ) ) {

      if( orig_tc->kind() != CORBA::tk_alias ) {
	
	CORBA::ULong num_elem = compute_num_array_elem( orig_tc );

	// Array slice
	o << "typedef ";
	emit_type_for_variable( orig, automatic );
	o << " " << name << "_slice";
	emit_array_slice_suffix( orig );
	o << ";" << endl;
	
	// _var for Array
	o << "typedef Array";
	o << (is_variable( orig_tc ) ? "Var" : "Fix") << "Var<";
	emit_type_for_variable( orig, automatic );
	o << "," << name << "_slice," << name << ",";
	o << num_elem << "> ";
	o << name << "_var;" << endl;
	
	// _out for Array
	o << "typedef " << name << "_var " << name << "_out;" << endl;
	
	// _forany for Array
	o << "typedef ArrayForAny<";
	emit_type_for_variable( orig, automatic );
	o << "," << name << "_slice," << name << ",";
	o << num_elem << "> ";
	o << name << "_forany;" << endl;
	
	// T_alloc for array
	o << "DECLARE_ARRAY_ALLOC(" << name << ",";
	emit_type_for_variable( orig, automatic );
	o << "," << name << "_slice," << num_elem << ")" << endl;

	// T_dup for array
	o << "DECLARE_ARRAY_DUP(" << name << ",";
	emit_type_for_variable( orig, automatic );
	o << "," << name << "_slice," << num_elem << ")" << endl;

	// T_free for array
	o << "DECLARE_ARRAY_FREE(" << name << ",";
	emit_type_for_variable( orig, automatic );
	o << "," << name << "_slice," << num_elem << ")" << endl;

	// T_copy for array (MICO extension)
	o << "DECLARE_ARRAY_COPY(" << name << ",";
	emit_type_for_variable( orig, automatic );
	o << "," << name << "_slice," << num_elem << ")" << endl;

	if (!_params.pseudo) {
	  o.switchStream( 1 );
	  CORBA::String_var abs_name = a->absolute_name();
	  abs_name = (const char *) ID(abs_name);
	  
	  // operator<<=()
	  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
	  o << abs_name << "_forany &_s );" << endl;
	  
	  // operator>>=()
	  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
	  o << abs_name << "_forany &_s );" << endl;
	  o.switchStream( 0 );
	}

      } else {
	// We have an alias of an array

	CORBA::IDLType_var unalias_orig = CORBA::IDLType::_duplicate( orig );
	while( unalias_orig->def_kind() == CORBA::dk_Alias ) {
	  CORBA::AliasDef_var tmp = CORBA::AliasDef::_narrow( unalias_orig );
	  assert( !CORBA::is_nil( tmp ) );
	  unalias_orig = tmp->original_type_def();
	}
	
	CORBA::TypeCode_var unalias_orig_tc = unalias_orig->type();
	CORBA::ULong num_elem = compute_num_array_elem( unalias_orig_tc );

	// Array slice
	o << "typedef ";
	emit_type_for_variable( orig, automatic );
	o << "_slice " << name << "_slice";
	o << ";" << endl;
	
	// _var for Array
	o << "typedef ";
	emit_type_for_variable( orig, automatic );
	o << "_var " << name << "_var";
	o << ";" << endl;
	
	// _out for Array
	o << "typedef " << name << "_var " << name << "_out;" << endl;
	
	// _forany for Array
	o << "typedef ";
	emit_type_for_variable( orig, automatic );
	o << "_forany " << name << "_forany";
	o << ";" << endl;
	
	// T_alloc for array
	o << "DECLARE_ARRAY_ALLOC(" << name << ",";
	emit_type_for_variable( unalias_orig, automatic );
	o << "," << name << "_slice," << num_elem << ")" << endl;
	
	// T_dup for array
	o << "DECLARE_ARRAY_DUP(" << name << ",";
	emit_type_for_variable( unalias_orig, automatic );
	o << "," << name << "_slice," << num_elem << ")" << endl;
	
	// T_free for array
	o << "DECLARE_ARRAY_FREE(" << name << ",";
	emit_type_for_variable( unalias_orig, automatic );
	o << "," << name << "_slice," << num_elem << ")" << endl;
      }
    }
  
    if( !is_base_type( orig_tc ) && !is_enum( orig_tc ) &&
	!is_array( orig_tc ) ) {
      // We have to generate _var and _out types as well
      // First a hack for VC++. Since it cannot instantiate templates
      // of templates properly, we force an instantiation through
      // a dummy variable declaration
      o << "#ifdef _WINDOWS" << endl;
      o << "static " << name << " _dummy_" << name << ";" << endl;
      o << "#endif" << endl;
      
      o << "typedef ";
      if( is_sequence( orig_tc ) ) {
	o << "TSeqVar<";
	emit_type_for_variable( orig, automatic );
	o << " >";
      } else {
	emit_type_for_variable( orig, automatic );
	o << "_var";
      }
      o << " " << name << "_var;" << endl;
      o << "typedef " << name << "_var " << name << "_out;" << endl << endl;
    }
  }

  if (_params.pseudo) {
    o << endl;
    return;
  }

  // Emit TypeCode for alias
  CORBA::Container_var def_in = a->defined_in();
  assert( !CORBA::is_nil( def_in ) );
  
  if( def_in->def_kind() == CORBA::dk_Repository )
    // We are on the toplevel
    o << "extern";
  else if( def_in->def_kind() == CORBA::dk_Module )
    // The current scope is not top level (inside a module)
    o << "MICO_EXPORT_VAR_DECL";
  else
    // The current scope is not top level (not inside a module)
    o << "static";
  o << " CORBA::TypeCodeConst _tc_" << name << ";" << endl << endl;
  
  o.switchStream( 1 );
  emit_marshal_prototype( a );
  o.switchStream( 0 );
}

/*
 * Emit definitions for PortableServer skeleton classes.
 *
 * The C++ mapping dictates that the skeleton class for an interface
 * B within module A becomes POA_A::B, which is in the POA_A namespace.
 */

void
CodeGenCPPCommon::emit_poa_obj (CORBA::IRObject_ptr obj)
{
  if (check_for_included_defn (obj))
    return;

  CORBA::Contained_var c = CORBA::Contained::_narrow ( obj );
  if( !CORBA::is_nil( c ) ) {
    CORBA::Container_var new_scope = c->defined_in();
    enter_scope( new_scope, "POA_" );
  }

  switch (obj->def_kind()) {
  case CORBA::dk_Interface:
    {
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( obj );
      emit_poa_skel (in);
      break;
    }
  }
}

void
CodeGenCPPCommon::emit_poa_skel (CORBA::InterfaceDef_ptr in)
{
  CORBA::String_var n = in->name();
  CORBA::String_var a = in->absolute_name();
  string className = (const char *) ID(n);
  string absName = (const char *) ID(a);
  absName = absName.substr (2);

  string prefix;
  CORBA::Container_var def_in = in->defined_in ();
  if (def_in->def_kind() == CORBA::dk_Repository) {
    prefix = (const char *) "POA_";
  }
  else {
    prefix = (const char *) "";
  }

  bool oldurn = use_rel_names (false);

  /*
   * PortableServer Skeleton class
   */

  CORBA::InterfaceDefSeq_var bc = in->base_interfaces ();

  o << "class " << prefix << className << " : ";
  if (bc->length() == 0) {
    o << "virtual public PortableServer::DynamicImplementation" << endl;
  } else {
    o << indent << endl;
    for (CORBA::ULong i=0; i<bc->length(); i++) {
      CORBA::String_var base_name = bc[i]->absolute_name();
      string absBaseName = ID (base_name);
      absBaseName = absBaseName.substr (2);
      o << "virtual public POA_" << absBaseName;
      if (i+1 != bc->length()) {
	o << ",";
      }
      o << endl;
    }
    o << exdent;
  }

  o << BL_OPEN;
  o << "public:" << indent << endl;
  o << "virtual ~" << prefix << className << " ();" << endl;
  o << absName << "_ptr _this ();" << endl;
  o << "bool dispatch (CORBA::ServerRequest_ptr);" << endl;
  o << "virtual bool _is_a (const char *);" << endl;
  o << "virtual void invoke (CORBA::ServerRequest_ptr);" << endl;
  o << "virtual CORBA::RepositoryId _primary_interface";
  o <<   " (const PortableServer::ObjectId &,";
  o <<    " PortableServer::POA_ptr);" << endl;
  o << endl;

  emitPrototypes (in, true);

  o << endl;
  o << exdent << "protected:" << indent << endl;
  o << prefix << className << " () {};" << endl << endl;
  o << exdent << "private:" << indent << endl;
  o << prefix << className << " (const " << prefix << className << " &);" << endl;
  o << "void operator= (const " << prefix << className << " &);" << endl;
  o << exdent << BL_CLOSE_SEMI << endl;

  /*
   * PortableServer Tie class
   */

  /*
   * XXX Don't generate Tie classes for non-toplevel interfaces;
   *     templates in struct's generate compiler errors
   */

#ifndef HAVE_NAMESPACE
  if (def_in->def_kind() != CORBA::dk_Repository) {
    use_rel_names (oldurn);
    return;
  }
#endif

  o << "template<class T>" << endl;
  o << "class " << prefix << className << "_tie : " << indent << endl;
  if (bc->length() == 0) {
    o << "virtual public POA_Tie_Base<T>," << endl;
  }
  else {
    for (CORBA::ULong i=0; i<bc->length(); i++) {
      CORBA::String_var base_name = bc[i]->absolute_name();
      string absBaseName = ID (base_name);
      absBaseName = absBaseName.substr (2);
      o << "virtual public POA_" << absBaseName << "_tie<T>," << endl;
    }
  }
  o << "virtual public " << prefix << className << exdent << endl;
  o << BL_OPEN;
  o << "public:" << indent << endl;

  o << prefix << className << "_tie (T &t)" << indent << endl;
  o << ": POA_Tie_Base<T> (&t, PortableServer::POA::_nil(), FALSE)"
    << exdent << endl << "{}" << endl;

  o << prefix << className << "_tie (T &t, PortableServer::POA_ptr _poa)"
    << indent << endl;
  o << ": POA_Tie_Base<T> (&t, _poa, FALSE)"
    << exdent << endl << "{}" << endl;

  o << prefix << className << "_tie (T *t, CORBA::Boolean _rel = TRUE)"
    << indent << endl;
  o << ": POA_Tie_Base<T> (t, PortableServer::POA::_nil(), _rel)"
    << exdent << endl << "{}" << endl;

  o << prefix << className << "_tie (T *t, PortableServer::POA_ptr _poa, CORBA::Boolean _rel = TRUE)"
    << indent << endl;
  o << ": POA_Tie_Base<T> (t, _poa, _rel)"
    << exdent << endl << "{}" << endl;
  o << "virtual ~" << prefix << className << "_tie ();" << endl;
  o << endl;
  o << "PortableServer::POA_ptr _default_POA ();" << endl;
  o << endl;

  emitPrototypes (in, false);

  o << endl;
  o << exdent << "protected:" << indent << endl;
  o << prefix << className << "_tie () {};" << endl << endl;
  o << exdent << "private:" << indent << endl;
  o << prefix << className << "_tie (const "
    << prefix << className << "_tie &);" << endl;
  o << "void operator= (const " << prefix << className << "_tie &);" << endl;
  o << exdent << BL_CLOSE_SEMI << endl;

  /*
   * PortableServer Tie Implementations
   */

  o << "template<class T>" << endl;
  o << prefix << className << "_tie<T>::~"
    << prefix << className << "_tie ()" << endl;
  o << BL_OPEN << BL_CLOSE << endl;

  o << "template<class T>" << endl;
  o << "PortableServer::POA_ptr" << endl;
  o << prefix << className << "_tie<T>::_default_POA ()" << endl;
  o << BL_OPEN;
  o << "if (!CORBA::is_nil (poa)) " << BL_OPEN;
  o << "return PortableServer::POA::_duplicate (poa);" << endl;
  o << BL_CLOSE;
  o << "return PortableServer::DynamicImplementation::_default_POA ();"
    << endl;
  o << BL_CLOSE << endl;

  emit_poa_ties (in, prefix, className);
  use_rel_names (oldurn);
}

void
CodeGenCPPCommon::emit_poa_ties (CORBA::InterfaceDef_ptr in,
				 string & prefix,
				 string & className)
{
  CORBA::InterfaceDef::FullInterfaceDescription_var desc;
  desc = in->describe_interface();

  // Emit attributes
  for (CORBA::ULong j0=0; j0<desc->attributes.length(); j0++) {
    CORBA::AttributeDescription attr;
    attr = desc->attributes[j0];
    CORBA::IDLType_var type_id = lookup_attribute_by_id (attr.id);

    o << "template<class T>" << endl;
    emit_type_for_result (type_id);
    o << endl;
    o << prefix << className << "_tie<T>::" << ID(attr.name) << " ()" << endl;
    o << BL_OPEN;
    o << "return ptr->" << ID(attr.name) << " ();" << endl;
    o << BL_CLOSE << endl;

    if (attr.mode == CORBA::ATTR_NORMAL) {
      o << "template<class T>" << endl;
      o << "void" << endl;
      o << prefix << className << "_tie<T>::" << ID(attr.name) << " (";
      emit_type_for_param (type_id, CORBA::PARAM_IN);
      o << " value)" << endl;
      o << BL_OPEN;
      o << "ptr->" << ID(attr.name) << " (value);" << endl;
      o << BL_CLOSE << endl;
    }
  }

  // Emit operations
  for (CORBA::ULong j1=0; j1<desc->operations.length(); j1++) {
    CORBA::OperationDescription op;
    op = desc->operations[j1];
    CORBA::IDLType_var type_id = lookup_result_by_id (op.id);
    CORBA::TypeCode_var tc = type_id->type();

    o << "template<class T>" << endl;
    emit_type_for_result (type_id);
    o << endl;
    o << prefix << className << "_tie<T>::" << ID(op.name) << " (";

    CORBA::ParDescriptionSeq p = op.parameters;
    CORBA::ULong k;
    for (k=0; k<p.length(); k++) {
      if( k != 0 )
	o << ", ";
      emit_type_for_param (p[k].type_def, p[k].mode);
      o << " " << ID(p[k].name);
    }

    // Do we have a context?
    CORBA::ContextIdSeq& ctx = op.contexts;
    
    if (ctx.length() > 0) {
      if (p.length() > 0)
	o << ", ";
      o << "CORBA::Context_ptr _ctx";
    }

    o << ")" << endl;
    o << BL_OPEN;

    if (tc->kind() != CORBA::tk_void) {
      o << "return ";
    }

    o << "ptr->" << ID(op.name) << " (";
    
    for (k=0; k<p.length(); k++) {
      if (k != 0)
	o << ", ";
      o << ID(p[k].name);
    }

    if (ctx.length() > 0) {
      if (p.length() > 0)
	o << ", ";
      o << "_ctx";
    }

    o << ");" << endl;
    o << BL_CLOSE << endl;
  }
}

