/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */


/*
 * The GNU compiler chokes on this:
 *
 * struct S {
 *   char x[2][3];
 *   S( const S &s )
 *   {
 *     x = s.x;
 *   }
 * };
 *
 * It works, when the const-ness is cast away. The appropriate
 * locations, which are influenced by this bug, are marked with "XXXarray"
 */


#include <CORBA.h>
#include <fstream.h>
#include <ctype.h>
#include <stdio.h>
#include "codegen-c++-stub.h"
#include <mico/template_impl.h>
#include <mico/util.h>


CodeGenCPPStub::CodeGenCPPStub( DB &db, IDLParam &params )
  : CodeGen( db.get_idl_container() ),
    CodeGenCPPUtil( db, params )
{
}



void CodeGenCPPStub::emit_stub()
{
  set_name_prefix( CORBA::Container::_nil() );

  for( CORBA::ULong i = 0; i < _idl_objs.length(); i++ ) {
    emit_idl_obj( _idl_objs[ i ] );
  }
}

void CodeGenCPPStub::emit_idl_obj( CORBA::IRObject_ptr obj )
{
  if( check_for_included_defn( obj ) )
    return;
    
  switch( obj->def_kind() ) {
  case CORBA::dk_Interface:
    {
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( obj );
      emit_Interface( in );
      break;
    }
  case CORBA::dk_Struct:
    {
      CORBA::StructDef_var s = CORBA::StructDef::_narrow( obj );
      emit_Struct( s );
      break;
    }
  case CORBA::dk_Union:
    {
      CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( obj );
      emit_Union( u );
      break;
    }
  case CORBA::dk_Constant:
    {
      CORBA::ConstantDef_var cst = CORBA::ConstantDef::_narrow( obj );
      emit_Constant( cst );
      break;
    }
  case CORBA::dk_Exception:
    {
      CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( obj );
      emit_Exception( e );
      break;
    }
  case CORBA::dk_Enum:
    {
      CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( obj );
      emit_Enum( e );
      break;
    }
  case CORBA::dk_Alias:
    {
      CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( obj );
      emit_Alias( a );
      break;
    }
  case CORBA::dk_String:
    {
      CORBA::StringDef_var s = CORBA::StringDef::_narrow( obj );
      emit_StringDef( s );
      break;
    }
  case CORBA::dk_Wstring:
    {
      CORBA::WstringDef_var ws = CORBA::WstringDef::_narrow( obj );
      emit_WstringDef( ws );
      break;
    }
  case CORBA::dk_Sequence:
    {
      CORBA::SequenceDef_var s = CORBA::SequenceDef::_narrow( obj );
      if(!is_base_sequence( s ))
        emit_SequenceDef( s );
      break;
    }
  case CORBA::dk_Fixed:
    {
      CORBA::FixedDef_var f = CORBA::FixedDef::_narrow( obj );
      emit_FixedDef( f );
      break;
    }
  case CORBA::dk_Array:
    {
      CORBA::ArrayDef_var a = CORBA::ArrayDef::_narrow( obj );
      emit_ArrayDef( a );
      break;
    }
  default:
    assert( 0 );
  }
}

void
CodeGenCPPStub::emit_StringDef( CORBA::StringDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_String );

  if( _params.pseudo )
    return;

  CORBA::TypeCode_var type_tc = s->type();
	  
  // Must be a bounded string
  assert( type_tc->length() > 0 );

  string name = "string_";
  name += xdec (type_tc->length());
  emit_typecode_def (CORBA::Container::_nil(), "", name, type_tc);
  o << endl;
}

void
CodeGenCPPStub::emit_WstringDef( CORBA::WstringDef_ptr ws )
{
  assert( ws->def_kind() == CORBA::dk_Wstring );

  if( _params.pseudo )
    return;
  
  CORBA::TypeCode_var type_tc = ws->type();
	  
  // Must be a bounded wstring
  assert( type_tc->length() > 0 );

  string name = "wstring_";
  name += xdec (type_tc->length());
  emit_typecode_def (CORBA::Container::_nil(), "", name, type_tc);
  o << endl;
}

void
CodeGenCPPStub::emit_SequenceDef( CORBA::SequenceDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_Sequence );

  if( _params.pseudo )
    return;
  
  CORBA::TypeCode_var type_tc = s->type();

  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  emit_sequence( s );
  o << " &_s )" << endl;
  o << BL_OPEN;
  // We have no TypeCode for recursive sequences
  if( !type_tc->is_recursive_seq() ) {
    o << "static CORBA::TypeCodeConst _tc ="
      << endl << indent;
    emit_type_code( type_tc );
    o << ";" << exdent << endl;
    o << "_a.type( _tc );" << endl;
  }
  o << "if( !_a.seq_put_begin( _s.length() ) )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "for( CORBA::ULong _i = 0; _i < _s.length(); _i++ )"
    << indent << endl;
  o << "if( !(_a <<= ";

  ostrstream ostr;
  {
    Output tmp;
    o.save_streams( tmp );
    o.start_output( ostr );
    o << "((";
    emit_sequence( s );
    o << "&)_s)[ _i ]";
    o.stop_output();
    o.restore_streams( tmp );
  }
  ostr << ends;
  string var = ostr.str();
  ostr.rdbuf()->freeze( 0 );
	    
  CORBA::IDLType_var seq_elem = s->element_type_def();
  emit_any_param( seq_elem, (char *) var.c_str(), true );
  o << ") )" << endl;
  o << indent << "return FALSE;" << exdent << exdent << endl;
  o << "return _a.seq_put_end();" << endl;
  o << BL_CLOSE << endl;
	    
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  emit_sequence( s );
  o << " &_s )" << endl;
  o << BL_OPEN;
  o << "CORBA::ULong _len;" << endl << endl;
  o << "if( !_a.seq_get_begin( _len ) )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "_s.length( _len );" << endl;
  o << "for( CORBA::ULong _i = 0; _i < _len; _i++ )"
    << indent << endl;
  o << "if( !(_a >>= ";
  emit_any_param( seq_elem, "_s[ _i ]", false );
  o << ") )" << endl;
  o << indent << "return FALSE;" << exdent << exdent << endl;
  o << "return _a.seq_get_end();" << endl;
  o << BL_CLOSE << endl << endl;

  emit_marshaller( s );
}

void
CodeGenCPPStub::emit_FixedDef( CORBA::FixedDef_ptr f )
{
  assert( f->def_kind() == CORBA::dk_Fixed );

  if( _params.pseudo )
    // XXX should still emit _var and _out type for Fixed
    return;
  
  CORBA::TypeCode_var type_tc = f->type();

  string name = "fixed_";
  name += xdec (type_tc->fixed_digits());
  name += "_";
  if( type_tc->fixed_scale() < 0 ) {
      name += "n";
      name += xdec (-type_tc->fixed_scale());
  } else {
      name += xdec (type_tc->fixed_scale());
  }
  emit_typecode_def (CORBA::Container::_nil(), "", name, type_tc);
  o << endl;

  // operator <<=
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, ";
  emit_fixed( f );
  o << " &_f )" << endl << BL_OPEN;
  o << "FixedBase::FixedValue_var _val = _f.to_digits();" << endl;
  o << "return( _a <<= CORBA::Any::from_fixed( _val, ";
  o << type_tc->fixed_digits() << ", ";
  o << type_tc->fixed_scale() << " ) );" << endl;
  o << BL_CLOSE << endl;
	    
  // operator >>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  emit_fixed( f );
  o << " &_f )" << endl << BL_OPEN;
  o << "FixedBase::FixedValue_var _val;" << endl;
  o << "if( !( _a >>= CORBA::Any::to_fixed( _val, ";
  o << type_tc->fixed_digits() << ", ";
  o << type_tc->fixed_scale() << " ) ) )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "_f.from_digits( _val );" << endl;
  o << "return TRUE;" << endl;
  o << BL_CLOSE << endl;

  emit_marshaller( f );
}

void
CodeGenCPPStub::emit_ArrayDef( CORBA::ArrayDef_ptr a )
{
  assert( a->def_kind() == CORBA::dk_Array );

  if( _params.pseudo )
    return;
  
  CORBA::TypeCode_var type_tc = a->type();

  //XXXarray, second argument should be const
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, ";
  emit_array( a );
  o << " (&_s)";
  emit_array_suffix( a );
  o << " )" << endl;
  o << BL_OPEN;
  o << "static CORBA::TypeCodeConst _tc ="
    << endl << indent;
  emit_type_code( type_tc );
  o << ";" << exdent << endl;
  o << "_a.type( _tc );" << endl;
  emit_array_marshalling( a, true, 0 );
  o << "return TRUE;" << endl;
  o << BL_CLOSE << endl;
	    
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  emit_array( a );
  o << " (&_s)";
  emit_array_suffix( a );
  o << " )" << endl;
  o << BL_OPEN;
  emit_array_marshalling( a, false, 0 );
  o << "return TRUE;" << endl;
  o << BL_CLOSE << endl << endl;

  // SII marshaller
  emit_marshaller( a );
}

void
CodeGenCPPStub::emit_marshaller_suffix( CORBA::IRObject_ptr obj )
{
  CORBA::IDLType_var t = CORBA::IDLType::_narrow( obj );
  if( CORBA::is_nil( t ) )
    return;
  
  CORBA::TypeCode_var tc = t->type();
  if( is_string( tc ) || is_objref( tc ) || is_type_code( tc ) )
    o << ".inout()";
}

void
CodeGenCPPStub::emit_marshaller( CORBA::IRObject_ptr obj )
{
  if( !_params.sii )
    return;

  string mangled_name;
  
  gen_mangled_name( mangled_name, obj );
  
  CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( obj );
  if( !CORBA::is_nil( a ) ) {
    // For aliases, we just use the marshaller of the original type
#ifdef WINDOWS_TC
    o.switchStream( 1 );
    o << mangled_name << " = ";
    CORBA::IDLType_var orig = a->original_type_def();
    string marshaller;
    gen_mangled_name( marshaller, orig );
    if( marshaller[ 1 ] == 'm' ) {
      o << "_get" << marshaller << "()";
    } else {
      emit_marshaller_ref( orig );
    }
    o << ";" << endl;
    o.switchStream( 0 );

    o << "CORBA::StaticTypeInfo* _get";
    emit_marshaller_ref( obj );
    o << "()" << endl << BL_OPEN;
    o << "return ";
    if( marshaller[ 1 ] == 'm' ) {
      o << "_get" << marshaller << "()";
    } else {
      emit_marshaller_ref( orig );
    }
    o << ";" << endl;
    o << BL_CLOSE << endl;

#else
#error Must use WINDOWS_TC here
#endif
    return;
  }

  // Mangled name must be something like "_marshaller_*"
  assert( mangled_name[ 0 ] == '_' && mangled_name[ 1 ] == 'm' );
  
  // Change lower case 'm' to upper case 'M' to get class name
  mangled_name[ 1 ] = 'M';

  o << "class " << mangled_name
    << " : public CORBA::StaticTypeInfo ";
  o << BL_OPEN;
  o << indent;

  // generic typedef
  CORBA::TypeCode_var tc;

  o << "typedef ";
  CORBA::IDLType_var t = CORBA::IDLType::_narrow( obj );
  if( !CORBA::is_nil( t ) ) {
    
    tc = t->type();
  
    if( is_sequence( tc ) ) {
      emit_sequence( t );
      o << " _T";
    } else if( tc->kind() == CORBA::tk_fixed ) {
      emit_fixed( t );
      o << " _T";
    } else if( is_array( tc ) ) {
      CORBA::ArrayDef_var arr = CORBA::ArrayDef::_narrow( obj );
      assert( !CORBA::is_nil( arr ) );

      CORBA::IDLType_ptr t = arr->element_type_def();
      emit_type_for_variable( t, automatic );
      o << " _T";
      emit_array_suffix( t );
    } else {
      emit_idl_type_name( t );
      o << " _T";
    }
  } else {
    
    CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( obj );
    assert( !CORBA::is_nil( e ) );
    tc = CORBA::_tc_null;
    CORBA::String_var name = e->absolute_name();
    o << ID(name) << " _T";
  }

  o << ";" << endl;

  o << exdent << "public:" << endl << indent;

  // create
  o << "StaticValueType create() const" << endl;
  o << BL_OPEN;
  o << "return (StaticValueType) ";
  if( is_array( tc ) ) {
    CORBA::ArrayDef_var a = CORBA::ArrayDef::_narrow( obj );
    assert( !CORBA::is_nil( a ) );
    CORBA::ULong len = a->length();
    o << "new _T[ " << len << " ]";
  } else {
    o << "new _T";
    if( is_objref( tc ) )
      o << "( 0 )";
  }
  o << ";" << endl;
  o << BL_CLOSE << endl;
  
  // copy
  o << "StaticValueType copy( const StaticValueType v ) const" << endl;
  o << BL_OPEN;
  if( is_array( tc ) ) {
    CORBA::ArrayDef_var a = CORBA::ArrayDef::_narrow( obj );
    assert( !CORBA::is_nil( a ) );
    CORBA::ULong len = a->length();
    o << "_T* c = new _T[ " << len << " ];" << endl;
    o << "for( int i = 0; i < " << len << "; i++ )" << endl;
    o << indent << "c[ i ] = ((_T *) v)[ i ];" << exdent << endl;
    o << "return (StaticValueType) c";
  } else {
    o << "return (StaticValueType) new _T( ";
    if( is_objref( tc ) ) {
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( obj );
      CORBA::String_var name = in->absolute_name();
      o << ID(name) << "::_duplicate( *(_T*) v ) )";
    } else {
      o << "*(_T*)v )";
    }
  }
  o << ";" << endl;
  o << BL_CLOSE << endl;
  
  // assign
  o << "void assign( StaticValueType d, "
    << " const StaticValueType s ) const" << endl;
  o << BL_OPEN;
  if( is_array( tc ) ) {
    CORBA::ArrayDef_var a = CORBA::ArrayDef::_narrow( obj );
    assert( !CORBA::is_nil( a ) );
    CORBA::ULong len = a->length();
    o << "for( int i = 0; i < " << len << "; i++ )" << endl;
    o << indent << "((_T *) d)[ i ] = ((_T *) s)[ i ]" << exdent;
  } else {
    o << "*(_T*) d = ";
    if( is_objref( tc ) ) {
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( obj );
      CORBA::String_var name = in->absolute_name();
      o << ID(name) << "::_duplicate( *(_T*) s )";
    } else {
      o << "*(_T*) s";
    }
  }
  o << ";" << endl;
  o << BL_CLOSE << endl;
  
  // free
  o << "void free( StaticValueType v ) const" << endl;
  o << BL_OPEN;
  if( is_array( tc ) ) {
    o << "delete[] (_T *) v";
  } else {
    if( is_objref( tc ) ) {
      o << "CORBA::release( *(_T *) v );" << endl;
    }
    o << "delete (_T*) v";
  }
  o << ";" << endl;
  o << BL_CLOSE << endl;
  
  // demarshal
  o << "CORBA::Boolean demarshal( CORBA::DataDecoder &dc, "
    << "StaticValueType v ) const" << endl;
  o << BL_OPEN;

  switch( obj->def_kind() ) {
  case CORBA::dk_Struct:
    {
      CORBA::StructDef_var s = CORBA::StructDef::_narrow( obj );
      CORBA::StructMemberSeq_var mem = s->members();
      
      o << "return" << endl << indent;
      o << "dc.struct_begin() &&" << endl;
      for( CORBA::ULong i = 0; i < mem->length(); i++ ) {
	emit_marshaller_ref( mem[ i ].type_def );
	o << "->demarshal( dc, &((_T*)v)->" << mem[ i ].name;
	emit_marshaller_suffix( mem[i].type_def );
	o << " ) &&" << endl;
      }
      o << "dc.struct_end();" << endl << exdent;
      break;
    }
  case CORBA::dk_Union:
    {
      CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( obj );
      CORBA::UnionMemberSeq_var mem = u->members();
      CORBA::IDLType_var disc = u->discriminator_type_def();
      
      o << "if( !dc.union_begin() )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      o << "if( !";
      emit_marshaller_ref( disc );
      o << "->demarshal( dc, &((_T *) v)->_discriminator";
      emit_marshaller_suffix( disc );
      o << " ) )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      
      o << "switch( ((_T *) v)->_discriminator ) " << BL_OPEN;
      for( CORBA::ULong i = 0; i < mem->length(); i++ ) {
	CORBA::UnionMember &m = mem[ i ];
	CORBA::TypeCode_var tc = m.label.type();
	if( tc->kind() == CORBA::tk_octet ) {
	  o << "default";
	} else {
	  o << "case ";
	  emit_union_case_label( disc, m, mem );
	}
	o << ":" << indent << endl;
	o << "if( !";
	emit_marshaller_ref( m.type_def );
	o << "->demarshal( dc, &((_T *) v)->_m." << ID(m.name);
	emit_marshaller_suffix( m.type_def );
	o << " ) )" << endl;
	o << indent << "return FALSE;" << exdent << endl;
	o << "break;" << exdent << endl;
      }
      o << BL_CLOSE;
      o << "return dc.union_end();" << endl;
      break;
    }
  case CORBA::dk_Enum:
    {
      o << "CORBA::ULong ul;" << endl;
      o << "if( !dc.enumeration( ul ) )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      o << "*(_T*) v = (_T) ul;" << endl;
      o << "return TRUE;" << endl;
      break;
    }
  case CORBA::dk_Sequence:
    {
      CORBA::SequenceDef_var s = CORBA::SequenceDef::_narrow( obj );
      CORBA::IDLType_var elem = s->element_type_def();

      o << "CORBA::ULong len;" << endl;
      o << "if( !dc.seq_begin( len ) )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      o << "((_T *) v)->length( len );" << endl;
      o << "for( CORBA::ULong i = 0; i < len; i++ ) " << BL_OPEN;
      o << "if( !";
      emit_marshaller_ref( elem );
      o << "->demarshal( dc, &(*(_T*)v)[i]";
      emit_marshaller_suffix( elem );
      o << " ) )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      o << BL_CLOSE;
      o << "return dc.seq_end();" << endl;
      break;
    }
  case CORBA::dk_Array:
    {
      CORBA::ArrayDef_var a = CORBA::ArrayDef::_narrow( obj );
      CORBA::IDLType_var elem = a->element_type_def();
      CORBA::ULong len = a->length();
      
      o << "if( !dc.arr_begin() )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      o << "for( CORBA::ULong i = 0; i < " << len << "; i++ ) " << BL_OPEN;
      o << "if( !";
      emit_marshaller_ref( elem );
      o << "->demarshal( dc, &((_T*)v)[i]";
      emit_marshaller_suffix( elem );
      o << " ) )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      o << BL_CLOSE;
      o << "return dc.arr_end();" << endl;
      break;
    }
  case CORBA::dk_Fixed:
    {
      CORBA::FixedDef_var f = CORBA::FixedDef::_narrow( obj );
      o << "FixedBase::FixedValue val;" << endl;
      o << "if( !dc.get_fixed( val, " << f->digits() << ", "
	<< f->scale() << " ) )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      o << "((_T *) v)->from_digits( val );" << endl;
      o << "return TRUE;" << endl;
      break;
    }
  case CORBA::dk_Interface:
    {
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( obj );
      o << "CORBA::Object_ptr obj;" << endl;
      o << "if( !CORBA::_stc_Object->demarshal( dc, &obj ) )" << endl;
      o << indent << "return FALSE;" << exdent << endl;
      o << "*(_T *) v = ";
      CORBA::String_var name = in->absolute_name();
      o << ID(name) << "::_narrow( obj );" << endl;
      o << "CORBA::release( obj );" << endl;
      o << "return TRUE;" << endl;
      break;
    }
  case CORBA::dk_Exception:
    {
      CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( obj );
      CORBA::StructMemberSeq_var mem = e->members();
      
      o << "string repoid;" << endl;
      o << "return" << endl << indent;
      o << "dc.except_begin( repoid ) &&" << endl;
      for( CORBA::ULong i = 0; i < mem->length(); i++ ) {
	emit_marshaller_ref( mem[ i ].type_def );
	o << "->demarshal( dc, &((_T*)v)->" << mem[ i ].name;
	emit_marshaller_suffix( mem[i].type_def );
	o << " ) &&" << endl;
      }
      o << "dc.except_end();" << endl << exdent;
      break;
    }
  default:
    assert( 0 );
  }
  o << BL_CLOSE << endl;
  
  // marshal
  o << "void marshal( CORBA::DataEncoder &ec, StaticValueType v ) const"
    << endl;
  o << BL_OPEN;

  switch( obj->def_kind() ) {
  case CORBA::dk_Struct:
    {
      CORBA::StructDef_var s = CORBA::StructDef::_narrow( obj );
      CORBA::StructMemberSeq_var mem = s->members();
      
      o << "ec.struct_begin();" << endl;
      for( CORBA::ULong i = 0; i < mem->length(); i++ ) {
	emit_marshaller_ref( mem[ i ].type_def );
	o << "->marshal( ec, &((_T*)v)->" << mem[ i ].name;
	emit_marshaller_suffix( mem[i].type_def );
	o << " );" << endl;
      }
      o << "ec.struct_end();" << endl;
      break;
    }
  case CORBA::dk_Union:
    {
      CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( obj );
      CORBA::UnionMemberSeq_var mem = u->members();
      CORBA::IDLType_var disc = u->discriminator_type_def();
      
      o << "ec.union_begin();" << endl;
      emit_marshaller_ref( disc );
      o << "->marshal( ec, &((_T *) v)->_discriminator";
      emit_marshaller_suffix( disc );
      o << " );" << endl;
      
      o << "switch( ((_T *) v)->_discriminator ) " << BL_OPEN;
      for( CORBA::ULong i = 0; i < mem->length(); i++ ) {
	CORBA::UnionMember &m = mem[ i ];
	CORBA::TypeCode_var tc = m.label.type();
	if( tc->kind() == CORBA::tk_octet ) {
	  o << "default";
	} else {
	  o << "case ";
	  emit_union_case_label( disc, m, mem );
	}
	o << ":" << indent << endl;
	emit_marshaller_ref( m.type_def );
	o << "->marshal( ec, &((_T *) v)->_m." << ID(m.name);
	emit_marshaller_suffix( m.type_def );
	o << " );" << endl;
	o << "break;" << exdent << endl;
      }
      o << BL_CLOSE;
      o << "ec.union_end();" << endl;
      break;
    }
  case CORBA::dk_Enum:
    {
      o << "ec.enumeration( (CORBA::ULong) *(_T *) v );" << endl;
      break;
    }
  case CORBA::dk_Sequence:
    {
      CORBA::SequenceDef_var s = CORBA::SequenceDef::_narrow( obj );
      CORBA::IDLType_var elem = s->element_type_def();
      
      o << "CORBA::ULong len = ((_T *) v)->length();" << endl;
      o << "ec.seq_begin( len );" << endl;
      o << "for( CORBA::ULong i = 0; i < len; i++ )" << endl;
      o << indent;
      emit_marshaller_ref( elem );
      o << "->marshal( ec, &(*(_T*)v)[i]";
      emit_marshaller_suffix( elem );
      o << " );" << endl;
      o << exdent;
      o << "ec.seq_end();" << endl;
      break;
    }
  case CORBA::dk_Array:
    {
      CORBA::ArrayDef_var a = CORBA::ArrayDef::_narrow( obj );
      CORBA::IDLType_var elem = a->element_type_def();
      CORBA::ULong len = a->length();
      
      o << "ec.arr_begin();" << endl;
      o << "for( CORBA::ULong i = 0; i < " << len << "; i++ )" << endl;
      o << indent;
      emit_marshaller_ref( elem );
      o << "->marshal( ec, &((_T*)v)[i]";
      emit_marshaller_suffix( elem );
      o << " );" << endl;
      o << exdent;
      o << "ec.arr_end();" << endl;
      break;
    }
  case CORBA::dk_Fixed:
    {
      CORBA::FixedDef_var f = CORBA::FixedDef::_narrow( obj );
      o << "FixedBase::FixedValue_var val = "
	<< "((_T *) v)->to_digits(); " << endl;
      o << "ec.put_fixed( val, " << f->digits() << ", "
	<< f->scale() << " );" << endl;
      break;
    }
  case CORBA::dk_Interface:
    {
      o << "CORBA::Object_ptr obj = *(_T *) v;" << endl;
      o << "CORBA::_stc_Object->marshal( ec, &obj );" << endl;
      break;
    }
  case CORBA::dk_Exception:
    {
      CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( obj );
      CORBA::StructMemberSeq_var mem = e->members();
      CORBA::String_var id = e->id();
      
      o << "ec.except_begin( \"" << id << "\" );" << endl;
      for( CORBA::ULong i = 0; i < mem->length(); i++ ) {
	emit_marshaller_ref( mem[ i ].type_def );
	o << "->marshal( ec, &((_T*)v)->" << mem[ i ].name;
	emit_marshaller_suffix( mem[i].type_def );
	o << " );" << endl;
      }
      o << "ec.except_end();" << endl;
      break;
    }
  default:
    assert( 0 );
  }
  o << BL_CLOSE << exdent << BL_CLOSE_SEMI << endl;

  // Generate Marshaller instance
#ifdef WINDOWS_TC
  o.switchStream( 1 );
  mangled_name[ 1 ] = 'm';
  o << mangled_name << " = _get";
  emit_marshaller_ref( obj );
  o << "();" << endl;

  o.switchStream( 0 );
#else
#error Must use WINDOWS_TC here
#endif

  o << "CORBA::StaticTypeInfo* _get";
  emit_marshaller_ref( obj );
  o << "()" << endl << BL_OPEN;
  mangled_name[ 1 ] = 'M';
  o << "return new " << mangled_name << ";" << endl;
  o << BL_CLOSE << endl;
}


void CodeGenCPPStub::emit_dii_stub_method( string &absClassName,
					   string &dispatch_name,
					   CORBA::OperationDescription &op,
					   CORBA::IDLType_ptr result )
{
  emit_type_for_result( result );
  o << " " << absClassName << _STUB << "::" << ID(op.name) << "(";
  CORBA::ParDescriptionSeq &p = op.parameters;
  CORBA::ULong k;
  for( k = 0; k != p.length(); k++ ) {
    if( k != 0 )
      o << ",";
    o << " ";
    emit_type_for_param( p[ k ].type_def, p[ k ].mode );
    o << " " << ID(p[ k ].name);
  }

  // Do we have a context?
  CORBA::ContextIdSeq& ctx = op.contexts;
    
  if( ctx.length() > 0 ) {
    if( p.length() > 0 )
      o << ", ";
    o << "CORBA::Context_ptr _ctx";
  }
    
  if( k != 0 || ctx.length() > 0 )
    o << " ";
  o << ")" << endl;
  o << BL_OPEN;

  // Generate wrapper variables for all array parameters
  // but not for variable out arrays
  for( k = 0; k != p.length(); k++ ) {
    if( is_array( p[ k ].type ) ) {
      if( !is_variable( p[ k ].type ) || p[ k ].mode != CORBA::PARAM_OUT ) {
	emit_type_for_variable( p[ k ].type_def, automatic );
	o << "_forany _forany_" << ID(p[ k ].name);
	o << "( ";
	if( p[ k ].mode == CORBA::PARAM_IN ) {
	  o << "(";
	  emit_idl_type_name( p[ k ].type_def );
	  o << ") ";
	}
	o << ID(p[ k ].name) << " );" << endl;
      }
    }
  }
  
  // Generate marshalling code
  o << "CORBA::Request_var _req = this->_request( \"" << dispatch_name <<
    "\" );" << endl;

  // Marshall context
  if( ctx.length() > 0 ) {
    o << "_req->ctx( _ctx );" << endl;
    for( k = 0; k < ctx.length(); k++ )
      o << "_req->contexts()->add( \"" << ctx[ k ] << "\" );" << endl;
  }

  // Marshall actual parameters
  for( k = 0; k < p.length(); k++ ) {
    o << "_req->add_";
    switch( p[ k ].mode ) {
    case CORBA::PARAM_IN:
      o << "in";
      break;
    case CORBA::PARAM_OUT:
      o << "out";
      break;
    case CORBA::PARAM_INOUT:
      o << "inout";
      break;
    default:
      assert( 0 );
    }
    o << "_arg( \"" << ID(p[ k ].name) << "\" )";
    switch( p[ k ].mode ) {
    case CORBA::PARAM_IN:
    case CORBA::PARAM_INOUT:
      o << " <<= ";
      if( is_string( p[ k ].type ) ) {
	o << "CORBA::Any::from_string( (char *) " << ID(p[ k ].name) << ", ";
	CORBA::TypeCode_var tc = resolve_alias( p[ k ].type );
	o << tc->length() << " )";
      } else if( is_array( p[ k ].type ) ) {
	o << "_forany_" << ID(p[ k ].name);
      } else {
	emit_any_param( p[ k ].type_def, p[ k ].name, true );
      }
      o << ";" << endl;
      break;
    case CORBA::PARAM_OUT:
      // Write out reference to TypeCode
      o << ";" << endl;
      o << "_req->arguments()->item( " << k;
      o << " )->value()->type( ";
      emit_type_code_name( p[ k ].type_def );
      o << " );" << endl;
    }
  }
  o << "_req->result()->value()->type( ";
  emit_type_code_name( result );
  o << " );" << endl;

  if( op.mode == CORBA::OP_NORMAL )
    o << "_req->invoke();" << endl;
  else
    o << "_req->send_oneway();" << endl;

  if( !_params.no_exceptions )
    o << "#ifdef HAVE_EXCEPTIONS" << endl;
    
  if( !_params.no_exceptions && op.mode == CORBA::OP_NORMAL ) {
    o << "if( _req->env()->exception() ) " << BL_OPEN;
    o << "CORBA::Exception *_ex = _req->env()->exception();" << endl;
    o << "CORBA::UnknownUserException *_uuex = ";
    o << "CORBA::UnknownUserException::_narrow( _ex );" << endl;
    o << "if( _uuex ) " << BL_OPEN;
    
    CORBA::ExcDescriptionSeq exs = op.exceptions;
    for( CORBA::ULong i = 0; i < exs.length(); i++ ) {
      CORBA::Contained_var con = _repo->lookup_id( exs[ i ].id );
      assert( !CORBA::is_nil( con ) );
      CORBA::String_var tmp = con->absolute_name();
      CORBA::String_var ex_abs_name = (const char *) ID(tmp);
      string ex_scope;
      ex_scope = ex_abs_name;
      int idx = ex_scope.find_last_of( ":" );
      assert( idx > 0 );
      ex_scope = ex_scope.substr( 0, idx - 1 );
      if( ex_scope == "::" )
	ex_scope = "";

      o << "if( !strcmp( _uuex->_except_repoid(), \"";
      o << exs[ i ].id;
      o << "\" ) ) " << BL_OPEN;
      o << "CORBA::Any &_a = _uuex->exception( " << ex_scope << "::_tc_";
      o << exs[ i ].name;
      o << " );" << endl;
      o << ex_abs_name << " _user_ex;" << endl;
      o << "_a >>= _user_ex;" << endl;
      o << "mico_throw( _user_ex );" << endl;
      o << BL_CLOSE;
    }
    
    o << "mico_throw( CORBA::UNKNOWN() );" << endl;
    o << exdent << "} else {" << indent << endl;
    o << "mico_throw( *_ex );" << endl;
    o << BL_CLOSE << BL_CLOSE;
  } else if( !_params.no_exceptions && op.mode == CORBA::OP_ONEWAY ) {
    o << BL_OPEN;
    o << "CORBA::Exception *_ex;" << endl;
    o << "if( (_ex = _req->env()->exception()) )" << endl;
    o << indent << "mico_throw( *_ex );" << exdent << endl;
    o << BL_CLOSE;
  } else {
    o << BL_OPEN;
    o << "CORBA::Exception *_ex;" << endl;
    o << "if( (_ex = _req->env()->exception()) )" << endl;
    o << indent << "CORBA::Exception::_throw_failed( _ex );" << exdent << endl;
    o << BL_CLOSE;
  }
  
  if( !_params.no_exceptions ) {
    o << "#else" << endl;
    o << BL_OPEN;
    o << "CORBA::Exception *_ex;" << endl;
    o << "if( (_ex = _req->env()->exception()) )" << endl;
    o << indent << "CORBA::Exception::_throw_failed( _ex );" << exdent << endl;
    o << BL_CLOSE;
    o << "#endif" << endl;
  }
  
  // Write back OUT and INOUT parameters
  for( k = 0; k < p.length(); k++ ) {
    switch( p[ k ].mode ) {
    case CORBA::PARAM_IN:
      // Nothing to do
      break;
    case CORBA::PARAM_OUT:
    case CORBA::PARAM_INOUT:
      // We need to allocate some memory to store the result
      if( is_variable( p[ k ].type ) && is_array( p[ k ].type ) &&
	  p[ k ].mode == CORBA::PARAM_OUT ) {
	o << ID(p[ k ].name) << " = ";
	emit_idl_type_name( p[ k ].type_def );
	o << "_alloc();" << endl;
	emit_type_for_variable( p[ k ].type_def, automatic );
	o << "_forany _forany_" << ID(p[ k ].name);
	o << "( " << ID(p[ k ].name) << " );" << endl;
      }
      
      if( is_variable( p[ k ].type ) && !is_string( p[ k ].type ) &&
	  !is_wstring( p[ k ].type ) && !is_objref( p[ k ].type ) &&
	  !is_array( p[ k ].type ) && p[ k ].mode == CORBA::PARAM_OUT ) {
	o << ID(p[ k ].name) << " = new ";
	if( is_any( p[ k ].type ) )
	  o << "CORBA::Any";
	else {
	  CORBA::Contained_var con =
	    CORBA::Contained::_narrow( p[ k ].type_def );
	  assert( !CORBA::is_nil( con ) );
	  CORBA::String_var n = con->absolute_name();
	  o << ID(n);
	}
	o << ";" << endl;
      }
      
      o << "*_req->arguments()->item( " << k << " )->value() >>= ";
      if( is_string( p[ k ].type ) ) {
	o << "CORBA::Any::to_string( " << ID(p[ k ].name) << ", ";
	CORBA::TypeCode_var tc = resolve_alias( p[ k ].type );
	o << tc->length() << " )";
      } else if( is_array( p[ k ].type ) ) {
	o << "_forany_" << ID(p[ k ].name );
      } else {
	string n;
    
	if( !is_objref( p[ k ].type ) && is_variable( p[ k ].type ) &&
	    p[ k ].mode == CORBA::PARAM_OUT )
	  n = "*";
	n += ID(p[ k ].name);
	emit_any_param( p[ k ].type_def, (char *) n.c_str(), false );
      }
      o << ";" << endl;
      break;
    }
  }
  
  // Write back result (if there is one)
  CORBA::TypeCode_var tc_result = result->type();
  if( tc_result->kind() != CORBA::tk_void ) {
    if( is_array( tc_result ) ) {
      emit_type_for_variable( result, automatic );
      o << "_slice* _res = ";
      emit_idl_type_name( result );
      o << "_alloc();" << endl;
      emit_type_for_variable( result, automatic );
      o << "_forany _res_forany( _res );" << endl;
    } else {
      emit_type_for_variable( result, manual );
      o << " _res";
      if( !is_objref( tc_result ) && !is_string( tc_result ) &&
	  !is_wstring( tc_result ) && is_variable( tc_result ) ) {
	o << " = new ";
	if( is_any( tc_result ) )
	  o << "CORBA::Any";
	else {
	  CORBA::Contained_var con = CORBA::Contained::_narrow( result );
	  assert( !CORBA::is_nil( con ) );
	  CORBA::String_var n = con->absolute_name();
	  o << ID(n);
	}
      }
      o << ";" << endl;
    }
    
    o << "*_req->result()->value() >>= ";
    
    if( is_array( tc_result ) ) {
      o << "_res_forany";
    } else {
      string res;
    
      if( !is_objref( tc_result ) && !is_string( tc_result ) &&
	  !is_wstring( tc_result ) && is_variable( tc_result ) )
	res = "*";
      res += "_res";
      emit_any_param( result, (char *) res.c_str(), false );
    }
    o << ";" << endl;
    o << "return _res;" << endl;
  }
  o << BL_CLOSE << endl << endl;
}


void CodeGenCPPStub::emit_sii_stub_method( string &absClassName,
					   string &dispatch_name,
					   CORBA::OperationDescription &op,
					   CORBA::IDLType_ptr result )
{
  emit_type_for_result( result );
  o << " " << absClassName << _STUB << "::" << ID(op.name) << "(";
  CORBA::ParDescriptionSeq &p = op.parameters;
  CORBA::ULong k;
  for( k = 0; k != p.length(); k++ ) {
    if( k != 0 )
      o << ",";
    o << " ";
    emit_type_for_param( p[ k ].type_def, p[ k ].mode );
    o << " " << ID(p[ k ].name);
  }

  // Do we have a context?
  CORBA::ContextIdSeq& ctx = op.contexts;
    
  if( ctx.length() > 0 ) {
    if( p.length() > 0 )
      o << ", ";
    o << "CORBA::Context_ptr _ctx";
  }
    
  if( k != 0 || ctx.length() > 0 )
    o << " ";
  o << ")" << endl;
  o << BL_OPEN;

  // Generate StaticAny's for arguments and result
  for( k = 0; k < p.length(); k++ ) {
    o << "CORBA::StaticAny _" << ID(p[ k ].name) << "( ";
    emit_marshaller_ref( p[ k ].type_def );
    if( p[k].mode != CORBA::PARAM_OUT || !is_variable_for_sii( p[k].type ) ) {
      o << ", &" << ID(p[ k ].name);
    }
    o << " );" << endl;
  }

  // Generate variable for result (if there is one)
  CORBA::TypeCode_var tc_result = result->type();
  if( tc_result->kind() != CORBA::tk_void ) {
    if( !is_variable_for_sii( tc_result ) ) {
      if( is_array( tc_result ) ) {
        emit_type_for_variable( result, automatic );
	o << "_slice* _res;" << endl;
      } else {
        emit_type_for_variable( result, manual );
	o << " _res;" << endl;
      }
      o << "CORBA::StaticAny __res( ";
      emit_marshaller_ref( result );
      o << ", &_res );";
    } else {
      o << "CORBA::StaticAny __res( ";
      emit_marshaller_ref( result );
      o << " );";
    }
    o << endl << endl;
  }
  
  // Generate marshalling code
  o << "CORBA::StaticRequest __req( this, \"" << dispatch_name
    << "\" );" << endl;

  // Marshall context
  if( ctx.length() > 0 ) {
    o << "__req.set_context( _ctx );" << endl;
    o << "CORBA::ContextList_var _clist = new CORBA::ContextList;" << endl;
    for( k = 0; k < ctx.length(); k++ )
      o << "_clist->add( \"" << ctx[ k ] << "\" );" << endl;
    o << "__req.set_context_list( _clist );" << endl;
  }

  // Marshall actual parameters
  for( k = 0; k < p.length(); k++ ) {
    o << "__req.add_";
    switch( p[ k ].mode ) {
    case CORBA::PARAM_IN:
      o << "in";
      break;
    case CORBA::PARAM_OUT:
      o << "out";
      break;
    case CORBA::PARAM_INOUT:
      o << "inout";
      break;
    default:
      assert( 0 );
    }
    o << "_arg( &_" << ID(p[ k ].name) << " );" << endl;
  }
  if (tc_result->kind() != CORBA::tk_void)
    o << "__req.set_result( &__res );" << endl;
  o << endl;

  if( op.mode == CORBA::OP_NORMAL )
    o << "__req.invoke();" << endl << endl;
  else
    o << "__req.oneway();" << endl << endl;

  if( !_params.no_exceptions )
    o << "#ifdef HAVE_EXCEPTIONS" << endl;
    
  if( !_params.no_exceptions && op.mode == CORBA::OP_NORMAL ) {
    o << "if( __req.exception() ) " << BL_OPEN;
    o << "CORBA::Exception *_ex = __req.exception();" << endl;
    o << "CORBA::UnknownUserException *_uuex = ";
    o << "CORBA::UnknownUserException::_narrow( _ex );" << endl;
    o << "if( _uuex ) " << BL_OPEN;
    
    CORBA::ExcDescriptionSeq exs = op.exceptions;
    for( CORBA::ULong i = 0; i < exs.length(); i++ ) {
      CORBA::Contained_var con = _repo->lookup_id( exs[ i ].id );
      assert( !CORBA::is_nil( con ) );
      CORBA::String_var tmp = con->absolute_name();
      CORBA::String_var ex_abs_name = (const char *) ID(tmp);
      string ex_scope;
      ex_scope = ex_abs_name;
      int idx = ex_scope.find_last_of( ":" );
      assert( idx > 0 );
      ex_scope = ex_scope.substr( 0, idx - 1 );
      if( ex_scope == "::" )
	ex_scope = "";

      o << "if( !strcmp( _uuex->_except_repoid(), \"";
      o << exs[ i ].id;
      o << "\" ) ) " << BL_OPEN;
      o << "CORBA::StaticAny &_a = _uuex->exception( ";
      emit_marshaller_ref( con );
      o << " );" << endl;
      o << "mico_throw( *(" << ex_abs_name << "*)_a.value() );" << endl;
      o << BL_CLOSE;
    }
    
    o << "mico_throw( CORBA::UNKNOWN() );" << endl;
    o << exdent << "} else {" << indent << endl;
    o << "mico_throw( *_ex );" << endl;
    o << BL_CLOSE << BL_CLOSE;
  } else if( !_params.no_exceptions && op.mode == CORBA::OP_ONEWAY ) {
    o << BL_OPEN;
    o << "CORBA::Exception *_ex;" << endl;
    o << "if( (_ex = __req.exception()) )" << endl;
    o << indent << "mico_throw( *_ex );" << exdent << endl;
    o << BL_CLOSE;
  } else {
    o << BL_OPEN;
    o << "CORBA::Exception *_ex;" << endl;
    o << "if( (_ex = __req.exception()) )" << endl;
    o << indent << "CORBA::Exception::_throw_failed( _ex );" << exdent << endl;
    o << BL_CLOSE;
  }
  
  if( !_params.no_exceptions ) {
    o << "#else" << endl;
    o << BL_OPEN;
    o << "if( __req.exception() )" << endl;
    o << indent << "CORBA::Exception::_throw_failed( __req.exception() );"
      << exdent << endl;
    o << BL_CLOSE;
    o << "#endif" << endl;
  }

  // Write back out parameters
  for( k = 0; k < p.length(); k++ ) {
    if( p[k].mode == CORBA::PARAM_OUT && is_variable_for_sii( p[k].type ) ) {
      o << ID(p[ k ].name) << " = ";
      o << "(";
      emit_idl_type_name( p[k].type_def );
      if( is_array( p[k].type ) )
	o << "_slice";
      o << "*) _" << ID(p[k].name) << "._retn();" << endl;
    }
  }

  // Write back result (if there is one)
  if( tc_result->kind() != CORBA::tk_void ) {
    o << "return ";
    if( is_variable_for_sii( tc_result ) ) {
      o << "(";
      emit_idl_type_name( result );
      if( is_array( tc_result ) )
	o << "_slice";
      o << "*) __res._retn();";
    } else {
      o << "_res;";
    }
    o << endl;
  }

  o << BL_CLOSE << endl << endl;
}



void CodeGenCPPStub::emit_Interface( CORBA::InterfaceDef_ptr in )
{
  assert( in->def_kind() == CORBA::dk_Interface );

  CORBA::ContainedSeq_var c;
  c = in->contents( CORBA::dk_all, 1 );
  for( CORBA::ULong i = 0; i < c->length(); i++ ) {

    if( check_for_included_defn( c[ i ] ) )
      continue;
    
    switch( c[ i ]->def_kind() ) {
    case CORBA::dk_Struct:
      {
	CORBA::StructDef_var s = CORBA::StructDef::_narrow( c[ i ] );
	emit_Struct( s );
	break;
      }
    case CORBA::dk_Union:
      {
	CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( c[ i ] );
	emit_Union( u );
	break;
      }
    case CORBA::dk_Exception:
      {
	CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( c[ i ] );
	emit_Exception( e );
	break;
      }
    case CORBA::dk_Enum:
      {
	CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( c[ i ] );
	emit_Enum( e );
	break;
      }
    case CORBA::dk_Operation:
    case CORBA::dk_Attribute:
      break;
    case CORBA::dk_Constant:
      {
	CORBA::ConstantDef_var co = CORBA::ConstantDef::_narrow( c[ i ] );
	emit_Constant( co );
	break;
      }
    case CORBA::dk_Alias:
      {
	CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( c[ i ] );
	emit_Alias( a );
	break;
      }
    default:
      assert( 0 );
    }
  }

  CORBA::String_var id = in->id();

  CORBA::String_var tmp;
  tmp = in->name();
  string className;
  className = ID(tmp);
  tmp = in->absolute_name();
  string absClassName;
  absClassName = ID(tmp);
  absClassName = absClassName.substr( 2 );
  
  o << endl;
  o << "// Stub interface " << className << endl;

  // dtor
  o << absClassName << "::~" << className << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;

  // _duplicate class method
  o << absClassName << "_ptr " << absClassName;
  o << "::_duplicate( " << className << "_ptr _obj )" << endl;
  o << BL_OPEN;
  o << "if( !CORBA::is_nil( _obj ) )" << endl;
  o << indent << "_obj->_ref();" << exdent << endl;
  o << "return _obj;" << endl;
  o << BL_CLOSE << endl;

  /*
   * _narrow_helper virtual method:
   *
   * searched upwards in the inheritance hierarchy starting at the
   * real class of the object that is to be narrowed until it finds
   * the class we want to narrow to or until top is reached.
   *
   * this is used to do a downcast from a virtual base class.
   */
  o << "void *" << absClassName;
  o << "::_narrow_helper( const char *_repoid )" << endl;
  o << BL_OPEN;
  o << "if( strcmp( _repoid, \"" << id << "\" ) == 0 )" << endl;
  o << indent << "return (void *)this;" << exdent << endl;
  CORBA::InterfaceDefSeq_var bases = in->base_interfaces();
  if( bases->length() != 0 ) {
    // We have base interfaces
    set_name_prefix( in );
    for( CORBA::ULong i = 0; i < bases->length(); i++ ) {
      o << BL_OPEN;
      o << "void *_p;" << endl;
      o << "if( (_p = ";
      emit_rel_name( bases[ i ] );
      o << "::_narrow_helper( _repoid )))" << endl;
      o << indent << "return _p;" << exdent << endl;
      o << BL_CLOSE;
    }
    set_name_prefix( CORBA::Container::_nil() );
  }
  o << "return NULL;" << endl;
  o << BL_CLOSE << endl;

  /*
   * _narrow_helper2 class method
   *
   * searches downwards in the inheritance hierarchy starting at the
   * class we want to narrow to until it finds the class of the object
   * we want to narrow or until bottom is reached. Returns true
   * if _narrow succeeds
   */

  if (!_params.pseudo) {
    o << "bool " << absClassName;
    o << "::_narrow_helper2( CORBA::Object_ptr _obj )" << endl;
    o << BL_OPEN;
    o << "if( strcmp( _obj->_repoid(), \"" << id << "\" ) == 0) {" << endl;
    o << indent << "return true;" << exdent << endl;
    o << "}" << endl;
    o << "for( vector<CORBA::Narrow_proto>::size_type _i = 0;" << endl;
    o << "     _narrow_helpers && _i < _narrow_helpers->size(); _i++ ) ";
    o << BL_OPEN;
    o << "bool _res = (*(*_narrow_helpers)[ _i ])( _obj );" << endl;
    o << "if( _res )" << endl;
    o << indent << "return true;" << exdent << endl;
    o << BL_CLOSE;
    o << "return false;" << endl;
    o << BL_CLOSE << endl;
  }

  /*
   * _narrow class method
   *
   * first see if we can just do a downcast by using _narrow_helper.
   * if that fails generate a stub for it using the _narrow_helper2
   * method. if that fails too call _obj->_is_a (_repoid) and create a
   * stub if _is_a() returns true. otherwise return _nil().
   */
  o << absClassName << "_ptr " << absClassName;
  o << "::_narrow( CORBA::Object_ptr _obj )" << endl;
  o << BL_OPEN;
  if (!_params.pseudo) {
    o << absClassName << "_ptr _o;" << endl;
  }
  o << "if( !CORBA::is_nil( _obj ) ) {" << indent << endl;
  o << "void *_p;" << endl;
  o << "if( (_p = _obj->_narrow_helper( \"" << id << "\" )))" << endl;
  o << indent
    << "return _duplicate( (" << absClassName << "_ptr) _p );"
    << exdent << endl;
  if (!_params.pseudo) {
    o << "if( _narrow_helper2( _obj ) ) " << BL_OPEN;
    o << "_o = new " << absClassName;
    o << _STUB << ";" << endl;
    o << "_o->";
    bool need_corba_scope;
    need_corba_scope = absClassName.length() < 7 ||
                       absClassName.substr( 0, 7 ) != "CORBA::";
    if( need_corba_scope )
      // Adding the scope "CORBA::" is needed due to a g++ 2.7.2 bug
      // But simply writing CORBA::Object::operator=() everytime would
      // make VC++ unhappy...  sigh!
      o << "CORBA::";
    o << "Object::operator=( *_obj );" << endl;
    o << "return _o;" << endl;
    o << BL_CLOSE;
  }
  o << BL_CLOSE;
  assert (!_params.pseudo || !_params.query_ir_for_narrow);
  if( _params.query_ir_for_narrow ) {
    o << "if( _obj->_is_a_remote( \"" << id << "\" ) ) " << BL_OPEN;
    o << "_o = new " << absClassName << _STUB << ";" << endl;
    o << "_o->";
    bool need_corba_scope;
    need_corba_scope = absClassName.length() < 7 ||
                       absClassName.substr( 0, 7 ) != "CORBA::";
    if( need_corba_scope )
      // Adding the scope "CORBA::" is needed due to a g++ 2.7.2 bug
      // But simply writing CORBA::Object::operator=() everytime would
      // make VC++ unhappy...  sigh!
      o << "CORBA::";
    o << "Object::operator=( *_obj );" << endl;
    o << "return _o;" << endl;
    o << BL_CLOSE;
  }
  o << "return _nil();" << endl;
  o << BL_CLOSE << endl;

  // _nil class method
  o << absClassName << "_ptr " << absClassName;
  o << "::_nil()" << endl;
  o << BL_OPEN;
  o << "return NULL;" << endl;
  o << BL_CLOSE << endl;

  if (_params.pseudo) {
    o << endl;
    return;
  }

  // Generate attributes
  CORBA::InterfaceDef::FullInterfaceDescription_var desc;
  desc = in->describe_interface();
  for( CORBA::ULong j = 0; j < desc->attributes.length(); j++ ) {
    CORBA::AttributeDescription attr;
    attr = desc->attributes[ j ];

    CORBA::OperationDescription op;
    op.name = (const char *) attr.name;
    op.id = (const char *) "";  // Not needed
    op.defined_in = attr.defined_in;
    op.version = (const char*) attr.version;
    op.result = CORBA::TypeCode::_nil();  // Not needed
    op.mode = CORBA::OP_NORMAL;
    // op.contexts does not need to be set for attributes
    op.parameters.length( 0 );  // No parameters for _get
    // op.exceptions does not need to be set for attributes

    string dispatch_name = string("_get_") + (const char *)attr.name;
    
    CORBA::IDLType_var idl_type = lookup_attribute_by_id( attr.id );
    if( _params.sii )
      emit_sii_stub_method( absClassName, dispatch_name, op, idl_type );
    else
      emit_dii_stub_method( absClassName, dispatch_name, op, idl_type );

    if( attr.mode == CORBA::ATTR_NORMAL ) {
      // We also need the set version of the attribute
      CORBA::ParameterDescription attr_param;
      attr_param.name = (const char *) "_value";
      attr_param.type = attr.type;
      attr_param.type_def = lookup_attribute_by_id( attr.id );
      attr_param.mode = CORBA::PARAM_IN;
      
      op.name = (const char *) attr.name;
      op.id = (const char *) "";  // Not needed
      op.defined_in = attr.defined_in;
      op.version = (const char*) attr.version;
      op.result = CORBA::TypeCode::_nil();  // Not needed
      op.mode = CORBA::OP_NORMAL;
      // op.contexts does not need to be set for attributes
      op.parameters.length( 1 );
      op.parameters[ 0 ] = attr_param;
      // op.exceptions does not need to be set for attributes
      
      dispatch_name = string("_set_") + (const char *)attr.name;
    
      idl_type = _repo->get_primitive( CORBA::pk_void );
      if( _params.sii )
	emit_sii_stub_method( absClassName, dispatch_name, op, idl_type );
      else
	emit_dii_stub_method( absClassName, dispatch_name, op, idl_type );
    }
  }

  // dtor
  o << absClassName << _STUB << "::~" << className << _STUB << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;

  // Generate methods
  for( CORBA::ULong j0 = 0; j0 < desc->operations.length(); j0++ ) {
    CORBA::OperationDescription op;
    op = desc->operations[ j0 ];
    string dispatch_name = (const char *) op.name;

    CORBA::IDLType_var idl_type = lookup_result_by_id( op.id );
    if( _params.sii )
      emit_sii_stub_method( absClassName, dispatch_name, op, idl_type );
    else
      emit_dii_stub_method( absClassName, dispatch_name, op, idl_type );
  }

  // Get base for absolute name
  tmp = in->absolute_name();
  string abs;
  abs = ID(tmp);
  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );

  // set all _narrow_helper2 methods in their respective
  // base classes
  if( bases->length() != 0 ) {
    o << "struct _global_init_";
    emit_mangled_name( (char *) absClassName.c_str() );
    o << " " << BL_OPEN;
    o << "_global_init_";
    emit_mangled_name( (char *) absClassName.c_str() );
    o << "()" << endl;
    o << BL_OPEN;
    for( CORBA::ULong i = 0; i < bases->length(); i++ ) {
      o << "if( ";
      CORBA::String_var abs = bases[ i ]->absolute_name();
      o << abs << "::_narrow_helpers == NULL )" << indent << endl;
      o << abs << "::_narrow_helpers = new vector<CORBA::Narrow_proto>;"
	<< exdent << endl;
      o << abs << "::_narrow_helpers->push_back( ";
      o << absClassName << "::_narrow_helper2 );" << endl;
    }
    o << BL_CLOSE;
    o << exdent << "} __global_init_";
    emit_mangled_name( (char *) absClassName.c_str() );
    o << ";" << endl << endl;
  }
    
  // Make instance for static member variable _narrow_helpers
  emit_var_def ("vector<CORBA::Narrow_proto> *", in, "_narrow_helpers", "");

  // Emit TypeCode for interface
  tmp = in->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::TypeCode_var tc_in = in->type();

  CORBA::Container_var container = in->defined_in();
  emit_typecode_def (container, abs, name.in(), tc_in);
  o << endl;

#if 0
  // operator <<=
  o << "CORBA::Boolean" << endl;
  o << "operator<<=( CORBA::Any &_a, const ";
  emit_idl_type_name( in );
  o << " _obj )" << endl;
  o << BL_OPEN;
  o << "return (_a <<= CORBA::Any::from_object( _obj, \"";
  o << name << "\" ));" << endl;
  o << BL_CLOSE << endl;

  // operator >>=
  o << "CORBA::Boolean" << endl;
  o << "operator>>=( const CORBA::Any &_a, ";
  emit_idl_type_name( in );
  o << " &_obj )" << endl;
  o << BL_OPEN;
  o << "CORBA::Object_ptr _o;" << endl;
  o << "if( !(_a >>= CORBA::Any::to_object( _o )) )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "if( CORBA::is_nil( _o ) ) " << BL_OPEN;
  tmp = in->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "_obj = " << abs_name;
  o << "::_nil();" << endl;
  o << "return TRUE;" << endl;
  o << BL_CLOSE;
  o << "_obj = " << abs_name << "::_narrow( _o );" << endl;
  o << "CORBA::release( _o );" << endl;
  o << "return TRUE;" << endl;
  o << BL_CLOSE << endl;

  // Emit SII marshaller
  emit_marshaller( in );
#else
  emit_Interface2( in );
#endif
}

/*
 * splitted emit_Interface() to work around gcc 2.7.2 internal
 * compiler error. kay.
 */
void CodeGenCPPStub::emit_Interface2( CORBA::InterfaceDef_ptr in )
{
  CORBA::String_var tmp = in->name();
  CORBA::String_var name = (const char *) ID(tmp);

  // operator <<=
  o << "CORBA::Boolean" << endl;
  o << "operator<<=( CORBA::Any &_a, const ";
  emit_idl_type_name( in );
  o << " _obj )" << endl;
  o << BL_OPEN;
  o << "return (_a <<= CORBA::Any::from_object( _obj, \"";
  o << name << "\" ));" << endl;
  o << BL_CLOSE << endl;

  // operator >>=
  o << "CORBA::Boolean" << endl;
  o << "operator>>=( const CORBA::Any &_a, ";
  emit_idl_type_name( in );
  o << " &_obj )" << endl;
  o << BL_OPEN;
  o << "CORBA::Object_ptr _o;" << endl;
  o << "if( !(_a >>= CORBA::Any::to_object( _o )) )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "if( CORBA::is_nil( _o ) ) " << BL_OPEN;
  tmp = in->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "_obj = " << abs_name;
  o << "::_nil();" << endl;
  o << "return TRUE;" << endl;
  o << BL_CLOSE;
  o << "_obj = " << abs_name << "::_narrow( _o );" << endl;
  o << "CORBA::release( _o );" << endl;
  o << "return TRUE;" << endl;
  o << BL_CLOSE << endl;

  // Emit SII marshaller
  emit_marshaller( in );
}


/*
 * Generates default constructor, copy constructor and assignment
 * operator for structs. These have to be generated due to a g++
 * bug (g++ doesn't expand templates properly for the implicit
 * generated versions of these methods). Note that according to
 * the CORBA2 specs these shouldn't be defined (see [16.9(1)])
 */
void
CodeGenCPPStub::emit_StructDefaultMembers( char *name,
					   char *abs,
					   CORBA::StructMemberSeq_var &m )
{
  o << "#ifdef HAVE_EXPLICIT_STRUCT_OPS" << endl;

  // Default constructor
  o << abs << name << "::" << name << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;

  // Copy constructor
  o << abs << name << "::" << name << "( const " << name << "& _s )" << endl;
  o << BL_OPEN;
  for( CORBA::ULong i0 = 0; i0 < m->length(); i0++ ) {
    o << ID(m[ i0 ].name) << " = ((" << name << "&)_s).";
    o << ID(m[ i0 ].name) << ";" << endl;
  }
  o << BL_CLOSE << endl;

  // Destructor (moving this out of .h file reduces template drag in a lot)
  o << abs << name << "::~" << name << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;
  
  // Assignment operator
  o << abs << name << "&" << endl;
  o << abs << name << "::operator=( const " << name << "& _s )" << endl;
  o << BL_OPEN;
  for( CORBA::ULong i1 = 0; i1 < m->length(); i1++ ) {
    o << ID(m[ i1 ].name) << " = ((" << name << "&)_s).";
    o << ID(m[ i1 ].name) << ";" << endl;
  }
  o << "return *this;" << endl;
  o << BL_CLOSE;

  o << "#endif" << endl << endl;
}



void CodeGenCPPStub::emit_Struct( CORBA::StructDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_Struct );

  // Get base for absolute name
  CORBA::String_var tmp = s->absolute_name();
  string abs;
  abs = ID(tmp);
  tmp = s->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::TypeCode_var tc = s->type();

  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );

  // Emit TypeCode for struct
  if( !_params.pseudo ) {
    if( tc->is_freestanding() ) {
      CORBA::Container_var container = s->defined_in();
      emit_typecode_def (container, abs, name.in(), tc);
      o << endl;
    }
  }

  // Emit default ctor, copy ctor, assignment op
  CORBA::StructMemberSeq_var m = s->members();
  emit_StructDefaultMembers (name, (char *)abs.c_str(), m);

  if (_params.pseudo) {
    return;
  }

  // operator<<=
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs << name << " &_s )" << endl;
  o << BL_OPEN;
  if( tc->is_freestanding() )
    o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
  o << "return (_a.struct_put_begin() &&" << indent << endl;
  CORBA::StructMemberSeq_var members = s->members();
  for( CORBA::ULong i = 0; i < members->length(); i++ ) {
    o << "(_a <<= ";
    string v = "((";
    v += abs;
    v += name;
    v += "&)_s).";
    v += ID(members[ i ].name);
    emit_any_param( members[ i ].type_def, (char *) v.c_str(), true );
    o << ") &&" << endl;
  }
  o << "_a.struct_put_end() );" << exdent << endl;
  o << BL_CLOSE << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs << name << " &_s )" << endl;
  o << BL_OPEN;
  o << "return (_a.struct_get_begin() &&" << indent << endl;
  for( CORBA::ULong i0 = 0; i0 < members->length(); i0++ ) {
    o << "(_a >>= ";
    string v = "_s.";
    v += ID(members[ i0 ].name);
    emit_any_param( members[ i0 ].type_def, (char *) v.c_str(), false );
    o << ") &&" << endl;
  }
  o << "_a.struct_get_end() );" << exdent << endl;
  o << BL_CLOSE << endl;

  // SII marshaller
  emit_marshaller( s );
}



/*
 * Generates default constructor, copy constructor and assignment
 * operator for unions. These have to be generated due to a g++
 * bug (g++ doesn't expand templates properly for the implicit
 * generated versions of these methods). Note that according to
 * the CORBA2 specs these shouldn't be defined (see [16.9(1)])
 */
void
CodeGenCPPStub::emit_UnionDefaultMembers( char *name,
					  char *abs,
					  CORBA::UnionMemberSeq_var &m )
{
  o << "#ifdef HAVE_EXPLICIT_STRUCT_OPS" << endl;

  // Default constructor
  o << abs << name << "::" << name << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;

  // Copy constructor
  CORBA::String_var last_name;
  last_name = (const char *) "";
  o << abs << name << "::" << name << "( const " << name << "&_u )" << endl;
  o << BL_OPEN;
  o << "_discriminator = _u._discriminator;" << endl;
  for( CORBA::ULong i0 = 0; i0 < m->length(); i0++ ) {
    if( strcmp( last_name, m[ i0 ].name ) == 0 )
      continue;
    last_name = m[ i0 ].name;
    o << "_m." << ID(m[ i0 ].name) << " = ((" << name << "&)_u)._m.";
    o << ID(m[ i0 ].name) << ";" << endl;
  }
  o << BL_CLOSE << endl;

  // Destructor (moving this out of .h file reduces template drag in a lot)
  o << abs << name << "::~" << name << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;
  
  // Assignment operator
  o << abs << name << "&" << endl;
  o << abs << name << "::operator=( const " << name << "&_u )" << endl;
  o << BL_OPEN;
  o << "_discriminator = _u._discriminator;" << endl;
  last_name = (const char *) "";
  for( CORBA::ULong i1 = 0; i1 < m->length(); i1++ ) {
    if( strcmp( last_name, m[ i1 ].name ) == 0 )
      continue;
    last_name = m[ i1 ].name;
    o << "_m." << ID(m[ i1 ].name) << " = ((" << name << "&)_u)._m.";
    o << ID(m[ i1 ].name) << ";" << endl;
  }
  o << "return *this;" << endl;
  o << BL_CLOSE;

  o << "#endif" << endl << endl;
}


    
void CodeGenCPPStub::emit_union_case_label( CORBA::IDLType_ptr discr,
					    CORBA::UnionMember &m,
					    CORBA::UnionMemberSeq &members )
{
  CORBA::TypeCode_var tc = m.label.type();
  if( is_octet( tc ) ) {
    string label = generate_union_default_case_label( discr, members );
    // Test if union has already all case members listed
    assert( label != "" );
    o << label;
    return;
  }
  
  tc = discr->type();
  if( is_enum( tc ) ) {
    CORBA::String_var s;
    CORBA::ULong idx;
    CORBA::Boolean r = m.label.enum_get( idx );
    assert (r);
    s = (const char *) tc->unalias()->member_name( idx );
    CORBA::Contained_var c = CORBA::Contained::_narrow( discr );
    assert( !CORBA::is_nil( c ) );
    CORBA::String_var tmp = c->absolute_name();
    string abs;
    abs = ID(tmp);
    int colon_pos = abs.rfind( "::" );
    if( colon_pos != -1 )
      abs = abs.substr( 2, colon_pos );
    o << abs << ID(s);
  } else {
    o << &m.label;
  }
}
    


void CodeGenCPPStub::emit_Union( CORBA::UnionDef_ptr u )
{
  assert( u->def_kind() == CORBA::dk_Union );

  // Get base for absolute name
  CORBA::String_var tmp = u->absolute_name();
  string abs;
  abs = ID(tmp);
  tmp = u->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::TypeCode_var tc = u->type();
  CORBA::UnionMemberSeq_var members = u->members();
  CORBA::IDLType_var discr = u->discriminator_type_def();
  
  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );

  if( !_params.pseudo ) {
    // Emit TypeCode for union
    if( tc->is_freestanding() ) {
      CORBA::Container_var container = u->defined_in();
      emit_typecode_def (container, abs, name.in(), tc);
      o << endl;
    }
  }

  // Emit default ctor, copy ctor, assignment op
  emit_UnionDefaultMembers( name, (char *) abs.c_str(), members );

  // Emit discriminator functions
  o << "void " << abs << name << "::_d( ";
  emit_type_for_param( discr, CORBA::PARAM_IN );
  o << " _p )" << endl;
  o << BL_OPEN;
  o << "_discriminator = _p;" << endl;
  o << BL_CLOSE << endl;
  
  emit_type_for_result( discr );
  o << " " << abs << name << "::_d() const" << endl;
  o << BL_OPEN;
  o << "return _discriminator;" << endl;
  o << BL_CLOSE << endl;

  // Generate member accessor functions
  CORBA::String_var last_name = (const char *) "";
  bool has_explicit_default = false;
  for( CORBA::ULong i = 0; i < members->length(); i++ ) {
    CORBA::UnionMember &m = members[ i ];
    CORBA::TypeCode_var tc = members[ i ].label.type();
    if( is_octet( tc ) )
      has_explicit_default = true;

    if( strcmp( last_name, m.name ) == 0 )
      continue;
    last_name = m.name;
    
    CORBA::IDLType_var m_type = CORBA::IDLType::_duplicate( m.type_def );
    CORBA::String_var m_name = (const char *) ID(m.name);
    tc = m_type->type();
    if( is_string( tc ) ) {
      // Accessor functions for string
      o << "void " << abs << name << "::" << m_name << "( char *_p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o << BL_CLOSE << endl;
      
      o << "void " << abs << name << "::"
	<< m_name << "( const char *_p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o << BL_CLOSE << endl;
      
      o << "void " << abs << name << "::"
	<< m_name << "( const CORBA::String_var &_p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o << BL_CLOSE << endl;
      
      o << "const char *" << abs << name << "::"
	<< m_name << "() const" << endl;
      o << BL_OPEN;
      o << "return _m." << m_name << ";" << endl;
      o << BL_CLOSE << endl;
    } else if( is_wstring( tc ) ) {
      // Accessor functions for wstring
      o << "void " << abs << name << "::" << m_name << "( wchar_t *_p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o << BL_CLOSE << endl;
      
      o << "void " << abs << name << "::"
	<< m_name << "( const wchar_t *_p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o << BL_CLOSE << endl;
      
      o << "void " << abs << name << "::"
	<< m_name << "( const CORBA::WString_var &_p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o << BL_CLOSE << endl;
      
      o << "const wchar_t *" << abs << name << "::"
	<< m_name << "() const" << endl;
      o << BL_OPEN;
      o << "return _m." << m_name << ";" << endl;
      o << BL_CLOSE << endl;
    } else if( is_array( tc ) && !is_alias( tc ) ) {
      // We have an anonymous array. See [16.10(7)]
      o << "void " << abs << name << "::" << m_name << "( ";
      emit_array( m_type );
      o << " _p";
      emit_array_suffix( m_type );
      o << " )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      emit_array( m.type_def );
      o << "* _dest = (";
      emit_array( m.type_def );
      o << "*) _m." << m_name << ";" << endl;
      emit_array( m.type_def );
      o << "* _src = (";
      emit_array( m.type_def );
      o << "*) _p;" << endl;
      o << "for( int _i = 0; _i < ";
      o << compute_num_array_elem( tc );
      o << "; _i++ )" << endl;
      o << indent << "*_dest++ = *_src++;" << exdent << endl;
      o << BL_CLOSE << endl;
      
      o << abs << name << "::_" << m_name << "_slice* ";
      o << abs << name << "::" << m_name << "()" << endl;
      o << BL_OPEN;
      emit_array( m.type_def );
      o << "* _res;" << endl;
      emit_array( m.type_def );
      o << "* _dest;" << endl;
      o << "_res = _dest = (";
      emit_array( m_type );
      o << "*) new ";
      emit_array( m.type_def );
      emit_array_suffix( m.type_def );
      o << ";" << endl;
      emit_array( m.type_def );
      o << "* _src = (";
      emit_array( m.type_def );
      o << "*) _m." << m_name << ";" << endl;
      o << "for( int _i = 0; _i < ";
      o << compute_num_array_elem( tc );
      o << "; _i++ )" << endl;
      o << indent << "*_dest++ = *_src++;" << exdent << endl;
      o << "return (_" << m_name << "_slice*) _res;" << endl;
      o << BL_CLOSE << endl;
    } else if( is_array( tc ) || is_objref( tc ) ||
	       is_type_code( tc ) || is_base_type( tc ) ) {
      // Accessor functions for array or object reference
      o << "void " << abs << name << "::" << m_name << "( ";
      if( is_array( tc ) )
	// Mapping for arrays differs from standard PARAM_IN mapping!
	emit_idl_type_name( m_type );
      else
	emit_type_for_param( m_type, CORBA::PARAM_IN );
      o << " _p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      if( is_array( tc ) ) {
	emit_array( m.type_def );
	o << "* _dest = (";
	emit_array( m.type_def );
	o << "*) _m." << m_name << ";" << endl;
	emit_array( m.type_def );
	o << "* _src = (";
	emit_array( m.type_def );
	o << "*) _p;" << endl;
	o << "for( int _i = 0; _i < ";
	CORBA::TypeCode_var orig_tc = resolve_alias( tc );
	o << compute_num_array_elem( orig_tc );
	o << "; _i++ )" << endl;
	o << indent << "*_dest++ = *_src++;" << exdent << endl;
      } else {
	o << "_m." << m_name << " = ";
	if( is_objref( tc ) ) {
	  if( m.type_def->def_kind() == CORBA::dk_Primitive ) {
	    // Gotta be "Object"
	    o << "CORBA::Object::_duplicate( _p );" << endl;
	  } else {
	    CORBA::Contained_var c = CORBA::Contained::_narrow( m_type );
	    assert( !CORBA::is_nil( c ) );
	    CORBA::String_var objref_name = c->absolute_name();
	    o << ID(objref_name) << "::_duplicate( _p );" << endl;
	  }
	} else if( is_type_code( tc ) ) {
	  o << "CORBA::TypeCode::_duplicate( _p );" << endl;
	} else
	  o << "_p;" << endl;
      }
      o << BL_CLOSE << endl;

      emit_type_for_result( m_type );
      o << " " << abs << name << "::" << m_name << "() const" << endl;
      o << BL_OPEN;
      if( is_array( tc ) ) {
	o << "return ";
	emit_idl_type_name( m_type );
	o << "_dup( _m." << m_name << " );" << endl;
      } else
	o << "return _m." << m_name << ";" << endl;
      o << BL_CLOSE << endl;
    } else if( is_sequence( tc ) && !is_alias( tc ) ) {
      // Anonymous sequence
      o << "void " << abs << name << "::" << m_name << "( const ";
      emit_sequence( m_type );
      o << "& _p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o << BL_CLOSE << endl;

      o << "const ";
      emit_sequence( m_type );
      o << "& " << abs << name << "::" << m_name << "() const" << endl;
      o << BL_OPEN;
      o << "return _m." << m_name << ";" << endl;
      o << BL_CLOSE << endl;

      emit_sequence( m_type );
      o << "& " << abs << name << "::" << m_name << "()" << endl;
      o << BL_OPEN;
      o << "return _m." << m_name << ";" << endl;
      o << BL_CLOSE << endl;
    } else {
      // Gotta be struct, union or sequence
      o << "void " << abs << name << "::" << m_name << "( ";
      emit_type_for_param( m_type, CORBA::PARAM_IN );
      o << " _p )" << endl;
      o << BL_OPEN;
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o << BL_CLOSE << endl;

      o << "const ";
      emit_idl_type_name( m_type );
      o << "& " << abs << name << "::" << m_name << "() const" << endl;
      o << BL_OPEN;
      o << "return (";
      emit_idl_type_name( m_type );
      o << "&) _m." << m_name << ";" << endl;
      o << BL_CLOSE << endl;

      emit_idl_type_name( m_type );
      o << "& " << abs << name << "::" << m_name << "()" << endl;
      o << BL_OPEN;
      o << "return _m." << m_name << ";" << endl;
      o << BL_CLOSE << endl;
    }
  }
  if( !has_explicit_default ) {
    // This union has no explicit default case label
    string default_label =
      generate_union_default_case_label( discr, members );
    if( default_label != "" ) {
      o << "void " << abs << name << "::_default()" << endl;
      o << BL_OPEN;
      o << "_discriminator = " << default_label << ";" << endl;
      o << BL_CLOSE << endl;
    }
  }

  if (_params.pseudo) {
    return;
  }

  // operator<<=
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs << name << " &_u )" << endl;
  o << BL_OPEN;
  if( tc->is_freestanding() )
    o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
  o << "if (!_a.union_put_begin())" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "if( !(_a <<= ";
  string v = "((";
  v += abs;
  v += name;
  v += "&)_u)._discriminator";
  emit_any_param( discr, (char *) v.c_str(), true );
  o << ") )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "switch( _u._d() ) " << BL_OPEN;
  for( CORBA::ULong i0 = 0; i0 < members->length(); i0++ ) {
    CORBA::UnionMember &m = members[ i0 ];

    CORBA::TypeCode_var tc = members[ i0 ].label.type();
    o << "case ";
    emit_union_case_label( discr, m, members );
    o << ":" << indent << endl;
    o << "if( !_a.union_put_selection( " << i0 << " ) )" << endl;
    o << indent << "return FALSE;" << exdent << endl;
    o << "if( !(_a <<= ";
    v = "((";
    v += abs;
    v += name;
    v += "&)_u)._m.";
    v += ID(m.name);
    emit_any_param( m.type_def, (char *) v.c_str(), true );
    o << ") )" << endl;
    o << indent << "return FALSE;" << exdent << endl;
    o << "break;" << exdent << endl;
  }
  o << "default:" << endl;
  o << indent << "break;" << exdent << endl;
  o << BL_CLOSE;
  o << "return _a.union_put_end();" << endl;
  o << BL_CLOSE << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs << name << " &_u )" << endl;
  o << BL_OPEN;
  o << "if( !_a.union_get_begin() )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "if( !(_a >>= ";
  v = "_u._discriminator";
  emit_any_param( discr, (char *) v.c_str(), false );
  o << ") )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "switch( _u._discriminator ) " << BL_OPEN;
  for( CORBA::ULong i1 = 0; i1 < members->length(); i1++ ) {
    CORBA::UnionMember &m = members[ i1 ];

    CORBA::TypeCode_var tc = members[ i1 ].label.type();
    if( tc->kind() == CORBA::tk_octet )
      o << "default";
    else {
      o << "case ";
      emit_union_case_label( discr, m, members );
    }
    o << ":" << indent << endl;
    o << "if( !_a.union_get_selection( " << i1 << " ) )" << endl;
    o << indent << "return FALSE;" << exdent << endl;
    o << "if( !(_a >>= ";
    v = "_u._m.";
    v += ID(m.name);
    emit_any_param( m.type_def, (char *) v.c_str(), false );
    o << ") )" << endl;
    o << indent << "return FALSE;" << exdent << endl;
    o << "break;" << exdent << endl;
  }
  if( !has_explicit_default ) {
    o << "default:" << indent << endl;
    o << "break;" << exdent << endl;
  }
  o << BL_CLOSE;
  o << "return _a.union_get_end();" << endl;
  o << BL_CLOSE << endl;

  // SII marshaller
  emit_marshaller( u );
}



void CodeGenCPPStub::emit_Constant( CORBA::ConstantDef_ptr c )
{
  assert( c->def_kind() == CORBA::dk_Constant );

  CORBA::Container_var def_in = c->defined_in();
  if( def_in->def_kind() == CORBA::dk_Repository )
    // This ConstantDef was defined on global scope, we
    // need to do nothing.
    return;

  CORBA::IDLType_var type = c->type_def();

  string const_type;
  {
    ostrstream ostr;
    Output save;
    o.save_streams (save);
    o.start_output (ostr);

    o << "const ";
    emit_base_type (type);

    o.stop_output ();
    o.restore_streams (save);

    ostr << ends;
    const_type = ostr.str();
    ostr.rdbuf()->freeze (0);
  }

  string const_init;
  {
    ostrstream ostr;
    Output save;
    o.save_streams (save);
    o.start_output (ostr);

    o << " = ";
    if( type->def_kind() == CORBA::dk_Fixed )
      o << "(CORBA::LongDouble) ";
    CORBA::Any* a = c->value();
    CORBA::TypeCode_var tc = a->type();
    if( tc->kind() == CORBA::tk_wstring || tc->kind() == CORBA::tk_wchar )
      o << "L";
    o << a;
    delete a;

    o.stop_output ();
    o.restore_streams (save);

    ostr << ends;
    const_init = ostr.str();
    ostr.rdbuf()->freeze (0);
  }

  CORBA::String_var const_name = c->name();
  CORBA::Container_var const_scope = c->defined_in();

  emit_var_def (const_type, const_scope, const_name.in(), const_init);
}



void CodeGenCPPStub::emit_Exception( CORBA::ExceptionDef_ptr e )
{
  assert( e->def_kind() == CORBA::dk_Exception );

  // Get base for absolute name
  CORBA::String_var tmp = e->absolute_name();
  string abs;
  abs = ID(tmp);
  tmp = e->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::TypeCode_var tc = e->type();

  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );

  CORBA::String_var repoid = e->id();

  if( !_params.pseudo ) {
    // Emit TypeCode for exception
    CORBA::Container_var container = e->defined_in();
    emit_typecode_def (container, abs, name.in(), tc);
    o << endl;
  }

  // Emit default ctor, copy ctor, assignment op
  CORBA::StructMemberSeq_var m = e->members();
  emit_StructDefaultMembers( name, (char *)abs.c_str(), m );

  if (!_params.pseudo) {
    // operator<<=
    o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
    o << abs << name << " &_e )" << endl;
    o << BL_OPEN;
    o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
    o << "return (_a.except_put_begin( \"" << repoid << "\" ) &&"
      << indent << endl;
    CORBA::StructMemberSeq_var members = e->members();
    for( CORBA::ULong i = 0; i < members->length(); i++ ) {
      o << "(_a <<= ";
      string v = "((";
      v += abs;
      v += name;
      v += "&)_e).";
      v += ID(members[ i ].name);
      emit_any_param( members[ i ].type_def, (char *) v.c_str(), true );
      o << ") &&" << endl;
    }
    o << "_a.except_put_end() );" << exdent << endl;
    o << BL_CLOSE << endl;
  
    // operator>>=
    o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
    o << abs << name << " &_e )" << endl;
    o << BL_OPEN;
    o << "CORBA::String_var _repoid;" << endl;
    o << "return (_a.except_get_begin( _repoid ) &&" << indent << endl;
    for( CORBA::ULong i0 = 0; i0 < members->length(); i0++ ) {
      o << "(_a >>= ";
      string v = "_e.";
      v += ID(members[ i0 ].name);
      emit_any_param( members[ i0 ].type_def, (char *) v.c_str(), false );
      o << ") &&" << endl;
    }
    o << "_a.except_get_end() );" << exdent << endl;
    o << BL_CLOSE << endl;
  }

  o << "void " << abs << name << "::_throwit() const" << endl;
  o << BL_OPEN;
  if( !_params.no_exceptions ) {
    o << "#ifdef HAVE_EXCEPTIONS" << endl;
    o << "throw " << name << "_var( (" << abs << name << "*)_clone() );"
      << endl;
    o << "#else" << endl;
    o << "CORBA::Exception::_throw_failed( _clone() );" << endl;
    o << "#endif" << endl;
  } else {
    o << "CORBA::Exception::_throw_failed( _clone() );" << endl;
  }
  o << BL_CLOSE << endl;
  
  o << "const char *" << abs << name << "::_repoid() const" << endl;
  o << BL_OPEN;
  o << "return \"" << repoid << "\";" << endl;
  o << BL_CLOSE << endl;

  if (!_params.pseudo) {
    o << "void " << abs << name
      << "::_encode( CORBA::DataEncoder &_en ) const" << endl;
    o << BL_OPEN;
    if( _params.sii ) {
      emit_marshaller_ref( e );
      o << "->marshal( _en, (void*) this );" << endl;
    } else {
      o << "CORBA::Any _a;" << endl;
      o << "_a <<= *this;" << endl;
      o << "_a.marshal( _en );" << endl;
    }
    o << BL_CLOSE << endl;
  }
  
  o << "CORBA::Exception *" << abs << name << "::_clone() const" << endl;
  o << BL_OPEN;
  o << "return new " << name << "( *this );" << endl;
  o << BL_CLOSE << endl;
  
  o << abs << name << " *" << abs << name;
  o << "::_narrow( CORBA::Exception *_ex )" << endl;
  o << BL_OPEN;
  o << "if( _ex && !strcmp( _ex->_repoid(), \"" << repoid << "\" ) )" << endl;
  o << indent << "return (" << name << " *) _ex;" << exdent << endl;
  o << "return NULL;" << endl;
  o << BL_CLOSE << endl;

  emit_marshaller( e );
}



void CodeGenCPPStub::emit_Enum( CORBA::EnumDef_ptr e )
{
  assert( e->def_kind() == CORBA::dk_Enum );

  // Get base for absolute name
  CORBA::String_var tmp;
  tmp = e->absolute_name();
  string abs;
  abs = ID(tmp);
  
  tmp = e->name();
  CORBA::String_var name = (const char *) ID(tmp);
  
  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );

  if (_params.pseudo) {
    return;
  }

  // Emit TypeCode for enum
  CORBA::Container_var container = e->defined_in();
  CORBA::TypeCode_var tc = e->type();
  emit_typecode_def (container, abs, name.in(), tc);
  o << endl;

  // operator<<=
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs << name << " &_e )" << endl;
  o << BL_OPEN;
  o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
  o << "return (_a.enum_put( (CORBA::ULong) _e ));" << endl;
  o << BL_CLOSE << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs << name << " &_e )" << endl;
  o << BL_OPEN;
  o << "CORBA::ULong _ul;" << endl;
  o << "if( !_a.enum_get( _ul ) )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "_e = (" << abs << name << ") _ul;" << endl;
  o << "return TRUE;" << endl;
  o << BL_CLOSE << endl;

  emit_marshaller( e );
}



void CodeGenCPPStub::emit_Alias( CORBA::AliasDef_ptr a )
{
  assert( a->def_kind() == CORBA::dk_Alias );

  // Get base for absolute name
  CORBA::String_var tmp = a->absolute_name();
  string abs;
  abs = ID(tmp);
  
  tmp = a->name();
  CORBA::String_var name = (const char *) ID(tmp);
  
  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );

  if (_params.pseudo) {
    return;
  }

  // Emit TypeCode for alias
  CORBA::Container_var container = a->defined_in();
  CORBA::TypeCode_var tc = a->type();
  emit_typecode_def (container, abs, name.in(), tc);
  o << endl;

  // We don't need to generate the operator<<=() and operator>>=()
  // marshalling functions. The C++ will automatically pick the
  // appropriate operators of the original types (i.e. the original
  // type of the alias). These operators are only defined for
  // arrays. The T_forany definition for arrays, as described in
  // [16.12(5)], helps to disambiguate arrays with the same
  // structure of the type.
  // The problem with all other aliases is, that the TypeCode isn't
  // set properly. The TypeCode for these aliases always refer to
  // the element_type_def. This is clearly a problem within the
  // CORBA standard. There ought to be a T_forany for every alias.

  CORBA::TypeCode_var orig_tc = tc->content_type();
  
  if( is_array( tc ) && !is_alias( orig_tc ) ) {
    // We have an alias to an array and the original type of the
    // alias is itself not an alias. Emit the marshalling
    // operators
    CORBA::AliasDef_var resolv = CORBA::AliasDef::_duplicate( a );
    CORBA::IDLType_var idl_type = resolv->original_type_def();
    CORBA::TypeCode_var idl_type_tc = idl_type->type();
    while( is_alias( idl_type_tc ) ) {
      resolv = CORBA::AliasDef::_narrow( idl_type );
      idl_type = resolv->original_type_def();
      idl_type_tc = idl_type->type();
    }
    CORBA::ArrayDef_var arr = CORBA::ArrayDef::_narrow( idl_type );
    assert( !CORBA::is_nil( arr ) );
    
    o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
    o << abs << name << "_forany &_const_s )" << endl;
    o << BL_OPEN;
    o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
    o << abs << name << "_forany &_s = "
      << "(" << abs << name << "_forany &) _const_s;" << endl;
    emit_array_marshalling( arr, true, 0 );
    o << "return TRUE;" << endl;
    o << BL_CLOSE << endl;
    
    o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
    o << abs << name << "_forany &_s )" << endl;
    o << BL_OPEN;
    emit_array_marshalling( arr, false, 0 );
    o << "return TRUE;" << endl;
    o << BL_CLOSE << endl << endl;
  }

  emit_marshaller( a );
}

void
CodeGenCPPStub::emit_var_def (const string &type,
			      CORBA::Container_ptr _scope,
			      const string &_name,
			      const string &init)
{
    CORBA::Contained_var scope = CORBA::Contained::_narrow (_scope);
    string name = _name;
    vector<string> modules;

    while (!CORBA::is_nil (scope)) {
	CORBA::String_var s = scope->name();
	if (scope->def_kind() != CORBA::dk_Module) {
	    name = string(s) + "::" + name;
	} else {
	    modules.push_back (s.in());
	}
	CORBA::Container_var container = scope->defined_in();
	scope = CORBA::Contained::_narrow (container);
    }
    if (name.find ("::") == 0)
	name = name.substr (2);

    // mapping is the same for structs and namespaces if outside of modules
    if (modules.size() == 0) {
	o << type << " " << ID(name.c_str()) << init << ";" << endl;
	return;
    }

    o << "#ifdef HAVE_NAMESPACE" << endl;

    for (mico_vec_size_type i = 0; i < modules.size(); ++i) {
	o << "namespace " << ID(modules[modules.size()-1-i].c_str()) << " { ";
    }
    o << type << " " << ID(name.c_str()) << init << ";";
    for (mico_vec_size_type j = 0; j < modules.size(); ++j) {
	o << " };";
    }
    o << endl;

    o << "#else" << endl;

    o << type << " ";
    for (mico_vec_size_type k = 0; k < modules.size(); ++k) {
	o << ID(modules[modules.size()-1-k].c_str()) << "::";
    }
    o << ID(name.c_str()) << init << ";" << endl;

    o << "#endif" << endl;
}

void
CodeGenCPPStub::emit_typecode_def (CORBA::Container_ptr scope,
				   const string &abs,
				   const string &_name,
				   CORBA::TypeCode_ptr tc)
{
  string name = string("_tc_") + _name;
#ifdef WINDOWS_TC
  o.switchStream( 1 );
  o << abs << name << " = ";
  emit_type_code( tc );
  o << ";" << endl;
  o.switchStream( 0 );
  emit_var_def ("CORBA::TypeCodeConst", scope, name, "");
#else
  string init;
  {
    ostrstream ostr;
    Output save;
    o.save_streams (save);
    o.start_output (ostr);

    o << " = ";
    emit_type_code (tc);

    o.stop_output ();
    o.restore_streams (save);

    ostr << ends;
    init = ostr.str();
    ostr.rdbuf()->freeze (0);
  }
  emit_var_def ("CORBA::TypeCodeConst", scope, name, init);
#endif
}
