/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */


/*
 * The GNU compiler chokes on this:
 *
 * struct S {
 *   char x[2][3];
 *   S( const S &s )
 *   {
 *     x = s.x;
 *   }
 * };
 *
 * It works, when the const-ness is cast away. The appropriate
 * locations, which are influenced by this bug, are marked with "XXXarray"
 */


#include <CORBA.h>
#include <fstream.h>
#include <ctype.h>
#include <stdio.h>
#include "codegen-c++-util.h"
#include <mico/template_impl.h>
#include <mico/util.h>


/*
 * The class CPPTypeFolder removes all those anonymous types from a
 * sequence of IRObjects, whose type in the C++ language would be
 * the same. This is necessary to avoid multiple definitions for
 * the same type. E.g., the IDL specification:
 *
 *   struct S {
 *     //...
 *   };
 *
 *   typedef S S_Alias;
 *   typedef sequence<S> seq1;
 *   typedef sequence<S_Alias> seq2;;
 *
 * only code for sequence<S> should be generated, but not for 
 * sequence<S_Alias>, because under the C++ mapping, those IDL
 * types denote the same C++ type.
 */
CPPTypeFolder::CPPTypeFolder( IRObjectSeq& objs )
{
  do_type_folding( objs );
}

/*
 * Resolves an alias (IDL-typedef) to its original type
 */
CORBA::TypeCode_ptr
CPPTypeFolder::resolve_alias( CORBA::TypeCode_ptr tc )
{
  CORBA::TypeCode_var tc_content;
  tc_content = CORBA::TypeCode::_duplicate( tc );
  while( tc_content->kind() == CORBA::tk_alias )
    tc_content = tc_content->content_type();
  return CORBA::TypeCode::_duplicate( tc_content );
}

/*
 * Returns true if TypeCode is a base type
 */
bool
CPPTypeFolder::is_base_type( CORBA::TypeCode_ptr tc )
{
  CORBA::TypeCode_var tc_resolv = resolve_alias( tc );
  switch( tc_resolv->kind() ) {
  case CORBA::tk_boolean:
  case CORBA::tk_char:
  case CORBA::tk_wchar:
  case CORBA::tk_octet:
  case CORBA::tk_ushort:
  case CORBA::tk_ulong:
  case CORBA::tk_short:
  case CORBA::tk_long:
  case CORBA::tk_longlong:
  case CORBA::tk_ulonglong:
  case CORBA::tk_float:
  case CORBA::tk_double:
  case CORBA::tk_longdouble:
  case CORBA::tk_string:
  case CORBA::tk_wstring:
  case CORBA::tk_fixed:
    return true;
  default:
    break;
  }
  return false;
}

/*
 * Returns true if TypeCode is of IDL-type Object or an alias of an Object
 */
bool
CPPTypeFolder::is_object( CORBA::TypeCode_ptr tc )
{
  CORBA::TypeCode_var tc_resolve = resolve_alias( tc );

  if( tc_resolve->kind() == CORBA::tk_objref ) {
    CORBA::String_var name = tc_resolve->name();
    if( strcmp( name, "Object" ) == 0 )
      return true;
  }
  return false;
}

/*
 * Resolves the parameter type of a sequence to its original type
 */
CORBA::TypeCode_ptr
CPPTypeFolder::resolve_sequence( CORBA::TypeCode_ptr tc_orig )
{
  CORBA::TypeCode_var tc = resolve_alias( tc_orig );
  
  if( tc->kind() != CORBA::tk_sequence )
    return CORBA::TypeCode::_duplicate( tc );
  
  CORBA::TypeCode_ptr tc_resolved;
  CORBA::TypeCode_var tc_content = tc->content_type();
  CORBA::TypeCode_var tc_resolve_seq = resolve_sequence( tc_content );
  tc_resolved =
      CORBA::TypeCode::create_sequence_tc( tc->length(),
					   tc_resolve_seq )->mk_constant();
  return tc_resolved;
}

/*
 * This method checks, if marshalling code needs to be generated for
 * the sequence or not (return == 'true' means no code needs to be
 * generated. This is done by checking if "seq" is contained in
 * "_seq_types". If not, it returns false and add "seq" to
 * "_seq_types". If yes, true is returned. The definition of
 * 'contained' is governed by the C++ type inference rule, whereupon:
 * sequence<string> and typedef string id; sequence<id> both refer to
 * the same C++ type. There is also no code generated for sequence<T>
 * where T is a basic data type for unbounded sequences. Those
 * marshalling routines are already part of libmico.a
 */
bool
CPPTypeFolder::check_for_sequence( CORBA::TypeCode_ptr seq,
				   CORBA::Boolean no_base_types )
{
  assert( seq->kind() == CORBA::tk_sequence );

  CORBA::TypeCode_var tc_reduced = resolve_sequence( seq );

#if 0
  // leave in seq of base types to emit prototypes for <<= and >>= operators
  if( no_base_types && seq->length() == 0 ) {
    CORBA::TypeCode_var tc_content = tc_reduced->content_type();
    if( (is_base_type( tc_content ) && tc_content->kind() != CORBA::tk_fixed)
	|| is_object( tc_content ) ){
      return true;
    }
  }
#endif

  for( CORBA::ULong i = 0; i < _seq_types.length(); i++ ) {
    if( tc_reduced->equal( _seq_types[ i ] ) )
      return true;
  }
  int len = _seq_types.length();
  _seq_types.length( len + 1 );
  _seq_types[ len ] = CORBA::TypeCode::_duplicate( tc_reduced );
  return false;
}



/*
 * This method checks, if "arr" is contained in "_arr_types". If not,
 * it returns false and adds "arr" to "_arr_types". If yes, true is
 * returned.
 */
bool
CPPTypeFolder::check_for_array( CORBA::TypeCode_ptr arr )
{
  assert( arr->kind() == CORBA::tk_array );

  for( CORBA::ULong i = 0; i < _arr_types.length(); i++ ) {
    if( arr->equal( _arr_types[ i ] ) )
      return true;
  }
  int len = _arr_types.length();
  _arr_types.length( len + 1 );
  _arr_types[ len ] = CORBA::TypeCode::_duplicate( arr );
  return false;
}



/*
 * This method checks, if "fix" is contained in "_fix_types". If not,
 * it returns false and adds "fix" to "_fix_types". If yes, true is
 * returned.
 */
bool
CPPTypeFolder::check_for_fixed( CORBA::TypeCode_ptr fix )
{
  assert( fix->kind() == CORBA::tk_fixed );

  for( CORBA::ULong i = 0; i < _fix_types.length(); i++ ) {
    if( fix->equal( _fix_types[ i ] ) )
      return true;
  }
  int len = _fix_types.length();
  _fix_types.length( len + 1 );
  _fix_types[ len ] = CORBA::TypeCode::_duplicate( fix );
  return false;
}

void
CPPTypeFolder::do_type_folding( IRObjectSeq& objs )
{
  for( CORBA::ULong i = 0; i < objs.length(); i++ ) {
    CORBA::IDLType_var t = CORBA::IDLType::_narrow( objs[ i ] );
    if( !CORBA::is_nil( t ) ) {

      bool erase;
      
      CORBA::TypeCode_var tc = t->type();
      switch( t->def_kind() ) {
      case CORBA::dk_Sequence:
	erase = check_for_sequence( tc );
	break;
      case CORBA::dk_Array:
	erase = check_for_array( tc );
	break;
      case CORBA::dk_Fixed:
	erase = check_for_fixed( tc );
	break;
      default:
	erase = false;
      }
      
      if( erase ) {
	// We can remove the current anonymous type
	for( CORBA::ULong j = i; j < objs.length() - 1; j++ ) {
	  CORBA::IRObject_ptr tmp = CORBA::IRObject::_duplicate( objs[ j + 1 ] );
	  objs[ j ] = tmp;
	}
	objs.length( objs.length() - 1 );
	i--;
      }
    }
  }
}


//----------------------------------------------------------------------


CodeGenCPPUtil::CodeGenCPPUtil( DB &db, IDLParam& params )
  : CodeGen( db.get_idl_container() ),
    _params( params )
{
  _db = &db;
  _emit_rel_names = true;
}



/*
 * The following operation checks if an IDL-identifier happens to
 * be a C++ keyword. If not, then the same identifier is returned. If
 * the identifier is a C++ keyword, then "_cxx_" is prepended
 * to avoid compiler error. See [20.1.2]
 */
char *CodeGenCPPUtil::ID( const char* ident )
{
  static char* keywords[] =
  { "and", "and_eq", "asm", "auto", "bitand", "bitor", "bool",
    "break", "case", "catch", "char", "class", "compl", "const",
    "const_cast", "continue", "default", "delete", "do", "double",
    "dynamic_cast", "else", "enum", "explicit", "extern", "false",
    "float", "for", "friend", "goto", "if", "inline", "int", "long",
    "mutable", "namespace", "new", "not", "not_eq", "operator", "or",
    "or_eq", "private", "protected", "public", "register",
    "reinterpret_cast", "return", "short", "signed", "sizeof", "static",
    "static_cast", "struct", "switch", "template", "this", "throw",
    "true", "try", "typedef", "typeid", "typename", "union", "unsigned",
    "using", "virtual", "void", "volatile", "wchar_t", "while", "xor",
    "xor_eq" 
  };

  static char *buf = NULL;

  string in = ident;
  string out;
  string current_id;
  
  while( in != "" ) {
    int colon_pos = in.find ( "::" );
    if( colon_pos == 0 ) {
      out += "::";
      in = in.substr( 2 );
      continue;
    }
    
    if( colon_pos != -1 ) {
      current_id = in.substr( 0, colon_pos );
      in = in.substr( colon_pos );
    } else {
      current_id = in;
      in = "";
    }
  
    for( size_t i = 0; i < sizeof( keywords ) / sizeof( char* ); i++ ) {
      if( current_id == keywords[ i ] ) {
	out += "_cxx_";
	break;
      }
    }
    out += current_id;
  }
  CORBA::string_free( buf );
  buf = CORBA::string_dup( (char *) out.c_str() );
  return buf;
}


void CodeGenCPPUtil::set_name_prefix( CORBA::Container_ptr c )
{
  if( CORBA::is_nil( c ) || c->def_kind() == CORBA::dk_Repository ) {
    _name_prefix = "::";
    return;
  }
  
  CORBA::Contained_var con = CORBA::Contained::_narrow( c );
  assert( !CORBA::is_nil( con ) );
  
  CORBA::String_var abs_name = con->absolute_name();
  _name_prefix = abs_name;
}

void CodeGenCPPUtil::save_name_prefix()
{
  _name_prefix_bak = _name_prefix;
}

void CodeGenCPPUtil::restore_name_prefix()
{
  _name_prefix = _name_prefix_bak;
}

void
CodeGenCPPUtil::enter_scope( CORBA::Container_ptr c,
			     string toplevel_prefix )
{
  vector<CORBA::Contained_var> trace_current_scope;
  vector<CORBA::Contained_var> trace_new_scope;
  CORBA::Container_var         scope;
  bool                         nl;

  // Decompose the current scope to a sequence of Contained
  // and store it in trace_current_scope
  scope = _current_scope;
  while( scope->def_kind() != CORBA::dk_Repository ) {
    CORBA::Contained_ptr scope_contained = CORBA::Contained::_narrow( scope );
    assert( !CORBA::is_nil( scope_contained ) );
    trace_current_scope.insert( trace_current_scope.begin(), scope_contained );
    scope = scope_contained->defined_in();
  }
  
  // Decompose the target scope to a sequence of Contained
  // and store it in trace_new_scope
  scope = CORBA::Container::_duplicate( c );
  while( scope->def_kind() != CORBA::dk_Repository ) {
    CORBA::Contained_ptr scope_contained = CORBA::Contained::_narrow( scope );
    assert( !CORBA::is_nil( scope_contained ) );
    trace_new_scope.insert( trace_new_scope.begin(), scope_contained );
    scope = scope_contained->defined_in();
  }
  
  // Compare both sequences and look for the first difference
  vector<CORBA::Contained_var>::iterator i, j;
  for( i = trace_current_scope.begin(), j = trace_new_scope.begin();
       i != trace_current_scope.end() && j != trace_new_scope.end() &&
	 (*i)->_is_equivalent( *j );
       i++, j++ ) ;
  
  // Backtrace and close all modules up to the difference found above
  nl = false;
  while( i != trace_current_scope.end() ) {

    CORBA::Container_var def_in = (*i)->defined_in();
    CORBA::String_var tmp = (*i)->name();
    CORBA::String_var name = (const char *) ID(tmp);
    
    if( def_in->def_kind() == CORBA::dk_Repository ) {
      o << "#endif // !defined(MICO_NO_TOPLEVEL_MODULES) || ";
      o << "defined(MICO_MODULE_" << toplevel_prefix << name << ")";
      o << endl << endl;
      o << "#ifndef MICO_NO_TOPLEVEL_MODULES" << endl;
      o << endl << "};" << endl;
      o << "#endif" << endl << endl;
    } else {
      o << "};" << endl;
    }
    
    nl = true;
    i++;
  }
  if( nl )
    o << endl << endl;

  // Open modules up to the target scope
  nl = false;
  while( j != trace_new_scope.end() ) {

    CORBA::Container_var def_in = (*j)->defined_in();
    CORBA::String_var tmp = (*j)->name();
    CORBA::String_var name = (const char *) ID(tmp);
    
    if( def_in->def_kind() == CORBA::dk_Repository ) {
      o << "#ifndef MICO_NO_TOPLEVEL_MODULES" << endl;
      o << "MICO_NAMESPACE_DECL " << toplevel_prefix << name << " {" << endl;
      o << "#endif" << endl << endl;
      o << "#if !defined(MICO_NO_TOPLEVEL_MODULES) || ";
      o << "defined(MICO_MODULE_" << toplevel_prefix << name << ")";
      o << endl << endl;
    } else {
      o << "MICO_NAMESPACE_DECL " << name << " {" << endl << endl;
    }
    
    CORBA::Container_var m  = CORBA::Container::_narrow( *j );
    assert( !CORBA::is_nil( m ) );

    // Emit all forward declarations for this scope
    emitForwardDcl( m );
    nl = true;
    j++;
  }
  if( nl )
    o << endl;
  
  _current_scope = CORBA::Container::_duplicate( c );

}

void
CodeGenCPPUtil::emitForwardDcl( CORBA::Container_ptr con )
{
  // Check if we already emitted the forward decls for this scope
  CORBA::String_var repo_id;
  if( con->def_kind() == CORBA::dk_Repository ) {
    // Generate pseudo repo id
    repo_id = (const char *) "__pseudo_repo_id";
  } else {
    CORBA::Contained_var c = CORBA::Contained::_narrow( con );
    assert( !CORBA::is_nil( c ) );
    repo_id = c->id();
  }
  if( _forward_dcls.find( string( repo_id ) ) != _forward_dcls.end() )
    return;
  
  _forward_dcls.insert( string( repo_id ) );
  
  CORBA::ContainedSeq_var c;
  c = con->contents( CORBA::dk_Interface, 1 );
  
  for( CORBA::ULong i = 0; i < c->length(); i++ ) {
    assert( c[ i ]->def_kind() == CORBA::dk_Interface );

    if (check_for_included_defn (c[i]))
      continue;
    
    CORBA::String_var tmp = c[ i ]->name();
    CORBA::String_var className = (const char *) ID(tmp);
    
    o << "class " << className << ";" << endl;
    o << "typedef " << className << " *" << className <<
      "_ptr;" << endl;
    o << "typedef " << className << "_ptr " << className <<
      "Ref;" << endl;
    o << "typedef ObjVar<" << className << "> " << className <<
      "_var;" << endl;
    o << "typedef " << className << "_var " << className << "_out;" << endl;
    o << endl;
  }
  o << endl;
}

bool CodeGenCPPUtil::use_rel_names( bool rn )
{
  bool cur = _emit_rel_names;
  _emit_rel_names = rn;
  return cur;
}

void CodeGenCPPUtil::emit_rel_name( CORBA::Object_ptr obj )
{
  CORBA::Contained_var con = CORBA::Contained::_narrow( obj );
  assert( !CORBA::is_nil( con ) );
  
  CORBA::String_var name_tmp = con->absolute_name();
  string name;
  name = name_tmp;

  /*
   * First check that we don't have the case that:
   * _name_prefix = ::A::B
   *         name = ::B
   * we need to keep 'name' as an absolute name, otherwise
   * it would refer to the wrong B
   */
  int i;
  for( i = 0; i + name.length() <= _name_prefix.length(); i++ ) {
    if( _name_prefix.substr( i, name.length() ) == name ) {
      o << ID((char*)name.c_str());
      return;
    }
  }
  
  int l = _name_prefix.length() < name.length() ?
               _name_prefix.length() : name.length();
  
  for( i = 0; i < l; i++ ) {
    if( _name_prefix[ i ] != name[ i ] )
      break;
  }
  
  if( i == name.length() ) {
    /*
     * We have the case that:
     *   _name_prefix = ::A::B::C::D
     *           name = ::A::B
     * in this case 'name' should be just 'B'
     */
    int j = name.find_last_of( ":" );
    name = name.substr( j + 1 );
  } else if( i == _name_prefix.length() ) {
    /*
     * We have the case that:
     * _name_prefix = ::A::B
     *         name = ::A::B::C::D
     * in this case name should be 'C::D'
     */
    if( name[ i ] != ':' )
      while( name[ i - 1 ] != ':' ) i--;
    name = name.substr( i );
  } else {
    /*
     * We have the case that:
     * _name_prefix = ::A::B::C
     *         name = ::A::D::E
     * in this case name should be 'D::E'
     */
    while( name[ --i ] != ':' ) ;
    name = name.substr( i + 1 );
  }

  if( name[ 0 ] == ':' )
    name = name.substr( 2 );
  
  o << ID((char*)name.c_str());
}


void CodeGenCPPUtil::emit_type_name( CORBA::Object_ptr obj )
{
  if (_emit_rel_names) {
    emit_rel_name (obj);
    return;
  }

  CORBA::Contained_var con = CORBA::Contained::_narrow( obj );
  assert( !CORBA::is_nil( con ) );
  
  CORBA::String_var name_tmp = con->absolute_name();
  string name;
  name = name_tmp;

  name = (const char *) ID(name_tmp);
  name = name.substr (2);
  o << name;
  return;
}

void CodeGenCPPUtil::emit_mangled_name( char* s )
{
  string str;
  int j;
  
  str = ID(s);
  while( 1 ) {
    j = str.find( "::" );
    if( j < 0 )
      break;
    str = str.replace( j, 2, "_" );
  }
  o << str;
}



/*
 * This method determines, if code has to be be generated by the definition
 * refered to "obj". No code needs to be generated if "obj" has been defined
 * in an included file. In this case we only emit an apropriate #include
 * statement. We use the map _repoid_fname_map to determine, whether "obj"
 * has been defined in an included file.
 */
bool CodeGenCPPUtil::check_for_included_defn( CORBA::IRObject_ptr obj )
{
  string id;
  
  CORBA::Contained_var con = CORBA::Contained::_narrow( obj );
  if( !CORBA::is_nil( con ) ) {
    CORBA::String_var tmp = con->id();
    id = tmp;
  } else {
    // Check if this is an anonymous type
    CORBA::IDLType_var idl_type = CORBA::IDLType::_narrow( obj );
    if( CORBA::is_nil( idl_type ) )
      // if it is not an IDLType, then it can't be an anonymous type
      return false;
    id = _db->gen_pseudo_repoid( idl_type );
  }
  
  if( _db->repoid_defined( id ) ) {
    // "obj" was defined in an include file. Don't emit code for it
    string fn;
    fn = _db->get_fname_for_repoid( id );
    string::size_type pos = fn.rfind( "." );
    if( (int)pos >= 0 && pos < fn.length() )
      fn = fn.substr (0, pos);
    fn += ".";
    fn += _params.h_suffix;
    if( _include_files.count( fn ) == 0 ) {
      _include_files.insert( fn );
      // For the first time we have to generate a #include directive
      if (_params.relative_paths) {
	// emit filename relative to specified include dirs
	for (vector<string>::size_type i = 0; i < _params.inc_paths.size(); ++i) {
	  if (fn.find (_params.inc_paths[i]) == 0) {
	    fn = fn.substr (_params.inc_paths[i].length());
	    while (fn[0] == '/')
	      fn = fn.substr (1);
	    break;
	  }
	}
      }
      o << "#ifdef MICO_IN_GENERATED_CODE" << endl;

      o << "#include ";
      o << ( _params.relative_paths ? "<" : "\"" );
      o << fn;
      o << ( _params.relative_paths ? ">" : "\"" );
      o << endl;

      o << "#else" << endl;

      o << "#define MICO_IN_GENERATED_CODE" << endl;
      o << "#include ";
      o << ( _params.relative_paths ? "<" : "\"" );
      o << fn;
      o << ( _params.relative_paths ? ">" : "\"" );
      o << endl;
      o << "#undef MICO_IN_GENERATED_CODE" << endl;

      o << "#endif" << endl << endl;
    }
    return true;
  }
  return false;
}


bool CodeGenCPPUtil::is_base_sequence (CORBA::SequenceDef_ptr sdef)
{
    CORBA::TypeCode_var seq_tc = sdef->type();

    CORBA::TypeCode_ptr tc = seq_tc->unalias();
    assert (tc->kind() == CORBA::tk_sequence);

    if (tc->length() != 0)
        return false;

    CORBA::TypeCode_var content_tc = tc->content_type();
    tc = content_tc->unalias();

    return (is_base_type(tc) && tc->kind() != CORBA::tk_fixed) || is_object(tc);
}

bool CodeGenCPPUtil::is_variable_for_sii (CORBA::TypeCode_ptr tc)
{
    return !is_string (tc) && !is_wstring (tc) && !is_objref (tc) &&
	!is_type_code (tc) && is_variable (tc);
}

void CodeGenCPPUtil::emit_any_param( CORBA::IDLType_ptr p,
				     char *var_name, bool is_from )
{
  const char *wrapper = is_from ? "CORBA::Any::from_" : "CORBA::Any::to_";

  CORBA::IDLType_var resolved_alias = CORBA::IDLType::_duplicate( p );
  while( resolved_alias->def_kind() == CORBA::dk_Alias ) {
    CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( resolved_alias );
    resolved_alias = a->original_type_def();
  }
  
  CORBA::PrimitiveDef_var prim = CORBA::PrimitiveDef::_narrow( resolved_alias );
  if( !CORBA::is_nil( prim ) && prim->kind() == CORBA::pk_objref ) {
    o << wrapper << "object( " << ID(var_name);
    if( is_from )
      o << ", \"Object\"";
    o << " )";
    return;
  }
  
  CORBA::TypeCode_var tc = p->type();
  CORBA::TypeCode_var tc_resolv = resolve_alias( tc );
  
  CORBA::TCKind tk = tc_resolv->kind();

  if( is_array( tc ) && is_alias( tc ) ) {
    if( is_from ) {
      emit_idl_type_name( p );
      o << "_forany( ";
      o << var_name << " )";
    } else {
      o << var_name;
    }
    return;
  }
  
  if( tk == CORBA::tk_objref ) {
    o << "(";
    emit_type_for_variable( p, manual );
    if( !is_from )
      o << "&";
    o << ") ";
  }

  if( tk == CORBA::tk_char )
    o << wrapper << "char( " << ID(var_name) << " )";
  else if( tk == CORBA::tk_wchar )
    o << wrapper << "wchar( " << ID(var_name) << " )";
  else if( tk == CORBA::tk_octet )
    o << wrapper << "octet( " << ID(var_name) << " )";
  else if( tk == CORBA::tk_boolean )
    o << wrapper << "boolean( " << ID(var_name) << " )";
  else if( tk == CORBA::tk_string ) {
    if( tc_resolv->length() == 0 ) {
      // Unbounded string
      o << ID(var_name);
    } else {
      // Bounded string
      o << wrapper << "string( " << ID(var_name) << ", "
	<< tc_resolv->length() << " )";
    }
  } else if( tk == CORBA::tk_wstring ) {
    if( tc_resolv->length() == 0 ) {
      // Unbounded wstring
      o << ID(var_name);
    } else {
      // Bounded wstring
      o << wrapper << "wstring( " << ID(var_name) << ", "
	<< tc_resolv->length() << " )";
    }
  } else
    o << ID(var_name);
}



void CodeGenCPPUtil::emit_type_code( CORBA::TypeCode_ptr tc )
{
  string str = tc->stringify();
  
  bool beginning_of_line = true;
  
  for( string::size_type i = 0; i < str.length(); i++ ) {
    if( beginning_of_line ) {
      o << "\"";
      beginning_of_line = false;
    }
    string c;
    c += str[ i ];
    o << c;
    if( (i+1) % 60 == 0 ) {
      o << "\"" << endl;
      beginning_of_line = true;
    }
  }
  if( !beginning_of_line )
    o << "\"";
}



void CodeGenCPPUtil::emit_type_code_name( CORBA::IDLType_ptr t )
{
  CORBA::PrimitiveDef_var prim = CORBA::PrimitiveDef::_narrow( t );
  if( !CORBA::is_nil( prim ) && prim->kind() == CORBA::pk_objref ) {
    o << "CORBA::_tc_Object";
    return;
  }
  
  CORBA::TypeCode_var tc = t->type();
  switch( tc->kind() ) {
  case CORBA::tk_void:
    o << "CORBA::_tc_void";
    break;
  case CORBA::tk_short:
    o << "CORBA::_tc_short";
    break;
  case CORBA::tk_long:
    o << "CORBA::_tc_long";
    break;
  case CORBA::tk_ushort:
    o << "CORBA::_tc_ushort";
    break;
  case CORBA::tk_ulong:
    o << "CORBA::_tc_ulong";
    break;
  case CORBA::tk_longlong:
    o << "CORBA::_tc_longlong";
    break;
  case CORBA::tk_ulonglong:
    o << "CORBA::_tc_ulonglong";
    break;
  case CORBA::tk_float:
    o << "CORBA::_tc_float";
    break;
  case CORBA::tk_double:
    o << "CORBA::_tc_double";
    break;
  case CORBA::tk_longdouble:
    o << "CORBA::_tc_longdouble";
    break;
  case CORBA::tk_boolean:
    o << "CORBA::_tc_boolean";
    break;
  case CORBA::tk_char:
    o << "CORBA::_tc_char";
    break;
  case CORBA::tk_wchar:
    o << "CORBA::_tc_wchar";
    break;
  case CORBA::tk_octet:
    o << "CORBA::_tc_octet";
    break;
  case CORBA::tk_any:
    o << "CORBA::_tc_any";
    break;
  case CORBA::tk_TypeCode:
    o << "CORBA::_tc_TypeCode";
    break;
  case CORBA::tk_Principal:
    o << "CORBA::_tc_Principal";
    break;
  case CORBA::tk_objref:
  case CORBA::tk_struct:
  case CORBA::tk_union:
  case CORBA::tk_enum:
  case CORBA::tk_alias:
    {
      string abs_name;
      CORBA::String_var tmp;
      switch( t->def_kind() ) {
      case CORBA::dk_Interface:
	{
	  CORBA::InterfaceDef_ptr in =
	    CORBA::InterfaceDef::_narrow( t );
	  tmp = in->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      case CORBA::dk_Struct:
	{
	  CORBA::StructDef_ptr s = 
	    CORBA::StructDef::_narrow( t );
	  tmp = s->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      case CORBA::dk_Union:
	{
	  CORBA::UnionDef_ptr u = 
	    CORBA::UnionDef::_narrow( t );
	  tmp = u->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      case CORBA::dk_Enum:
	{
	  CORBA::EnumDef_ptr e =
	    CORBA::EnumDef::_narrow( t );
	  tmp = e->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      case CORBA::dk_Alias:
	{
	  CORBA::AliasDef_ptr a =
	    CORBA::AliasDef::_narrow( t );
	  tmp = a->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      default:
	assert( 0 );
      }
      int i = abs_name.rfind( "::" );
      assert( i >= 0 );
      abs_name.insert( i + 2, "_tc_" );
      abs_name = abs_name.substr( 2 );
      o << abs_name;
      break;
    }
  case CORBA::tk_string:
    if( tc->length() == 0 )
      o << "CORBA::_tc_string";
    else
      o << "_tc_string_" << tc->length();
    break;
  case CORBA::tk_wstring:
    if( tc->length() == 0 )
      o << "CORBA::_tc_wstring";
    else
      o << "_tc_wstring_" << tc->length();
    break;
  case CORBA::tk_fixed:
    o << "_tc_fixed_" << tc->fixed_digits() << "_";
    if( tc->fixed_scale() < 0 )
      o << "n" << -( tc->fixed_scale() );
    else
      o << tc->fixed_scale();
    break;
  case CORBA::tk_sequence:
    // Sequences are anonymous types and don't have a _tc_* name
    // They are only referred to indirectly (through aliases e.g.)
    assert( 0 );
  case CORBA::tk_array:
    assert( 0 );
  default:
    assert( 0 );
  }
}



void CodeGenCPPUtil::emit_idl_type_name( CORBA::IDLType_ptr t )
{
  CORBA::String_var tmp;
  string abs_name;

  CORBA::TypeCode_var tc = t->type();
  
  switch( t->def_kind() ) {
  case CORBA::dk_Struct:
  case CORBA::dk_Union:
  case CORBA::dk_Enum:
  case CORBA::dk_Alias:
    {
      emit_type_name (t);
      if( is_objref( tc ) )
	// Must be IDL-Type "Object"
        o << "_ptr";
      break;
    }
  case CORBA::dk_Interface:
    {
      emit_type_name (t);
      o << "_ptr";
      break;
    }
  case CORBA::dk_Primitive:
    {
      // Gotta be an any, objref or a TypeCode
      CORBA::PrimitiveDef_var p = CORBA::PrimitiveDef::_narrow( t );
      if( p->kind() == CORBA::pk_any )
	o << "CORBA::Any";
      else if( p->kind() == CORBA::pk_objref )
	o << "CORBA::Object_ptr";
      else if( p->kind() == CORBA::pk_TypeCode )
	o << "CORBA::TypeCode_ptr";
      else {
	cout << p->kind() << endl;
	assert( 0 );
      }
      break;
    }
  default:
    cout << t->def_kind() << endl;
    assert( 0 );
  }
}
  


/*
 * Emit a TypeCode to be used in a signature of a C++ method.
 * The ParameterMode is necessary for this. See Table 24 in
 * [16.18(13)]
 */
void CodeGenCPPUtil::emit_type_for_param( CORBA::IDLType_ptr t,
					  CORBA::ParameterMode mode )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  // Mapping for arrays
  if( is_array( tc ) ) {
    if( mode == CORBA::PARAM_IN )
      o << "const ";
    emit_idl_type_name( t );
    if( mode == CORBA::PARAM_OUT && is_variable( tc ) )
      o << "_slice*&";
    return;
  }
  
  // Basic argument mapping [16.18(12)]
  if( is_string( tc ) ) {
    if( mode == CORBA::PARAM_IN )
      o << "const char*";
    else
      o << "char*&";
    return;
  }
    
  if( emit_base_type( t ) ) {
    if( mode != CORBA::PARAM_IN )
      o << "&";
    return;
  }

  if( is_enum( tc ) || is_objref( tc ) ) {
    emit_idl_type_name( t );
    if( mode != CORBA::PARAM_IN )
      o << "&";
    return;
  }

  if( is_type_code( tc ) ) {
    o << "CORBA::TypeCode_ptr";
    if( mode != CORBA::PARAM_IN )
      o << "&";
    return;
  }

  if( mode == CORBA::PARAM_IN )
    o << "const ";

  switch( tc->kind() ) {
  case CORBA::tk_struct:
  case CORBA::tk_union:
  case CORBA::tk_alias:
  case CORBA::tk_any:
    emit_idl_type_name( t );
    if( mode == CORBA::PARAM_OUT && is_variable( tc ) )
      o << "*";
    o << "&";
    break;
  default:
    cout << tc->kind() << endl;
    assert( 0 );
  }
}



/*
 * Emit a TypeCode to be used as a result in a signature of a
 * C++ method. See Table 24 in [16.18(13)]
 */
void CodeGenCPPUtil::emit_type_for_result( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  if( is_array( tc ) ) {
    // Mapping for arrays
    emit_idl_type_name( t );
    o << "_slice*";
    return;
  }

  if( is_string( tc ) ) {
    o << "char*";
    return;
  }
  if( is_wstring( tc ) ) {
    o << "CORBA::WChar*";
    return;
  }
  
  if( !emit_base_type( t ) ) {
    switch( tc->kind() ) {
    case CORBA::tk_objref:
    case CORBA::tk_struct:
    case CORBA::tk_union:
    case CORBA::tk_alias:
    case CORBA::tk_enum:
      emit_idl_type_name( t );
      break;
    case CORBA::tk_any:
      o << "CORBA::Any";
      break;
    case CORBA::tk_TypeCode:
      o << "CORBA::TypeCode_ptr";
      break;
    default:
      cout << "TCKind: " << tc->kind() << endl;
      assert( 0 );
    }
  }
  
  // Basic argument mapping [16.18(12)]
  if( is_variable( tc ) && !is_objref( tc ) )
    o << "*";
}



void CodeGenCPPUtil::emit_sequence( CORBA::IDLType_ptr type )
{
  assert( type->def_kind() == CORBA::dk_Sequence );
  CORBA::SequenceDef_var s = CORBA::SequenceDef::_narrow( type );
  CORBA::TypeCode_var seq_tc = s->type();
  CORBA::IDLType_var idl_type = s->element_type_def();

  if( s->bound() != 0 )
    o << "Bounded";

  CORBA::TypeCode_var tc = idl_type->type();
  if( is_array( tc ) || seq_tc->is_recursive_seq() )
    o << "SequenceIndTmpl<";
  else
    o << "SequenceTmpl<";
  
  if (is_char (tc))
    o << "CORBA::CharWrapper";
  else if( is_wchar( tc ) )
    o << "CORBA::WCharWrapper";
  else if (is_octet (tc))
    o << "CORBA::OctetWrapper";
  else if (is_bool (tc))
    o << "CORBA::BooleanWrapper";
  else if( is_array( tc ) ) {
    CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( idl_type );
    assert( !CORBA::is_nil( a ) );
    CORBA::IDLType_var orig = a->original_type_def();
    CORBA::TypeCode_var orig_tc = orig->type();
    CORBA::ULong num_elem = compute_num_array_elem( orig_tc );
    CORBA::String_var tmp = a->absolute_name();
    CORBA::String_var name = (const char *) ID(tmp);
    emit_type_for_variable( orig, automatic );
    o << "," << name << "," << num_elem;
  } else if( seq_tc->is_recursive_seq() ) {
    emit_type_for_variable( idl_type, automatic );
    o << ",";
    emit_type_for_variable( idl_type, automatic );
    o << ",1";
  } else
    emit_type_for_variable( idl_type, automatic );

  if( s->bound() != 0 )
    o << "," << s->bound();
  if( tc->kind() == CORBA::tk_fixed && s->bound() == 0 )
    o << " ";
  if( tc->kind() == CORBA::tk_sequence )
    o << " ";
  o << ">";
}
  


void CodeGenCPPUtil::emit_array( CORBA::IDLType_ptr type )
{
  assert( type->def_kind() == CORBA::dk_Array ||
	  type->def_kind() == CORBA::dk_Alias );

  CORBA::IDLType_var idl_type = CORBA::IDLType::_duplicate( type );

  CORBA::AliasDef_var alias;
  while( idl_type->def_kind() == CORBA::dk_Alias ) {
    alias = CORBA::AliasDef::_narrow( idl_type );
    assert( !CORBA::is_nil( alias ) );
    idl_type = alias->original_type_def();
  }
  assert( idl_type->def_kind() == CORBA::dk_Array );
  
  CORBA::ArrayDef_var arr;
  while( idl_type->def_kind() == CORBA::dk_Array ) {
    arr = CORBA::ArrayDef::_narrow( idl_type );
    assert( !CORBA::is_nil( arr ) );
    idl_type = arr->element_type_def();
  }
  emit_type_for_variable( idl_type, automatic );
}



void CodeGenCPPUtil::emit_fixed( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  o << "Fixed<" << tc->fixed_digits() << "," << tc->fixed_scale() << ">";
}



void CodeGenCPPUtil::emit_array_suffix( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  while( tc->kind() == CORBA::tk_array ) {
    o << "[ " << tc->length() << " ]";
    tc = tc->content_type();
  }
}



void CodeGenCPPUtil::emit_array_slice_suffix( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  tc = tc->content_type();
  while( tc->kind() == CORBA::tk_array ) {
    o << "[ " << tc->length() << " ]";
    tc = tc->content_type();
  }
}



CORBA::ULong CodeGenCPPUtil::compute_num_array_elem( CORBA::TypeCode_ptr tc )
{
  if( tc->kind() != CORBA::tk_array )
    return 1;
  CORBA::ULong len = tc->length();
  CORBA::TypeCode_var nested_tc = tc->content_type();
  return len * compute_num_array_elem( nested_tc );
}



/*
 * Emit a TypeCode to be used in a C++ variable declaration.
 */
void CodeGenCPPUtil::emit_type_for_variable( CORBA::IDLType_ptr t,
					     Storage s )
{
  CORBA::String_var tmp;
  
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  if( is_string( tc ) ) {
    if( is_alias( tc ) ) {
      emit_type_name (t);
      if( s == automatic )
	o << "_var";
    } else {
      if( s == manual )
	o << "char*";
      else
	o << "CORBA::String_var";
    }
    return;
  }

  if( is_wstring( tc ) ) {
    if( is_alias( tc ) ) {
      CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( t );
      emit_type_name (t);
      if( s == automatic )
	o << "_var";
    } else {
      if( s == manual )
	o << "CORBA::WChar*";
      else
	o << "CORBA::WString_var";
    }
    return;
  }

  CORBA::PrimitiveDef_var prim = CORBA::PrimitiveDef::_narrow( t );
  if( !CORBA::is_nil( prim ) && prim->kind() == CORBA::pk_objref ) {
    o << "CORBA::Object";
    if( s == manual )
      o << "_ptr";
    else
      o << "_var";
    return;
  }
  
  if( is_objref( tc ) ) {
    // This can be an object reference or an alias to "Object"
    emit_type_name (t);
    if( s == manual )
      o << "_ptr";
    else
      o << "_var";
    return;
  }

  if( !emit_base_type( t ) ) {
    switch( tc->kind() ) {
    case CORBA::tk_objref:
    case CORBA::tk_struct:
    case CORBA::tk_union:
    case CORBA::tk_enum:
    case CORBA::tk_alias:
    case CORBA::tk_any:
      emit_idl_type_name( t );
      break;
    case CORBA::tk_sequence:
      emit_sequence( t );
      break;
    case CORBA::tk_TypeCode:
      o << "CORBA::TypeCode";
      if( s == manual )
	o << "_ptr";
      else
	o << "_var";
      break;
    case CORBA::tk_array:
      emit_array( t );
      if( s == manual )
	o << "_slice";
      // the case "s == automatic" is taken care of
      // in emit_array_suffix
      break;
    default:
      assert( 0 );
    }
  }
  if( s == manual && is_variable( tc ) )
    o << "*";
}



bool CodeGenCPPUtil::emit_base_type( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  // Mapping of Base Data Types [16.5]
  switch( tc->kind() ) {
  case CORBA::tk_void:
    o << "void";
    break;
  case CORBA::tk_boolean:
    o << "CORBA::Boolean";
    break;
  case CORBA::tk_char:
    o << "CORBA::Char";
    break;
  case CORBA::tk_wchar:
    o << "CORBA::WChar";
    break;
  case CORBA::tk_octet:
    o << "CORBA::Octet";
    break;
  case CORBA::tk_ushort:
    o << "CORBA::UShort";
    break;
  case CORBA::tk_ulong:
    o << "CORBA::ULong";
    break;
  case CORBA::tk_short:
    o << "CORBA::Short";
    break;
  case CORBA::tk_long:
    o << "CORBA::Long";
    break;
  case CORBA::tk_longlong:
    o << "CORBA::LongLong";
    break;
  case CORBA::tk_ulonglong:
    o << "CORBA::ULongLong";
    break;
  case CORBA::tk_float:
    o << "CORBA::Float";
    break;
  case CORBA::tk_double:
    o << "CORBA::Double";
    break;
  case CORBA::tk_longdouble:
    o << "CORBA::LongDouble";
    break;
  case CORBA::tk_string:
    o << "char*";
    break;
  case CORBA::tk_wstring:
    o << "CORBA::WChar*";
    break;
  case CORBA::tk_fixed:
    o << "Fixed<" << tc->fixed_digits() << "," << tc->fixed_scale() << ">";
    break;
  case CORBA::tk_alias:
    if( is_base_type( tc ) ) {
      emit_idl_type_name( t );
      return true;
    } else
      return false;
    break;
  default:
    return false;
  }
  return true;
}



void CodeGenCPPUtil::emit_array_marshalling( CORBA::ArrayDef_ptr arr,
					     bool is_from,
					     int level )
{
  const char *direction = is_from ? "put" : "get";
  
  o << "if( !_a.array_" << direction << "_begin() )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
  o << "for( CORBA::ULong _i" << level << " = 0; _i" << level << " < "
    << arr->length() << "; _i" << level << "++ ) {" << indent << endl;
  CORBA::IDLType_var elem = arr->element_type_def();
  if( elem->def_kind() == CORBA::dk_Array ) {
    CORBA::ArrayDef_var nested_arr = CORBA::ArrayDef::_narrow( elem );
    assert( !CORBA::is_nil( nested_arr ) );
    emit_array_marshalling( nested_arr, is_from, ++level );
  } else {
    o << "if( !(_a ";
    if( is_from )
      o << " <<= ";
    else
      o << " >>= ";
    string var;
    var = "_s";
    for( int i = 0; i <= level; i++ ) {
      var += "[ _i";
      var += xdec( i );
      var += " ]";
    }
    emit_any_param( elem, (char *) var.c_str(), is_from );
    o << ") )" << endl;
    o << indent << "return FALSE;" << exdent << endl;
  }
  o << BL_CLOSE;
  
  o << "if( !_a.array_" << direction << "_end() )" << endl;
  o << indent << "return FALSE;" << exdent << endl;
}



bool CodeGenCPPUtil::is_union_case_defined( CORBA::UnionMemberSeq& mem,
					    CORBA::Any &any )
{
  for( CORBA::ULong i = 0; i < mem.length(); i++ )
    if( mem[ i ].label == any )
      return true;
  return false;
}



string
CodeGenCPPUtil::generate_union_default_case_label( CORBA::IDLType_ptr discr,
						   CORBA::UnionMemberSeq &mem )
{
  string label;
  label = "";
  
  CORBA::TypeCode_var tc = discr->type();
  if( is_enum( tc ) ) {
    CORBA::IDLType_var t = CORBA::IDLType::_duplicate( discr );
    while( t->def_kind() == CORBA::dk_Alias ) {
      CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( t );
      t = a->original_type_def();
    }
    CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( t );
    assert( !CORBA::is_nil( e ) );
    CORBA::EnumMemberSeq_var e_mem = e->members();
    for( CORBA::ULong i = 0; i < e_mem->length(); i++ ) {
      CORBA::Any any;
      any.type( tc->unalias() );
      CORBA::Boolean r = any.enum_put( i );
      assert (r);
      if( !is_union_case_defined( mem, any ) ) {
	label = e_mem[ i ];
	break;
      }
    }
    return label;
  }
  
  tc = resolve_alias( tc );
  if( tc->kind() == CORBA::tk_long ) {
    CORBA::Long l = 0;
    do {
      CORBA::Any any;
      any <<= l;
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (l);
	break;
      }
    } while (++l);
    return label;
  }

  if( tc->kind() == CORBA::tk_ulong ) {
    CORBA::ULong ul = 0;
    do {
      CORBA::Any any;
      any <<= ul;
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (ul);
	break;
      }
    } while (++ul);
    return label;
  }

  if( tc->kind() == CORBA::tk_longlong ) {
    CORBA::LongLong ll = 0;
    do {
      CORBA::Any any;
      any <<= ll;
      if( !is_union_case_defined( mem, any ) ) {
	  label = xdec (ll);
	break;
      }
    } while (++ll);
    return label;
  }

  if( tc->kind() == CORBA::tk_ulonglong ) {
    CORBA::ULongLong ull = 0;
    do {
      CORBA::Any any;
      any <<= ull;
      if( !is_union_case_defined( mem, any ) ) {
	  label = xdec (ull);
	break;
      }
    } while (++ull);
    return label;
  }

  if( tc->kind() == CORBA::tk_char ) {
    CORBA::Char c = 0;
    do {
      CORBA::Any any;
      any <<= CORBA::Any::from_char( c );
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (c);
	break;
      }
    } while (++c);
    return label;
  }

  if( tc->kind() == CORBA::tk_wchar ) {
    CORBA::WChar c = 0;
    do {
      CORBA::Any any;
      any <<= CORBA::Any::from_wchar( c );
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (c);
	break;
      }
    } while (++c);
    return label;
  }

  if( tc->kind() == CORBA::tk_short ) {
    CORBA::Short s = 0;
    do {
      CORBA::Any any;
      any <<= s;
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (s);
	break;
      }
    } while (++s);
    return label;
  }

  if( tc->kind() == CORBA::tk_ushort ) {
    CORBA::UShort us = 0;
    do {
      CORBA::Any any;
      any <<= us;
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (us);
	break;
      }
    } while (++us);
    return label;
  }

  if( tc->kind() == CORBA::tk_boolean ) {
    CORBA::Any any;
    any <<= CORBA::Any::from_boolean( TRUE );
    if( is_union_case_defined( mem, any ) ) {
      any <<= CORBA::Any::from_boolean( FALSE );
      if( !is_union_case_defined( mem, any ) )
	label = "FALSE";
    } else
      label = "TRUE";
    return label;
  }

  return label;
}


void
CodeGenCPPUtil::emit_marshaller_ref( CORBA::IRObject_ptr obj )
{
  string mangled_name;
  
  gen_mangled_name( mangled_name, obj );
  o << mangled_name;
}

void
CodeGenCPPUtil::gen_mangled_name( string &str,
				  CORBA::IRObject_ptr obj,
				  bool first_call_to_gen )
{
  switch( obj->def_kind() ) {
  case CORBA::dk_Primitive:
    {
      CORBA::PrimitiveDef_var p = CORBA::PrimitiveDef::_narrow( obj );
      CORBA::TypeCode_var tc = p->type();
      if( first_call_to_gen )
	str += "CORBA::_stc_";
      switch( tc->kind() ) {
      case CORBA::tk_void:
	str += "void";
	break;
      case CORBA::tk_boolean:
	str += "boolean";
	break;
      case CORBA::tk_char:
	str += "char";
	break;
      case CORBA::tk_wchar:
	str += "wchar";
	break;
      case CORBA::tk_octet:
	str += "octet";
	break;
      case CORBA::tk_ushort:
	str += "ushort";
	break;
      case CORBA::tk_ulong:
	str += "ulong";
	break;
      case CORBA::tk_short:
	str += "short";
	break;
      case CORBA::tk_long:
	str += "long";
	break;
      case CORBA::tk_longlong:
	str += "longlong";
	break;
      case CORBA::tk_ulonglong:
	str += "ulonglong";
	break;
      case CORBA::tk_float:
	str += "float";
	break;
      case CORBA::tk_double:
	str += "double";
	break;
      case CORBA::tk_longdouble:
        str += "longdouble";
	break;
      case CORBA::tk_string:
	str += "string";
	break;
      case CORBA::tk_wstring:
	str += "wstring";
	break;
      case CORBA::tk_objref:
	str += "Object";
	break;
      case CORBA::tk_TypeCode:
	str += "TypeCode";
	break;
      case CORBA::tk_any:
	str += "any";
	break;
      default:
	assert( 0 );
      }
      break;
    }
  case CORBA::dk_Sequence:
    {
      CORBA::SequenceDef_var s = CORBA::SequenceDef::_narrow( obj );
      CORBA::ULong bound = s->bound();

      if( first_call_to_gen && is_base_sequence( s ) )
	str += "CORBA::_stcseq_";
      
      if( first_call_to_gen && !is_base_sequence( s ) )
	str += "_marshaller__seq_";

      if( !first_call_to_gen )
	str += "_seq_";
      
      if( bound > 0 ) {
	str += xdec( bound );
	str += "b_";
      }
      
      CORBA::IDLType_var elem = s->element_type_def();

      // Need to resolve Alias in case it points to a PrimitiveDef
      // (and therefore is a base sequence)
      while( elem->def_kind() == CORBA::dk_Alias ) {
	CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( elem );
	elem = a->original_type_def();
      }
      
      if( elem->def_kind() != CORBA::dk_Primitive )
	elem = s->element_type_def();

      gen_mangled_name( str, elem, false );
      break;
    }
  case CORBA::dk_Array:
    {
      CORBA::ArrayDef_var a = CORBA::ArrayDef::_narrow( obj );
      if( first_call_to_gen )
	str += "_marshaller_";
      str += "_a";
      str += xdec( a->length() );
      str += "_";
      CORBA::IDLType_var elem = a->element_type_def();
      gen_mangled_name( str, elem, false );
      break;
    }
  case CORBA::dk_Fixed:
    {
      CORBA::FixedDef_var f = CORBA::FixedDef::_narrow( obj );
      if( first_call_to_gen )
	str += "_marshaller_";
      str += "_fixed_";
      str += xdec( f->digits() );
      if( f->scale() < 0 ) {
	str += "_n";
	str += xdec( -( f->scale() ) );
      } else {
	str += "_";
	str += xdec( f->scale() );
      }
      break;
    }
  case CORBA::dk_String:
    if( first_call_to_gen )
      str += "CORBA::_stc_";
    str += "string";
    break;
  case CORBA::dk_Wstring:
    if( first_call_to_gen )
      str += "CORBA::_stc_";
    str += "wstring";
    break;
  case CORBA::dk_Struct:
  case CORBA::dk_Union:
  case CORBA::dk_Alias:
  case CORBA::dk_Enum:
  case CORBA::dk_Interface:
  case CORBA::dk_Exception:
    {
      CORBA::Contained_var c = CORBA::Contained::_narrow( obj );
      CORBA::String_var n = c->absolute_name();
      string name;
      name = ID(n);
      for( int i = name.find ("::"); i >= 0; i = name.find ("::") )
	name.replace( i, 2, "_" );
      name = name.substr( 1 );
      if( first_call_to_gen )
	str += "_marshaller_";
      str += name;
      break;
    }
  default:
    assert( 0 );
  }
}
