/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

/*
 * The GNU compiler chokes on this:
 *
 * struct S {
 *   char x[2][3];
 *   S( const S &s )
 *   {
 *     x = s.x;
 *   }
 * };
 *
 * It works, when the const-ness is cast away. The appropriate
 * locations, which are influenced by this bug, are marked with "XXXarray"
 */


#include <CORBA.h>
#include <fstream.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "codegen-c.h"
#include <mico/template_impl.h>
#include <mico/util.h>


static char *copyright[] = {
  "/*",
  " *  MICO --- a free CORBA implementation",
  " *  Copyright (C) 1997-98 Kay Roemer & Arno Puder",
  " *",
  " *  This file was automatically generated. DO NOT EDIT!",
  " */"
};

const int copyright_size = sizeof( copyright ) / sizeof( char * );



CodeGenC::CodeGenC( DB &db, bool gen_exceptions, bool query_ir_for_narrow,
			string &cpp_suffix, string &h_suffix )
  : CodeGen( db.get_idl_container() )
{
  _db = &db;
  _gen_exceptions = gen_exceptions;
  _query_ir_for_narrow = query_ir_for_narrow;
  _cpp_suffix = cpp_suffix;
  _h_suffix = h_suffix;
}


void CodeGenC::emit( string &fn )
{
  string fnbase = fn;
  
  for( string::size_type i = 0; i < fnbase.length(); i++ ) {
    fnbase[ i ] = isalnum( fnbase[ i ] ) ? tolower( fnbase[ i ] ): '_';
  }
  
  string fnHeader = fnbase + "-c." + _h_suffix;
  string fnStub   = fnbase + "-stub." + _cpp_suffix;
  string fnServer = fnbase + "-skel." + _cpp_suffix;
  
  ofstream header( fnHeader.c_str() );

  for( int i0 = 0; i0 < copyright_size; i0++ )
    header << copyright[ i0 ] << endl;

  //CTypeFolder folder( _idl_objs );
  
  header << endl;
  header << "#if !defined(__" << fnbase << "_H__)" << endl;
  header << "#define __" << fnbase << "_H__" << endl;
  header << endl;
  header << "#include \"CORBA-c.h\"" << endl;

  o.start_output( header );
  o.switchStream( 0 );
  emitHeader();
  o.stop_output();

#if 0  
  header << endl;
  header << "#include <mico/template_impl.h>" << endl;
  header << endl;
#endif
  header << "#endif /* __" << fnbase << "_H__ */" << endl;

  ofstream stub( fnStub.c_str() );

  for( int i1 = 0; i1 < copyright_size; i1++ )
    stub << copyright[ i1 ] << endl;
  
  stub << endl;
  stub << "#include <CORBA.h>" << endl;
  stub << "#include <mico/throw.h>" << endl;
  stub << "#include <mico/template_impl.h>" << endl;
  stub << "#define CPP_WRAPPER 1" << endl;
  stub << "extern \"C\" {" << endl;
  stub << "#include \"" << fnHeader << '"' << endl << endl;
  stub << "//--------------------------------------------------------" << endl;
  stub << "//  Implementation of stubs" << endl;
  stub << "//--------------------------------------------------------" << endl;

  o.start_output( stub );
  emitClient();
  o.stop_output();

  stub << "} /* end extern \"C\" */" << endl;

  ofstream skel( fnServer.c_str() );
  for( int i2 = 0; i2 < copyright_size; i2++ )
    skel << copyright[ i2 ] << endl;
  
  skel << endl;
  skel << "#include <CORBA.h>" << endl;
  skel << "#include <mico/throw.h>" << endl;
  skel << "#include <mico/template_impl.h>" << endl;
  skel << "#define CPP_WRAPPER 1" << endl;
  skel << "extern \"C\" {" << endl;
  skel << "#include \"" << fnHeader << '"' << endl << endl;
  skel << "//--------------------------------------------------------" << endl;
  skel << "//  Implementation of skeletons" << endl;
  skel << "//--------------------------------------------------------" << endl;

  o.start_output( skel );
  emitServer();
  o.stop_output();

  skel << "} /* end extern \"C\" */" << endl;
}

char *
CodeGenC::get_absolute_name (CORBA::Contained_ptr cont, const char *sep)
{
    CORBA::String_var _s = cont->absolute_name();
    string s = (const char *)_s;

    for (int i = s.find ("::"); i >= 0; i = s.find ("::")) {
        if (i == 0)
            s.replace (i, 2, "");
        else
            s.replace (i, 2, sep);
    }
    return CORBA::string_dup (s.c_str());
}

/*
 * The following operation checks if an IDL-identifier happens to
 * be a C++ keyword. If not, then the same identifier is returned. If
 * the identifier is a C++ keyword, then a underscore "_" is prepended
 * to avoid compiler error. See [16.1.1(4)]
 */
char *CodeGenC::ID( char* ident )
{
  static char* keywords[] =
  { "and", "and_eq", "asm", "auto", "bitand", "bitor", "bool",
    "break", "case", "catch", "char", "class", "compl", "const",
    "const_cast", "continue", "default", "delete", "do", "double",
    "dynamic_cast", "else", "enum", "explicit", "extern", "false",
    "float", "for", "friend", "goto", "if", "inline", "int", "long",
    "mutable", "namespace", "new", "not", "not_eq", "operator", "or",
    "or_eq", "private", "protected", "public", "register",
    "reinterpret_cast", "return", "short", "signed", "sizeof", "static",
    "static_cast", "struct", "switch", "template", "this", "throw",
    "true", "try", "typedef", "typeid", "typename", "union", "unsigned",
    "using", "virtual", "void", "volatile", "wchar_t", "while", "xor",
    "xor_eq" 
  };

  static char *buf = NULL;

  string in = ident;
  string out;
  string current_id;
  
  while( in != "" ) {
    int colon_pos = in.find( "::" );
    if( colon_pos == 0 ) {
      out += "::";
      in = in.substr( 2 );
      continue;
    }
    
    if( colon_pos != -1 ) {
      current_id = in.substr( 0, colon_pos );
      in = in.substr( colon_pos );
    } else {
      current_id = in;
      in = "";
    }
  
    for( size_t i = 0; i < sizeof( keywords ) / sizeof( char* ); i++ ) {
      if( current_id == keywords[ i ] ) {
	out += "_";
	break;
      }
    }
    out += current_id;
  }
  CORBA::string_free( buf );
  buf = CORBA::string_dup( (char *) out.c_str() );
  return buf;
}

void CodeGenC::emit_mangled_name( char* s )
{
  string str;
  int j;
  
  str = ID(s);
  while( 1 ) {
    j = str.find( "::" );
    if( j < 0 )
      break;
    str = str.replace( j, 2, "_" );
  }
  o << str;
}

/*
 * This method determines, if code has to be be generated by the definition
 * refered to "obj". No code needs to be generated if "obj" has been defined
 * in an included file. In this case we only emit an apropriate #include
 * statement. We use the map _repoid_fname_map to determine, whether "obj"
 * has been defined in an included file.
 */
bool CodeGenC::check_for_included_defn( CORBA::IRObject_ptr obj )
{
  string id;
  
  CORBA::Contained_var con = CORBA::Contained::_narrow( obj );
  if( !CORBA::is_nil( con ) ) {
    CORBA::String_var tmp = con->id();
    id = tmp;
  } else {
    // Check if this is an anonymous type
    CORBA::IDLType_var idl_type = CORBA::IDLType::_narrow( obj );
    if( CORBA::is_nil( idl_type ) )
      // if it is not an IDLType, then it can't be an anonymous type
      return false;
    id = _db->gen_pseudo_repoid( idl_type );
  }
  
  if( _db->repoid_defined( id ) ) {
    // "obj" was defined in an include file. Don't emit code for it
    string fn;
    fn = _db->get_fname_for_repoid( id );
    string::size_type pos = fn.rfind( "." );
    if( (int)pos >= 0 && pos < fn.length() )
      fn = fn.substr (0, pos);
    fn += ".h";
    if( _include_files.count( fn ) == 0 ) {
      // For the first time we have to generate a #include directive
      o << "#include \"" << fn << "\"" << endl << endl;
      _include_files.insert( fn );
    }
    return true;
  }
  return false;
}

void CodeGenC::emitPrototypes( CORBA::InterfaceDef_ptr in,
				 bool as_pure_virtual )
{
  CORBA::InterfaceDef::FullInterfaceDescription_var desc;
  desc = in->describe_interface();
  CORBA::String_var myctx = get_absolute_name(in, "_");

  // Emit attributes
  for( CORBA::ULong j = 0; j < desc->attributes.length(); j++ ) {
    CORBA::AttributeDescription attr;
    attr = desc->attributes[ j ];
    CORBA::IDLType_var type_id = lookup_attribute_by_id( attr.id );
    o << "extern ";
    emit_type_for_result( type_id );
    
    o << " " << myctx << "__get_" << ID( attr.name )
      << "(CORBA_Object _handle, CORBA_Environment *ev);" << endl;

    o << "extern ";
    emit_type_for_result( type_id );
    o << " server_" << myctx << "__get_" << ID( attr.name )
      << "(CORBA_Object _handle, CORBA_Environment *ev);" << endl;

    if( attr.mode == CORBA::ATTR_NORMAL ) {
      o << "extern void " << myctx << "__set_" << ID(attr.name) << "(CORBA_Object _handle, ";
      emit_type_for_param( type_id, CORBA::PARAM_IN );
      o << " value, CORBA_Environment *ev);" << endl;
      o << "extern void server_" << myctx << "__set_" << ID(attr.name) << "(CORBA_Object _handle, ";
      emit_type_for_param( type_id, CORBA::PARAM_IN );
      o << " value, CORBA_Environment *ev);" << endl;
    }
  }

  // Emit operations
  for( CORBA::ULong j0 = 0; j0 < desc->operations.length(); j0++ ) {
    CORBA::OperationDescription op;
    op = desc->operations[ j0 ];
#if 0
    if( as_pure_virtual )
      o << "virtual ";
#endif
    CORBA::IDLType_var type_id = lookup_result_by_id( op.id );
    emit_type_for_result( type_id );
    o << " " << myctx << "_" << ID(op.name) << "(CORBA_Object _handle, ";
    CORBA::ParDescriptionSeq p = op.parameters;
    CORBA::ULong k;
    for( k = 0; k < p.length(); k++ ) {
      emit_type_for_param( p[ k ].type_def, p[ k ].mode );
      o << " " << ID(p[ k ].name);
      o << ", ";
    }

    // Do we have a context?
    CORBA::ContextIdSeq& ctx = op.contexts;
    
    if( ctx.length() > 0 ) {
      if( p.length() > 0 )
	o << ", ";
      o << "CORBA::Context_ptr _ctx";
    }
    
    if( k != 0 || ctx.length() > 0 )
      o << " ";
    else if(p.length() > 0)
      o << ", ";
    o << "CORBA_Environment *ev)";
    o << ";" << endl;

    // Redo server-end thing

    o << "extern ";
    emit_type_for_result( type_id );
    o << " server_" << myctx << "_" << ID(op.name)
      << "(CORBA_Object _handle, ";
    p = op.parameters;
    for( k = 0; k < p.length(); k++ ) {
      emit_type_for_param( p[ k ].type_def, p[ k ].mode );
      o << " " << ID(p[ k ].name);
      o << ", ";
    }

    // Do we have a context?
    ctx = op.contexts;
    
    if( ctx.length() > 0 ) {
      if( p.length() > 0 )
	o << ", ";
      o << "CORBA::Context_ptr _ctx";
    }
    
    if( k != 0 || ctx.length() > 0 )
      o << " ";
    else if(p.length() > 0)
      o << ", ";
    o << "CORBA_Environment *ev)";
    o << ";" << endl;

  }
}


void CodeGenC::emitLocalDecls( CORBA::InterfaceDef_ptr in )
{
  CORBA::ContainedSeq_var c;
  c = in->contents( CORBA::dk_all, 1 );
  for( CORBA::ULong i = 0; i < c->length(); i++ ) {
    switch( c[ i ]->def_kind() ) {
    case CORBA::dk_Operation:
    case CORBA::dk_Attribute:
      // Operation & attribute declarations are generated in emitPrototypes()
      break;
    case CORBA::dk_Struct:
      {
	CORBA::StructDef_var s = CORBA::StructDef::_narrow( c[ i ] );
	emitHeaderStruct( s );
	break;
      }
    case CORBA::dk_Union:
      {
	CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( c[ i ] );
	emitHeaderUnion( u );
	break;
      }
    case CORBA::dk_Constant:
      {
	CORBA::ConstantDef_var co = CORBA::ConstantDef::_narrow( c[ i ] );
	emitHeaderConstant( co );
	break;
      }
    case CORBA::dk_Exception:
      {
	CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( c[ i ] );
	emitHeaderException( e );
	break;
      }
    case CORBA::dk_Enum:
      {
	CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( c[ i ] );
	emitHeaderEnum( e );
	break;
      }
    case CORBA::dk_Alias:
      {
	CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( c[ i ] );
	emitHeaderAlias( a );
	break;
      }
    default:
      assert( 0 );
    }
  }
}

void CodeGenC::emitForwardDcl()
{
  for( CORBA::ULong i = 0; i < _idl_objs.length(); i++ ) {

    CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( _idl_objs[i] );
    
    if( !CORBA::is_nil( in ) ) {
      
      if( check_for_included_defn( in ) )
	continue;

      CORBA::String_var aname = get_absolute_name(in, "_");
      o << "typedef CORBA_Object " << aname << ";" << endl;
    }
  }
  
  o << endl;
}

void CodeGenC::emitHeader()
{
  emitForwardDcl();

  for( CORBA::ULong i = 0; i < _idl_objs.length(); i++ ) {
    emit_header_idl_obj( _idl_objs[ i ] );
  }
}

void CodeGenC::emit_header_idl_obj( CORBA::IRObject_ptr obj )
{
  if( check_for_included_defn( obj ) )
    return;
    
  switch( obj->def_kind() ) {
  case CORBA::dk_Interface:
    {
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( obj );
      emitHeaderInterface( in );
      break;
    }
  case CORBA::dk_Struct:
    {
      CORBA::StructDef_var s = CORBA::StructDef::_narrow( obj );
      emitHeaderStruct( s );
      break;
    }
  case CORBA::dk_Union:
    {
      CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( obj );
      emitHeaderUnion( u );
      break;
    }
  case CORBA::dk_Constant:
    {
      CORBA::ConstantDef_var co = CORBA::ConstantDef::_narrow( obj );
      emitHeaderConstant( co );
      break;
    }
  case CORBA::dk_Exception:
    {
      CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( obj );
      emitHeaderException( e );
      break;
    }
  case CORBA::dk_Enum:
    {
      CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( obj );
      emitHeaderEnum( e );
      break;
    }
  case CORBA::dk_Alias:
    {
      CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( obj );
      emitHeaderAlias( a );
      break;
    }
  case CORBA::dk_String:
    {
      CORBA::StringDef_var s = CORBA::StringDef::_narrow( obj );
      emitHeaderStringDef( s );
      break;
    }
  case CORBA::dk_Wstring:
    {
      // Not implemented
      assert( 0 );
    }
  case CORBA::dk_Sequence:
    {
      // Not implemented
      assert( 0 );
    }
  case CORBA::dk_Fixed:
    {
      // Not implemented
      assert( 0 );
    }
  case CORBA::dk_Array:
    {
      // Not implemented
      assert( 0 );
    }
  default:
    assert( 0 );
  }
}

void
CodeGenC::emitHeaderStringDef( CORBA::StringDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_String );

  CORBA::TypeCode_var type_tc = s->type();
	  
  // Must be a bounded string
  assert( type_tc->length() > 0 );
  o << "extern CORBA_TypeCode_ptr _tc_string_";
  o << type_tc->length();
  o << ";" << endl << endl;
}


/*
 * Resolves an alias (IDL-typedef) to its original type
 */
CORBA::TypeCode_ptr
CodeGenC::resolve_alias( CORBA::TypeCode_ptr tc )
{
  CORBA::TypeCode_var tc_content;
  tc_content = CORBA::TypeCode::_duplicate( tc );
  while( tc_content->kind() == CORBA::tk_alias )
    tc_content = tc_content->content_type();
  return CORBA::TypeCode::_duplicate( tc_content );
}

/*
 * Resolves the parameter type of a sequence to its original type
 */
CORBA::TypeCode_ptr
CodeGenC::resolve_sequence( CORBA::TypeCode_ptr tc_orig )
{
  CORBA::TypeCode_var tc = resolve_alias( tc_orig );
  
  if( tc->kind() != CORBA::tk_sequence )
    return CORBA::TypeCode::_duplicate( tc );
  
  CORBA::TypeCode_ptr tc_resolved;
  CORBA::TypeCode_var tc_content = tc->content_type();
  CORBA::TypeCode_var tc_resolve_seq = resolve_sequence( tc_content );
  tc_resolved =
      CORBA::TypeCode::create_sequence_tc( tc->length(),
					   tc_resolve_seq )->mk_constant();
  return tc_resolved;
}

/*
 * Returns true if TypeCode is a base type
 */
bool
CodeGenC::is_base_type( CORBA::TypeCode_ptr tc )
{
  CORBA::TypeCode_var tc_resolv = resolve_alias( tc );
  switch( tc_resolv->kind() ) {
  case CORBA::tk_boolean:
  case CORBA::tk_char:
  case CORBA::tk_wchar:
  case CORBA::tk_octet:
  case CORBA::tk_ushort:
  case CORBA::tk_ulong:
  case CORBA::tk_short:
  case CORBA::tk_long:
  case CORBA::tk_longlong:
  case CORBA::tk_ulonglong:
  case CORBA::tk_float:
  case CORBA::tk_double:
  case CORBA::tk_longdouble:
  case CORBA::tk_string:
  case CORBA::tk_wstring:
  case CORBA::tk_fixed:
    return true;
  default:
    break;
  }
  return false;
}

/*
 * Returns true if TypeCode is of IDL-type Object or an alias of an Object
 */
bool
CodeGenC::is_object( CORBA::TypeCode_ptr tc )
{
  CORBA::TypeCode_var tc_resolve = resolve_alias( tc );

  if( tc_resolve->kind() == CORBA::tk_objref ) {
    CORBA::String_var name = tc_resolve->name();
    if( strcmp( name, "Object" ) == 0 )
      return true;
  }
  return false;
}

/*
 * This method checks, if marshalling code needs to be generated for
 * the sequence or not (return == 'true' means no code needs to be
 * generated. This is done by checking if "seq" is contained in
 * "types". If not, it returns false and add "seq" to "types". If yes,
 * true is returned. The definition of 'contained' is governed by the
 * C++ type inference rule, whereupon: sequence<string> and typedef
 * string id; sequence<id> both refer to the same C++ type. There is
 * also no code generated for sequence<T> where T is a basic data type
 * for unbounded sequences. Those marshalling routines are already
 * part of libmico.a
 */
bool CodeGenC::check_for_sequence( TypeCodeSeq& types,
				     CORBA::TypeCode_ptr seq,
				     CORBA::Boolean no_base_types )
{
  assert( seq->kind() == CORBA::tk_sequence );

  CORBA::TypeCode_var tc_reduced = resolve_sequence( seq );

  if( no_base_types && seq->length() == 0 ) {
    CORBA::TypeCode_var tc_content = tc_reduced->content_type();
    if( is_base_type( tc_content ) )
      return true;
  }

  for( CORBA::ULong i = 0; i < types.length(); i++ ) {
    if( tc_reduced->equal( types[ i ] ) )
      return true;
  }
  int len = types.length();
  types.length( len + 1 );
  types[ len ] = CORBA::TypeCode::_duplicate( tc_reduced );
  return false;
}

/*
 * This method checks, if "arr" is contained in "types". If not,
 * it returns false and adds "arr" to "types". If yes, true is
 * returned.
 */
bool CodeGenC::check_for_array( TypeCodeSeq& types,
				  CORBA::TypeCode_ptr arr )
{
  assert( arr->kind() == CORBA::tk_array );

  for( CORBA::ULong i = 0; i < types.length(); i++ ) {
    if( arr->equal( types[ i ] ) )
      return true;
  }
  int len = types.length();
  types.length( len + 1 );
  types[ len ] = CORBA::TypeCode::_duplicate( arr );
  return false;
}


void CodeGenC::emitHeaderInterface( CORBA::InterfaceDef_ptr in )
{
  assert( in->def_kind() == CORBA::dk_Interface );
  CORBA::String_var n = in->name();
  string className = (const char *) ID(n);
  
  o << endl;

  //-----COMMON------------------------------------------------------
  o << "/* Common definitions for interface " << className << " */ " << endl;

#ifdef CPP_CODE  
  o << "class " << className << " : ";
 
  CORBA::InterfaceDefSeq_var bases = in->base_interfaces();
  if( bases->length() == 0 ) {
    // This interface has no base interfaces
    o << "virtual public CORBA::Object" << endl;
  } else {
    // This interface *has* base interfaces
    o << endl;
    o.enterLevel();
    for( int i = 0; i < bases->length(); i++ ) {
      o << "virtual public ";
      CORBA::String_var base_name = bases[ i ]->absolute_name();
      o << ID(base_name);
      if( i + 1 != bases->length() )
	o << ",";
      o << endl;
    }
    o.leaveLevel();
  }
#endif

#ifdef CPP_CODE  
  o << "{" << endl;
  o.enterLevel();
  
  o << "public:" << endl;
  o.enterLevel();
  o << "virtual ~" << className << "();" << endl;

  o << "static " << className << "_ptr _duplicate( ";
  o << className << "_ptr obj );" << endl;
  o << "static " << className << "_ptr _narrow( CORBA::Object_ptr obj );"
    << endl;
  o << "static " << className << "_ptr _nil();" << endl << endl;
  
  o << "static vector<CORBA::Narrow_proto> *_narrow_helpers;" << endl;
  o << "virtual void *_narrow_helper( const char *repoid );" << endl;
  o << "static bool _narrow_helper2( CORBA::Object_ptr obj );" << endl << endl;
#endif

  // Emit local declarations
  emitLocalDecls( in );
  
  // Generate prototypes
  emitPrototypes( in, true );

#ifdef CPP_CODE  
  // Generate epilogue
  o.leaveLevel();
  o << "protected:" << endl;
  o.enterLevel();
  o << className << "() {};" << endl;
  o.leaveLevel();
  o << "private:" << endl;
  o.enterLevel();
  o << className << "( const " << className << "& );" << endl;
  o << "void operator=( const " << className << "& );" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "};" << endl << endl;

#endif

  CORBA::String_var absname = get_absolute_name(in, "_");
  o << "extern " << absname << " " << absname << "_CreateTrue(void);" << endl;
  o << "extern " << absname << " server_" << absname << "_CreateTrue(" << absname << " _handle);" << endl;
  // Emit TypeCode for interface
  CORBA::Container_var def_in = in->defined_in();
  o << "extern CORBA_TypeCode_ptr _tc_" << absname << ";" << endl << endl;

#ifdef CPP_CODE
  // operator<<= for interface
  o.switchStream( 1 );
  o << "CORBA::Boolean operator<<=( CORBA::Any &a, const ";
  emit_idl_type_name( in );
  o << " obj );" << endl;
    
  // operator>>= for interface
  o << "CORBA::Boolean operator>>=( const CORBA::Any &a, ";
  emit_idl_type_name( in );
  o << " &obj );" << endl << endl;
#endif

  o.switchStream( 0 );

#ifdef CPP_CODE  
  //-----STUB--------------------------------------------------------
  o << "/* Stub for interface " << className << " */" << endl;

  o << "class " << className << _STUB << 
    " : virtual public " << className;

  if( bases->length() == 0 ) {
    // This interface has no base interfaces
    o << endl;
  } else {
    // This interface *has* base interfaces
    o << "," << endl;
    o.enterLevel();
    for( int i = 0; i < bases->length(); i++ ) {
      o << "virtual public ";
      CORBA::String_var base_name = bases[ i ]->absolute_name();
      o << ID(base_name);
      o << _STUB;
      if( i + 1 != bases->length() )
	o << ",";
      o << endl;
    }
    o.leaveLevel();
  }
  
  o << "{" << endl;
  o.enterLevel();
  
  o << "public:" << endl;
  o.enterLevel();
  o << "virtual ~" << className << _STUB << "();" << endl;

#endif
  
#ifdef CPP_CODE  
  // Generate epilogue
  o.leaveLevel();
  o << "private:" << endl;
  o.enterLevel();
  o << "void operator=( const " << className << _STUB << "& );" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "};" << endl << endl;
#endif

#ifdef CPP_CODE  
  //-----SKELETON----------------------------------------------------
  // Generate skeleton class for interface [18.3.1(6)], [18.4.5]
  o << "class " << className << _SKEL << " :" << endl;
  o.enterLevel();
  o << "virtual public MethodDispatcher," << endl;
  o << "virtual public " << className << endl;
  o.leaveLevel();
  o << "{" << endl;
  o.enterLevel();
  o << "public:" << endl;
  o.enterLevel();
  o << className << _SKEL;
  o << "( const CORBA::BOA::ReferenceData & = CORBA::BOA::ReferenceData() );";
  o << endl;
  o << "virtual ~" << className << _SKEL << "();" << endl;
  o << className << _SKEL << "( CORBA::Object_ptr obj );" << endl;
  o << "virtual bool dispatch( CORBA::ServerRequest_ptr _req, "
    "CORBA::Environment &_env );" << endl;
  o << className << "_ptr _this();" << endl << endl;
  
  o.leaveLevel();
  o.leaveLevel();
  o << "};" << endl << endl;
#endif
}

/*
 * Generates default constructor, copy constructor and assignment
 * operator for structs and unions. These have to be generated due
 * to a g++ bug (g++ doesn't expand templates properly for the implicit
 * generated versions of these methods). Note that according to
 * the CORBA2 specs these shouldn't be defined (see [16.9(1)])
 */
void
CodeGenC::emitHeaderDefaultMembers( char *name )
{
#ifdef CPP_CODE
  // Default constructor
  o << name << "();" << endl;

  // Destructor
  o << "~" << name << "();" << endl;

  // Copy constructor
  o << name << "( const " << name << "& s );" << endl;
  
  // Assignment operator
  o << name << "& operator=( const " << name << "& s );" << endl;
#endif
}
  
void CodeGenC::emitHeaderStruct( CORBA::StructDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_Struct );

  CORBA::String_var tmp = s->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::StructMemberSeq_var m = s->members();
  CORBA::TypeCode_var tc = s->type();

  if( !tc->is_freestanding() ) {
    // This struct contains somewhere nested a recursive sequence
    // which refers not to this struct. We therefore need to emit
    // forward declarations of those outer level structs because
    // the SequenceIndTmpl will make reference to them.
    for( CORBA::ULong i = 0; i < m->length(); i++ ) {
      if( m[ i ].type->is_recursive_seq() ) {
	o << "struct ";
	CORBA::SequenceDef_var seq =
	  CORBA::SequenceDef::_narrow( m[ i ].type_def );
	assert( !CORBA::is_nil( seq ) );
	CORBA::IDLType_var elem = seq->element_type_def();
	CORBA::Contained_var con = CORBA::Contained::_narrow( elem );
	assert( !CORBA::is_nil( con ) );
	tmp = con->name();
	o << ID(tmp) << ";" << endl;
      }
    }
  }
      
  o << "struct " << name << " {" << endl;
  o.enterLevel();
  emitHeaderDefaultMembers( name );
  for( CORBA::ULong i = 0; i < m->length(); i++ ) {
    emit_type_for_variable( m[ i ].type_def, automatic );
    o << " " << ID(m[ i ].name);
    emit_array_suffix( m[ i ].type_def );
    o << ";" << endl;
  }
  o.leaveLevel();
  o << "};" << endl << endl;

#ifdef CPP_CODE
  // Emit _var type
  o << "typedef TVar<" << name << "> " << name << "_var;";
  o << endl << endl;

#endif  
  // Emit TypeCode for struct
  if( tc->is_freestanding() ) {
    CORBA::String_var aname = get_absolute_name (s, "_");
    o << "extern CORBA_TypeCode_ptr _tc_" << aname << ";" << endl << endl;
  }

#ifdef CPP_CODE  
  // operator<<=
  o.switchStream( 1 );
  tmp = s->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs_name << " &_s );" << endl;
    
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs_name << " &_s );" << endl << endl;
  o.switchStream( 0 );
#endif
}

void CodeGenC::emitHeaderUnion( CORBA::UnionDef_ptr u )
{
  assert( u->def_kind() == CORBA::dk_Union );

  CORBA::String_var tmp = u->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::UnionMemberSeq_var members = u->members();
  CORBA::TypeCode_var union_tc = u->type();

#ifdef CPP_CODE  
  if( !union_tc->is_freestanding() ) {
    // This union contains somewhere nested a recursive sequence
    // which refers not to this union. We therefore need to emit
    // forward declarations of those outer level unions because
    // the SequenceIndTmpl will make reference to them.
    for( int i = 0; i < members->length(); i++ ) {
      if( members[ i ].type->is_recursive_seq() ) {
	o << "class ";
	CORBA::SequenceDef_var seq =
	  CORBA::SequenceDef::_narrow( members[ i ].type_def );
	assert( !CORBA::is_nil( seq ) );
	CORBA::IDLType_var elem = seq->element_type_def();
	CORBA::Contained_var con = CORBA::Contained::_narrow( elem );
	assert( !CORBA::is_nil( con ) );
	tmp = con->name();
	o << ID(tmp) << ";" << endl;
      }
    }
  }
#endif

  o << "typedef struct {" << endl;
  o.enterLevel();
#ifdef CPP_CODE
  o << "public:" << endl;
#endif
  o.enterLevel();

  CORBA::IDLType_var discr = u->discriminator_type_def();

  // Emit declaration for the discriminator
  emit_type_for_variable( discr, automatic );
  o << " _d;" << endl << endl;

  // Generate member variables. Note: these can't be put in a C++
  // union as one would probably expect. The reason is that MICO
  // struct's are mapped to C++ struct's with explicit default
  // constructors and destructors (this is necessary due to a
  // g++ bug). But struct's with constructors are not allowed
  // in an union.
  CORBA::String_var last_name;
  last_name = (const char *) "";
  o << "union {" << endl;
  o.enterLevel();

  bool has_explicit_default = false;
  
  for( CORBA::ULong i = 0; i < members->length(); i++ ) {
    CORBA::UnionMember& m = members[ i ];
    CORBA::TypeCode_var tc = members[ i ].label.type();
    if( is_octet( tc ) )
      has_explicit_default = true;

    if( strcmp( last_name, m.name ) == 0 )
      // We already generated a delclaration for this name
      continue;
    emit_type_for_variable( m.type_def, automatic );
    o << " " << ID(m.name);
    emit_array_suffix( m.type_def );
    o << ";" << endl;
    last_name = m.name;
  }
  o.leaveLevel();
  o << "} _u;" << endl << endl;
#ifdef CPP_CODE  
  o.leaveLevel();
  o << "public:" << endl;
  o. enterLevel();
  
  emitHeaderDefaultMembers( name );
  o << endl;
#endif

#ifdef CPP_CODE  
  // Emit discriminator functions
  o << "void _d( ";
  emit_type_for_param( discr, CORBA::PARAM_IN );
  o << " _p );" << endl;
  emit_type_for_result( discr );
  o << " _d() const;" << endl << endl;
#endif

#ifdef CPP_CODE
  // Generate member accessor functions
  last_name = (const char *) "";
  for( int i = 0; i < members->length(); i++ ) {
    CORBA::UnionMember &m = members[ i ];

    if( strcmp( last_name, m.name ) == 0 )
      continue;
    last_name = m.name;
    
    CORBA::IDLType_var m_type = CORBA::IDLType::_duplicate( m.type_def );
    CORBA::String_var m_name = (const char *) ID(m.name);
    CORBA::TypeCode_var tc = m_type->type();
    if( is_string( tc ) ) {
      // Accessor functions for string
      o << "void " << m_name << "( char *_p );" << endl;
      o << "void " << m_name << "( const char *_p );" << endl;
      o << "void " << m_name << "( const CORBA::String_var &_p );" << endl;
      o << "const char *" << m_name << "() const;" << endl << endl;
    } else if( is_array( tc ) && !is_alias( tc ) ) {
      // We have an anonymous array. See [16.10(7)]
      o << "typedef ";
      emit_array( m_type );
      o << " " << "_" << m_name << "_slice";
      emit_array_slice_suffix( m_type );
      o << ";" << endl;
      o << "void " << m_name << "( ";
      emit_array( m_type );
      o << " _p";
      emit_array_suffix( m_type );
      o << " );" << endl;
      o << "_" << m_name << "_slice* " << m_name << "();" << endl << endl;
    } else if( is_array( tc ) || is_objref( tc ) ||
	       is_type_code( tc ) || is_base_type( tc ) ) {
      // Accessor functions for array or object reference
      o << "void " << m_name << "( ";
      if( is_array( tc ) )
	// Mapping for arrays differs from standard PARAM_IN mapping!
	emit_idl_type_name( m_type );
      else
	emit_type_for_param( m_type, CORBA::PARAM_IN );
      o << " _p );" << endl;
      emit_type_for_result( m_type );
      o << " " << m_name << "();" << endl << endl;
    } else if( is_sequence( tc ) && !is_alias( tc ) ) {
      // An array, which is not referred to by an alias. Note
      // That this is a special case, as emit_type_for_param
      // does not handle this situation
      o << "void " << m_name << "( const ";
      emit_sequence( m_type );
      o << "& _p );" << endl;
      o << "const ";
      emit_sequence( m_type );
      o << "& " << m_name << "() const;" << endl;
      emit_sequence( m_type );
      o << "& " << m_name << "();" << endl;
    } else {
      // Gotta be struct, union or sequence
      o << "void " << m_name << "( ";
      emit_type_for_param( m_type, CORBA::PARAM_IN );
      o << " _p );" << endl;
      o << "const ";
      emit_idl_type_name( m_type );
      o << "& " << m_name << "() const;" << endl;
      emit_idl_type_name( m_type );
      o << "& " << m_name << "();" << endl << endl;
    }
  }
  if( !has_explicit_default ) {
    // This union has no explicit default case label
    string default_label =
      generate_union_default_case_label( discr, members );
    if( default_label != "" ) {
      // We have a case label left for an implicit default member
      o << "void _default();" << endl << endl;
    }
  }
#endif

  o.leaveLevel();
  o.leaveLevel();
  CORBA::String_var aname = get_absolute_name (u, "_");
  o << "} " << aname << ";" << endl << endl;

#ifdef CPP_CODE
  // Emit _var type
  o << "typedef TVar<" << name << "> " << name << "_var;";
  o << endl << endl;
#endif
  
  // Emit TypeCode for union
  if( union_tc->is_freestanding() ) {
    aname = get_absolute_name (u, "_");
    o << "extern CORBA_TypeCode_ptr _tc_" << aname << ";" << endl << endl;
  }

#ifdef CPP_CODE  
  // operator<<=
  o.switchStream( 1 );
  tmp = u->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs_name << " &_u );" << endl;
    
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs_name << " &_u );" << endl << endl;
  o.switchStream( 0 );
#endif
}

void CodeGenC::emitHeaderConstant( CORBA::ConstantDef_ptr c )
{
  assert( c->def_kind() == CORBA::dk_Constant );
  CORBA::Any* a = c->value();
  CORBA::String_var tmp = c->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::String_var aname = get_absolute_name (c, "_");
  o << "#define " << aname << " " << a << endl;
  delete a;
}

void CodeGenC::emitHeaderException( CORBA::ExceptionDef_ptr e )
{
  assert( e->def_kind() == CORBA::dk_Exception );

  CORBA::String_var tmp = e->name();
  CORBA::String_var name = (const char *) ID(tmp);

#ifdef CPP_CODE  
  o << "struct " << name << " : public CORBA::UserException {" << endl;
  o.enterLevel();
  CORBA::StructMemberSeq_var m = e->members();
  emitHeaderDefaultMembers( name );

  o << "void _throwit() const;" << endl;
  o << "const char *_repoid() const;" << endl;
  o << "void _encode( CORBA::DataEncoder &en ) const;" << endl;
  o << name << " *_clone() const;" << endl;
  o << "static " << name << " *_narrow( CORBA::Exception *ex );" << endl;
  
  for( int i = 0; i < m->length(); i++ ) {
    emit_type_for_variable( m[ i ].type_def, automatic );
    o << " " << ID(m[ i ].name);
    emit_array_suffix( m[ i ].type_def );
    o << ";" << endl;
  }
  o.leaveLevel();
  o << "};" << endl << endl;

  // Emit _var type
  o << "typedef ExceptVar<" << name << "> " << name << "_var;";
  o << endl << endl;
  
  // Emit TypeCode for struct
  CORBA::Container_var def_in = e->defined_in();
  CORBA::String_var aname = get_absolute_name (e, "_");
  o << "extern CORBA_TypeCode_ptr _tc_" << aname << ";" << endl << endl;

  // operator<<=
  o.switchStream( 1 );
  tmp = e->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "CORBA::Boolean operator<<=( CORBA::Any &a, const ";
  o << abs_name << " &e );" << endl;
    
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &a, ";
  o << abs_name << " &e );" << endl << endl;
  o.switchStream( 0 );
#endif
}

void CodeGenC::emitHeaderEnum( CORBA::EnumDef_ptr e )
{
  CORBA::String_var tstr, tstr2;
  char *tstr3;
  assert( e->def_kind() == CORBA::dk_Enum );

  tstr = get_absolute_name(e, "_");
  CORBA::String_var tmp = e->name();
  CORBA::String_var name = (const char *) ID(tmp);
  o << "enum " << tstr << " {" << endl;
  o.enterLevel();

  CORBA::EnumMemberSeq_var m = e->members();

  CORBA::String_var ename = e->name();
  tstr2 = CORBA::string_dup (tstr);
  tstr3 = (char *)tstr2 + strlen(tstr2) - strlen(ename);
  *tstr3 = '\0';
  
  for( CORBA::ULong i = 0; i < m->length(); i++ ) {
    if( i != 0 )
      o << "," << endl;
    o << tstr2 << ID(m[ i ]);
    if( i == 0 )
      o << " = 0";
  }
  o << endl;
  o.leaveLevel();
  o << "};" << endl << endl;

  // Emit TypeCode for enum
  o << "extern CORBA_TypeCode_ptr _tc_" << tstr << ";" << endl << endl;

#ifdef CPP_CODE
  // operator<<=
  o.switchStream( 1 );
  tmp = e->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "CORBA::Boolean operator<<=( CORBA::Any &a, ";
  o << abs_name << " &e );" << endl << endl;
    
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &a, ";
  o << abs_name << " &e );" << endl << endl;
  o.switchStream( 0 );
#endif
}

void CodeGenC::emitHeaderAlias( CORBA::AliasDef_ptr a )
{
  assert( a->def_kind() == CORBA::dk_Alias );
  
  CORBA::IDLType_var orig = a->original_type_def();
  assert( !CORBA::is_nil( orig ) );
  
  CORBA::TypeCode_var orig_tc = orig->type();

  CORBA::String_var tmp = a->name();
  CORBA::String_var a_absname = get_absolute_name(a, "_");
  CORBA::String_var name = (const char *) ID(tmp);

  if( is_string( orig_tc ) ) {
    o << "typedef char* " << a_absname << ";" << endl;
#ifdef CPP_CODE
    o << "typedef CORBA_String_var ";
    o << name << "_var;" << endl << endl;
#endif
  } else if( is_objref( orig_tc ) ) {
    CORBA::String_var abs_name;
    
    if( orig_tc->kind() == CORBA::tk_alias ) {
      // we have an alias of an alias
      CORBA::AliasDef_var alias = CORBA::AliasDef::_narrow( orig );
      assert( !CORBA::is_nil( alias ) );
      abs_name = alias->absolute_name();
    } else {
      // gotta be an interface
      assert( orig_tc->kind() == CORBA::tk_objref );
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( orig );
      if( !CORBA::is_nil( in ) ) {
	// Must be a regular object reference
	abs_name = get_absolute_name(in, "_");
      } else {
	// We have an "Object" object reference
	abs_name = (const char *) "CORBA_Object";
      }
    }
    abs_name = (const char *) ID(abs_name);

    o << "typedef " << abs_name << " "
      << a_absname << ";" << endl;
#ifdef CPP_CODE
    o << "typedef " << abs_name << "_ptr " << name << "_ptr;" << endl;
    o << "typedef " << abs_name << "Ref " << name << "Ref;" << endl;
    o << "typedef " << abs_name << "_var " << name << "_var;" << endl;
#endif
    o << endl;
  } else {
    o << "typedef ";
    emit_type_for_variable( orig, automatic );
    o << " " << a_absname;
    emit_array_suffix( orig );
    o << ";" << endl;
    
    if( is_array( orig_tc ) ) {
      // Array slice
      o << "typedef ";
      emit_type_for_variable( orig, automatic );
      o << " " << a_absname << "_slice";
      emit_array_slice_suffix( orig );
      o << ";" << endl;
      o << "extern " << a_absname << "_slice *"
	<< a_absname << "__alloc(void);" << endl;
#ifdef CPP_CODE      
      // _var for Array
      CORBA::ULong num_elem = compute_num_array_elem( orig_tc );
      o << "typedef ArrayVar<";
      emit_type_for_variable( orig, automatic );
      o << "," << name << "_slice," << name << ",";
      o << num_elem << "> ";
      o << name << "_var;" << endl;
      
      // _forany for Array
      o << "typedef ArrayForAny<";
      emit_type_for_variable( orig, automatic );
      o << "," << name << "_slice," << name << ",";
      o << num_elem << "> ";
      o << name << "_forany;" << endl;

      // T_alloc for array
      o << "DECLARE_ARRAY_ALLOC(" << name << ",";
      emit_type_for_variable( orig, automatic );
      o << "," << name << "_slice," << num_elem << ")" << endl;
      
      // T_dup for array
      o << "DECLARE_ARRAY_DUP(" << name << ",";
      emit_type_for_variable( orig, automatic );
      o << "," << name << "_slice," << num_elem << ")" << endl;
      
      // T_free for array
      o << "DECLARE_ARRAY_FREE(" << name << ",";
      emit_type_for_variable( orig, automatic );
      o << "," << name << "_slice," << num_elem << ")" << endl;

      o.switchStream( 1 );
      CORBA::String_var abs_name = a->absolute_name();
      abs_name = (const char *) ID(abs_name);
      
      // operator<<=()
      o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
      o << abs_name << "_forany &_s );" << endl;

      // operator>>=()

      o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
      o << abs_name << "_forany &_s );" << endl;
      o.switchStream( 0 );
#endif
    }

#ifdef CPP_CODE
    if( !is_base_type( orig_tc ) && !is_enum( orig_tc ) &&
	!is_array( orig_tc ) ) {
      // We have to generate a _var type as well
      o << "typedef ";
      if( is_sequence( orig_tc ) ) {
	o << "TSeqVar<";
	emit_type_for_variable( orig, automatic );
	o << " >";
      } else {
	emit_type_for_variable( orig, automatic );
	o << "_var";
      }
      o << " " << name << "_var;" << endl << endl;
    }
#endif
  }
  
  // Emit TypeCode for alias
  CORBA::Container_var def_in = a->defined_in();
  assert( !CORBA::is_nil( def_in ) );
  o << "extern CORBA_TypeCode_ptr _tc_" << a_absname << ";" << endl << endl;
}

void CodeGenC::emitClient()
{
  for( CORBA::ULong i = 0; i < _idl_objs.length(); i++ ) {

    CORBA::IRObject_var obj = _idl_objs[ i ];
    
    if( check_for_included_defn( obj ) )
      continue;
    
    switch( obj->def_kind() ) {
    case CORBA::dk_Interface:
      {
	CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( obj );
	emitClientInterface( in );
	break;
      }
    case CORBA::dk_Struct:
      {
	CORBA::StructDef_var s = CORBA::StructDef::_narrow( obj );
	emitClientStruct( s );
	break;
      }
    case CORBA::dk_Union:
      {
	CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( obj );
	emitClientUnion( u );
	break;
      }
    case CORBA::dk_Constant:
      // We don't need to do anything for constants
      break;
    case CORBA::dk_Exception:
      {
	CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( obj );
	emitClientException( e );
	break;
      }
    case CORBA::dk_Enum:
      {
	CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( obj );
	emitClientEnum( e );
	break;
      }
    case CORBA::dk_Alias:
      {
	CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( obj );
	emitClientAlias( a );
	break;
      }
    case CORBA::dk_String:
      {
	CORBA::StringDef_var s = CORBA::StringDef::_narrow( obj );
	emitClientStringDef( s );
	break;
      }
    case CORBA::dk_Wstring:
      {
	// Not implemented
	assert( 0 );
      }
    case CORBA::dk_Sequence:
      {
	// Not implemented
	assert( 0 );
      }
    case CORBA::dk_Fixed:
      {
	// Not implemented
	assert( 0 );
      }
    case CORBA::dk_Array:
      {
	// Not implemented
	assert( 0 );
      }
    default:
      assert( 0 );
    }
  }
}

void
CodeGenC::emitClientStringDef( CORBA::StringDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_String );

  CORBA::TypeCode_var type_tc = s->type();
	  
  // Must be a bounded string
  assert( type_tc->length() > 0 );
  o << "CORBA_TypeCode_ptr _tc_string_";
  o << type_tc->length() << " = ";
  emit_type_code( type_tc );
  o << ";" << endl << endl;
}


void CodeGenC::emit_array_marshalling( CORBA::ArrayDef_ptr arr,
				       bool is_from,
				       int level )
{
  const char *direction = is_from ? "put" : "get";
  
  o << "if( !_a.array_" << direction << "_begin() )" << endl;
  o.enterLevel();
  o << "return FALSE;" << endl;
  o.leaveLevel();
  o << "for( CORBA::ULong _i" << level << " = 0; _i" << level << " < "
    << arr->length() << "; _i" << level << "++ ) {" << endl;
  o.enterLevel();
  CORBA::IDLType_var elem = arr->element_type_def();
  if( elem->def_kind() == CORBA::dk_Array ) {
    CORBA::ArrayDef_var nested_arr = CORBA::ArrayDef::_narrow( elem );
    assert( !CORBA::is_nil( nested_arr ) );
    emit_array_marshalling( nested_arr, is_from, ++level );
  } else {
    o << "if( !(_a ";
    if( is_from )
      o << " <<= ";
    else
      o << " >>= ";
    string var;
    var = "_s";
    for( int i = 0; i <= level; i++ ) {
      var += "[ _i";
      char buf[ 20 ];
      sprintf( buf, "%d", i );
      var += buf;
      var += " ]";
    }
    emitAnyParam( elem, (char *) var.c_str(), is_from );
    o << ") )" << endl;
    o.enterLevel();
    o << "return FALSE;" << endl;
    o.leaveLevel();
  }
  o.leaveLevel();
  o << "}" << endl;
  
  o << "if( !_a.array_" << direction << "_end() )" << endl;
  o.enterLevel();
  o << "return FALSE;" << endl;
  o.leaveLevel();
}

void CodeGenC::emit_stub_method( string &absClassName,
				 string &dispatch_name,
				 CORBA::OperationDescription &op,
				 CORBA::IDLType_ptr result )
{
  emit_type_for_result( result );
  o << " " << absClassName << "_" << ID(op.name) << "(";
  o << "CORBA_Object _handle, ";
  CORBA::ParDescriptionSeq &p = op.parameters;
  CORBA::ULong k;
  for( k = 0; k != p.length(); k++ ) {
    emit_type_for_param( p[ k ].type_def, p[ k ].mode );
    o << " " << ID(p[ k ].name) << ", ";
  }

  // Do we have a context?
  CORBA::ContextIdSeq& ctx = op.contexts;
    
  if( ctx.length() > 0 ) {
    o << "CORBA_Context_ptr _ctx";
    o << ", ";
  }
    
  o << " CORBA_Environment *ev)" << endl;
  o << "{" << endl;
  o.enterLevel();

  // Generate wrapper variables for all array parameters
  // but not for variable out arrays
  for( k = 0; k != p.length(); k++ ) {
    if( is_array( p[ k ].type ) ) {
      if( !is_variable( p[ k ].type ) || p[ k ].mode != CORBA::PARAM_OUT ) {
	emit_type_for_variable( p[ k ].type_def, automatic );
	o << "_forany _forany_" << ID(p[ k ].name);
	o << "( ";
	if( p[ k ].mode == CORBA::PARAM_IN ) {
	  o << "(";
	  emit_idl_type_name( p[ k ].type_def );
	  o << ") ";
	}
	o << ID(p[ k ].name) << " );" << endl;
      }
    }
  }
  
  // Generate marshalling code
  o << "CORBA::Request_var _req = _handle->_request( \"" << dispatch_name <<
    "\" );" << endl;

  // Marshall context
  if( ctx.length() > 0 ) {
    o << "_req->ctx( _ctx );" << endl;
    for( k = 0; k < ctx.length(); k++ )
      o << "_req->contexts()->add( \"" << ctx[ k ] << "\" );" << endl;
  }

  // Marshall actual parameters
  for( k = 0; k < p.length(); k++ ) {
    o << "_req->add_";
    switch( p[ k ].mode ) {
    case CORBA::PARAM_IN:
      o << "in";
      break;
    case CORBA::PARAM_OUT:
      o << "out";
      break;
    case CORBA::PARAM_INOUT:
      o << "inout";
      break;
    default:
      assert( 0 );
    }
    o << "_arg( \"" << ID(p[ k ].name) << "\" )";
    switch( p[ k ].mode ) {
    case CORBA::PARAM_IN:
    case CORBA::PARAM_INOUT:
      o << " <<= ";
      if( is_string( p[ k ].type ) ) {
	o << "CORBA::Any::from_string( (char *) ";
	if(p[k].mode != CORBA::PARAM_IN)
	  o << "*";
	o << ID(p[ k ].name) << ", ";
	CORBA::TypeCode_var tc = resolve_alias( p[ k ].type );
	o << tc->length() << " )";
      } else if( is_array( p[ k ].type ) ) {
	o << "_forany_" << ID(p[ k ].name);
      } else {
	emitAnyParam( p[ k ].type_def, p[ k ].name, true );
      }
      o << ";" << endl;
      break;
    case CORBA::PARAM_OUT:
      // Write out reference to TypeCode
      o << ";" << endl;
      o << "_req->arguments()->item( " << k;
      o << " )->value()->type( ";
      emit_type_code_name( p[ k ].type_def );
      o << " );" << endl;
    }
  }
  o << "_req->result()->value()->type( ";
  emit_type_code_name( result );
  o << " );" << endl;

  if( op.mode == CORBA::OP_NORMAL )
    o << "_req->invoke();" << endl;
  else
    o << "_req->send_oneway();" << endl;

  if( _gen_exceptions )
    o << "#ifdef HAVE_EXCEPTIONS" << endl;
    
  if( _gen_exceptions && op.mode == CORBA::OP_NORMAL ) {
    o << "if( _req->env()->exception() ) {" << endl;
    o.enterLevel();
    o << "CORBA::Exception *_ex = _req->env()->exception();" << endl;
    o << "CORBA::UnknownUserException *_uuex = ";
    o << "CORBA::UnknownUserException::_narrow( _ex );" << endl;
    o << "if( _uuex ) {" << endl;
    o.enterLevel();
    
    CORBA::ExcDescriptionSeq exs = op.exceptions;
    for( CORBA::ULong i = 0; i < exs.length(); i++ ) {
      CORBA::Contained_var con = _repo->lookup_id( exs[ i ].id );
      assert( !CORBA::is_nil( con ) );
      CORBA::String_var tmp = con->absolute_name();
      CORBA::String_var ex_abs_name = (const char *) ID(tmp);
      string ex_scope;
      ex_scope = ex_abs_name;
      int idx = ex_scope.find_last_of( ":" );
      assert( idx > 0 );
      ex_scope = ex_scope.substr( 0, idx - 1 );
      if( ex_scope == "::" )
	ex_scope = "";

      o << "if( !strcmp( _uuex->_except_repoid(), \"";
      o << exs[ i ].id;
      o << "\" ) ) {" << endl;
      o.enterLevel();
      o << "CORBA::Any &_a = _uuex->exception( " << ex_scope << "::_tc_";
      o << exs[ i ].name;
      o << " );" << endl;
      o << ex_abs_name << " _user_ex;" << endl;
      o << "_a >>= _user_ex;" << endl;
      o << "mico_throw( _user_ex );" << endl;
      o.leaveLevel();
      o << "}" << endl;
    }
    
    o << "mico_throw( CORBA::UNKNOWN() );" << endl;
    o.leaveLevel();
    o << "} else {" << endl;
    o.enterLevel();
    o << "mico_throw( *_ex );" << endl;
    o.leaveLevel();
    o << "}" << endl;
    o.leaveLevel();
    o << "}" << endl;
  } else if( _gen_exceptions && op.mode == CORBA::OP_ONEWAY ) {
    o << "if( CORBA::Exception *_ex = _req->env()->exception() )" << endl;
    o.enterLevel();
    o << "mico_throw( *_ex );" << endl;
    o.leaveLevel();
  } else {
    o << "if( CORBA::Exception *_ex = _req->env()->exception() )" << endl;
    o.enterLevel();
    o << "CORBA::Exception::_throw_failed( _ex );" << endl;
    o.leaveLevel();
  }
  
  if( _gen_exceptions ) {
    o << "#else" << endl;
    o << "if( CORBA::Exception *_ex = _req->env()->exception() )" << endl;
    o.enterLevel();
    o << "CORBA::Exception::_throw_failed( _ex );" << endl;
    o.leaveLevel();
    o << "#endif" << endl;
  }
  
  // Write back OUT and INOUT parameters
  for( k = 0; k < p.length(); k++ ) {
    switch( p[ k ].mode ) {
    case CORBA::PARAM_IN:
      // Nothing to do
      break;
    case CORBA::PARAM_OUT:
    case CORBA::PARAM_INOUT:
      // We need to allocate some memory to store the result
      if( is_variable( p[ k ].type ) && is_array( p[ k ].type ) &&
	  p[ k ].mode == CORBA::PARAM_OUT ) {
	o << ID(p[ k ].name) << " = ";
	emit_idl_type_name( p[ k ].type_def );
	o << "_alloc();" << endl;
	emit_type_for_variable( p[ k ].type_def, automatic );
	o << "_forany _forany_" << ID(p[ k ].name);
	o << "( " << ID(p[ k ].name) << " );" << endl;
      }
      
      if( is_variable( p[ k ].type ) && !is_string( p[ k ].type ) &&
	  !is_wstring( p[ k ].type ) && !is_objref( p[ k ].type ) &&
	  !is_array( p[ k ].type ) && p[ k ].mode == CORBA::PARAM_OUT ) {
	o << ID(p[ k ].name) << " = new ";
	if( is_any( p[ k ].type ) )
	  o << "CORBA::Any";
	else {
	  CORBA::Contained_var con =
	    CORBA::Contained::_narrow( p[ k ].type_def );
	  assert( !CORBA::is_nil( con ) );
	  CORBA::String_var n = con->absolute_name();
	  o << ID(n);
	}
	o << ";" << endl;
      }
      
      o << "*_req->arguments()->item( " << k << " )->value() >>= ";
      if( is_string( p[ k ].type ) ) {
	o << "CORBA::Any::to_string( ";
	if(p[k].mode != CORBA::PARAM_IN)
	  o << "*";
	o << ID(p[ k ].name) << ", ";
	CORBA::TypeCode_var tc = resolve_alias( p[ k ].type );
	o << tc->length() << " )";
      } else if( is_array( p[ k ].type ) ) {
	o << "_forany_" << ID(p[ k ].name );
      } else {
	string n;

	if( !is_objref( p[ k ].type ) /* && is_variable( p[ k ].type ) */
	    && p[ k ].mode != CORBA::PARAM_IN )
	  n = "*";
	n += ID(p[ k ].name);
	emitAnyParam( p[ k ].type_def, (char *) n.c_str(), false );
      }
      o << ";" << endl;
      break;
    }
  }
  
  // Write back result (if there is one)
  CORBA::TypeCode_var tc_result = result->type();
  if( tc_result->kind() != CORBA::tk_void ) {
    if( is_array( tc_result ) ) {
      emit_type_for_variable( result, automatic );
      o << "_slice* _res = ";
      emit_idl_type_name( result );
      o << "_alloc();" << endl;
      emit_type_for_variable( result, automatic );
      o << "_forany _res_forany( _res );" << endl;
    } else {
      emit_type_for_variable( result, manual );
      o << " _res";
      if( !is_objref( tc_result ) && !is_string( tc_result ) &&
	  !is_wstring( tc_result ) && is_variable( tc_result ) ) {
	o << " = new ";
	if( is_any( tc_result ) )
	  o << "CORBA::Any";
	else {
	  CORBA::Contained_var con = CORBA::Contained::_narrow( result );
	  assert( !CORBA::is_nil( con ) );
	  CORBA::String_var n = con->absolute_name();
	  o << ID(n);
	}
      }
      o << ";" << endl;
    }
    
    o << "*_req->result()->value() >>= ";
    
    if( is_array( tc_result ) ) {
      o << "_res_forany";
    } else {
      string res;
    
      if( !is_objref( tc_result ) && !is_string( tc_result ) &&
	  !is_wstring( tc_result ) && is_variable( tc_result ) )
	res = "*";
      res += "_res";
      emitAnyParam( result, (char *) res.c_str(), false );
    }
    o << ";" << endl;
    o << "return _res;" << endl;
  }
  o.leaveLevel();
  o << "}" << endl << endl << endl;
}


void CodeGenC::emitClientInterface( CORBA::InterfaceDef_ptr in )
{
  assert( in->def_kind() == CORBA::dk_Interface );

  CORBA::ContainedSeq_var c;
  c = in->contents( CORBA::dk_all, 1 );
  for( CORBA::ULong i = 0; i < c->length(); i++ ) {

    if( check_for_included_defn( c[ i ] ) )
      continue;
    
    switch( c[ i ]->def_kind() ) {
    case CORBA::dk_Struct:
      {
	CORBA::StructDef_var s = CORBA::StructDef::_narrow( c[ i ] );
	emitClientStruct( s );
	break;
      }
    case CORBA::dk_Union:
      {
	CORBA::UnionDef_var u = CORBA::UnionDef::_narrow( c[ i ] );
	emitClientUnion( u );
	break;
      }
    case CORBA::dk_Exception:
      {
	CORBA::ExceptionDef_var e = CORBA::ExceptionDef::_narrow( c[ i ] );
	emitClientException( e );
	break;
      }
    case CORBA::dk_Enum:
      {
	CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( c[ i ] );
	emitClientEnum( e );
	break;
      }
    case CORBA::dk_Operation:
    case CORBA::dk_Attribute:
    case CORBA::dk_Constant:
      break;
    case CORBA::dk_Alias:
      {
	CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( c[ i ] );
	emitClientAlias( a );
	break;
      }
    default:
      assert( 0 );
    }
  }

  CORBA::String_var id = in->id();

  CORBA::String_var tmp;
  tmp = in->name();
  string className;
  className = ID(tmp);
  tmp = in->absolute_name();
  string absClassName;
  absClassName = ID(tmp);
  absClassName = absClassName.substr( 2 );
  
  o << endl;
#ifdef CPP_CODE
  o << "// Stub interface " << className << endl;

  // dtor
  o << absClassName << "::~" << className << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;

  // _duplicate class method
  o << absClassName << "_ptr " << absClassName;
  o << "::_duplicate( " << className << "_ptr _obj )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "if( !CORBA::is_nil( _obj ) )" << endl;
  o.enterLevel();
  o << "_obj->_ref();" << endl;
  o.leaveLevel();
  o << "return _obj;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;

  /*
   * _narrow_helper virtual method:
   *
   * searched upwards in the inheritance hierarchy starting at the
   * real class of the object that is to be narrowed until it finds
   * the class we want to narrow to or until top is reached.
   *
   * this is used to do a downcast from a virtual base class.
   */
  o << "void *" << absClassName;
  o << "::_narrow_helper( const char *_repoid )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "if( strcmp( _repoid, \"" << id << "\" ) == 0 )" << endl;
  o.enterLevel();
  o << "return (void *)this;" << endl;
  o.leaveLevel();
  CORBA::InterfaceDefSeq_var bases = in->base_interfaces();
  if( bases->length() != 0 ) {
    // We have base interfaces
    for( int i = 0; i < bases->length(); i++ ) {
      tmp = bases[ i ]->absolute_name();
      string base_name;
      base_name = ID(tmp);
      base_name = base_name.substr( 2 );
      o << "if( void *_p = "
	<< base_name << "::_narrow_helper( _repoid ))" << endl;
      o.enterLevel();
      o << "return _p;" << endl;
      o.leaveLevel();
    }
  }
  o << "return NULL;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;

  /*
   * _narrow_helper2 class method
   *
   * searches downwards in the inheritance hierarchy starting at the
   * class we want to narrow to until it finds the class of the object
   * we want to narrow or until bottom is reached. Returns true
   * if _narrow succeeds
   */
  o << "bool " << absClassName;
  o << "::_narrow_helper2( CORBA::Object_ptr _obj )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "if( strcmp( _obj->_repoid(), \"" << id << "\" ) == 0) {" << endl;
  o.enterLevel();
  o << "return true;" << endl;
  o.leaveLevel();
  o << "}" << endl;
  o << "for( int _i = 0; _narrow_helpers && _i < _narrow_helpers->size();";
  o << " _i++ ) {" << endl;
  o.enterLevel();
  o << "bool _res = (*(*_narrow_helpers)[ _i ])( _obj );" << endl;
  o << "if( _res )" << endl;
  o.enterLevel();
  o << "return true;" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "}" << endl;
  o << "return false;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;

  /*
   * _narrow class method
   *
   * first see if we can just do a downcast by using _narrow_helper.
   * if that fails generate a stub for it using the _narrow_helper2
   * method. if that fails too call _obj->_is_a (_repoid) and create a
   * stub if _is_a() returns true. otherwise return _nil().
   */
  o << absClassName << "_ptr " << absClassName;
  o << "::_narrow( CORBA::Object_ptr _obj )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << absClassName << "_ptr _o;" << endl;
  o << "if( !CORBA::is_nil( _obj ) ) {" << endl;
  o.enterLevel();
  o << "if( void *_p = _obj->_narrow_helper( \"" << id << "\" ))" << endl;
  o.enterLevel();
  o << "return _duplicate( (" << absClassName << "_ptr) _p );" << endl;
  o.leaveLevel();
  o << "if( _narrow_helper2( _obj ) ) {" << endl;
  o.enterLevel();
  o << "_o = new " << absClassName;
  o << _STUB << ";" << endl;
  o << "_o->CORBA::Object::operator=( *_obj );" << endl;
  o << "return _o;" << endl;
  o.leaveLevel();
  o << "}" << endl;
  o.leaveLevel();
  o << "}" << endl;
  if( _query_ir_for_narrow ) {
    o << "if( _obj->is_a_remote( \"" << id << "\" ) ) {" << endl;
    o.enterLevel();
    o << "_o = new " << absClassName << _STUB << ";" << endl;
    o << "_o->CORBA::Object::operator=( *_obj );" << endl;
    o << "return _o;" << endl;
    o.leaveLevel();
    o << "}" << endl;
  }
  o << "return _nil();" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;

  // _nil class method
  o << absClassName << "_ptr " << absClassName;
  o << "::_nil()" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "return NULL;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
#endif

  // Generate attributes
  CORBA::InterfaceDef::FullInterfaceDescription_var desc;
  desc = in->describe_interface();
  string buf;
  for( CORBA::ULong j = 0; j < desc->attributes.length(); j++ ) {
    CORBA::AttributeDescription attr;
    attr = desc->attributes[ j ];

    CORBA::OperationDescription op;
    buf = string("_get_") + (const char *)attr.name;
    op.name = CORBA::string_dup (buf.c_str());
    op.id = (const char *) "";  // Not needed
    op.defined_in = attr.defined_in;
    op.version = (const char*) attr.version;
    op.result = CORBA::TypeCode::_nil();  // Not needed
    op.mode = CORBA::OP_NORMAL;
    // op.contexts does not need to be set for attributes
    op.parameters.length( 0 );  // No parameters for _get
    // op.exceptions does not need to be set for attributes

    string dispatch_name = string("_get_") + (const char *)attr.name;
    
    CORBA::IDLType_var idl_type = lookup_attribute_by_id( attr.id );
    emit_stub_method( absClassName, dispatch_name, op, idl_type );

    if( attr.mode == CORBA::ATTR_NORMAL ) {
      // We also need the set version of the attribute
      CORBA::ParameterDescription attr_param;
      attr_param.name = (const char *) "_value";
      attr_param.type = attr.type;
      attr_param.type_def = lookup_attribute_by_id( attr.id );
      attr_param.mode = CORBA::PARAM_IN;

      buf = string("_set_") + (const char *)attr.name;
      op.name = CORBA::string_dup (buf.c_str());
      op.id = (const char *) "";  // Not needed
      op.defined_in = attr.defined_in;
      op.version = (const char*) attr.version;
      op.result = CORBA::TypeCode::_nil();  // Not needed
      op.mode = CORBA::OP_NORMAL;
      // op.contexts does not need to be set for attributes
      op.parameters.length( 1 );
      op.parameters[ 0 ] = attr_param;
      // op.exceptions does not need to be set for attributes
      
      dispatch_name = string("_set_") + (const char *)attr.name;
    
      idl_type = _repo->get_primitive( CORBA::pk_void );
      emit_stub_method( absClassName, dispatch_name, op, idl_type );
    }
  }

#ifdef CPP_CODE
  // dtor
  o << absClassName << _STUB << "::~" << className << _STUB << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;
#endif

  // Generate methods
  for( CORBA::ULong j0 = 0; j0 < desc->operations.length(); j0++ ) {
    CORBA::OperationDescription op;
    op = desc->operations[ j0 ];
    string dispatch_name = (const char *) op.name;

    CORBA::IDLType_var idl_type = lookup_result_by_id( op.id );
    emit_stub_method( absClassName, dispatch_name, op, idl_type );
    
  }

  // Get base for absolute name
  tmp = get_absolute_name(in, "_");
  string abs;
  abs = ID(tmp);
#ifdef XXX_DUNNO
  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );
#endif

#ifdef CPP_CODE
  // set all _narrow_helper2 methods in their respective
  // base classes
  if( bases->length() != 0 ) {
    o << "struct _global_init_";
    emit_mangled_name( (char *) absClassName.c_str() );
    o << " {" << endl;
    o.enterLevel();
    o << "_global_init_";
    emit_mangled_name( (char *) absClassName.c_str() );
    o << "()" << endl;
    o << "{" << endl;
    o.enterLevel();
    for( int i = 0; i < bases->length(); i++ ) {
      o << "if( ";
      CORBA::String_var abs = bases[ i ]->absolute_name();
      o << abs << "::_narrow_helpers == NULL )" << endl;
      o.enterLevel();
      o << abs << "::_narrow_helpers = new vector<CORBA::Narrow_proto>;" << endl;
      o.leaveLevel();
      o << abs << "::_narrow_helpers->push_back( ";
      o << absClassName << "::_narrow_helper2 );" << endl;
    }
    o.leaveLevel();
    o << "}" << endl;
    o.leaveLevel();
    o << "} __global_init_";
    emit_mangled_name( (char *) absClassName.c_str() );
    o << ";" << endl << endl;
  }
    
  // Make instance for static member variable _narrow_helpers
  o << "vector<CORBA::Narrow_proto> *" << absClassName
    << "::_narrow_helpers;" << endl << endl;
#endif
  
  // Emit TypeCode for interface
  tmp = in->name();
  CORBA::String_var name = (const char *) ID(tmp);
  o << "CORBA_TypeCode_ptr " << "_tc_" << abs;
  o << " = ";
  CORBA::TypeCode_var tc_in = in->type();
  emit_type_code( tc_in );
  o << ";" << endl << endl;

#ifdef CPP_CODE
  // operator <<=
  o << "CORBA::Boolean" << endl;
  o << "operator<<=( CORBA::Any &_a, const ";
  emit_idl_type_name( in );
  o << " _obj )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "return (_a <<= CORBA::Any::from_object( _obj, \"";
  o << name << "\" ));" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
  
  // operator >>=
  o << "CORBA::Boolean" << endl;
  o << "operator>>=( const CORBA::Any &_a, ";
  emit_idl_type_name( in );
  o << " &_obj )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "CORBA::Object_ptr _o;" << endl;
  o << "if( !(_a >>= CORBA::Any::to_object( _o )) )" << endl;
  o.enterLevel();
  o << "return FALSE;" << endl;
  o.leaveLevel();

  o << "if( CORBA::is_nil( _o ) ) {" << endl;
  o.enterLevel();
  tmp = in->absolute_name();
  CORBA::String_var abs_name = (const char *) ID(tmp);
  o << "_obj = " << abs_name;
  o << "::_nil();" << endl;
  o << "return TRUE;" << endl;
  o.leaveLevel();
  o << "}" << endl;
  o << "_obj = " << abs_name << "::_narrow( _o );" << endl;
  o << "CORBA::release( _o );" << endl;
  o << "return TRUE;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
#endif
}

void CodeGenC::emit_type_code( CORBA::TypeCode_ptr tc )
{
  o << "(new CORBA::TypeCode(" << endl;
  o.enterLevel();

  string str = tc->stringify();
  
  bool beginning_of_line = true;
  
  for( string::size_type i = 0; i < str.length(); i++ ) {
    if( beginning_of_line ) {
      o << "\"";
      beginning_of_line = false;
    }
    string c;
    c += str[ i ];
    o << c;
    if( (i+1) % 60 == 0 ) {
      o << "\"" << endl;
      beginning_of_line = true;
    }
  }
  if( !beginning_of_line )
    o << "\"";
  o << " ))->mk_constant()";
  o.leaveLevel();
}

void CodeGenC::emit_type_code_name( CORBA::IDLType_ptr t )
{
  CORBA::PrimitiveDef_var prim = CORBA::PrimitiveDef::_narrow( t );
  if( !CORBA::is_nil( prim ) && prim->kind() == CORBA::pk_objref ) {
    o << "CORBA::_tc_Object";
    return;
  }
  
  CORBA::TypeCode_var tc = t->type();
  switch( tc->kind() ) {
  case CORBA::tk_void:
    o << "CORBA::_tc_void";
    break;
  case CORBA::tk_short:
    o << "CORBA::_tc_short";
    break;
  case CORBA::tk_long:
    o << "CORBA::_tc_long";
    break;
  case CORBA::tk_ushort:
    o << "CORBA::_tc_ushort";
    break;
  case CORBA::tk_ulong:
    o << "CORBA::_tc_ulong";
    break;
  case CORBA::tk_float:
    o << "CORBA::_tc_float";
    break;
  case CORBA::tk_double:
    o << "CORBA::_tc_double";
    break;
  case CORBA::tk_boolean:
    o << "CORBA::_tc_boolean";
    break;
  case CORBA::tk_char:
    o << "CORBA::_tc_char";
    break;
  case CORBA::tk_octet:
    o << "CORBA::_tc_octet";
    break;
  case CORBA::tk_any:
    o << "CORBA::_tc_any";
    break;
  case CORBA::tk_TypeCode:
    o << "CORBA::_tc_TypeCode";
    break;
  case CORBA::tk_Principal:
    o << "CORBA::_tc_Principal";
    break;
  case CORBA::tk_objref:
  case CORBA::tk_struct:
  case CORBA::tk_union:
  case CORBA::tk_enum:
  case CORBA::tk_alias:
    {
      string abs_name;
      CORBA::String_var tmp;
      switch( t->def_kind() ) {
      case CORBA::dk_Interface:
	{
	  CORBA::InterfaceDef_ptr in =
	    CORBA::InterfaceDef::_narrow( t );
	  tmp = in->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      case CORBA::dk_Struct:
	{
	  CORBA::StructDef_ptr s = 
	    CORBA::StructDef::_narrow( t );
	  tmp = s->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      case CORBA::dk_Union:
	{
	  CORBA::UnionDef_ptr u = 
	    CORBA::UnionDef::_narrow( t );
	  tmp = u->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      case CORBA::dk_Enum:
	{
	  CORBA::EnumDef_ptr e =
	    CORBA::EnumDef::_narrow( t );
	  tmp = e->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      case CORBA::dk_Alias:
	{
	  CORBA::AliasDef_ptr a =
	    CORBA::AliasDef::_narrow( t );
	  tmp = a->absolute_name();
	  abs_name = ID(tmp);
	  break;
	}
      default:
	assert( 0 );
      }
      int i = abs_name.rfind( "::" );
      assert( i >= 0 );
      abs_name.insert( i + 2, "_tc_" );
      abs_name = abs_name.substr( 2 );
      o << abs_name;
      break;
    }
  case CORBA::tk_string:
    if( tc->length() == 0 )
      o << "CORBA::_tc_string";
    else
      o << "_tc_string_" << tc->length();
    break;
  case CORBA::tk_sequence:
    // Sequences are anonymous types and don't have a _tc_* name
    // They are only referred to indirectly (through aliases e.g.)
    assert( 0 );
  case CORBA::tk_array:
    assert( 0 );
  default:
    assert( 0 );
  }
}

/*
 * Generates default constructor, copy constructor and assignment
 * operator for structs. These have to be generated due to a g++
 * bug (g++ doesn't expand templates properly for the implicit
 * generated versions of these methods). Note that according to
 * the CORBA2 specs these shouldn't be defined (see [16.9(1)])
 */
void
CodeGenC::emitClientStructDefaultMembers( char *name,
					  char *abs,
					  CORBA::StructMemberSeq_var &m )
{
  o << name << "* " << name << "__alloc(void)" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "return (" << name << " *) malloc(sizeof(" << name << "));" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;


#ifdef CPP_CODE
  // Default constructor
  o << abs << name << "::" << name << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;

  // Copy constructor
  o << abs << name << "::" << name << "( const " << name << "& _s )" << endl;
  o << "{" << endl;
  o.enterLevel();
  for( int i = 0; i < m->length(); i++ ) {
    o << ID(m[ i ].name) << " = ((" << name << "&)_s).";
    o << ID(m[ i ].name) << ";" << endl;
  }
  o.leaveLevel();
  o << "}" << endl << endl;

  // Destructor (moving this out of .h file reduces template drag in a lot)
  o << abs << name << "::~" << name << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;
  
  // Assignment operator
  o << abs << name << "&" << endl;
  o << abs << name << "::operator=( const " << name << "& _s )" << endl;
  o << "{" << endl;
  o.enterLevel();
  for( int i = 0; i < m->length(); i++ ) {
    o << ID(m[ i ].name) << " = ((" << name << "&)_s).";
    o << ID(m[ i ].name) << ";" << endl;
  }
  o << "return *this;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
#endif
}

void CodeGenC::emitClientStruct( CORBA::StructDef_ptr s )
{
  assert( s->def_kind() == CORBA::dk_Struct );

  // Get base for absolute name
  CORBA::String_var tmp = get_absolute_name(s, "_");
  string abs;
  abs = ID(tmp);
  tmp = s->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::TypeCode_var tc = s->type();

  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );

  // Emit TypeCode for struct
  if( tc->is_freestanding() ) {
    o << "CORBA_TypeCode_ptr " << "_tc_" << abs;
    o << " = ";
    emit_type_code( tc );
    o << ";" << endl << endl;
  }
  
  // Emit default ctor, copy ctor, assignment op
  CORBA::StructMemberSeq_var m = s->members();
  emitClientStructDefaultMembers (name, (char *)abs.c_str(), m);


#ifdef CPP_CODE
  // operator<<=
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs << name << " &_s )" << endl;
  o << "{" << endl;
  o.enterLevel();
  if( tc->is_freestanding() )
    o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
  o << "return (_a.struct_put_begin() &&" << endl;
  o.enterLevel();
  CORBA::StructMemberSeq_var members = s->members();
  for( int i = 0; i < members->length(); i++ ) {
    o << "(_a <<= ";
    string v = "((";
    v += abs;
    v += name;
    v += "&)_s).";
    v += ID(members[ i ].name);
    emitAnyParam( members[ i ].type_def, (char *) v.c_str(), true );
    o << ") &&" << endl;
  }
  o << "_a.struct_put_end() );" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "}" << endl << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs << name << " &_s )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "return (_a.struct_get_begin() &&" << endl;
  o.enterLevel();
  for( int i = 0; i < members->length(); i++ ) {
    o << "(_a >>= ";
    string v = "_s.";
    v += ID(members[ i ].name);
    emitAnyParam( members[ i ].type_def, (char *) v.c_str(), false );
    o << ") &&" << endl;
  }
  o << "_a.struct_get_end() );" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "}" << endl << endl;
#endif
}

/*
 * Generates default constructor, copy constructor and assignment
 * operator for unions. These have to be generated due to a g++
 * bug (g++ doesn't expand templates properly for the implicit
 * generated versions of these methods). Note that according to
 * the CORBA2 specs these shouldn't be defined (see [16.9(1)])
 */
void
CodeGenC::emitClientUnionDefaultMembers( char *name,
					 char *abs,
					 CORBA::UnionMemberSeq_var &m )
{
#ifdef CPP_CODE
  // Default constructor
  o << abs << name << "::" << name << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;

  // Copy constructor
  CORBA::String_var last_name;
  last_name = (const char *) "";
  o << abs << name << "::" << name << "( const " << name << "&_u )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "_discriminator = _u._discriminator;" << endl;
  for( int i = 0; i < m->length(); i++ ) {
    if( strcmp( last_name, m[ i ].name ) == 0 )
      continue;
    last_name = m[ i ].name;
    o << "_m." << ID(m[ i ].name) << " = ((" << name << "&)_u)._m.";
    o << ID(m[ i ].name) << ";" << endl;
  }
  o.leaveLevel();
  o << "}" << endl << endl;

  // Destructor (moving this out of .h file reduces template drag in a lot)
  o << abs << name << "::~" << name << "()" << endl;
  o << "{" << endl;
  o << "}" << endl << endl;
  
  // Assignment operator
  o << abs << name << "&" << endl;
  o << abs << name << "::operator=( const " << name << "&_u )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "_discriminator = _u._discriminator;" << endl;
  last_name = (const char *) "";
  for( int i = 0; i < m->length(); i++ ) {
    if( strcmp( last_name, m[ i ].name ) == 0 )
      continue;
    last_name = m[ i ].name;
    o << "_m." << ID(m[ i ].name) << " = ((" << name << "&)_u)._m.";
    o << ID(m[ i ].name) << ";" << endl;
  }
  o << "return *this;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
#endif
}

bool CodeGenC::is_union_case_defined( CORBA::UnionMemberSeq& mem,
					CORBA::Any &any )
{
  for( CORBA::ULong i = 0; i < mem.length(); i++ )
    if( mem[ i ].label == any )
      return true;
  return false;
}

string
CodeGenC::generate_union_default_case_label( CORBA::IDLType_ptr discr,
					       CORBA::UnionMemberSeq &mem )
{
  string label;
  label = "";
  
  CORBA::TypeCode_var tc = discr->type();
  if( is_enum( tc ) ) {
    CORBA::IDLType_var t = CORBA::IDLType::_duplicate( discr );
    while( t->def_kind() == CORBA::dk_Alias ) {
      CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( t );
      t = a->original_type_def();
    }
    CORBA::EnumDef_var e = CORBA::EnumDef::_narrow( t );
    assert( !CORBA::is_nil( e ) );
    CORBA::EnumMemberSeq_var e_mem = e->members();
    for( CORBA::ULong i = 0; i < e_mem->length(); i++ ) {
      CORBA::Any any;
      any.type( tc->unalias() );
      CORBA::Boolean r = any.enum_put( i );
      assert (r);
      if( !is_union_case_defined( mem, any ) ) {
	label = e_mem[ i ];
	break;
      }
    }
    return label;
  }
  
  tc = resolve_alias( tc );
  if( tc->kind() == CORBA::tk_long ) {
    CORBA::Long l = 0;
    do {
      CORBA::Any any;
      any <<= l;
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (l);
	break;
      }
    } while (++l);
    return label;
  }

  if( tc->kind() == CORBA::tk_ulong ) {
    CORBA::ULong ul = 0;
    do {
      CORBA::Any any;
      any <<= ul;
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (ul);
	break;
      }
    } while (++ul);
    return label;
  }

  if( tc->kind() == CORBA::tk_char ) {
    CORBA::Char c = 0;
    do {
      CORBA::Any any;
      any <<= CORBA::Any::from_char( c );
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (c);
	break;
      }
    } while (++c);
    return label;
  }

  if( tc->kind() == CORBA::tk_short ) {
    CORBA::Short s = 0;
    do {
      CORBA::Any any;
      any <<= s;
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (s);
	break;
      }
    } while (++s);
    return label;
  }

  if( tc->kind() == CORBA::tk_ushort ) {
    CORBA::UShort us = 0;
    do {
      CORBA::Any any;
      any <<= us;
      if( !is_union_case_defined( mem, any ) ) {
	label = xdec (us);
	break;
      }
    } while (++us);
    return label;
  }

  if( tc->kind() == CORBA::tk_boolean ) {
    CORBA::Any any;
    any <<= CORBA::Any::from_boolean( TRUE );
    if( is_union_case_defined( mem, any ) ) {
      any <<= CORBA::Any::from_boolean( FALSE );
      if( !is_union_case_defined( mem, any ) )
	label = "FALSE";
    } else
      label = "TRUE";
    return label;
  }

  return label;
}
    
void CodeGenC::emit_union_case_label( CORBA::IDLType_ptr discr,
				      CORBA::UnionMember &m,
				      CORBA::UnionMemberSeq &members )
{
  CORBA::TypeCode_var tc = m.label.type();
  if( is_octet( tc ) ) {
    string label = generate_union_default_case_label( discr, members );
    // Test if union has already all case members listed
    assert( label != "" );
    o << label;
    return;
  }
  
  tc = discr->type();
  if( is_enum( tc ) ) {
    CORBA::String_var s;
    CORBA::ULong idx;
    CORBA::Boolean r = m.label.enum_get( idx );
    assert (r);
    s = (const char *) tc->unalias()->member_name( idx );
    CORBA::Contained_var c = CORBA::Contained::_narrow( discr );
    assert( !CORBA::is_nil( c ) );
    CORBA::String_var tmp = get_absolute_name(c, "_");
    string abs;
    abs = ID(tmp);
#ifdef XXX_I_DUNNO_IF_THIS_SHOULD_GO_IN_HERE
    int colon_pos = abs.rfind( "_" );
    if( colon_pos != -1 )
      abs = abs.substr( 2, colon_pos );
#endif
    o << abs << ID(s);
  } else {
    o << &m.label;
  }
}
    
void CodeGenC::emitClientUnion( CORBA::UnionDef_ptr u )
{
  assert( u->def_kind() == CORBA::dk_Union );

  // Get base for absolute name
  CORBA::String_var tmp = get_absolute_name(u, "_");
  string abs;
  abs = ID(tmp);
  tmp = u->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::TypeCode_var tc = u->type();
  CORBA::UnionMemberSeq_var members = u->members();
  CORBA::IDLType_var discr = u->discriminator_type_def();

#ifdef XXX_DUNNO  
  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );
#endif

  // Emit TypeCode for union
  if( tc->is_freestanding() ) {
    o << "CORBA_TypeCode_ptr " << "_tc_" << abs;
    o << " = ";
    emit_type_code( tc );
    o << ";" << endl << endl;
  }

  // Emit default ctor, copy ctor, assignment op
  emitClientUnionDefaultMembers( name, (char *) abs.c_str(), members );

#ifdef CPP_CODE
  // Emit discriminator functions
  o << "void " << abs << name << "::_d( ";
  emit_type_for_param( discr, CORBA::PARAM_IN );
  o << " _p )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "_discriminator = _p;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
  
  emit_type_for_result( discr );
  o << " " << abs << name << "::_d() const" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "return _discriminator;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;

  // Generate member accessor functions
  CORBA::String_var last_name = (const char *) "";
  bool has_explicit_default = false;
  for( int i = 0; i < members->length(); i++ ) {
    CORBA::UnionMember &m = members[ i ];
    CORBA::TypeCode_var tc = members[ i ].label.type();
    if( is_octet( tc ) )
      has_explicit_default = true;

    if( strcmp( last_name, m.name ) == 0 )
      continue;
    last_name = m.name;
    
    CORBA::IDLType_var m_type = CORBA::IDLType::_duplicate( m.type_def );
    CORBA::String_var m_name = (const char *) ID(m.name);
    tc = m_type->type();
    if( is_string( tc ) ) {
      // Accessor functions for string
      o << "void " << abs << name << "::" << m_name << "( char *_p )" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
      
      o << "void " << abs << name << "::"
	<< m_name << "( const char *_p )" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
      
      o << "void " << abs << name << "::"
	<< m_name << "( const CORBA::String_var &_p )" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
      
      o << "const char *" << abs << name << "::"
	<< m_name << "() const" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "return _m." << m_name << ";" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
    } else if( is_array( tc ) && !is_alias( tc ) ) {
      // We have an anonymous array. See [16.10(7)]
      o << "void " << abs << name << "::" << m_name << "( ";
      emit_array( m_type );
      o << " _p";
      emit_array_suffix( m_type );
      o << " )" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      emit_array( m.type_def );
      o << "* _dest = (";
      emit_array( m.type_def );
      o << "*) _m." << m_name << ";" << endl;
      emit_array( m.type_def );
      o << "* _src = (";
      emit_array( m.type_def );
      o << "*) _p;" << endl;
      o << "for( int _i = 0; _i < ";
      o << compute_num_array_elem( tc );
      o << "; _i++ )" << endl;
      o.enterLevel();
      o << "*_dest++ = *_src++;" << endl;
      o.leaveLevel();
      o.leaveLevel();
      o << "}" << endl << endl;
      
      o << abs << name << "::_" << m_name << "_slice* ";
      o << abs << name << "::" << m_name << "()" << endl;
      o << "{" << endl;
      o.enterLevel();
      emit_array( m.type_def );
      o << "* _res;" << endl;
      emit_array( m.type_def );
      o << "* _dest;" << endl;
      o << "_res = _dest = (";
      emit_array( m_type );
      o << "*) new ";
      emit_array( m.type_def );
      emit_array_suffix( m.type_def );
      o << ";" << endl;
      emit_array( m.type_def );
      o << "* _src = (";
      emit_array( m.type_def );
      o << "*) _m." << m_name << ";" << endl;
      o << "for( int _i = 0; _i < ";
      o << compute_num_array_elem( tc );
      o << "; _i++ )" << endl;
      o.enterLevel();
      o << "*_dest++ = *_src++;" << endl;
      o.leaveLevel();
      o << "return (_" << m_name << "_slice*) _res;" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
    } else if( is_array( tc ) || is_objref( tc ) ||
	       is_type_code( tc ) || is_base_type( tc ) ) {
      // Accessor functions for array or object reference
      o << "void " << abs << name << "::" << m_name << "( ";
      if( is_array( tc ) )
	// Mapping for arrays differs from standard PARAM_IN mapping!
	emit_idl_type_name( m_type );
      else
	emit_type_for_param( m_type, CORBA::PARAM_IN );
      o << " _p )" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      if( is_array( tc ) ) {
	emit_array( m.type_def );
	o << "* _dest = (";
	emit_array( m.type_def );
	o << "*) _m." << m_name << ";" << endl;
	emit_array( m.type_def );
	o << "* _src = (";
	emit_array( m.type_def );
	o << "*) _p;" << endl;
	o << "for( int _i = 0; _i < ";
	o << compute_num_array_elem( tc );
	o << "; _i++ )" << endl;
	o.enterLevel();
	o << "*_dest++ = *_src++;" << endl;
	o.leaveLevel();
      } else {
	o << "_m." << m_name << " = ";
	if( is_objref( tc ) ) {
	  if( m.type_def->def_kind() == CORBA::dk_Primitive ) {
	    // Gotta be "Object"
	    o << "CORBA::Object::_duplicate( _p );" << endl;
	  } else {
	    CORBA::Contained_var c = CORBA::Contained::_narrow( m_type );
	    assert( !CORBA::is_nil( c ) );
	    CORBA::String_var objref_name = c->absolute_name();
	    o << ID(objref_name) << "::_duplicate( _p );" << endl;
	  }
	} else if( is_type_code( tc ) ) {
	  o << "CORBA::TypeCode::_duplicate( _p );" << endl;
	} else
	  o << "_p;" << endl;
      }
      o.leaveLevel();
      o << "}" << endl << endl;

      emit_type_for_result( m_type );
      o << " " << abs << name << "::" << m_name << "()" << endl;
      o << "{" << endl;
      o.enterLevel();
      if( is_array( tc ) ) {
	o << "return ";
	emit_idl_type_name( m_type );
	o << "_dup( _m." << m_name << " );" << endl;
      } else
	o << "return _m." << m_name << ";" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
    } else if( is_sequence( tc ) && !is_alias( tc ) ) {
      // Anonymous sequence
      o << "void " << abs << name << "::" << m_name << "( const ";
      emit_sequence( m_type );
      o << "& _p )" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;

      o << "const ";
      emit_sequence( m_type );
      o << "& " << abs << name << "::" << m_name << "() const" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "return _m." << m_name << ";" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;

      emit_sequence( m_type );
      o << "& " << abs << name << "::" << m_name << "()" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "return _m." << m_name << ";" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
    } else {
      // Gotta be struct, union or sequence
      o << "void " << abs << name << "::" << m_name << "( ";
      emit_type_for_param( m_type, CORBA::PARAM_IN );
      o << " _p )" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "_discriminator = ";
      emit_union_case_label( discr, m, members );
      o << ";" << endl;
      o << "_m." << m_name << " = _p;" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;

      o << "const ";
      emit_idl_type_name( m_type );
      o << "& " << abs << name << "::" << m_name << "() const" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "return (";
      emit_idl_type_name( m_type );
      o << "&) _m." << m_name << ";" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;

      emit_idl_type_name( m_type );
      o << "& " << abs << name << "::" << m_name << "()" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "return _m." << m_name << ";" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
    }
  }
  if( !has_explicit_default ) {
    // This union has no explicit default case label
    string default_label =
      generate_union_default_case_label( discr, members );
    if( default_label != "" ) {
      o << "void " << abs << name << "::_default()" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "_discriminator = " << default_label << ";" << endl;
      o.leaveLevel();
      o << "}" << endl << endl;
    }
  }

  // operator<<=
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs << name << " &_u )" << endl;
  o << "{" << endl;
  o.enterLevel();
  if( tc->is_freestanding() )
    o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
  o << "if (!_a.union_put_begin())" << endl;
  o.enterLevel();
  o << "return FALSE;" << endl;
  o.leaveLevel();
  o << "if( !(_a <<= ";
  string v = "((";
  v += abs;
  v += name;
  v += "&)_u)._discriminator";
  emitAnyParam( discr, (char *) v.c_str(), true );
  o << ") )" << endl;
  o.enterLevel();
  o << "return FALSE;" << endl;
  o.leaveLevel();
  o << "switch( _u._d() ) {" << endl;
  o.enterLevel();
  for( int i = 0; i < members->length(); i++ ) {
    CORBA::UnionMember &m = members[ i ];

    CORBA::TypeCode_var tc = members[ i ].label.type();
    o << "case ";
    emit_union_case_label( discr, m, members );
    o << ":" << endl;
    o.enterLevel();
    o << "if( !_a.union_put_selection( " << i << " ) )" << endl;
    o.enterLevel();
    o << "return FALSE;" << endl;
    o.leaveLevel();
    o << "if( !(_a <<= ";
    v = "((";
    v += abs;
    v += name;
    v += "&)_u)._m.";
    v += ID(m.name);
    emitAnyParam( m.type_def, (char *) v.c_str(), true );
    o << ") )" << endl;
    o.enterLevel();
    o << "return FALSE;" << endl;
    o.leaveLevel();
    o << "break;" << endl;
    o.leaveLevel();
  }
  o << "default:" << endl;
  o.enterLevel();
  o << "break;" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "}" << endl;
  o << "return _a.union_put_end();" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs << name << " &_u )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "if( !_a.union_get_begin() )" << endl;
  o.enterLevel();
  o << "return FALSE;" << endl;
  o.leaveLevel();
  o << "if( !(_a >>= ";
  v = "_u._discriminator";
  emitAnyParam( discr, (char *) v.c_str(), false );
  o << ") )" << endl;
  o.enterLevel();
  o << "return FALSE;" << endl;
  o.leaveLevel();
  o << "switch( _u._discriminator ) {" << endl;
  o.enterLevel();
  for( int i = 0; i < members->length(); i++ ) {
    CORBA::UnionMember &m = members[ i ];

    CORBA::TypeCode_var tc = members[ i ].label.type();
    o << "case ";
    emit_union_case_label( discr, m, members );
    o << ":" << endl;
    o.enterLevel();
    o << "if( !_a.union_get_selection( " << i << " ) )" << endl;
    o.enterLevel();
    o << "return FALSE;" << endl;
    o.leaveLevel();
    o << "if( !(_a >>= ";
    v = "_u._m.";
    v += ID(m.name);
    emitAnyParam( m.type_def, (char *) v.c_str(), false );
    o << ") )" << endl;
    o.enterLevel();
    o << "return FALSE;" << endl;
    o.leaveLevel();
    o << "break;" << endl;
    o.leaveLevel();
  }
  // ZZZ do not overwrite transmitted discriminator value
  o << "default:" << endl;
  o.enterLevel();
  string default_label =
    generate_union_default_case_label( discr, members );
  if( !has_explicit_default && default_label != "" )
    o << "_u._default();" << endl;
  o << "break;" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "}" << endl;
  o << "return _a.union_get_end();" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
#endif
}

void CodeGenC::emitClientException( CORBA::ExceptionDef_ptr e )
{
  assert( e->def_kind() == CORBA::dk_Exception );

  // Get base for absolute name
  CORBA::String_var tmp = e->absolute_name();
  string abs;
  abs = ID(tmp);
  tmp = e->name();
  CORBA::String_var name = (const char *) ID(tmp);
  CORBA::TypeCode_var tc = e->type();

  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );

  CORBA::String_var repoid = e->id();
  
  // Emit TypeCode for exception
  o << "CORBA::TypeCode_ptr " << abs << "_tc_" << name;
  o << " = ";
  emit_type_code( tc );
  o << ";" << endl << endl;

  // Emit default ctor, copy ctor, assignment op
  CORBA::StructMemberSeq_var m = e->members();
  emitClientStructDefaultMembers( name, (char *)abs.c_str(), m );

  // operator<<=
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
  o << abs << name << " &_e )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
  o << "return (_a.except_put_begin( \"" << repoid << "\" ) &&" << endl;
  o.enterLevel();
  CORBA::StructMemberSeq_var members = e->members();
  for( CORBA::ULong i = 0; i < members->length(); i++ ) {
    o << "(_a <<= ";
    string v = "((";
    v += abs;
    v += name;
    v += "&)_e).";
    v += ID(members[ i ].name);
    emitAnyParam( members[ i ].type_def, (char *) v.c_str(), true );
    o << ") &&" << endl;
  }
  o << "_a.except_put_end() );" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "}" << endl << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs << name << " &_e )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "CORBA::String_var _repoid;" << endl;
  o << "return (_a.except_get_begin( _repoid ) &&" << endl;
  o.enterLevel();
  for( CORBA::ULong i0 = 0; i0 < members->length(); i0++ ) {
    o << "(_a >>= ";
    string v = "_e.";
    v += ID(members[ i0 ].name);
    emitAnyParam( members[ i0 ].type_def, (char *) v.c_str(), false );
    o << ") &&" << endl;
  }
  o << "_a.except_get_end() );" << endl;
  o.leaveLevel();
  o.leaveLevel();
  o << "}" << endl << endl;

  o << "void " << abs << name << "::_throwit() const" << endl;
  o << "{" << endl;
  o.enterLevel();
  if( _gen_exceptions )
    o << "throw " << name << "_var( _clone() );" << endl;
  else
    o << "CORBA::Exception::_throw_failed( _clone() );" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
  
  o << "const char *" << abs << name << "::_repoid() const" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "return \"" << repoid << "\";" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
  
  o << "void " << abs << name << "::_encode( CORBA::DataEncoder &_en ) const" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "CORBA::Any _a;" << endl;
  o << "_a <<= *this;" << endl;
  o << "_a.marshal( _en );" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
  
  o << abs << name << " *" << abs << name << "::_clone() const" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "return new " << name << "( *this );" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
  
  o << abs << name << " *" << abs << name;
  o << "::_narrow( CORBA::Exception *_ex )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "if( _ex && !strcmp( _ex->_repoid(), \"" << repoid << "\" ) )" << endl;
  o.enterLevel();
  o << "return (" << name << " *) _ex;" << endl;
  o.leaveLevel();
  o << "return NULL;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl << endl;
}

void CodeGenC::emitClientEnum( CORBA::EnumDef_ptr e )
{
  assert( e->def_kind() == CORBA::dk_Enum );

  // Get base for absolute name
  CORBA::String_var tmp;
  tmp = get_absolute_name(e, "_");
  string abs;
  abs = ID(tmp);
  
  tmp = e->name();
  CORBA::String_var name = (const char *) ID(tmp);

#ifdef XXX_DUNNO  
  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );
#endif

  // Emit TypeCode for enum
  o << "CORBA_TypeCode_ptr " << "_tc_" << abs;
  o << " = ";
  CORBA::TypeCode_var tc = e->type();
  emit_type_code( tc );
  o << ";" << endl << endl;

#ifdef CPP_CODE
  // operator<<=
  o << "CORBA::Boolean operator<<=( CORBA::Any &_a, ";
  o << abs << name << " &_e )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
  o << "return (_a.enum_put( (CORBA::ULong) _e ));" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
  
  // operator>>=
  o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
  o << abs << name << " &_e )" << endl;
  o << "{" << endl;
  o.enterLevel();
  o << "CORBA::ULong _ul;" << endl;
  o << "if( !_a.enum_get( _ul ) )" << endl;
  o.enterLevel();
  o << "return FALSE;" << endl;
  o.leaveLevel();
  o << "_e = (" << abs << name << ") _ul;" << endl;
  o << "return TRUE;" << endl;
  o.leaveLevel();
  o << "}" << endl << endl;
#endif
}

void CodeGenC::emitClientAlias( CORBA::AliasDef_ptr a )
{
  assert( a->def_kind() == CORBA::dk_Alias );

  // Get base for absolute name
  CORBA::String_var tmp = get_absolute_name(a, "_");
  string abs;
  abs = ID(tmp);
  
  tmp = a->name();
  CORBA::String_var name = (const char *) ID(tmp);

#ifdef XXX_DUNNO  
  int colon_pos = abs.rfind( "::" );
  if( colon_pos != -1 )
    abs = abs.substr( 2, colon_pos );
#endif

  // Emit TypeCode for alias
  o << "CORBA_TypeCode_ptr " << "_tc_" << abs;
  o << " = ";
  CORBA::TypeCode_var tc = a->type();
  emit_type_code( tc );
  o << ";" << endl << endl;

  if( is_array ( tc ) )
    {
      o << abs << "_slice *" << abs << "__alloc(void)" << endl;
      o << "{" << endl;
      o.enterLevel();
      o << "return (" << abs << "_slice *) malloc(sizeof(" << abs << "_slice));" << endl;
      o.leaveLevel();
      o << "}" << endl;      
    }

#ifdef CPP_CODE
  // We don't need to generate the operator<<=() and operator>>=()
  // marshalling functions. The C++ will automatically pick the
  // appropriate operators of the original types (i.e. the original
  // type of the alias). These operators are only defined for
  // arrays. The T_forany definition for arrays, as described in
  // [16.12(5)], helps to disambiguate arrays with the same
  // structure of the type.
  // The problem with all other aliases is, that the TypeCode isn't
  // set properly. The TypeCode for these aliases always refer to
  // the element_type_def. This is clearly a problem within the
  // CORBA standard. There ought to be a T_forany for every alias.

  if( is_array( tc ) ) {
    // We have an alias to an array. Emit the marshalling
    // operators
    CORBA::AliasDef_var resolv = CORBA::AliasDef::_duplicate( a );
    CORBA::IDLType_var idl_type = resolv->original_type_def();
    CORBA::TypeCode_var idl_type_tc = idl_type->type();
    while( is_alias( idl_type_tc ) ) {
      resolv = CORBA::AliasDef::_narrow( idl_type );
      idl_type = resolv->original_type_def();
      idl_type_tc = idl_type->type();
    }
    CORBA::ArrayDef_var arr = CORBA::ArrayDef::_narrow( idl_type );
    assert( !CORBA::is_nil( arr ) );
    
    o << "CORBA::Boolean operator<<=( CORBA::Any &_a, const ";
    o << abs << name << "_forany &_const_s )" << endl;
    o << "{" << endl;
    o.enterLevel();
    o << "_a.type( " << abs << "_tc_" << name << " );" << endl;
    o << abs << name << "_forany &_s = "
      << "(" << abs << name << "_forany &) _const_s;" << endl;
    emit_array_marshalling( arr, true, 0 );
    o << "return TRUE;" << endl;
    o.leaveLevel();
    o << "}" << endl << endl;
    
    o << "CORBA::Boolean operator>>=( const CORBA::Any &_a, ";
    o << abs << name << "_forany &_s )" << endl;
    o << "{" << endl;
    o.enterLevel();
    emit_array_marshalling( arr, false, 0 );
    o << "return TRUE;" << endl;
    o.leaveLevel();
    o << "}" << endl << endl << endl;
  }
#endif
}

void CodeGenC::emitAnyParam( CORBA::IDLType_ptr p,
			       char *var_name, bool is_from )
{
  const char *wrapper = is_from ? "CORBA::Any::from_" : "CORBA::Any::to_";

  CORBA::IDLType_var resolved_alias = CORBA::IDLType::_duplicate( p );
  while( resolved_alias->def_kind() == CORBA::dk_Alias ) {
    CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( resolved_alias );
    resolved_alias = a->original_type_def();
  }
  
  CORBA::PrimitiveDef_var prim = CORBA::PrimitiveDef::_narrow( resolved_alias );
  if( !CORBA::is_nil( prim ) && prim->kind() == CORBA::pk_objref ) {
    o << wrapper << "object( " << ID(var_name);
    if( is_from )
      o << ", \"Object\"";
    o << " )";
    return;
  }
  
  CORBA::TypeCode_var tc = p->type();
  CORBA::TypeCode_var tc_resolv = resolve_alias( tc );
  
  CORBA::TCKind tk = tc_resolv->kind();

  if( is_array( tc ) && is_alias( tc ) ) {
    if( is_from ) {
      emit_idl_type_name( p );
      o << "_forany( ";
      o << var_name << " )";
    } else {
      o << var_name;
    }
    return;
  }
  
  if( tk == CORBA::tk_objref ) {
    o << "(";
    emit_type_for_variable( p, manual );
    if( !is_from )
      o << "&";
    o << ") ";
  }

  if( tk == CORBA::tk_char )
    o << wrapper << "char( " << ID(var_name) << " )";
  else if( tk == CORBA::tk_octet )
    o << wrapper << "octet( " << ID(var_name) << " )";
  else if( tk == CORBA::tk_boolean )
    o << wrapper << "boolean( " << ID(var_name) << " )";
  else if( tk == CORBA::tk_string ) {
    if( tc_resolv->length() == 0 ) {
      // Unbounded string
      o << ID(var_name);
    } else {
      // Bounded string
      o << wrapper << "string( " << ID(var_name) << ", "
	<< tc_resolv->length() << " )";
    }
  } else
    o << ID(var_name);
}

void CodeGenC::emit_method_dispatcher( string &absClassName,
				       string &dispatch_name,
				       CORBA::OperationDescription &op,
				       CORBA::IDLType_ptr result )
{
  string myctx = string("server_") + absClassName + string("_")
	+ dispatch_name;
  o << "if( strcmp( _req->op_name(), \"" << 
    ID( (char *) dispatch_name.c_str() ) << "\" ) == 0 ) {" << endl;
  o.enterLevel();
  CORBA::ParDescriptionSeq &p = op.parameters;
  CORBA::ULong k;
  
  // Emit variable declarations for all parameters as well as
  // the wrappers for the array parameters
  for( k = 0; k != p.length(); k++ ) {
    // Declaration for parameter
    if( p[ k ].mode != CORBA::PARAM_OUT ) {
      emit_type_for_variable( p[ k ].type_def, automatic );
      o << " " << ID(p[ k ].name);
      emit_array_suffix( p[ k ].type_def );
      o << ";" << endl;
    } else {
      if( is_array( p[ k ].type ) && is_variable( p[ k ].type ) ) {
	emit_type_for_variable( p[ k ].type_def, automatic );
	o << "_slice*";
      } else
	emit_type_for_variable( p[ k ].type_def, manual );
      o << " " << ID(p[ k ].name) << ";" << endl;
    }
    // Wrapper for array parameter
    if( is_array( p[ k ].type ) ) {
      if( !is_variable( p[ k ].type ) || p[ k ].mode != CORBA::PARAM_OUT ) {
	emit_type_for_variable( p[ k ].type_def, automatic );
	o << "_forany _forany_" << ID(p[ k ].name);
	o << "( " << ID(p[ k ].name) << " );" << endl;
      }
    }
  }
  if( k != 0 )
    o << endl;
  
  o << "CORBA::NVList_ptr _args;" << endl;
  o << "_orbnc()->create_list( " << p.length() << ", _args );" << endl;
  
  for( k = 0; k != p.length(); k++ ) {
    o << "_args->add( CORBA::";
    switch( p[ k ].mode ) {
    case CORBA::PARAM_IN:
      o << "ARG_IN";
      break;
    case CORBA::PARAM_OUT:
      o << "ARG_OUT";
      break;
    case CORBA::PARAM_INOUT:
      o << "ARG_INOUT";
      break;
    default:
      assert( 0 );
    }
    o << " );" << endl;
    o << "_args->item( " << k << " )->value()->type( ";
    emit_type_code_name( p[ k ].type_def );
    o << " );" << endl;
  }
  o << endl;
  
  o << "if (!_req->params( _args ))" << endl;
  o.enterLevel();
  o << "return true;" << endl << endl;
  o.leaveLevel();
  
  for( k = 0; k < p.length(); k++ ) {
    if( p[ k ].mode != CORBA::PARAM_OUT ) {
      o << "*_args->item( " << k << " )->value() >>= ";
      if( is_string( p[ k ].type ) ) {
	o << "CORBA::Any::to_string( " << ID(p[ k ].name) << ", ";
	CORBA::TypeCode_var tc_string = resolve_alias( p[ k ].type );
	o << tc_string->length() << " )";
      } else if( is_array( p[ k ].type ) ) {
	o << "_forany_" << ID(p[ k ].name);
      } else {
	emitAnyParam( p[ k ].type_def, p[ k ].name, false );
      }
      o << ";" << endl;
    }
  }

  // Do we have a result for this operation?
  CORBA::TypeCode_var tc_result = result->type();
  if( tc_result->kind() != CORBA::tk_void ) {
    if( is_array( tc_result ) ) {
      emit_type_for_variable( result, automatic );
      o << "_slice*";
    } else
      emit_type_for_variable( result, manual );
    o << " _res;" << endl;
  }

  if( _gen_exceptions && op.mode == CORBA::OP_NORMAL ) {
    o << "#ifdef HAVE_EXCEPTIONS" << endl;
    o << "try {" << endl;
    o << "#endif" << endl;
    o.enterLevel();
  }
  
  // Do we have a result for this operation?
  if( tc_result->kind() != CORBA::tk_void ) {
    o << "_res = ";
  }

  // Call the implementation
  o << myctx << "(this, ";
  for( k = 0; k < p.length(); k++ ) {
    o << ID(p[ k ].name) << ", ";
  }
  CORBA::ContextIdSeq& ctx = op.contexts;
  if( ctx.length() > 0 ) {
    o << "_req->ctx(), ";
  }
  if( k != 0 || ctx.length() > 0 )
    o << " ";
  o << "&_env);" << endl;

  if( _gen_exceptions && op.mode == CORBA::OP_NORMAL ) {
    o.leaveLevel();
    o << "#ifdef HAVE_EXCEPTIONS" << endl;
    CORBA::ExcDescriptionSeq exs = op.exceptions;
    for( CORBA::ULong i = 0; i < exs.length(); i++ ) {
      CORBA::Contained_var con = _repo->lookup_id( exs[ i ].id );
      assert( !CORBA::is_nil( con ) );
      CORBA::String_var tmp = con->absolute_name();
      CORBA::String_var ex_abs_name = (const char *) ID(tmp);
      o << "} catch( " << ex_abs_name << "_var &_ex ) {" << endl;
      o.enterLevel();
      o << "_req->exception( _ex->_clone() );" << endl;
      o << "return true;" << endl;
      o.leaveLevel();
    }
    o << "} catch( CORBA::SystemException_var &_ex ) {" << endl;
    o.enterLevel();
    o << "_req->exception( _ex->_clone() );" << endl;
    o << "return true;" << endl;
    o.leaveLevel();
    o << "} catch( ... ) {" << endl;
    o.enterLevel();
    o << "assert( 0 );" << endl;
    o.leaveLevel();
    o << "}" << endl;
    o << "#endif" << endl;
  }
  
  // Write back OUT and INOUT parameters
  for( k = 0; k < p.length(); k++ ) {
    if( p[ k ].mode != CORBA::PARAM_IN ) {
      // Do we need a wrapper for an out variable length array?
      if( is_variable( p[ k ].type ) && is_array( p[ k ].type ) &&
	  p[ k ].mode == CORBA::PARAM_OUT ) {
	// Yes
	emit_type_for_variable( p[ k ].type_def, automatic );
	o << "_forany _forany_" << ID(p[ k ].name);
	o << "( " << ID(p[ k ].name) << " );" << endl;
      }
      
      o << "*_args->item( " << k << " )->value() <<= ";
      if( is_string( p[ k ].type ) ) {
	o << "CORBA::Any::from_string( (char *) " << ID(p[ k ].name) << ", ";
	CORBA::TypeCode_var tc_string = resolve_alias( p[ k ].type );
	o << tc_string->length() << " );" << endl;
	if( p[ k ].mode == CORBA::PARAM_OUT )
	  o << "CORBA::string_free( " << ID(p[ k ].name) << " );" << endl;
      } else if( is_array( p[ k ].type ) ) {
	o << "_forany_" << ID(p[ k ].name) << ";" << endl;
      } else {
	string n;
	if( is_variable( p[ k ].type ) && !is_objref( p[ k ].type ) &&
	    p[ k ].mode == CORBA::PARAM_OUT )
	  n = "*";
	n += ID(p[ k ].name);
	emitAnyParam( p[ k ].type_def, (char *) n.c_str(), true );
	o << ";" << endl;
	if( is_objref( p[ k ].type ) && p[ k ].mode == CORBA::PARAM_OUT )
	  o << "CORBA::release( " << ID(p[ k ].name) << " );" << endl;
	if( is_variable( p[ k ].type ) && !is_objref( p[ k ].type ) &&
	    p[ k ].mode == CORBA::PARAM_OUT )
	  o << "delete " << ID(p[ k ].name) << ";" << endl;
      }
    }
  }
  
  // Write back result (if there was one)
  if( tc_result->kind() != CORBA::tk_void ) {
    o << "CORBA::Any *_any_res = new CORBA::Any;" << endl;
    // Do we have to generate a wrapper for an array result?
    if( is_array( tc_result ) ) {
      // Yes
      emit_type_for_variable( result, automatic );
      o << "_forany _res_forany( _res );" << endl;
    }
    o << "*_any_res <<= ";
    if( is_array( tc_result ) ) {
      o << "_res_forany;" << endl;
      emit_idl_type_name( result );
      o << "_free( _res );" << endl;
    } else {
      if( !is_objref( tc_result ) && !is_string( tc_result ) &&
	  !is_wstring( tc_result ) && is_variable( tc_result ) ) {
	o << "*_res;" << endl;
	o << "delete _res;" << endl;
      } else {
	emitAnyParam( result, "_res", true );
	o << ";" << endl;
	if( is_string( tc_result ) )
	  o << "CORBA::string_free( _res );" << endl;
	if( is_objref( tc_result ) )
	  o << "CORBA::release( _res );" << endl;
      }
    }
    o<< "_req->result( _any_res );" << endl;
  }
  
  o << "return true;" << endl;
  o.leaveLevel();
  o << "}" << endl;
}

void CodeGenC::emitServer()
{
  CORBA::IDLType_var idl_type;
  CORBA::ContainedSeq_var c;
  CORBA::String_var tmp;
  
  for( CORBA::ULong i0 = 0; i0 < _idl_objs.length(); i0++ ) {

    CORBA::InterfaceDef_var in;
    in = CORBA::InterfaceDef::_narrow( _idl_objs[ i0 ] );
    if( CORBA::is_nil( in ) )
      continue;
    
    if( check_for_included_defn( in ) )
      continue;
    
    tmp = in->name();
    string className;
    className = ID(tmp);
    tmp = in->absolute_name();
    string absClassName;
    absClassName = ID(tmp);
    absClassName = absClassName.substr( 2 );
    CORBA::String_var in_id = in->id();
    
    o << endl;
    o << "// Dynamic Implementation Routine for interface " <<
      className << endl;
    o << "typedef void * " << absClassName << "_ptr;" << endl;
    o << "class " << absClassName << _SKEL << ": virtual public MethodDispatcher, virtual public CORBA::Object {" << endl;
    o.enterLevel();
    o << "public:" << endl;
    o.enterLevel();
    o << "~" << className << _SKEL << "() {};" << endl;
    o << className << _SKEL
      << "( const CORBA::BOA::ReferenceData &_id = CORBA::BOA::ReferenceData());" << endl;
    o << className << _SKEL << "( CORBA::Object_ptr _obj );" << endl;
    o << "bool dispatch(CORBA::ServerRequest_ptr _req, CORBA::Environment &_env);"
      << endl;
    o << absClassName << "_ptr _this();" << endl;
    o.leaveLevel();
    o.leaveLevel();
    o << "};" << endl << endl;
    o << absClassName << _SKEL << "::" << className << _SKEL;
    o << "( const CORBA::BOA::ReferenceData &_id )" << endl;
    o << "{" << endl;
    o.enterLevel();
    o << "CORBA::ImplementationDef_var _impl =" << endl;
    o.enterLevel();
    o << "_find_impl( \"" << in_id << "\", \"";
    o << className << "\" );" << endl;
    o.leaveLevel();
    o << "assert( !CORBA::is_nil( _impl ) );" << endl;
    o << "_create_ref( _id," << endl;
    o.enterLevel();
    o << "CORBA::InterfaceDef::_nil()," << endl;
    o << "_impl," << endl;
    o << "\"" << in_id << "\" );" << endl;
    o.leaveLevel();
    o << "register_dispatcher( new InterfaceDispatcherWrapper<";
    o << className << _SKEL;
    o << ">( this ) );" << endl;
    o << "server_" << absClassName << "_CreateTrue(this);" << endl;
    o.leaveLevel();
    o << "}" << endl << endl;

    o << absClassName << _SKEL << "::" << className << _SKEL;
    o << "( CORBA::Object_ptr _obj )" << endl;
    o << "{" << endl;
    o.enterLevel();
    o << "CORBA::ImplementationDef_var _impl =" << endl;
    o.enterLevel();
    o << "_find_impl( \"" << in->id() << "\", \"";
    o << className << "\" );" << endl;
    o.leaveLevel();
    o << "assert( !CORBA::is_nil( _impl ) );" << endl;
    o << "_restore_ref( _obj," << endl;
    o.enterLevel();
    o << "CORBA::BOA::ReferenceData()," << endl;
    o << "CORBA::InterfaceDef::_nil()," << endl;
    o << "_impl );" << endl;
    o.leaveLevel();
    o << "register_dispatcher( new InterfaceDispatcherWrapper<";
    o << className << _SKEL;
    o << ">( this ) );" << endl;
    o.leaveLevel();
    o << "}" << endl << endl;

#if 0
    // dtor
    o << absClassName << _SKEL;
    o << "::~" << className << _SKEL << "()" << endl;
    o << "{" << endl;
    o << "}" << endl << endl;
#endif

    o << absClassName << " " << absClassName << "_CreateTrue(void)" << endl;
    o << "{" << endl;
    o.enterLevel();
    o << "return new " << absClassName << "_skel();" << endl;
    o.leaveLevel();
    o << "}" << endl << endl;

    // Generate method dispatcher
    o << "bool " << absClassName << _SKEL <<
      "::dispatch( CORBA::ServerRequest_ptr _req, CORBA::Environment &_env )"
      << endl;
    o << "{" << endl;
    o.enterLevel();

    CORBA::InterfaceDef::FullInterfaceDescription_var desc;
    desc = in->describe_interface();

    // Generate dispatcher for attributes
    for( CORBA::ULong j = 0; j < desc->attributes.length(); j++ ) {
      CORBA::AttributeDescription attr;
      attr = desc->attributes[ j ];

      CORBA::OperationDescription op;
      op.name = (const char *) attr.name;
      op.id = (const char *) "";  // Not needed
      op.defined_in = attr.defined_in;
      op.version = (const char*) attr.version;
      op.result = CORBA::TypeCode::_nil();  // Not needed
      op.mode = CORBA::OP_NORMAL;
      // op.contexts not needed for attributes
      op.parameters.length( 0 );  // No parameters for _get
      // op.exceptions not needed for attributes

      string dispatch_name = string("_get_") + (const char *)attr.name;
    
      idl_type = lookup_attribute_by_id( attr.id );
      emit_method_dispatcher( absClassName, dispatch_name, op, idl_type );

      if( attr.mode == CORBA::ATTR_NORMAL ) {
	// We also need the set version of the attribute
	CORBA::ParameterDescription attr_param;
	attr_param.name = (const char *) "_value";
	attr_param.type = attr.type;
	attr_param.type_def = idl_type;
	attr_param.mode = CORBA::PARAM_IN;
	
	op.name = (const char *) attr.name;
	op.id = (const char *) "";  // Not needed
	op.defined_in = attr.defined_in;
	op.version = (const char*) attr.version;
	op.result = CORBA::TypeCode::_nil();  // Not needed
	op.mode = CORBA::OP_NORMAL;
	// op.contexts not needed for attributes
	op.parameters.length( 1 );
	op.parameters[ 0 ] = attr_param;
	// op.exceptions not needed for attributes
	
	dispatch_name = string("_set_") + (const char *)attr.name;
	
	emit_method_dispatcher( absClassName, dispatch_name, op,
				_repo->get_primitive( CORBA::pk_void ) );
      }
    }

    // Generate dispatcher for methods
    for( CORBA::ULong j0 = 0; j0 < desc->operations.length(); j0++ ) {
      CORBA::OperationDescription op;
      op = desc->operations[ j0 ];
      string dispatch_name = (const char *) op.name;
      idl_type = lookup_result_by_id( op.id );
      emit_method_dispatcher( absClassName, dispatch_name, op, idl_type );
    }
    o << "return false;" << endl;
    o.leaveLevel();
    o << "}" << endl << endl;

    // _this method
    o << absClassName << "_ptr " << absClassName << _SKEL;
    o << "::_this()" << endl;
    o << "{" << endl;
    o.enterLevel();
    o << "return _duplicate( this );" << endl;
    o.leaveLevel();
    o << "}" << endl << endl;
  }
}

void CodeGenC::emit_idl_type_name( CORBA::IDLType_ptr t )
{
  CORBA::String_var tmp;
  string abs_name;

  CORBA::TypeCode_var tc = t->type();
  
  switch( t->def_kind() ) {
  case CORBA::dk_Struct:
  case CORBA::dk_Union:
  case CORBA::dk_Enum:
  case CORBA::dk_Alias:
    {
      CORBA::Contained_var con = CORBA::Contained::_narrow( t );
      tmp = get_absolute_name(con, "_");
      abs_name = ID(tmp);
      if( is_objref( tc ) )
	// Must be IDL-Type "Object"
	abs_name += "_ptr";
      break;
    }
  case CORBA::dk_Interface:
    {
      CORBA::InterfaceDef_var in = CORBA::InterfaceDef::_narrow( t );
      tmp = get_absolute_name(in, "_");
      abs_name = ID(tmp);
      abs_name += "_ptr";
      break;
    }
  case CORBA::dk_Primitive:
    {
      // Gotta be an any, objref or a TypeCode
      CORBA::PrimitiveDef_var p = CORBA::PrimitiveDef::_narrow( t );
      if( p->kind() == CORBA::pk_any )
	o << "CORBA_any";
      else if( p->kind() == CORBA::pk_objref )
	o << "CORBA_Object";
      else if( p->kind() == CORBA::pk_TypeCode )
	o << "CORBA_TypeCode_type";
      else {
	cout << p->kind() << endl;
	assert( 0 );
      }
      return;
    }
  default:
    cout << t->def_kind() << endl;
    assert( 0 );
  }
  if( abs_name[ 0 ] == ':' )
    abs_name = abs_name.substr( 2 );
  o << (char *) abs_name.c_str();
}
  
/*
 * Emit a TypeCode to be used in a signature of a C++ method.
 * The ParameterMode is necessary for this. See Table 24 in
 * [16.18(13)]
 */
void CodeGenC::emit_type_for_param( CORBA::IDLType_ptr t,
				      CORBA::ParameterMode mode )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  // Mapping for arrays
  if( is_array( tc ) ) {
    if( mode == CORBA::PARAM_IN )
      o << "const ";
    emit_idl_type_name( t );
    if( mode == CORBA::PARAM_OUT && is_variable( tc ) )
      o << "_slice**";
    return;
  }
  
  // Basic argument mapping [16.18(12)]
  if( mode == CORBA::PARAM_IN && is_string( tc ) )
    o << "const ";
    
  if( emit_base_type( t ) ) {
    if( mode != CORBA::PARAM_IN )
      o << "*";
    return;
  }

  if( is_enum( tc ) || is_objref( tc ) ) {
    emit_idl_type_name( t );
    if( mode != CORBA::PARAM_IN )
      o << "*";
    return;
  }

  if( is_type_code( tc ) ) {
    o << "CORBA::TypeCode_ptr";
    if( mode != CORBA::PARAM_IN )
      o << "*";
    return;
  }

  if( mode == CORBA::PARAM_IN )
    o << "const ";

  switch( tc->kind() ) {
  case CORBA::tk_struct:
  case CORBA::tk_union:
  case CORBA::tk_alias:
  case CORBA::tk_any:
    emit_idl_type_name( t );
    if( mode == CORBA::PARAM_OUT && is_variable( tc ) &&
	!is_string( tc ) && !is_wstring( tc ) )
      o << "*";
    o << "*";
    break;
  default:
    cout << tc->kind() << endl;
    assert( 0 );
  }
}

/*
 * Emit a TypeCode to be used as a result in a signature of a
 * C++ method. See Table 24 in [16.18(13)]
 */
void CodeGenC::emit_type_for_result( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  if( is_array( tc ) ) {
    // Mapping for arrays
    emit_idl_type_name( t );
    o << "_slice*";
    return;
  }
  
  if( !emit_base_type( t ) ) {
    switch( tc->kind() ) {
    case CORBA::tk_objref:
    case CORBA::tk_struct:
    case CORBA::tk_union:
    case CORBA::tk_alias:
    case CORBA::tk_enum:
      emit_idl_type_name( t );
      break;
    case CORBA::tk_any:
      o << "CORBA::Any";
      break;
    case CORBA::tk_TypeCode:
      o << "CORBA::TypeCode_ptr";
      break;
    default:
      cout << "TCKind: " << tc->kind() << endl;
      assert( 0 );
    }
  }
  
  // Basic argument mapping [16.18(12)]
  if( is_variable( tc ) && !is_string( tc ) && !is_wstring( tc ) &&
      !is_objref( tc ) )
    o << "*";
}

void CodeGenC::emit_sequence( CORBA::IDLType_ptr type,
			      char *ctx)
{
  assert( type->def_kind() == CORBA::dk_Sequence );
  CORBA::SequenceDef_var s = CORBA::SequenceDef::_narrow( type );
  CORBA::TypeCode_var seq_tc = s->type();
  CORBA::IDLType_var idl_type = s->element_type_def();
  CORBA::TypeCode_var tc = idl_type->type();

  /* XXX this doesn't strictly conform to CORBA specs,
     but I don't know a way to get the string expansion of the
     type of tc without a prefixing CORBA_ */

  o << "struct {" << endl;
  o.enterLevel();
  o << "CORBA_unsigned_long _maximum;" << endl;
  o << "CORBA_unsigned_long _length;" << endl;
  emit_type_for_variable( idl_type, automatic );
  o << "* _buffer;" << endl;
  o.leaveLevel();
  o << "}";
  

#ifdef CPP_CODE

  if( s->bound() != 0 )
    o << "Bounded";

  if( is_array( tc ) || seq_tc->is_recursive_seq() )
    o << "SequenceIndTmpl<";
  else
    o << "SequenceTmpl<";
  
  if (is_char (tc))
    o << "CORBA::CharWrapper";
  else if (is_octet (tc))
    o << "CORBA::OctetWrapper";
  else if (is_bool (tc))
    o << "CORBA::BooleanWrapper";
  else if( is_array( tc ) ) {
    CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( idl_type );
    assert( !CORBA::is_nil( a ) );
    CORBA::IDLType_var orig = a->original_type_def();
    CORBA::TypeCode_var orig_tc = orig->type();
    CORBA::ULong num_elem = compute_num_array_elem( orig_tc );
    CORBA::String_var tmp = a->absolute_name();
    CORBA::String_var name = (const char *) ID(tmp);
    emit_type_for_variable( orig, automatic );
    o << "," << name << "," << num_elem;
  } else if( seq_tc->is_recursive_seq() ) {
    emit_type_for_variable( idl_type, automatic );
    o << ",";
    emit_type_for_variable( idl_type, automatic );
    o << ",1";
  } else
    emit_type_for_variable( idl_type, automatic );

  if( s->bound() != 0 )
    o << "," << s->bound();
  o << ">";
#endif
}
  
void CodeGenC::emit_array( CORBA::IDLType_ptr type )
{
  assert( type->def_kind() == CORBA::dk_Array ||
	  type->def_kind() == CORBA::dk_Alias );

  CORBA::IDLType_var idl_type = CORBA::IDLType::_duplicate( type );

  CORBA::AliasDef_var alias;
  while( idl_type->def_kind() == CORBA::dk_Alias ) {
    alias = CORBA::AliasDef::_narrow( idl_type );
    assert( !CORBA::is_nil( alias ) );
    idl_type = alias->original_type_def();
  }
  assert( idl_type->def_kind() == CORBA::dk_Array );
  
  CORBA::ArrayDef_var arr;
  while( idl_type->def_kind() == CORBA::dk_Array ) {
    arr = CORBA::ArrayDef::_narrow( idl_type );
    assert( !CORBA::is_nil( arr ) );
    idl_type = arr->element_type_def();
  }
  emit_type_for_variable( idl_type, automatic );
}

void CodeGenC::emit_array_suffix( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  while( tc->kind() == CORBA::tk_array ) {
    o << "[ " << tc->length() << " ]";
    tc = tc->content_type();
  }
}

void CodeGenC::emit_array_slice_suffix( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  tc = tc->content_type();
  while( tc->kind() == CORBA::tk_array ) {
    o << "[ " << tc->length() << " ]";
    tc = tc->content_type();
  }
}

CORBA::ULong CodeGenC::compute_num_array_elem( CORBA::TypeCode_ptr tc )
{
  if( tc->kind() != CORBA::tk_array )
    return 1;
  CORBA::ULong len = tc->length();
  CORBA::TypeCode_var nested_tc = tc->content_type();
  return len * compute_num_array_elem( nested_tc );
}

/*
 * Emit a TypeCode to be used in a C variable declaration.
 */
void CodeGenC::emit_type_for_variable( CORBA::IDLType_ptr t,
					 Storage s )
{
  CORBA::String_var tmp;
  
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  if( is_string( tc ) ) {
    if( is_alias( tc ) ) {
      CORBA::AliasDef_var a = CORBA::AliasDef::_narrow( t );
      tmp = a->absolute_name();
      string abs_name;
      abs_name = ID(tmp);
      abs_name = abs_name.substr( 2 );
      o << abs_name;
      if( s == automatic )
	o << "_var";
    } else {
      if( s == manual )
	o << "char*";
      else
	o << "CORBA_String_var";
    }
    return;
  }

  CORBA::PrimitiveDef_var prim = CORBA::PrimitiveDef::_narrow( t );
  if( !CORBA::is_nil( prim ) && prim->kind() == CORBA::pk_objref ) {
    o << "CORBA_Object";
    if( s == manual )
      o << "_ptr";
    else
      o << "_var";
    return;
  }
  
  if( is_objref( tc ) ) {
    // This can be an object reference or an alias to "Object"
    CORBA::Contained_var con = CORBA::Contained::_narrow( t );
    assert( !CORBA::is_nil( con ) );
    tmp = con->absolute_name();
    string abs_name;
    abs_name = ID(tmp);
    abs_name = abs_name.substr( 2 );
    o << abs_name;
    if( s == manual )
      o << "_ptr";
    else
      o << "_var";
    return;
  }

  if( !emit_base_type( t ) ) {
    switch( tc->kind() ) {
    case CORBA::tk_objref:
    case CORBA::tk_struct:
    case CORBA::tk_union:
    case CORBA::tk_enum:
    case CORBA::tk_alias:
    case CORBA::tk_any:
      emit_idl_type_name( t );
      break;
    case CORBA::tk_sequence:
      emit_sequence( t );
      break;
    case CORBA::tk_TypeCode:
      o << "CORBA_TypeCode";
      if( s == manual )
	o << "_ptr";
      else
	o << "_var";
      break;
    case CORBA::tk_array:
      emit_array( t );
      if( s == manual )
	o << "_slice";
      // the case "s == automatic" is taken care of
      // in emit_array_suffix
      break;
    default:
      assert( 0 );
    }
  }
  if( s == manual && is_variable( tc ) )
    o << "*";
}

bool CodeGenC::emit_base_type( CORBA::IDLType_ptr t )
{
  CORBA::TypeCode_var tc = t->type();
  assert( !CORBA::is_nil( tc ) );

  // Mapping of Base Data Types [16.5]
  switch( tc->kind() ) {
  case CORBA::tk_void:
    o << "void";
    break;
  case CORBA::tk_boolean:
    o << "CORBA_boolean";
    break;
  case CORBA::tk_char:
    o << "CORBA_char";
    break;
  case CORBA::tk_octet:
    o << "CORBA_octet";
    break;
  case CORBA::tk_ushort:
    o << "CORBA_unsigned_short";
    break;
  case CORBA::tk_ulong:
    o << "CORBA_unsigned_long";
    break;
  case CORBA::tk_short:
    o << "CORBA_unsigned_short";
    break;
  case CORBA::tk_long:
    o << "CORBA_long";
    break;
  case CORBA::tk_float:
    o << "CORBA_float";
    break;
  case CORBA::tk_double:
    o << "CORBA_double";
    break;
  case CORBA::tk_string:
    o << "char*";
    break;
  case CORBA::tk_alias:
    if( is_base_type( tc ) ) {
      emit_idl_type_name( t );
      return true;
    } else
      return false;
    break;
  default:
    return false;
  }
  return true;
}

