/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include "const.h"
#include "error.h"


//-- Const --------------------------------------------------------------

Const::Const()
{
  _node = NULL;
}
  
Const::Const( CORBA::TypeCode_ptr target_type, CORBA::LongLong l,
	      ParseNode* node )
{
  CORBA::Any a;
  a <<= l;
  _value = do_coercion( target_type, a );
  _node = node;
}

Const::Const( CORBA::TypeCode_ptr target_type, char *s, ParseNode* node )
{
  CORBA::Any a;
  a <<= CORBA::Any::from_string( s, 0 );
  _value = do_coercion( target_type, a );
  _node = node;
}

Const::Const( CORBA::TypeCode_ptr target_type, CORBA::CharWrapper c,
	      ParseNode* node )
{
  CORBA::Any a;
  a <<= CORBA::Any::from_char( c );
  _value = do_coercion( target_type, a );
  _node = node;
}

Const::Const( CORBA::TypeCode_ptr target_type, CORBA::LongDouble ld,
	      ParseNode* node )
{
  CORBA::Any a;
  a <<= ld;
  _value = do_coercion( target_type, a );
  _node = node;
}

Const::Const( CORBA::TypeCode_ptr target_type, CORBA::BooleanWrapper b,
	      ParseNode* node )
{
  CORBA::Any a;
  a <<= CORBA::Any::from_boolean( b );
  _value = do_coercion( target_type, a );
  _node = node;
}

Const::Const( const Const& c )
{
  copy_const( c );
}

Const::Const( CORBA::Any *a, ParseNode* node )
{
  _value = *a;
  delete a;
  _node = node;
}

void Const::copy_const( const Const &c )
{
  _value = c._value;
  _node = c._node;
}

Const &Const::operator=( const Const& c )
{
  copy_const( c );
  return *this;
}

/*
 * Get the content of the Const-Object and deposit into an Any.
 * The parameter "target_type" denotes the desired type. We therefore
 * need to do some type- and range-checking.
 *    const unsigned short c1 = 10;
 *          --------------      --
 *                |             |_ type of the Const Object (long long)
 *                |_______________ desired target type
 */
CORBA::Any Const::do_coercion( CORBA::TypeCode_ptr target_type,
			       CORBA::Any& val )
{
  CORBA::TypeCode_var tc = val.type();
  if( tc->equal( target_type ) )
    // Nothing to do
    return val;
  CORBA::Any coerced_val;
  switch( tc->kind() ) {
  case CORBA::tk_char:
    {
      if( target_type->kind() != CORBA::tk_wchar )
	coercion_error( tc, target_type );
      CORBA::Char c;
      val >>= CORBA::Any::to_char( c );
      coerced_val <<= CORBA::Any::from_wchar( (CORBA::WChar) (c & 0xff) );
      break;
    }
  case CORBA::tk_string:
    {
      if( target_type->kind() != CORBA::tk_wstring )
	coercion_error( tc, target_type );
      char *s;
      val >>= CORBA::Any::to_string( s, 0 );
      CORBA::WChar* w;
      w = CORBA::wstring_alloc( strlen( s ) );
      for( size_t i = 0; i < strlen( s ); i++ )
	w[ i ] = s[ i ] & 0xff;
      w[ strlen( s ) ] = 0;
      coerced_val <<= CORBA::Any::from_wstring( w, 0 );
      CORBA::string_free( s );
      CORBA::wstring_free( w );
      break;
    }
  case CORBA::tk_fixed:
    {
      if( target_type->kind() != CORBA::tk_longdouble )
	coercion_error( tc, target_type );
      FixedBase f( tc->fixed_digits(), tc->fixed_scale() );
      FixedBase::FixedValue_var fv;
      val >>= CORBA::Any::to_fixed( fv, tc->fixed_digits(),
				    tc->fixed_scale() );
      f.from_digits( fv );
      
      coerced_val <<= f.to_longdouble();
      break;
    }
  case CORBA::tk_longdouble:
    {
      CORBA::LongDouble ld;
      val >>= ld;
      switch( target_type->kind() ) {
      case CORBA::tk_double:
	{
	  //XXX
	  //assert( ld >= MIN_DOUBLE && ld <= MAX_DOUBLE );
	  coerced_val <<= (CORBA::Double) ld;
	  break;
	}
      case CORBA::tk_float:
	{
	  //XXX
	  //assert( ld >= MIN_FLOAT && ld <= MAX_FLOAT );
	  coerced_val <<= (CORBA::Float) ld;
	  break;
	}
      case CORBA::tk_fixed:
	{
	  CORBA::UShort d = target_type->fixed_digits();
	  CORBA::Short  s = target_type->fixed_scale();
	  FixedBase f( d, s );
	  f.from_longdouble( ld );
	  FixedBase::FixedValue_var fv = f.to_digits();
	  coerced_val <<= CORBA::Any::from_fixed( fv, d, s );
	  break;
	}
      default:
	coercion_error( tc, target_type );
      }
      break;
    }
  case CORBA::tk_double:
    {
      CORBA::Double d;
      val >>= d;
      switch( target_type->kind() ) {
      case CORBA::tk_longdouble:
	coerced_val <<= (CORBA::LongDouble) d;
	break;
      case CORBA::tk_float:
	{
	  //XXX what value range do we need to check here?
	  //assert( d >= FLOAT_MIN && d <= FLOAT_MAX );
	  coerced_val <<= (CORBA::Float) d;
	  break;
	}
      default:
	coercion_error( tc, target_type );
      }
      break;
    }
  case CORBA::tk_float:
    {
      CORBA::Float f;
      val >>= f;
      switch( target_type->kind() ) {
      case CORBA::tk_longdouble:
	coerced_val <<= (CORBA::LongDouble) f;
	break;
      case CORBA::tk_double:
	coerced_val <<= (CORBA::Double) f;
	break;
      default:
	coercion_error( tc, target_type );
      }
      break;
    }
  case CORBA::tk_longlong:
    {
      CORBA::LongLong ll;
      val >>= ll;
      switch( target_type->kind() ) {
      case CORBA::tk_longdouble:
	coerced_val <<= (CORBA::LongDouble) ll;
	break;
      case CORBA::tk_double:
	coerced_val <<= (CORBA::Double) ll;
	break;
      case CORBA::tk_float:
	coerced_val <<= (CORBA::Float) ll;
	break;
      case CORBA::tk_ulonglong:
	{
	  //XXX
	  //assert( ll < MAX_ULONGLONG );
	  coerced_val <<= (CORBA::ULongLong) ll;
	  break;
	}
      case CORBA::tk_long:
	{
	  //XXX
	  //assert( ll >= MIN_LONG && ll <= MAX_LONG );
	  coerced_val <<= (CORBA::Long) ll;
	  break;
	}
      case CORBA::tk_ulong:
	{
	  //XXX
	  //assert( ll >= MIN_ULONG && MAX_ULONG );
	  coerced_val <<= (CORBA::ULong) ll;
	  break;
	}
      case CORBA::tk_short:
	assert( ll > -32769 && ll < 32768 );
	coerced_val <<= (CORBA::Short) ll;
	break;
      case CORBA::tk_ushort:
	assert( ll >= 0 && ll < 65536 );
	coerced_val <<= (CORBA::UShort) ll;
	break;
      default:
	coercion_error( tc, target_type );
      }
      break;
    }
  case CORBA::tk_long:
    {
      CORBA::Long l;
      val >>= l;
      switch( target_type->kind() ) {
      case CORBA::tk_longdouble:
	coerced_val <<= (CORBA::LongDouble) l;
	break;
      case CORBA::tk_float:
	coerced_val <<= (CORBA::Float) l;
	break;
      case CORBA::tk_double:
	coerced_val <<= (CORBA::Double) l;
	break;
      case CORBA::tk_longlong:
	coerced_val <<= (CORBA::LongLong) l;
	break;
      case CORBA::tk_ulonglong:
	coerced_val <<= (CORBA::ULongLong) l;
	break;
      case CORBA::tk_ulong:
	assert( l >= 0 );
	coerced_val <<= (CORBA::ULong) l;
	break;
      case CORBA::tk_short:
	assert( l > -32769 && l < 32768 );
	coerced_val <<= (CORBA::Short) l;
	break;
      case CORBA::tk_ushort:
	assert( l >= 0 && l < 65536 );
	coerced_val <<= (CORBA::UShort) l;
	break;
      default:
	coercion_error( tc, target_type );
      }
      break;
    }
  case CORBA::tk_ulong:
    {
      CORBA::ULong ul;
      val >>= ul;
      switch( target_type->kind() ) {
      case CORBA::tk_longdouble:
	coerced_val <<= (CORBA::LongDouble) ul;
	break;
      case CORBA::tk_float:
	coerced_val <<= (CORBA::Float) ul;
	break;
      case CORBA::tk_double:
	coerced_val <<= (CORBA::Double) ul;
	break;
      case CORBA::tk_longlong:
	coerced_val <<= (CORBA::LongLong) ul;
	break;
      case CORBA::tk_ulonglong:
	coerced_val <<= (CORBA::ULongLong) ul;
	break;
      case CORBA::tk_long:
	assert( ul < 65536 );
	coerced_val <<= (CORBA::Long) ul;
	break;
      case CORBA::tk_short:
	assert( ul < 32768 );
	coerced_val <<= (CORBA::Short) ul;
	break;
      case CORBA::tk_ushort:
	assert( ul < 32768 );
	coerced_val <<= (CORBA::UShort) ul;
	break;
      default:
	coercion_error( tc, target_type );
      }
      break;
    }
  case CORBA::tk_short:
    {
      CORBA::Short s;
      val >>= s;
      switch( target_type->kind() ) {
      case CORBA::tk_longdouble:
	coerced_val <<= (CORBA::LongDouble) s;
	break;
      case CORBA::tk_float:
	coerced_val <<= (CORBA::Float) s;
	break;
      case CORBA::tk_double:
	coerced_val <<= (CORBA::Double) s;
	break;
      case CORBA::tk_longlong:
	coerced_val <<= (CORBA::LongLong) s;
	break;
      case CORBA::tk_ulonglong:
	coerced_val <<= (CORBA::ULongLong) s;
	break;
      case CORBA::tk_long:
	coerced_val <<= (CORBA::Long) s;
	break;
      case CORBA::tk_ulong:
	coerced_val <<= (CORBA::ULong) s;
	break;
      case CORBA::tk_ushort:
	assert( s >= 0 );
	coerced_val <<= (CORBA::UShort) s;
	break;
      default:
	coercion_error( tc, target_type );
      }
      break;
    }
  case CORBA::tk_ushort:
    {
      CORBA::UShort us;
      val >>= us;
      switch( target_type->kind() ) {
      case CORBA::tk_longdouble:
	coerced_val <<= (CORBA::LongDouble) us;
	break;
      case CORBA::tk_float:
	coerced_val <<= (CORBA::Float) us;
	break;
      case CORBA::tk_double:
	coerced_val <<= (CORBA::Double) us;
	break;
      case CORBA::tk_longlong:
	coerced_val <<= (CORBA::LongLong) us;
	break;
      case CORBA::tk_ulonglong:
	coerced_val <<= (CORBA::ULongLong) us;
	break;
      case CORBA::tk_long:
	coerced_val <<= (CORBA::Long) us;
	break;
      case CORBA::tk_ulong:
	coerced_val <<= (CORBA::ULong) us;
	break;
      case CORBA::tk_short:
	assert( us < 32768 );
	coerced_val <<= (CORBA::Short) us;
	break;
      default:
	coercion_error( tc, target_type );
      }
      break;
    }
  default:
    coercion_error( tc, target_type );
  }
  return coerced_val;
}

bool Const::is_floating_pt()
{
  CORBA::TypeCode_var tc = _value.type();
  return tc->kind() == CORBA::tk_float || tc->kind() == CORBA::tk_double ||
    tc->kind() == CORBA::tk_longdouble || tc->kind() == CORBA::tk_fixed;
}

bool Const::is_signed()
{
  CORBA::TypeCode_var tc = _value.type();
  return tc->kind() == CORBA::tk_long || tc->kind() == CORBA::tk_short ||
    tc->kind() == CORBA::tk_longlong;
}

CORBA::Long Const::get_int()
{
  CORBA::Any val;
  val = get_value( CORBA::_tc_long );
  CORBA::Long l;
  CORBA::Boolean r = ( val >>= l );
  assert (r);
  return l;
}

CORBA::Any Const::get_value( CORBA::TypeCode_ptr target_type )
{
  return do_coercion( target_type, _value );
}

CORBA::Any Const::get_any()
{
  return _value;
}

ParseNode* Const::get_parse_node()
{
  return _node;
}

void Const::coercion_error( CORBA::TypeCode_ptr from, CORBA::TypeCode_ptr to )
{
  static char *kind[] = {
    "null",
    "void",
    "short",
    "long",
    "unsigned short",
    "unsigned long",
    "float",
    "double",
    "boolean",
    "char",
    "octet",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "string",
    "",
    "",
    "",
    "",
    "long long",
    "unsigned long long",
    "long double",
    "wchar",
    "wstring",
    "fixed"
  };
 
  if( _node != NULL )
    cerr << IDLError( 10, _node, kind[ from->kind() ], kind[ to->kind() ] ) << endl;
  else
    cerr << IDLError( 10, kind[ from->kind() ], kind[ to->kind() ] ) << endl;
  exit( 1 );
}

Const &Const::operator|( Const &c )
{
  CORBA::Any       a1, a2;
  CORBA::ULongLong c1, c2, r;
  
  a1 = do_coercion( CORBA::_tc_ulonglong, _value );
  a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
  CORBA::Boolean ret = ( a1 >>= c1 );
  assert (ret);
  ret = ( a2 >>= c2 );
  assert (ret);
  r = c1 | c2;
  _value <<= r;
  return *this;
}

Const &Const::operator^( Const &c )
{
  CORBA::Any       a1, a2;
  CORBA::ULongLong c1, c2, r;
  
  a1 = do_coercion( CORBA::_tc_ulonglong, _value );
  a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
  CORBA::Boolean ret = ( a1 >>= c1 );
  assert (ret);
  ret = ( a2 >>= c2 );
  assert (ret);
  r = c1 ^ c2;
  _value <<= r;
  return *this;
}

Const &Const::operator&( Const &c )
{
  CORBA::Any       a1, a2;
  CORBA::ULongLong c1, c2, r;
  
  a1 = do_coercion( CORBA::_tc_ulonglong, _value );
  a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
  CORBA::Boolean ret = ( a1 >>= c1 );
  assert (ret);
  ret = ( a2 >>= c2 );
  assert (ret);
  r = c1 & c2;
  _value <<= r;
  return *this;
}

Const &Const::operator>>( Const &c )
{
  CORBA::Any       a1, a2;
  CORBA::ULongLong c1, c2, r;
  
  a1 = do_coercion( CORBA::_tc_ulonglong, _value );
  a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
  CORBA::Boolean ret = ( a1 >>= c1 );
  assert (ret);
  ret = ( a2 >>= c2 );
  assert (ret);
  r = c1 >> c2;
  _value <<= r;
  return *this;
}

Const &Const::operator<<( Const &c )
{
  CORBA::Any       a1, a2;
  CORBA::ULongLong c1, c2, r;
  
  a1 = do_coercion( CORBA::_tc_ulonglong, _value );
  a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
  CORBA::Boolean ret = ( a1 >>= c1 );
  assert (ret);
  ret = ( a2 >>= c2 );
  assert (ret);
  r = c1 << c2;
  _value <<= r;
  return *this;
}

Const &Const::operator+( Const &c )
{
  CORBA::Any a1, a2;

  if( is_floating_pt() || c.is_floating_pt() ) {
    // Coerce operands to long double
    CORBA::LongDouble ld1, ld2, r;
    a1 = do_coercion( CORBA::_tc_longdouble, _value );
    a2 = do_coercion( CORBA::_tc_longdouble, c._value );
    CORBA::Boolean ret = ( a1 >>= ld1 );
    assert (ret);
    ret = ( a2 >>= ld2 );
    assert (ret);
    r = ld1 + ld2;
    _value <<= r;
  } else {
    if( is_signed() || c.is_signed() ) {
      // Coerce operands to signed long long
      CORBA::LongLong ll1, ll2, r;
      a1 = do_coercion( CORBA::_tc_longlong, _value );
      a2 = do_coercion( CORBA::_tc_longlong, c._value );
      CORBA::Boolean ret = ( a1 >>= ll1 );
      assert (ret);
      ret = ( a2 >>= ll2 );
      assert (ret);
      r = ll1 + ll2;
      _value <<= r;
    } else {
      // Coerce operands to unsigned long long
      CORBA::ULongLong ull1, ull2, r;
      a1 = do_coercion( CORBA::_tc_ulonglong, _value );
      a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
      CORBA::Boolean ret = ( a1 >>= ull1 );
      assert (ret);
      ret = ( a2 >>= ull2 );
      assert (ret);
      r = ull1 + ull2;
      _value <<= r;
    }
  }
  return *this;
}

Const &Const::operator-( Const &c )
{
  CORBA::Any a1, a2;

  if( is_floating_pt() || c.is_floating_pt() ) {
    // Coerce operands to long double
    CORBA::LongDouble ld1, ld2, r;
    a1 = do_coercion( CORBA::_tc_longdouble, _value );
    a2 = do_coercion( CORBA::_tc_longdouble, c._value );
    CORBA::Boolean ret = ( a1 >>= ld1 );
    assert (ret);
    ret = ( a2 >>= ld2 );
    assert (ret);
    r = ld1 - ld2;
    _value <<= r;
  } else {
    if( is_signed() || c.is_signed() ) {
      // Coerce operands to signed long long
      CORBA::LongLong ll1, ll2, r;
      a1 = do_coercion( CORBA::_tc_longlong, _value );
      a2 = do_coercion( CORBA::_tc_longlong, c._value );
      CORBA::Boolean ret = ( a1 >>= ll1 );
      assert (ret);
      ret = ( a2 >>= ll2 );
      assert (ret);
      r = ll1 - ll2;
      _value <<= r;
    } else {
      // Coerce operands to unsigned long long
      CORBA::ULongLong ull1, ull2, r;
      a1 = do_coercion( CORBA::_tc_ulonglong, _value );
      a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
      CORBA::Boolean ret = ( a1 >>= ull1 );
      assert (ret);
      ret = ( a2 >>= ull2 );
      assert (ret);
      r = ull1 - ull2;
      _value <<= r;
    }
  }
  return *this;
}

Const &Const::operator*( Const &c )
{
  CORBA::Any a1, a2;

  if( is_floating_pt() || c.is_floating_pt() ) {
    // Coerce operands to long double
    CORBA::LongDouble ld1, ld2, r;
    a1 = do_coercion( CORBA::_tc_longdouble, _value );
    a2 = do_coercion( CORBA::_tc_longdouble, c._value );
    CORBA::Boolean ret = ( a1 >>= ld1 );
    assert (ret);
    ret = ( a2 >>= ld2 );
    assert (ret);
    r = ld1 * ld2;
    _value <<= r;
  } else {
    if( is_signed() || c.is_signed() ) {
      // Coerce operands to signed long long
      CORBA::LongLong ll1, ll2, r;
      a1 = do_coercion( CORBA::_tc_longlong, _value );
      a2 = do_coercion( CORBA::_tc_longlong, c._value );
      CORBA::Boolean ret = ( a1 >>= ll1 );
      assert (ret);
      ret = ( a2 >>= ll2 );
      assert (ret);
      r = ll1 * ll2;
      _value <<= r;
    } else {
      // Coerce operands to unsigned long long
      CORBA::ULongLong ull1, ull2, r;
      a1 = do_coercion( CORBA::_tc_ulonglong, _value );
      a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
      CORBA::Boolean ret = ( a1 >>= ull1 );
      assert (ret);
      ret = ( a2 >>= ull2 );
      assert (ret);
      r = ull1 * ull2;
      _value <<= r;
    }
  }
  return *this;
}

Const &Const::operator/( Const &c )
{
  CORBA::Any a1, a2;

  if( is_floating_pt() || c.is_floating_pt() ) {
    // Coerce operands to long double
    CORBA::LongDouble ld1, ld2, r;
    a1 = do_coercion( CORBA::_tc_longdouble, _value );
    a2 = do_coercion( CORBA::_tc_longdouble, c._value );
    CORBA::Boolean ret = ( a1 >>= ld1 );
    assert (ret);
    ret = ( a2 >>= ld2 );
    assert (ret);
    assert( ld2 != 0.0 );
    r = ld1 / ld2;
    _value <<= r;
  } else {
    if( is_signed() || c.is_signed() ) {
      // Coerce operands to signed long long
      CORBA::LongLong ll1, ll2, r;
      a1 = do_coercion( CORBA::_tc_longlong, _value );
      a2 = do_coercion( CORBA::_tc_longlong, c._value );
      CORBA::Boolean ret = ( a1 >>= ll1 );
      assert (ret);
      ret = ( a2 >>= ll2 );
      assert (ret);
      assert( ll2 != 0 );
      r = ll1 / ll2;
      _value <<= r;
    } else {
      // Coerce operands to unsigned long long
      CORBA::ULongLong ull1, ull2, r;
      a1 = do_coercion( CORBA::_tc_ulonglong, _value );
      a2 = do_coercion( CORBA::_tc_ulonglong, c._value );
      CORBA::Boolean ret = ( a1 >>= ull1 );
      assert (ret);
      ret = ( a2 >>= ull2 );
      assert (ret);
      assert( ull2 != 0 );
      r = ull1 / ull2;
      _value <<= r;
    }
  }
  return *this;
}

Const &Const::operator%( Const &c )
{
  CORBA::Any      a1, a2;
  CORBA::LongLong c1, c2, r;
  
  a1 = do_coercion( CORBA::_tc_longlong, _value );
  a2 = do_coercion( CORBA::_tc_longlong, c._value );
  CORBA::Boolean ret = ( a1 >>= c1 );
  assert (ret);
  ret = ( a2 >>= c2 );
  assert (ret);
  r = c1 % c2;
  _value <<= r;
  return *this;
}

Const &Const::operator~()
{
  CORBA::Any       a1;
  CORBA::ULongLong c1, r;
  
  a1 = do_coercion( CORBA::_tc_ulonglong, _value );
  CORBA::Boolean ret = ( a1 >>= c1 );
  assert (ret);
  r = ~c1;
  _value <<= r;
  return *this;
}
