/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#ifdef _WINDOWS
#include <string.h>
#endif
#include <iostream.h>
#include <mico/impl.h>
#include <mico/util.h>
#include <mico/template_impl.h>

/*
 * whether or not to allow structural subtyping ala ODP-RM during
 * extraction from any...
 */
#define ALLOW_SUBTYPING


CORBA::Any::Any ()
{
    thetc = _tc_void;
    checker = new TypeCodeChecker ();
    ec = new MICO::SimpleEncoder ();
    dc = new MICO::SimpleDecoder (ec->buffer(), FALSE);
}

CORBA::Any::Any (TypeCode_ptr tc, Buffer *b, Boolean release)
{
    thetc = CORBA::TypeCode::_duplicate (tc);
    checker = new TypeCodeChecker ();
    ec = new MICO::SimpleEncoder (b, FALSE);
    dc = new MICO::SimpleDecoder (b, release);
}

CORBA::Any::Any (TypeCode_ptr tc, DataDecoder *_dc, DataEncoder *_ec)
{
    thetc = CORBA::TypeCode::_duplicate (tc);
    checker = new TypeCodeChecker ();
    ec = _ec;
    dc = _dc;
}

CORBA::Any::Any (const Any &a)
{
    copy (a);
}

CORBA::Any::Any (TypeCode_ptr tc, void *value, Boolean release)
{
    thetc = _tc_void;
    checker = new TypeCodeChecker ();
    ec = new MICO::SimpleEncoder ();
    dc = new MICO::SimpleDecoder (ec->buffer(), FALSE);

    replace (tc, value, release);
}

CORBA::Any::~Any ()
{
    free ();
}

inline CORBA::TypeCode_ptr
CORBA::Any::tc () const
{
    return thetc;
}

inline void
CORBA::Any::tc (TypeCode_ptr ntc)
{
    CORBA::release (thetc);
    thetc = ntc;
}

inline void
CORBA::Any::tc_if_changed (TypeCode_ptr ntc)
{
    if (thetc->equaltype (ntc)) {
	CORBA::release (ntc);
    } else {
	CORBA::release (thetc);
	thetc = ntc;
    }
}

CORBA::Any &
CORBA::Any::operator= (const Any &a)
{
    if (this != &a) {
	free ();
	copy (a);
    }
    return *this;
}

CORBA::Boolean
CORBA::Any::operator== (const Any &_a) const
{
    if (this == &_a)
	return TRUE;

    Any &me = (Any &)*this;
    Any &he = (Any &)_a;

    if (strcmp (me.ec->type(), he.ec->type()) ||
	strcmp (me.dc->type(), he.dc->type()))
	// different codec types, do a structural compare
	return equivalent (_a);

    if (!me.tc()->equaltype (he.tc()))
	return FALSE;

    return (me.length() == he.length() &&
	    !memcmp (me.value(), he.value(), me.length()));
}

CORBA::Boolean
CORBA::Any::equivalent (const Any &_a) const
{
    if (this == &_a)
	return TRUE;

    Any &me = (Any &)*this;
    Any &he = (Any &)_a;

    /*
     * XXX ugly, because we must not change the any's we are
     *     forced to use other TypeCodeCheckers and preserve
     *     the buffer positions...
     */
    TypeCodeChecker *tcc1 = me.checker;
    ULong pos1 = me.ec->buffer()->rpos();
    TypeCodeChecker ntcc1 (me.tc()); 
    me.checker = &ntcc1;

    TypeCodeChecker *tcc2 = he.checker;
    ULong pos2 = he.ec->buffer()->rpos();
    TypeCodeChecker ntcc2 (he.tc()); 
    he.checker = &ntcc2;

    me.prepare_read ();
    he.prepare_read ();

    Boolean retval = me.compare_any (he);

    me.checker = tcc1;
    me.ec->buffer()->rseek_beg (pos1);

    he.checker = tcc2;
    he.ec->buffer()->rseek_beg (pos2);

    return retval;
}

void
CORBA::Any::free ()
{
    delete ec;
    delete dc;
    delete checker;
    CORBA::release (thetc);
}

void
CORBA::Any::copy (const Any &a)
{
    thetc = TypeCode::_duplicate (a.thetc);
    checker = new TypeCodeChecker ();
    ec = a.ec->clone ();
    dc = a.dc->clone (ec->buffer(), FALSE);
}

void
CORBA::Any::prepare_write ()
{
    if (checker->completed()) {
	ec->buffer()->reset();
	checker->restart (tc());
    }
}

void
CORBA::Any::prepare_read ()
{
    if (checker->completed()) {
	dc->buffer()->rseek_beg (0);
	checker->restart (tc());
    }
}

CORBA::Boolean
CORBA::Any::copy_any (Any &a, Boolean recurse)
{
    TypeCode_ptr atc = a.checker->tc()->unalias();

    switch (atc->kind()) {
    case tk_null:
    case tk_void:
	return a.checker->basic (atc) && checker->basic (atc);

    case tk_short: {
	Short s;
	return (a >>= s) && (*this <<= s);
    }
    case tk_long: {
	Long l;
	return (a >>= l) && (*this <<= l);
    }
    case tk_longlong: {
	LongLong l;
	return (a >>= l) && (*this <<= l);
    }
    case tk_ushort: {
	UShort us;
	return (a >>= us) && (*this <<= us);
    }
    case tk_ulong: {
	ULong ul;
	return (a >>= ul) && (*this <<= ul);
    }
    case tk_ulonglong: {
	ULongLong ul;
	return (a >>= ul) && (*this <<= ul);
    }
    case tk_float: {
	Float f;
	return (a >>= f) && (*this <<= f);
    }
    case tk_double: {
	Double d;
	return (a >>= d) && (*this <<= d);
    }
    case tk_longdouble: {
	LongDouble d;
	return (a >>= d) && (*this <<= d);
    }
    case tk_boolean: {
	Boolean b;
	return (a >>= to_boolean (b)) && (*this <<= from_boolean (b));
    }
    case tk_char: {
	Char c;
	return (a >>= to_char (c)) && (*this <<= from_char (c));
    }
    case tk_wchar: {
	WChar c;
	return (a >>= to_wchar (c)) && (*this <<= from_wchar (c));
    }
    case tk_octet: {
	Octet o;
	return (a >>= to_octet (o)) && (*this <<= from_octet (o));
    }
    case tk_any: {
	Any any;
	return ((a >>= any) && (*this <<= any));
    }
    case tk_objref: {
	Object_var obj;
	String_var name;
	return ((a >>= to_object (obj, &(char *&)name)) &&
		(*this <<= from_object (obj, name)));
    }
    case tk_enum: {
	ULong en;
	return a.enum_get (en) && enum_put (en);
	break;
    }
    case tk_string: {
	String_var thestring;
	return ((a >>= to_string (thestring, atc->length())) &&
		(*this <<= from_string (thestring, atc->length())));
    }
    case tk_wstring: {
	WString_var thestring;
	return ((a >>= to_wstring (thestring, atc->length())) &&
		(*this <<= from_wstring (thestring, atc->length())));
    }
    case tk_fixed: {
        FixedBase::FixedValue_var val;
        UShort digits = atc->fixed_digits();
        Short scale = atc->fixed_scale();
	return ((a >>= to_fixed (val, digits, scale)) &&
		(*this <<= from_fixed (val, digits, scale)));
    }
    case tk_alias:
	assert (0);

    case tk_TypeCode: {
	TypeCode_var t;
	return ((a >>= t) && (*this <<= t));
    }
    case tk_Principal: {
	Principal_var p;
	return ((a >>= p) && (*this <<= p));
    }
    default:
	break;
    }
    if (!recurse)
	return FALSE;

    switch (atc->kind()) {
    case tk_except: {
	String_var repoid;
	if (!a.except_get_begin (repoid) || !except_put_begin (repoid))
	    return FALSE;
	for (CORBA::ULong i = 0; i < atc->member_count(); ++i) {
	    if (!copy_any (a))
		return FALSE;
	}
	return a.except_get_end() && except_put_end();
    }
    case tk_struct: {
	if (!a.struct_get_begin() || !struct_put_begin())
	    return FALSE;
	for (CORBA::ULong i = 0; i < atc->member_count(); ++i) {
	    if (!copy_any (a))
		return FALSE;
	}
	return a.struct_get_end() && struct_put_end();
    }
    case tk_union: {
	if (!a.union_get_begin() || !union_put_begin())
	    return FALSE;
	Any adisc;
	if (!a.any_get (adisc, FALSE))
	    return FALSE;
	adisc.prepare_read ();
	if (!copy_any (adisc, FALSE))
	    return FALSE;
	Long i = atc->member_index (adisc);
	if (i >= 0) {
	    if (!a.union_get_selection (i) || !union_put_selection (i))
		return FALSE;
	    if (!copy_any (a))
		return FALSE;
	}
	return a.union_get_end() && union_put_end();
    }
    case tk_sequence: {
	ULong seqlen;
	if (!a.seq_get_begin (seqlen) || !seq_put_begin (seqlen))
	    return FALSE;
	for (CORBA::ULong i = 0; i < seqlen; ++i) {
	    if (!copy_any (a))
		return FALSE;
	}
	return a.seq_get_end() && seq_put_end();
    }
    case tk_array: {
	ULong arrlen = atc->length();
	if (!a.array_get_begin() || !array_put_begin())
	    return FALSE;
	for (CORBA::ULong i = 0; i < arrlen; ++i) {
	    if (!copy_any (a))
		return FALSE;
	}
	return a.array_get_end() && array_put_end();
    }
    default:
	assert (0);
    }
    return FALSE;
}

CORBA::Boolean
CORBA::Any::compare_any (Any &a)
{
    TypeCode_ptr atc = a.checker->tc()->unalias();
    TypeCode_ptr mtc = checker->tc()->unalias();

    /*
     * TypeCode::max() returns the "maximum type" of the two typecodes
     * if both typecodes are arithmetic types. Otherwise it returns
     * atc.
     */
    switch (atc->maxtype(mtc)->kind()) {
    case tk_null:
    case tk_void:
	return a.checker->basic (atc) && checker->basic (atc);

    case tk_short: {
	Short s1, s2;
	return (a >>= s1) && (*this >>= s2) && s1 == s2;
    }
    case tk_long: {
	Long l1, l2;
	return (a >>= l1) && (*this >>= l2) && l1 == l2;
    }
    case tk_longlong: {
	LongLong l1, l2;
	return (a >>= l1) && (*this >>= l2) && l1 == l2;
    }
    case tk_ushort: {
	UShort us1, us2;
	return (a >>= us1) && (*this >>= us2) && us1 == us2;
    }
    case tk_ulong: {
	ULong ul1, ul2;
	return (a >>= ul1) && (*this >>= ul2) && ul1 == ul2;
    }
    case tk_ulonglong: {
	ULongLong ul1, ul2;
	return (a >>= ul1) && (*this >>= ul2) && ul1 == ul2;
    }
    case tk_float: {
	Float f1, f2;
	return (a >>= f1) && (*this >>= f2) && f1 == f2;
    }
    case tk_double: {
	Double d1, d2;
	return (a >>= d1) && (*this >>= d2) && d1 == d2;
    }
    case tk_longdouble: {
	LongDouble d1, d2;
	return (a >>= d1) && (*this >>= d2) && d1 == d2;
    }
    case tk_boolean: {
	Boolean b1, b2;
	return (a >>= to_boolean (b1)) && (*this >>= to_boolean (b2))
	    && b1 == b2;
    }
    case tk_char: {
	Char c1, c2;
	return (a >>= to_char (c1)) && (*this >>= to_char (c2)) && c1 == c2;
    }
    case tk_wchar: {
	WChar c1, c2;
	return (a >>= to_wchar (c1)) && (*this >>= to_wchar (c2)) && c1 == c2;
    }
    case tk_octet: {
	Octet o1, o2;
	return (a >>= to_octet (o1)) && (*this >>= to_octet (o2)) && o1 == o2;
    }
    case tk_any: {
	Any any1, any2;
	return (a >>= any1) && (*this >>= any2) && any1 == any2;
    }
    case tk_objref: {
	Object_var objptr1, objptr2;
	// XXX is Object::is_equivalent() appropriate ??
	return ((a >>= to_object (objptr1)) &&
		(*this >>= to_object (objptr2)) &&
		objptr1->_is_equivalent (objptr2));
    }
    case tk_enum: {
	ULong en1, en2;
	return a.enum_get (en1) && enum_get (en2) && en1 == en2;
	break;
    }
    case tk_string: {
	String_var str1, str2;
	return ((a >>= to_string (str1, atc->length())) &&
		(*this >>= to_string (str2, mtc->length())) &&
		!strcmp (str1, str2));
    }
    case tk_wstring: {
	WString_var str1, str2;
	return ((a >>= to_wstring (str1, atc->length())) &&
		(*this >>= to_wstring (str2, mtc->length())) &&
		!xwcscmp (str1, str2));
    }
    case tk_fixed: {
	FixedBase::FixedValue_var val1, val2;
        UShort digits = atc->fixed_digits();
        Short scale = atc->fixed_scale();
	return ((a >>= to_fixed (val1, digits, scale)) &&
		(*this >>= to_fixed (val2, digits, scale)) &&
                *val1 == *val2);
    }
    case tk_alias:
	assert (0);

    case tk_TypeCode: {
	TypeCode_ptr t1, t2;
	return ((a >>= t1) && (*this >>= t2) && t1->equal (t2));
    }
    case tk_Principal: {
	Principal_ptr p1, p2;
	return ((a >>= p1) && (*this >>= p2) && *p1 == *p2);
    }
    case tk_except: {
	String_var repoid1, repoid2;
	if (!a.except_get_begin (repoid1) || !except_get_begin (repoid2) ||
	    strcmp (repoid1, repoid2))
	    return FALSE;
	if (atc->member_count() != mtc->member_count())
	    return FALSE;
	for (CORBA::ULong i = 0; i < atc->member_count(); ++i) {
	    if (!compare_any (a))
		return FALSE;
	}
	return a.except_get_end() && except_get_end();
    }
    case tk_struct: {
	if (!a.struct_get_begin() || !struct_get_begin())
	    return FALSE;
	if (atc->member_count() != mtc->member_count())
	    return FALSE;
	for (CORBA::ULong i = 0; i < atc->member_count(); ++i) {
	    if (!compare_any (a))
		return FALSE;
	}
	return a.struct_get_end() && struct_get_end();
    }
    case tk_union: {
	if (!a.union_get_begin() || !union_get_begin())
	    return FALSE;
	Any adisc, mdisc;
	if (!a.any_get (adisc, FALSE))
	    return FALSE;
	if (!any_get (mdisc, FALSE))
	    return FALSE;
	if (!adisc.equivalent (mdisc))
	    return FALSE;
	Long i = atc->member_index (adisc);
	if (i != mtc->member_index (mdisc))
	    return FALSE;
	if (i >= 0) {
	    if (!a.union_get_selection (i) || !union_get_selection (i))
		return FALSE;
	    if (!compare_any (a))
		return FALSE;
	}
	return a.union_get_end() && union_get_end();
    }
    case tk_sequence: {
	ULong seqlen1, seqlen2;
	if (!a.seq_get_begin (seqlen1) || !seq_get_begin (seqlen2))
	    return FALSE;
	if (seqlen1 != seqlen2)
	    return FALSE;
	for (CORBA::ULong i = 0; i < seqlen1; ++i) {
	    if (!compare_any (a))
		return FALSE;
	}
	return a.seq_get_end() && seq_get_end();
    }
    case tk_array: {
	ULong arrlen1 = atc->length();
	ULong arrlen2 = mtc->length();
	if (arrlen1 != arrlen2)
	    return FALSE;
	if (!a.array_get_begin() || !array_get_begin())
	    return FALSE;
	for (CORBA::ULong i = 0; i < arrlen1; ++i) {
	    if (!compare_any (a))
		return FALSE;
	}
	return a.array_get_end() && array_get_end();
    }
    default:
	assert (0);
    }
    return FALSE;
}

void
CORBA::Any::replace (TypeCode_ptr tc, void *value, Boolean release)
{
    reset ();
    type (tc);

    if (value) {
	Buffer b (value);
	Any a (tc, &b, FALSE);

	prepare_write ();
	a.prepare_read ();
	CORBA::Boolean r = copy_any (a);
	assert (r);

	if (release)
	    // XXX ???
	    delete[] (CORBA::Octet *)value;
    }
}

CORBA::TypeCode_ptr
CORBA::Any::type () const
{
    return CORBA::TypeCode::_duplicate (tc());
}

void
CORBA::Any::type (TypeCode_ptr t)
{
    if (checker->level_count() == 0) {
        ec->buffer()->reset ();
        tc (CORBA::TypeCode::_duplicate (t));
        checker->restart (tc());

        switch (t->unalias()->kind()) {
        case CORBA::tk_void:
        case CORBA::tk_null:
            /*
             * make sure checker is completed, otherwise marshal()
             * will fail.
             */
            checker->finish();
            break;

	default:
	    break;
        }
    }
}

void
CORBA::Any::reset ()
{
    ec->buffer()->reset ();
    checker->restart ();
}

void
CORBA::Any::rewind ()
{
    dc->buffer()->rseek_beg (0);
    checker->restart ();
}

const void *
CORBA::Any::value () const
{
    return ec->buffer()->buffer();
}

CORBA::ULong
CORBA::Any::length () const
{
    return ec->buffer()->wpos();
}

void
CORBA::Any::encode (DataEncoder &e) const
{
    e.put_typecode (*tc());
    ((Any *)this)->marshal (e);
}

CORBA::Boolean
CORBA::Any::decode (DataDecoder &e)
{
    TypeCode_ptr t = TypeCode::create_basic_tc (tk_null);
    CORBA::Boolean r = e.get_typecode (*t);
    assert (r);
    r = demarshal (t, e);
    assert (r);
    CORBA::release (t);
    return TRUE;
}

void
CORBA::Any::marshal (DataEncoder &e)
{
    assert (checker->completed());
    Any a (tc(), new MICO::SimpleDecoder,
	   e.clone (e.buffer(), FALSE,
		    e.converter(), FALSE,
		    e.wconverter(), FALSE));

    // cannot use a.prepare_write() since that will clear the buffer ...
    a.checker->restart (a.tc());
    prepare_read ();
    CORBA::Boolean r = a.copy_any (*this);
    assert (r);
}

CORBA::Boolean
CORBA::Any::demarshal (TypeCode_ptr t, DataDecoder &d)
{
    reset ();
    type (t);
    Any a (t, d.clone (d.buffer(), FALSE,
		       d.converter(), FALSE,
		       d.wconverter(), FALSE),
	   new MICO::SimpleEncoder);

    // cannot use a.pepare_read() ...
    a.checker->restart (a.tc());
    prepare_write ();
    return copy_any (a);
}

void
CORBA::Any::from_static_any (TypeCode_ptr t, const StaticAny &sa)
{
    reset ();
    type (t);

    prepare_write ();
    ((StaticAny &)sa).marshal (*ec);
    checker->finish ();
}

CORBA::Boolean
CORBA::Any::to_static_any (StaticAny &sa) const
{
    Any &me = (Any &)*this;

    assert (me.checker->completed());
    me.prepare_read ();

    if (!sa.demarshal (*me.dc)) {
	me.rewind ();
	return FALSE;
    }
    me.checker->finish ();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (Short s)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_short);
    } else if (!checker->basic (_tc_short)) {
	reset ();
	return FALSE;
    }
    ec->put_short (s);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (Long l)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_long);
    } else if (!checker->basic (_tc_long)) {
	reset ();
	return FALSE;
    }
    ec->put_long (l);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (LongLong l)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_longlong);
    } else if (!checker->basic (_tc_longlong)) {
	reset ();
	return FALSE;
    }
    ec->put_longlong (l);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (UShort us)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_ushort);
    } else if (!checker->basic (_tc_ushort)) {
	reset ();
	return FALSE;
    }
    ec->put_ushort (us);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (ULong ul)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_ulong);
    } else if (!checker->basic (_tc_ulong)) {
	reset ();
	return FALSE;
    }
    ec->put_ulong (ul);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (ULongLong ul)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_ulonglong);
    } else if (!checker->basic (_tc_ulonglong)) {
	reset ();
	return FALSE;
    }
    ec->put_ulonglong (ul);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (Float f)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_float);
    } else if (!checker->basic (_tc_float)) {
	reset ();
	return FALSE;
    }
    ec->put_float (f);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (Double d)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_double);
    } else if (!checker->basic (_tc_double)) {
	reset ();
	return FALSE;
    }
    ec->put_double (d);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (LongDouble d)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_longdouble);
    } else if (!checker->basic (_tc_longdouble)) {
	reset ();
	return FALSE;
    }
    ec->put_longdouble (d);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (const Any &a)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_any);
    } else if (!checker->basic (_tc_any)) {
	reset ();
	return FALSE;
    }
    ec->put_any (a);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (const char *s)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_string);
    } else if (!checker->basic (_tc_string)) {
	reset ();
	return FALSE;
    }
    ec->put_string (s ? s : "");
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (const wchar_t *s)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_wstring);
    } else if (!checker->basic (_tc_wstring)) {
	reset ();
	return FALSE;
    }
    ec->put_wstring (s ? s : L"");
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (from_boolean b)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_boolean);
    } else if (!checker->basic (_tc_boolean)) {
	reset ();
	return FALSE;
    }
    ec->put_boolean (b.val);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (from_octet o)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_octet);
    } else if (!checker->basic (_tc_octet)) {
	reset ();
	return FALSE;
    }
    ec->put_octet (o.val);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (from_char c)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_char);
    } else if (!checker->basic (_tc_char)) {
	reset ();
	return FALSE;
    }
    ec->put_char (c.val);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (from_wchar c)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_wchar);
    } else if (!checker->basic (_tc_wchar)) {
	reset ();
	return FALSE;
    }
    ec->put_wchar (c.val);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (from_string s)
{
    if (s.bound > 0 && s.val && strlen (s.val) > s.bound)
        return FALSE;
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (TypeCode::create_string_tc (s.bound));
    } else {
	TypeCode_var tmp = TypeCode::create_string_tc (s.bound);
        if (!checker->basic (tmp)) {
            reset ();
            return FALSE;
        }
    }
    ec->put_string (s.val ? s.val : "");
    if (s.nocopy)
	CORBA::string_free (s.val);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (from_wstring s)
{
    if (s.bound > 0 && s.val && xwcslen (s.val) > s.bound)
        return FALSE;
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (TypeCode::create_wstring_tc (s.bound));
    } else {
	TypeCode_var tmp = TypeCode::create_wstring_tc (s.bound);
        if (!checker->basic (tmp)) {
            reset ();
            return FALSE;
        }
    }
    ec->put_wstring (s.val ? s.val : L"");
    if (s.nocopy)
	CORBA::wstring_free (s.val);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (from_fixed f)
{
    if ((CORBA::ULong)f.digits+1 != f.value.length())
        return FALSE;
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (TypeCode::create_fixed_tc (f.digits, f.scale));
    } else {
	TypeCode_var tmp = TypeCode::create_fixed_tc (f.digits, f.scale);
        if (!checker->basic (tmp)) {
            reset ();
            return FALSE;
        }
    }
    ec->put_fixed (f.value, f.digits, f.scale);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (from_object o)
{
    IOR *ior;
    const char *repoid;
    if (CORBA::is_nil (o.val)) {
        ior = new IOR;
	repoid = "";
    } else {
        ior = o.val->_ior();
	repoid = ior->objid();
    }
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (TypeCode::create_interface_tc (repoid, o.name));
    } else if (checker->tc()->kind() != tk_objref ||
	       !checker->basic (checker->tc())) {
	// XXX should compare checker->tc()->id() and repoid somehow ...
	if (CORBA::is_nil (o.val))
	    delete ior;
	reset ();
	return FALSE;
    }

    ec->put_ior (*ior);
    if (CORBA::is_nil (o.val))
        delete ior;
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (TypeCode_ptr t)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_TypeCode);
    } else if (!checker->basic (_tc_TypeCode)) {
	reset ();
	return FALSE;
    }
    ec->put_typecode (*t);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (Context_ptr ctx)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_Context);
    } else if (!checker->basic (_tc_Context)) {
	reset ();
	return FALSE;
    }
    ec->put_context (*ctx);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator<<= (Principal_ptr pr)
{
    if (checker->completed()) {
	ec->buffer()->reset();
	tc_if_changed (_tc_Principal);
    } else if (!checker->basic (_tc_Principal)) {
	reset ();
	return FALSE;
    }
    ec->put_principal (*pr);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_short (Short &s)
{
    return (checker->basic (_tc_short) && dc->get_short (s));
}

CORBA::Boolean
CORBA::Any::operator>>= (Short &s) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_short (s)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_long (Long &l)
{
    if (checker->basic (_tc_long)) {
	return dc->get_long (l);
    }
#ifdef ALLOW_SUBTYPING
    else {
	Short s;
	UShort us;
	if (get_short (s)) {
	    l = s;
	    return TRUE;
	} else if (get_ushort (us)) {
	    l = us;
	    return TRUE;
	}
    }
#endif
    return FALSE;
}

CORBA::Boolean
CORBA::Any::operator>>= (Long &l) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_long (l)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_longlong (LongLong &l)
{
    if (checker->basic (_tc_longlong)) {
	return dc->get_longlong (l);
    }
#ifdef ALLOW_SUBTYPING
    else {
	Long _l;
	ULong _ul;
	if (get_long (_l)) {
	    l = _l;
	    return TRUE;
	} else if (get_ulong (_ul)) {
	    l = _ul;
	    return TRUE;
	}
    }
#endif
    return FALSE;
}

CORBA::Boolean
CORBA::Any::operator>>= (LongLong &l) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_longlong (l)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_ushort (UShort &us)
{
    return (checker->basic (_tc_ushort) && dc->get_ushort (us));
}

CORBA::Boolean
CORBA::Any::operator>>= (UShort &us) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_ushort (us)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_ulong (ULong &ul)
{
    if (checker->basic (_tc_ulong)) {
	return dc->get_ulong (ul);
    }
#ifdef ALLOW_SUBTYPING
    else {
	UShort us;
	if (get_ushort (us)) {
	    ul = us;
	    return TRUE;
	}
    }
#endif
    return FALSE;
}

CORBA::Boolean
CORBA::Any::operator>>= (ULong &ul) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_ulong (ul)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_ulonglong (ULongLong &ul)
{
    if (checker->basic (_tc_ulonglong)) {
	return dc->get_ulonglong (ul);
    }
#ifdef ALLOW_SUBTYPING
    else {
	ULong _ul;
	if (get_ulong (_ul)) {
	    ul = _ul;
	    return TRUE;
	}
    }
#endif
    return FALSE;
}

CORBA::Boolean
CORBA::Any::operator>>= (ULongLong &ul) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_ulonglong (ul)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_float (Float &f)
{
    if (checker->basic (_tc_float)) {
	return dc->get_float (f);
    }
#ifdef ALLOW_SUBTYPING
    else {
	Short s;
	UShort us;
	if (get_short (s)) {
	    f = s;
	    return TRUE;
	} else if (get_ushort (us)) {
	    f = us;
	    return TRUE;
	}
    }
#endif
    return FALSE;
}

CORBA::Boolean
CORBA::Any::operator>>= (Float &f) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_float (f)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_double (Double &d)
{
    if (checker->basic (_tc_double)) {
	return dc->get_double (d);
    }
#ifdef ALLOW_SUBTYPING
    else {
	Long l;
	ULong ul;
	Float f;
	if (get_long (l)) {
	    d = l;
	    return TRUE;
	} else if (get_ulong (ul)) {
	    d = ul;
	    return TRUE;
	} else if (get_float (f)) {
	    d = f;
	    return TRUE;
	}
    }
#endif
    return FALSE;
}

CORBA::Boolean
CORBA::Any::operator>>= (Double &d) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_double (d)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::get_longdouble (LongDouble &d)
{
    if (checker->basic (_tc_longdouble)) {
	return dc->get_longdouble (d);
    }
#ifdef ALLOW_SUBTYPING
    else {
	LongLong _l;
	ULongLong _ul;
	Double _d;
	if (get_longlong (_l)) {
	    d = _l;
	    return TRUE;
	} else if (get_ulonglong (_ul)) {
#ifdef _WINDOWS
	    d = __int64( _ul );
#else
	    d = _ul;
#endif
	    return TRUE;
	} else if (get_double (_d)) {
	    d = _d;
	    return TRUE;
	}
    }
#endif
    return FALSE;
}

CORBA::Boolean
CORBA::Any::operator>>= (LongDouble &d) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.get_longdouble (d)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (Any &a) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.checker->basic (_tc_any) || !me.dc->get_any (a)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (char * &s) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    TypeCode_var tmp = TypeCode::create_string_tc (0);
    if (!me.checker->basic (tmp) || !me.dc->get_string (s)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (wchar_t * &s) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    TypeCode_var tmp = TypeCode::create_wstring_tc (0);
    if (!me.checker->basic (tmp) || !me.dc->get_wstring (s)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (to_boolean b) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.checker->basic (_tc_boolean) || !me.dc->get_boolean (b.ref)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (to_octet o) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.checker->basic (_tc_octet) || !me.dc->get_octet (o.ref)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (to_char c) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.checker->basic (_tc_char) || !me.dc->get_char (c.ref)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (to_wchar c) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.checker->basic (_tc_wchar) || !me.dc->get_wchar (c.ref)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (to_string s) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    TypeCode_var tmp = TypeCode::create_string_tc (s.bound);
    if (!me.checker->basic (tmp) ||
	!me.dc->get_string (s.ref)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (to_wstring s) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    TypeCode_var tmp = TypeCode::create_wstring_tc (s.bound);
    if (!me.checker->basic (tmp) ||
	!me.dc->get_wstring (s.ref)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (to_fixed f) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    TypeCode_var tmp = TypeCode::create_fixed_tc (f.digits, f.scale);
    FixedBase::FixedValue_var val = new FixedBase::FixedValue;
    if (!me.checker->basic (tmp) ||
	!me.dc->get_fixed (*(FixedBase::FixedValue *)val, f.digits, f.scale)) {
	me.rewind ();
	return FALSE;
    }
    f.value = val._retn();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (to_object o) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    IOR *ior = new IOR;

    if (me.checker->tc()->kind () != tk_objref || !me.dc->get_ior (*ior)) {
	delete ior;
	me.rewind ();
	return FALSE;
    }
    // XXX should check for zero number of profiles, says [10-16]
    if (strlen (ior->objid()) == 0) {
        o.ref = CORBA::Object::_nil();
        delete ior;
    } else {
        ORB_var orb = ORB_instance ("mico-local-orb");
        o.ref = orb->ior_to_object (ior);
    }
    if (o.name)
	*o.name = CORBA::string_dup (me.checker->tc()->name());
    // XXX should compare checker->tc()->id() and repoid somehow ...
    me.checker->basic (checker->tc());
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (TypeCode_ptr &t) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    t = TypeCode::create_basic_tc (tk_null);

    if (!me.checker->basic (_tc_TypeCode) || !me.dc->get_typecode (*t)) {
	delete t;
	me.rewind ();
        return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (Context_ptr &ctx) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    ctx = new Context ("");

    if (!me.checker->basic (_tc_Context) || !me.dc->get_context (*ctx)) {
	delete ctx;
	me.rewind ();
        return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::operator>>= (Principal_ptr &pr) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    pr = new Principal;

    if (!me.checker->basic (_tc_Principal) || !me.dc->get_principal (*pr)) {
	delete pr;
	me.rewind ();
        return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::struct_put_begin ()
{
    prepare_write ();
    if (!checker->struct_begin()) {
	reset ();
	return FALSE;
    }
    ec->struct_begin ();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::struct_put_end ()
{
    if (!checker->struct_end()) {
	reset ();
	return FALSE;
    }
    ec->struct_end ();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::union_put_begin ()
{
    prepare_write ();
    if (!checker->union_begin()) {
	reset ();
	return FALSE;
    }
    ec->union_begin ();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::union_put_selection (Long idx)
{
    if (!checker->union_selection (idx)) {
	reset ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::union_put_end ()
{
    if (!checker->union_end()) {
	reset ();
	return FALSE;
    }
    ec->union_end();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::enum_put (ULong val)
{
    prepare_write ();
    if (!checker->enumeration (val)) {
	reset ();
	return FALSE;
    }
    ec->enumeration (val);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::any_put (Any &a, Boolean recurse)
{
    prepare_write ();
    a.prepare_read ();
    if (!copy_any (a, recurse)) {
	reset ();
	a.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::seq_put_begin (ULong len)
{
    prepare_write ();
    if (!checker->seq_begin (len)) {
	reset ();
	return FALSE;
    }
    ec->seq_begin (len);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::seq_put_end ()
{
    if (!checker->seq_end()) {
	reset ();
	return FALSE;
    }
    ec->seq_end();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::array_put_begin ()
{
    prepare_write ();
    if (!checker->arr_begin()) {
	reset ();
	return FALSE;
    }
    ec->arr_begin ();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::array_put_end ()
{
    if (!checker->arr_end()) {
	reset ();
	return FALSE;
    }
    ec->arr_end();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::except_put_begin (const char *repoid)
{
    prepare_write ();
    if (!checker->except_begin()) {
	reset ();
	return FALSE;
    }
    ec->except_begin (repoid);
    return TRUE;
}

CORBA::Boolean
CORBA::Any::except_put_end ()
{
    if (!checker->except_end()) {
	reset ();
	return FALSE;
    }
    ec->except_end();
    return TRUE;
}

CORBA::Boolean
CORBA::Any::struct_get_begin () const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.checker->struct_begin() || !me.dc->struct_begin ()) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::struct_get_end () const
{
    Any &me = (Any &)*this;

#ifdef ALLOW_SUBTYPING
    if (me.checker->struct_end()) {
	if (!me.dc->struct_end()) {
	    me.rewind ();
	    return FALSE;
	}
	return TRUE;
    }
    if (!me.checker->inside (tk_struct)) {
	me.rewind ();
	return FALSE;
    }
    Any a;
    while (!me.checker->level_finished ()) {
	CORBA::Boolean r = any_get (a);
	assert (r);
    }
#endif
    if (!me.checker->struct_end() || !me.dc->struct_end()) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::union_get_begin () const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.checker->union_begin() || !me.dc->union_begin()) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::union_get_selection (Long idx) const
{
    Any &me = (Any &)*this;

    if (!me.checker->union_selection (idx)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::union_get_end () const
{
    Any &me = (Any &)*this;

    if (!me.checker->union_end() || !me.dc->union_end()) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::enum_get (ULong &val) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.dc->enumeration (val) || !me.checker->enumeration (val)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::any_get (Any &a, Boolean recurse) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    a.type (checker->tc());
    a.prepare_write ();
    if (!a.copy_any (me, recurse)) {
	me.rewind ();
	a.reset ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::seq_get_begin (ULong &len) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.dc->seq_begin (len) || !me.checker->seq_begin (len)) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::seq_get_end () const
{
    Any &me = (Any &)*this;

    if (!me.checker->seq_end() || !me.dc->seq_end()) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::array_get_begin () const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    if (!me.checker->arr_begin() || !me.dc->arr_begin()) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::array_get_end () const
{
    Any &me = (Any &)*this;

    if (!me.checker->arr_end() || !me.dc->arr_end()) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Any::except_get_begin (char *&repoid) const
{
    Any &me = (Any &)*this;

    me.prepare_read ();
    string s;
    if (!me.checker->except_begin() || !me.dc->except_begin (s)) {
	me.rewind ();
	return FALSE;
    }
    repoid = CORBA::string_dup (s.c_str());
    return TRUE;
}

CORBA::Boolean
CORBA::Any::except_get_end () const
{
    Any &me = (Any &)*this;

#ifdef ALLOW_SUBTYPING
    if (me.checker->except_end()) {
	if (!dc->except_end()) {
	    me.rewind ();
	    return FALSE;
	}
	return TRUE;
    }
    if (!me.checker->inside (tk_except)) {
	me.rewind ();
	return FALSE;
    }
    Any a;
    while (!me.checker->level_finished()) {
	CORBA::Boolean r = any_get (a);
        assert (r);
    }
#endif
    if (!me.checker->except_end() || !me.dc->except_end()) {
	me.rewind ();
	return FALSE;
    }
    return TRUE;
}
