/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <iomanip.h>
#include <mico/template_impl.h>

CORBA::Buffer::Buffer (void *b)
{
    // readonly buffer with given contents
    len = wptr = 0x7fffffff;
    rptr = 0;
    ralignbase = walignbase = 0;
    buf = (Octet *)b;
    readonly = TRUE;
}

CORBA::Buffer::Buffer (ULong sz)
{
    // read/write buffer with given initial size
    if (sz < MINSIZE)
        sz = MINSIZE;
    buf = alloc (sz);
    len = sz;
    rptr = wptr = 0;
    ralignbase = walignbase = 0;
    readonly = FALSE;
}

CORBA::Buffer::Buffer (const Buffer &b)
{
    buf = alloc (b.len);
    memcpy (buf, b.buf, b.len);
    len = b.len;
    rptr = b.rptr;
    wptr = b.wptr;
    ralignbase = b.ralignbase;
    walignbase = b.walignbase;
    readonly = FALSE;
}

CORBA::Buffer::~Buffer ()
{
    if (!readonly)
        free (buf);
}

CORBA::Buffer &
CORBA::Buffer::operator= (const Buffer &b)
{
    if (this != &b) {
	assert (!readonly && !b.readonly);
	free (buf);
	buf = alloc (b.len);
	memcpy (buf, b.buf, b.len);
	len = b.len;
	rptr = b.rptr;
	wptr = b.wptr;
	ralignbase = b.ralignbase;
	walignbase = b.walignbase;
    }
    return *this;
}

CORBA::Boolean
CORBA::Buffer::operator== (const Buffer &b)
{
    assert (!readonly && !b.readonly);
    return length() == b.length() && !memcmp (data(), b.data(), length());
}

CORBA::Octet *
CORBA::Buffer::alloc (ULong sz)
{
    Octet *p = (Octet *)::malloc (sz);
    assert (p);
    return p;
}

CORBA::Octet *
CORBA::Buffer::realloc (Octet *b, ULong osz, ULong nsz)
{
    Octet *p = (Octet *)::realloc ((void *)b, nsz);
    assert (p);
    return p;
}

void
CORBA::Buffer::free (Octet *b)
{
    ::free ((void *)b);
}

void
CORBA::Buffer::reset (ULong sz)
{
    rptr = 0;
    ralignbase = walignbase = 0;
    if (!readonly) {
        wptr = 0;
        if (sz < MINSIZE)
            sz = MINSIZE;
        if (len > sz) {
	    free (buf);
            buf = alloc (sz);
            len = sz;
        }
    }
}

void
CORBA::Buffer::resize (ULong needed)
{
    assert (!readonly);
    if (wptr + needed > len) {
        ULong nlen = (len < RESIZE_THRESH)
            ? (2*len)
            : (len + RESIZE_INCREMENT);
        if (wptr + needed > nlen)
            nlen = wptr + needed;
	buf = realloc (buf, len, nlen);
        len = nlen;
    }
}

CORBA::Boolean
CORBA::Buffer::ralign (ULong modulo)
{
    assert (modulo >= 1);
    assert (rptr >= ralignbase);
    ULong r = (rptr - ralignbase) + modulo - 1;
    r -= r % modulo;
    r += ralignbase;
    if (r > wptr)
	return FALSE;
    rptr = r;
    return TRUE;
}

CORBA::Boolean
CORBA::Buffer::peek (void *b, ULong blen)
{
    if (wptr - rptr < blen)
        return FALSE;
    memcpy (b, &buf[rptr], blen);
    return TRUE;
}

CORBA::Boolean
CORBA::Buffer::peek (Octet &o)
{
    if (wptr == rptr)
        return FALSE;
    o = buf[rptr];
    return TRUE;
}

CORBA::Boolean
CORBA::Buffer::get (Octet &o)
{
    if (wptr == rptr)
        return FALSE;
    o = buf[rptr++];
    return TRUE;
}

CORBA::Boolean
CORBA::Buffer::get (void *b, ULong l)
{
    if (wptr - rptr < l)
	return FALSE;
    memcpy (b, &buf[rptr], l);
    rptr += l;
    return TRUE;
}

CORBA::Boolean
CORBA::Buffer::get1 (void *p)
{
    if (wptr == rptr)
        return FALSE;
    *(Octet *)p = buf[rptr++];
    return TRUE;
}

CORBA::Boolean
CORBA::Buffer::get2 (void *p)
{
    assert (rptr >= ralignbase);
    ULong r = ((rptr - ralignbase + 1) & ~1L) + ralignbase;
    if (r+2 > wptr)
	return FALSE;
    rptr = r;
    if (!(rptr&1) && !(((long)p)&1)) {
	*(CORBA::Short *)p = (CORBA::Short &)buf[rptr];
	rptr += 2;
    } else {
	*((Octet * &)p)++ = buf[rptr++];
	*(Octet *)p = buf[rptr++];
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Buffer::get4 (void *p)
{
    assert (rptr >= ralignbase);
    ULong r = ((rptr - ralignbase + 3) & ~3L) + ralignbase;
    if (r+4 > wptr)
	return FALSE;
    rptr = r;
    if (!(rptr&3) && !(((long)p)&3)) {
	*(CORBA::Long *)p = (CORBA::Long &)buf[rptr];
	rptr += 4;
    } else {
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*(Octet *)p = buf[rptr++];
    }
    return TRUE;
}

CORBA::Boolean
CORBA::Buffer::get8 (void *p)
{
    assert (rptr >= ralignbase);
    ULong r = ((rptr - ralignbase + 7) & ~7L) + ralignbase;
    if (r+8 > wptr)
	return FALSE;
    rptr = r;
    if (!(rptr&7) && !(((long)p)&7)) {
	*(CORBA::LongLong *)p = (CORBA::LongLong &)buf[rptr];
	rptr += 8;
    } else {
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*(Octet *)p = buf[rptr++];
    }
    return TRUE;
}

// get 16 bytes with 8 byte alignment
CORBA::Boolean
CORBA::Buffer::get16 (void *p)
{
    assert (rptr >= ralignbase);
    ULong r = ((rptr - ralignbase + 7) & ~7L) + ralignbase;
    if (r+16 > wptr)
	return FALSE;
    rptr = r;
    if (!(rptr&7) && !(((long)p)&7)) {
	*((CORBA::LongLong * &)p)++ = (CORBA::LongLong &)buf[rptr];
	rptr += 8;
	*(CORBA::LongLong *)p = (CORBA::LongLong &)buf[rptr];
	rptr += 8;
    } else {
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*((Octet * &)p)++ = buf[rptr++];
	*(Octet *)p = buf[rptr++];
    }
    return TRUE;
}

void
CORBA::Buffer::walign (ULong modulo)
{
    assert (!readonly);
    assert (modulo >= 1);
    assert (wptr >= walignbase);
    ULong w = (wptr - walignbase) + modulo - 1;
    w -= w % modulo;
    w += walignbase;
    resize (w - wptr);
    while (wptr < w)
	buf[wptr++] = 0;
}

void
CORBA::Buffer::replace (const void *o, ULong blen)
{
    assert (!readonly);
    reset (blen);
    memcpy (&buf[wptr], o, blen);
    wptr += blen;
}

void
CORBA::Buffer::replace (Octet o)
{
    assert (!readonly);
    reset (1);
    buf[wptr++] = o;
}

void
CORBA::Buffer::put (const void *o, ULong l)
{
    assert (!readonly);
    if (wptr + l > len)
	resize (l);
    memcpy (&buf[wptr], o, l);
    wptr += l;
}

void
CORBA::Buffer::put (Octet o)
{
    assert (!readonly);
    if (wptr + 1 > len)
	resize (1);
    buf[wptr++] = o;
}

void
CORBA::Buffer::put1 (const void *p)
{
    assert (!readonly);
    if (wptr == len)
	resize (1);
    buf[wptr++] = *(const Octet *)p;
}

void
CORBA::Buffer::put2 (const void *p)
{
    assert (!readonly && wptr >= walignbase);
    if (wptr+3 > len)
	resize (3);
    CORBA::Octet *b = buf + wptr;
    if ((wptr - walignbase) & 1L)
	*b++ = 0;
    if (!(((long)b)&1) && !(((long)p)&1)) {
	*((CORBA::Short *&)b)++ = *(const CORBA::Short *)p;
    } else {
	*b++ = *((const Octet * &)p)++;
	*b++ = *(const Octet *)p;
    }
    wptr = b - buf;
}

void
CORBA::Buffer::put4 (const void *p)
{
    assert (!readonly && wptr >= walignbase);
    if (wptr+7 > len)
	resize (7);
    CORBA::Octet *b = buf + wptr;
    switch ((wptr - walignbase) & 3L) {
    case 1: *b++ = 0;
    case 2: *b++ = 0;
    case 3: *b++ = 0;
    }
    if (!(((long)b)&3) && !(((long)p)&3)) {
	*((CORBA::Long * &)b)++ = *(const CORBA::Long *)p;
    } else {
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *(const Octet *)p;
    }
    wptr = b - buf;
}

void
CORBA::Buffer::put8 (const void *p)
{
    assert (!readonly && wptr >= walignbase);
    if (wptr+15 > len)
	resize (15);
    CORBA::Octet *b = buf + wptr;
    switch ((wptr - walignbase) & 7L) {
    case 1: *b++ = 0;
    case 2: *b++ = 0;
    case 3: *b++ = 0;
    case 4: *b++ = 0;
    case 5: *b++ = 0;
    case 6: *b++ = 0;
    case 7: *b++ = 0;
    }
    if (!(((long)b)&7) && !(((long)p)&7)) {
	*((CORBA::LongLong *&)b)++ = *(const CORBA::LongLong *)p;
    } else {
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *(const Octet *)p;
    }
    wptr = b - buf;
}

// put 16 bytes with 8 byte alignment
void
CORBA::Buffer::put16 (const void *p)
{
    assert (!readonly && wptr >= walignbase);
    if (wptr+23 > len)
	resize (23);
    CORBA::Octet *b = buf + wptr;
    switch ((wptr - walignbase) & 7L) {
    case 1: *b++ = 0;
    case 2: *b++ = 0;
    case 3: *b++ = 0;
    case 4: *b++ = 0;
    case 5: *b++ = 0;
    case 6: *b++ = 0;
    case 7: *b++ = 0;
    }
    if (!(((long)b)&7) && !(((long)p)&7)) {
	*((CORBA::LongLong *&)b)++ = *((const CORBA::LongLong * &)p)++;
	*((CORBA::LongLong *&)b)++ = *(const CORBA::LongLong *)p;
    } else {
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *((const Octet * &)p)++;
	*b++ = *(const Octet *)p;
    }
    wptr = b - buf;
}

void
CORBA::Buffer::dump (ostream &o) const
{
    o << "--" << endl;
    o << "rptr = " << rptr << endl;
    o << "ralign = " << ralignbase << endl;
    o << "wptr = " << wptr << endl;
    o << "walign = " << walignbase << endl;

    for (CORBA::ULong i = rptr, cnt = 0; i < wptr; ++i, ++cnt) {
        if (cnt == 16) {
            o << endl;
            cnt = 0;
        }
        char b[5];
        sprintf (b, "%2x ", b[i] & 0xff);
        o << b;
    }
    o << endl;
}
