/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#ifndef _WINDOWS
#include <string.h>
#endif
#include <iostream.h>
#include <mico/impl.h>
#include <mico/intercept.h>
#include <mico/util.h>
#include <mico/template_impl.h>
#include <mico/os-misc.h>


/*****************************  NamedValue  ***************************/


CORBA::NamedValue::NamedValue (const char *name, const Any &val, Flags flags)
    : n (name),  v (val), f (flags)
{
}

CORBA::NamedValue::NamedValue (const char *name, Flags flags)
    : n (name), f (flags)
{
}

CORBA::NamedValue::NamedValue ()
    : n (""), f (0)
{
    
}

CORBA::NamedValue &
CORBA::NamedValue::operator= (const NamedValue &nv)
{
    n = nv.n;
    f = nv.f;
    v = nv.v;
    return *this;
}

CORBA::NamedValue::~NamedValue ()
{
}

CORBA::NamedValue_ptr
CORBA::NamedValue::_duplicate (NamedValue_ptr nv)
{
    if (!CORBA::is_nil (nv))
	nv->_ref();
    return nv;
}

CORBA::NamedValue_ptr
CORBA::NamedValue::_nil ()
{
    return NULL;
}

void
CORBA::release (NamedValue_ptr nv)
{
    if (!CORBA::is_nil (nv) && nv->_deref())
	delete nv;
}

CORBA::Boolean
CORBA::is_nil (NamedValue_ptr nv)
{
    return nv == NULL;
}


/*****************************  NVList  ***************************/


CORBA::NVList::NVList (ULong count)
    : cnt (count)
{
}

CORBA::NVList::~NVList ()
{
}

CORBA::NamedValue_ptr
CORBA::NVList::add (Flags flags)
{
    // assert (vec.size() < cnt);
    vec.push_back (new NamedValue ("", flags));
    return vec.back();
}

CORBA::NamedValue_ptr
CORBA::NVList::add_item (const char *name, Flags flags)
{
    // assert (vec.size() < cnt);
    vec.push_back (new NamedValue (name, flags));
    return vec.back();
}

CORBA::NamedValue_ptr
CORBA::NVList::add_value (const char *name, const Any &val, Flags flags)
{
    // assert (vec.size() < cnt);
    vec.push_back (new NamedValue (name, val, flags));
    return vec.back();
}

CORBA::NamedValue_ptr
CORBA::NVList::add_item_consume (char *name, Flags flags)
{
    // assert (vec.size() < cnt);
    vec.push_back (new NamedValue (name, flags));
    string_free (name);
    return vec.back();
}

CORBA::NamedValue_ptr
CORBA::NVList::add_value_consume (char *name, Any *val, Flags flags)
{
    // assert (vec.size() < cnt);
    vec.push_back (new NamedValue (name, *val, flags));
    delete val;
    string_free (name);
    return vec.back();
}

CORBA::NamedValue_ptr
CORBA::NVList::item (ULong idx)
{
    assert (idx < vec.size());
    return vec[idx];
}

CORBA::Status
CORBA::NVList::remove (ULong idx)
{
    assert (idx < vec.size());
    vec.erase (vec.begin() + idx);
    return 0;
}

CORBA::Boolean
CORBA::NVList::copy (NVList_ptr src, Flags f)
{
    NVList_ptr dst = this;

    int dst_count = dst->count ();
    int src_count = src->count ();

    NamedValue_ptr dst_nv, src_nv;

    for (int d = 0, s = 0;; ++d, ++s) {
	dst_nv = NamedValue::_nil();
	for ( ; d < dst_count; ++d) {
	    dst_nv = dst->item (d);
	    if (dst_nv->flags () & f)
		break;
	}
	src_nv = NamedValue::_nil();
	for ( ; s < src_count; ++s) {
	    src_nv = src->item (s);
	    if (src_nv->flags () & f)
		break;
	}
	if (s >= src_count || d >= dst_count) {
	    // assertion fails iff #params-to-copy in dst and src differ.
	    if (!(s >= src_count && d >= dst_count))
                return FALSE;
	    break;
	}
	// make sure we have the same type of args
	if (!((src_nv->flags() & f) == (dst_nv->flags() & f)))
            return FALSE;

	// XXX possibly avoid a copy here ...
	*dst_nv->value() = *src_nv->value();
    }
    return TRUE;
}

CORBA::NVList_ptr
CORBA::NVList::_duplicate (NVList_ptr nvl)
{
    if (!CORBA::is_nil (nvl))
	nvl->_ref();
    return nvl;
}

CORBA::NVList_ptr
CORBA::NVList::_nil ()
{
    return NULL;
}

void
CORBA::release (NVList_ptr nvl)
{
    if (!CORBA::is_nil (nvl) && nvl->_deref())
	delete nvl;
}

CORBA::Boolean
CORBA::is_nil (NVList_ptr nvl)
{
    return nvl == NULL;
}


/*****************************  Environment  ***************************/


CORBA::Environment::Environment (Principal_ptr pr)
{
    except = NULL;
    princ = CORBA::Principal::_duplicate (pr);
}

CORBA::Environment::~Environment ()
{
    if (except)
	delete except;
    CORBA::release (princ);
}

void
CORBA::Environment::exception (Exception *e)
{
    if (except)
	delete except;
    except = e;
}

void
CORBA::Environment::clear ()
{
    if (except)
	delete except;
    except = NULL;
}

CORBA::Environment_ptr
CORBA::Environment::_duplicate (Environment_ptr e)
{
    if (!CORBA::is_nil (e))
	e->_ref();
    return e;
}

CORBA::Environment_ptr
CORBA::Environment::_nil ()
{
    return NULL;
}

void
CORBA::release (Environment_ptr e)
{
    if (!CORBA::is_nil (e) && e->_deref())
	delete e;
}

CORBA::Boolean
CORBA::is_nil (Environment_ptr e)
{
    return e == NULL;
}


/*****************************  Principal  ***************************/


vector<CORBA::Octet> CORBA::Principal::_peer_info;

CORBA::Principal::Principal (CORBA::Transport *t)
{
    _rep = _peer_info;
    // XXX make sure this isnt released ...
    _transp = t;
}

CORBA::Principal::Principal (DataDecoder &dc, CORBA::Transport *t)
{
    // XXX make sure this isnt released ...
    _transp = t;
    CORBA::Boolean r = decode (dc);
    assert (r);
}

CORBA::Principal::~Principal ()
{
}

CORBA::Boolean
CORBA::Principal::operator== (const Principal &p) const
{
    return _rep == p._rep;
}

void
CORBA::Principal::encode (DataEncoder &ec) const
{
    ec.seq_begin (_rep.size());
    {
	if (_rep.size() > 0) {
	    assert (_rep.size() < 2 || &_rep[0] + 1 == &_rep[1]);
	    ec.put_octets (&_rep[0], _rep.size());
	}
    }
    ec.seq_end ();
}

#define check(exp) if (!(exp)) return FALSE

CORBA::Boolean
CORBA::Principal::decode (DataDecoder &dc)
{
    ULong len;
    check (dc.seq_begin (len));
    {
	_rep.erase (_rep.begin(), _rep.end());
	if (len > 0) {
	    _rep.insert (_rep.begin(),
			 dc.buffer()->data(),
			 dc.buffer()->data() + len);
	    dc.buffer()->rseek_rel (len);
	}
    }
    check (dc.seq_end ());
    return TRUE;
}

#undef check

CORBA::Principal::PropertyNameList_ptr
CORBA::Principal::list_properties ()
{
    PropertyNameList_ptr pl = new PropertyNameList;
    pl->length (3);

    (*pl)[0] = (const char *)"peer-info";
    (*pl)[1] = (const char *)"auth-method";
    (*pl)[2] = (const char *)"peer-address";
    return pl;
}

CORBA::Any*
CORBA::Principal::get_property (const char *prop_name)
{
    if (!strcmp ("peer-info", prop_name)) {
	Any *a = new Any;
        if (_rep.size() > 0) {
            OctetSeq os (_rep.size(), _rep.size(),
                         (OctetWrapper *)&_rep[0], FALSE);
            *a <<= os;
        } else {
            *a <<= OctetSeq();
        }
	return a;
    }
    if (!strcmp ("auth-method", prop_name)) {
	Any *a = new Any;
	*a <<= "basic";
	return a;
    }
    if (!strcmp ("peer-address", prop_name)) {
        Any *a = new Any;
        if (_transp) {
            *a <<= _transp->peer()->stringify().c_str();
        } else {
            *a <<= "";
        }
        return a;
    }
    return new Any;
}

void
CORBA::Principal::set_peer_info (const OctetSeq &i)
{
    _peer_info.erase (_peer_info.begin(), _peer_info.end());
    _peer_info.insert (_peer_info.begin(),
		       (const Octet *)&i[0],
		       (const Octet *)&i[0] + i.length());
}

CORBA::Principal_ptr
CORBA::Principal::_duplicate (Principal_ptr p)
{
    if (!CORBA::is_nil (p))
	p->_ref();
    return p;
}

CORBA::Principal_ptr
CORBA::Principal::_nil ()
{
    return NULL;
}

void
CORBA::release (Principal_ptr p)
{
    if (!CORBA::is_nil (p) && p->_deref())
	delete p;
}

CORBA::Boolean
CORBA::is_nil (Principal_ptr p)
{
    return p == NULL;
}


/*****************************  ContextList  ***************************/


CORBA::ContextList::ContextList ()
{
}

CORBA::ContextList::~ContextList ()
{
}

void
CORBA::ContextList::add (const char *ctxt)
{
    vec.push_back (ctxt);
}

void
CORBA::ContextList::add_consume (char *ctxt)
{
    vec.push_back (ctxt);
    string_free (ctxt);
}

const char *
CORBA::ContextList::item (ULong idx)
{
    assert (idx < vec.size());
    return vec[idx].c_str();
}

CORBA::Status
CORBA::ContextList::remove (ULong idx)
{
    assert (idx < vec.size());
    vec.erase (vec.begin() + idx);
    return 0;
}

CORBA::ContextList_ptr
CORBA::ContextList::_duplicate (ContextList_ptr c)
{
    if (!CORBA::is_nil (c))
	c->_ref();
    return c;
}

CORBA::ContextList_ptr
CORBA::ContextList::_nil ()
{
    return NULL;
}

void
CORBA::release (ContextList_ptr c)
{
    if (!CORBA::is_nil (c) && c->_deref())
	delete c;
}

CORBA::Boolean
CORBA::is_nil (ContextList_ptr c)
{
    return c == NULL;
}


/*****************************  ExceptionList  ***************************/


CORBA::ExceptionList::ExceptionList ()
{
}

CORBA::ExceptionList::~ExceptionList ()
{
}

void
CORBA::ExceptionList::add (TypeCode_ptr tc)
{
    vec.push_back (TypeCode::_duplicate (tc));
}

void
CORBA::ExceptionList::add_consume (TypeCode_ptr tc)
{
    vec.push_back (tc);
}

CORBA::TypeCode_ptr
CORBA::ExceptionList::item (ULong idx)
{
    assert (idx < vec.size());
    return TypeCode::_duplicate (vec[idx]);
}

CORBA::Status
CORBA::ExceptionList::remove (ULong idx)
{
    assert (idx < vec.size());
    vec.erase (vec.begin() + idx);
    return 0;
}

CORBA::ExceptionList_ptr
CORBA::ExceptionList::_duplicate (ExceptionList_ptr e)
{
    if (!CORBA::is_nil (e))
	e->_ref();
    return e;
}

CORBA::ExceptionList_ptr
CORBA::ExceptionList::_nil ()
{
    return NULL;
}

void
CORBA::release (ExceptionList_ptr e)
{
    if (!CORBA::is_nil (e) && e->_deref())
	delete e;
}

CORBA::Boolean
CORBA::is_nil (ExceptionList_ptr e)
{
    return e == NULL;
}


/*************************  RequestCallback  **************************/


CORBA::RequestCallback::~RequestCallback ()
{
}


/*****************************  Request  ******************************/


CORBA::Request::Request ()
{
    assert (0);
}

CORBA::Request::Request (Object_ptr o, Context_ptr c,
			 const char *op,
			 NVList_ptr arguments,
			 NamedValue_ptr result,
			 Flags f)
    : object (Object::_duplicate (o)),
      context (Context::_duplicate (c)), opname (op),
      args (NVList::_duplicate (arguments)),
      res (NamedValue::_duplicate (result)),
      flags (f)
{
    invoke_pending = FALSE;

    ORB_ptr orb = o->_orbnc();

    CORBA::Status r = orb->create_environment (environ);
    assert (r == 0);
    // XXX use IR to fill elist/clist
    r = orb->create_context_list (clist);
    assert (r == 0);
    r = orb->create_exception_list (elist);
    assert (r == 0);
    orbreq = new MICO::LocalRequest (this);
    iceptreq = Interceptor::ClientInterceptor::_create_request (
	object, opname, *orbreq->context(), this);
    cb = 0;
}

CORBA::Request::Request (Object_ptr o,
			 Context_ptr c,
			 const char *op,
			 NVList_ptr arguments,
			 NamedValue_ptr result,
			 ExceptionList_ptr el,
			 ContextList_ptr cl,
			 Flags f)
    : object (Object::_duplicate (o)),
      context (Context::_duplicate (c)), opname (op),
      args (NVList::_duplicate (arguments)),
      res (NamedValue::_duplicate (result)), flags (f),
      elist (ExceptionList::_duplicate (el)),
      clist (ContextList::_duplicate (cl))
{
    invoke_pending = FALSE;

    ORB_ptr orb = o->_orbnc();
    CORBA::Status r = orb->create_environment (environ);
    assert (r == 0);
    orbreq = new MICO::LocalRequest (this);
    iceptreq = Interceptor::ClientInterceptor::_create_request (
	object, opname, *orbreq->context(), this);
    cb = 0;
}

CORBA::Request::Request (Object_ptr o, const char *op)
    : object (Object::_duplicate (o)), opname (op)
{
    invoke_pending = FALSE;
	
    ORB_ptr orb = o->_orbnc();

    context = Context::_nil();
    CORBA::Status r = orb->create_environment (environ);
    assert (r == 0);
    r = orb->create_list (0, args);
    assert (r == 0);
    r = orb->create_named_value (res);
    assert (r == 0);
    // XXX use IR to fill elist/clist
    r = orb->create_context_list (clist);
    assert (r == 0);
    r = orb->create_exception_list (elist);
    assert (r == 0);
    flags = 0;
    orbreq = new MICO::LocalRequest (this);
    iceptreq = Interceptor::ClientInterceptor::_create_request (
	object, opname, *orbreq->context(), this);
    cb = 0;
}

CORBA::Request::~Request ()
{
    if (invoke_pending)
        object->_orbnc()->cancel (msgid);
    CORBA::release (context);
    CORBA::release (environ);
    CORBA::release (args);
    CORBA::release (res);
    CORBA::release (clist);
    CORBA::release (elist);
    CORBA::release (orbreq);
    CORBA::release (iceptreq);
}

void
CORBA::Request::decode_user_exception ()
{
    Exception *ex = environ->exception();
    if (ex && elist->count() > 0) {
	UnknownUserException *uuex = UnknownUserException::_narrow (ex);
	if (uuex) {
	    // decode UnknownUserException using ExceptionList ...
	    const char *repoid = uuex->_except_repoid ();
	    TypeCode_ptr tc;
	    for (CORBA::ULong i = 0; i < elist->count(); ++i) {
		tc = elist->item (i);
		if (!strcmp (repoid, tc->id())) {
		    uuex->exception (tc);
		    break;
		}
	    }
	}
    }
}

void
CORBA::Request::callback (ORB_ptr, ULong, ORBCallback::Event ev)
{
    assert (ev == ORBCallback::Invoke);
    if (cb) {
        cb->callback (this, RequestCallback::RequestDone);
        cb = 0;
    }
}

void
CORBA::Request::ctx (Context_ptr c)
{
    CORBA::release (context);
    context = Context::_duplicate (c);
}

CORBA::Any &
CORBA::Request::add_in_arg ()
{
    return *arguments()->add(ARG_IN)->value();
}

CORBA::Any &
CORBA::Request::add_in_arg (const char *name)
{
    return *arguments()->add_item(name, ARG_IN)->value();
}

CORBA::Any &
CORBA::Request::add_inout_arg ()
{
    return *arguments()->add(ARG_INOUT)->value();
}

CORBA::Any &
CORBA::Request::add_inout_arg (const char *name)
{
    return *arguments()->add_item(name, ARG_INOUT)->value();
}

CORBA::Any &
CORBA::Request::add_out_arg ()
{
    return *arguments()->add(ARG_OUT)->value();
}

CORBA::Any &
CORBA::Request::add_out_arg (const char *name)
{
    return *arguments()->add_item(name, ARG_OUT)->value();
}

void
CORBA::Request::set_return_type (TypeCode_ptr tc)
{
    result()->value()->type (tc);
}

CORBA::Any &
CORBA::Request::return_value ()
{
    return *result()->value();
}

CORBA::Status
CORBA::Request::invoke ()
{
    Status s;

    if ((s = send_deferred ()) != 0)
	return s;
    if ((s = get_response()) != 0)
	return s;
    return 0;
}

CORBA::Status
CORBA::Request::send_oneway ()
{
    if (!Interceptor::ClientInterceptor::
	_exec_initialize_request ((Interceptor::LWRequest_ptr)iceptreq,
				  environ)) {
	return 0;
    }

    ORB_ptr orb = object->_orbnc();
    orb->invoke_async (object, orbreq, Principal::_nil(), FALSE);

    Interceptor::ClientInterceptor::
	_exec_after_marshal ((Interceptor::LWRequest_ptr)iceptreq, environ);

    return 0;
}

CORBA::Status
CORBA::Request::send_deferred (RequestCallback *_cb)
{
    cb = _cb;

    if (!Interceptor::ClientInterceptor::
	_exec_initialize_request ((Interceptor::LWRequest_ptr)iceptreq,
				  environ)) {
	if (cb)
	    cb->callback (this, RequestCallback::RequestDone);
	cb = 0;
	return 0;
    }

    invoke_pending = TRUE;

    ORB_ptr orb = object->_orbnc();
    msgid = orb->invoke_async (object, orbreq, Principal::_nil(),
                               TRUE, cb ? this : 0);

    if (!Interceptor::ClientInterceptor::
	_exec_after_marshal ((Interceptor::LWRequest_ptr)iceptreq, environ)) {
	orb->cancel (msgid);
	invoke_pending = FALSE;
	if (cb)
	    cb->callback (this, RequestCallback::RequestDone);
	cb = 0;
	return 0;
    }

    return 0;
}

CORBA::Status
CORBA::Request::get_response (Boolean block)
{
    if (!invoke_pending)
	return 0;

    ORB_ptr orb = object->_orbnc();

    while (invoke_pending) {
	Object_var nobj;
	ORBRequest *dummy;

	if (block) {
	    CORBA::Boolean r = orb->wait (msgid);
	    assert (r);
	} else {
	    if (!orb->wait (msgid, 0))
		return 1;
	}

	// XXX called multiple times in case of Forward
	if (!Interceptor::ClientInterceptor::
	    _exec_before_unmarshal ((Interceptor::LWRequest_ptr)iceptreq,
				    environ)) {
	    orb->cancel (msgid);
	    invoke_pending = FALSE;
	    cb = 0;
	    return 0;
	}

	InvokeStatus rs = orb->get_invoke_reply (msgid, nobj, dummy);

	switch (rs) {
	case InvokeForward:
	    // XXX what if 'object' is not a stub ???
	    *(Object_ptr)object = *(Object_ptr)nobj;
	    msgid = orb->invoke_async (object, orbreq, Principal::_nil());
	    break;

	case InvokeOk:
	case InvokeSysEx:
	    invoke_pending = FALSE;
	    break;

	case InvokeUsrEx:
	    decode_user_exception ();
	    invoke_pending = FALSE;
	    break;

	default:
	    assert (0);
	}
    }

    if (!Interceptor::ClientInterceptor::
	_exec_finish_request ((Interceptor::LWRequest_ptr)iceptreq, environ)) {
	cb = 0;
	return 0;
    }

    cb = 0;
    return 0;
}

CORBA::Boolean
CORBA::Request::poll_response ()
{
    // XXX call get_response() to check for redirections...
    get_response (FALSE);
    return !invoke_pending;
}

CORBA::Request_ptr
CORBA::Request::_duplicate (Request_ptr r)
{
    if (!CORBA::is_nil (r))
	r->_ref();
    return r;
}

CORBA::Request_ptr
CORBA::Request::_nil ()
{
    return NULL;
}

void
CORBA::release (Request_ptr r)
{
    if (!CORBA::is_nil (r) && r->_deref())
	delete r;
}

CORBA::Boolean
CORBA::is_nil (Request_ptr r)
{
    return r == NULL;
}


/************************* LocalRequest ******************************/


MICO::LocalRequest::LocalRequest ()
    : req (CORBA::Request::_nil()), have_except (FALSE), have_result (FALSE)
{
}

MICO::LocalRequest::LocalRequest (CORBA::Request_ptr r)
    : req (r), have_except (FALSE), have_result (FALSE)
{
}

MICO::LocalRequest::~LocalRequest ()
{
}

const char *
MICO::LocalRequest::op_name()
{
    return req->operation();
}

CORBA::Boolean
MICO::LocalRequest::get_in_args (CORBA::NVList_ptr iparams,
				 CORBA::Context_ptr &ctx)
{
    if (!iparams->copy (req->arguments(), CORBA::ARG_IN|CORBA::ARG_INOUT))
        return FALSE;
    ctx = CORBA::Context::_duplicate (req->ctx());
    return TRUE;
}

CORBA::Boolean
MICO::LocalRequest::get_in_args (StaticAnyList *iparams,
				 CORBA::Context_ptr &ctx)
{
    CORBA::NVList_ptr args = req->arguments();

    if (args->count() != iparams->size())
	return FALSE;

    CORBA::NamedValue_ptr nv;
    for (mico_vec_size_type i0 = 0; i0 < iparams->size(); ++i0) {
	nv = args->item(i0);
	if ((*iparams)[i0]->flags() != nv->flags())
	    return FALSE;
	if ((*iparams)[i0]->flags() & (CORBA::ARG_IN|CORBA::ARG_INOUT)) {
	    if (!nv->value()->to_static_any (*(*iparams)[i0]))
		return FALSE;
	}
    }
    ctx = CORBA::Context::_duplicate (req->ctx());
    return TRUE;
}

CORBA::Boolean
MICO::LocalRequest::get_in_args (CORBA::DataEncoder *ec)
{
    CORBA::NVList_ptr nv = req->arguments();
    for (CORBA::ULong i = 0; i < nv->count(); ++i) {
	if (nv->item(i)->flags() & (CORBA::ARG_IN|CORBA::ARG_INOUT)) {
	    nv->item(i)->value()->marshal (*ec);
	}
    }
    if (!CORBA::is_nil (req->ctx())) {
	ec->put_context (*req->ctx(), req->contexts());
    }
    return TRUE;
}

CORBA::Boolean
MICO::LocalRequest::get_out_args (CORBA::Any *res, CORBA::NVList_ptr oparams,
				  CORBA::Exception *&except)
{
    if (have_except) {
	CORBA::Exception *e = req->env()->exception();
	assert (e);
	except = e->_clone ();
    } else {
	except = 0;
	if (have_result && res)
	    *res = *req->result()->value();
	if (!oparams->copy (req->arguments(), CORBA::ARG_OUT|CORBA::ARG_INOUT))
            return FALSE;
    }
    return TRUE;
}

CORBA::Boolean
MICO::LocalRequest::get_out_args (CORBA::StaticAny *res,
				  StaticAnyList *oparams,
				  CORBA::Exception *&except)
{
    if (have_except) {
	CORBA::Exception *e = req->env()->exception();
	assert (e);
	except = e->_clone ();
	return TRUE;
    }

    except = 0;

    CORBA::NVList_ptr args = req->arguments();

    if (args->count() != oparams->size())
	return FALSE;

    if (have_result && res) {
	if (!req->result()->value()->to_static_any (*res))
	    return FALSE;
    }

    CORBA::NamedValue_ptr nv;
    for (CORBA::ULong i0 = 0; i0 < args->count(); ++i0) {
	nv = args->item(i0);
	if ((*oparams)[i0]->flags() != nv->flags())
	    return FALSE;
	if ((*oparams)[i0]->flags() & (CORBA::ARG_OUT|CORBA::ARG_INOUT)) {
	    if (!nv->value()->to_static_any (*(*oparams)[i0]))
		return FALSE;
	}
    }
    return TRUE;
}

CORBA::Boolean
MICO::LocalRequest::get_out_args (CORBA::DataEncoder *ec,
				  CORBA::Boolean &is_except)
{
    if (have_except) {
	is_except = TRUE;
	CORBA::Exception *e = req->env()->exception();
	assert (e);
	e->_encode (*ec);
    } else {
	is_except = FALSE;
	if (have_result) {
	    req->result()->value()->marshal (*ec);
	}
	CORBA::NVList_ptr nv = req->arguments();
	for (CORBA::ULong i = 0; i < nv->count(); ++i) {
	    if (nv->item(i)->flags() & (CORBA::ARG_OUT|CORBA::ARG_INOUT)) {
		nv->item(i)->value()->marshal (*ec);
	    }
	}
    }
    return TRUE;
}

CORBA::Boolean
MICO::LocalRequest::set_out_args (CORBA::Any *res, CORBA::NVList_ptr oparams)
{
    if (res) {
	have_result = TRUE;
	*req->result()->value() = *res;
    }
    return req->arguments()->copy (oparams, CORBA::ARG_OUT|CORBA::ARG_INOUT);
}

CORBA::Boolean
MICO::LocalRequest::set_out_args (CORBA::StaticAny *res,
				  StaticAnyList *oparams)
{
    CORBA::NVList_ptr args = req->arguments();

    if (args->count() != oparams->size())
	return FALSE;

    CORBA::TypeCode_var tc;

    if (res) {
	have_result = TRUE;
	tc = req->result()->value()->type();
	req->result()->value()->from_static_any (tc, *res);
    }

    CORBA::NamedValue_ptr nv;
    for (mico_vec_size_type i0 = 0; i0 < oparams->size(); ++i0) {
	nv = args->item(i0);
	if ((*oparams)[i0]->flags() != nv->flags())
	    return FALSE;
	if ((*oparams)[i0]->flags() & (CORBA::ARG_OUT|CORBA::ARG_INOUT)) {
	    tc = nv->value()->type();
	    nv->value()->from_static_any (tc, *(*oparams)[i0]);
	}
    }
    return TRUE;
}

void
MICO::LocalRequest::set_out_args (CORBA::Exception *except)
{
    have_except = TRUE;
    req->env()->exception (except->_clone());
}

CORBA::Boolean
MICO::LocalRequest::set_out_args (CORBA::DataDecoder *dc, CORBA::Boolean is_ex)
{
    CORBA::Boolean ret;
    CORBA::TypeCode_var tc;

    if (is_ex) {
	have_except = TRUE;
	CORBA::Exception *e = CORBA::Exception::_decode (*dc);
	assert (e);
	req->env()->exception (e);
    } else {
	have_result = TRUE;

	tc = req->result()->value()->type();
	ret = req->result()->value()->demarshal (tc, *dc);
	if (!ret)
            return FALSE;

	CORBA::NVList_ptr nv = req->arguments();
	for (CORBA::ULong i = 0; i < nv->count(); ++i) {
	    if (nv->item(i)->flags() & (CORBA::ARG_OUT|CORBA::ARG_INOUT)) {
		tc = nv->item(i)->value()->type();
		ret = nv->item(i)->value()->demarshal (tc, *dc);
                if (!ret)
                    return FALSE;
	    }
	}
    }
    return TRUE;
}

CORBA::Boolean
MICO::LocalRequest::copy_out_args (CORBA::ORBRequest *r)
{
    if (this == r)
	return TRUE;

    // copy service context list
    copy_svc (r);

    CORBA::Exception *ex = 0;
    if (!r->get_out_args (req->result()->value(), req->arguments(), ex))
        return FALSE;
    if (ex) {
	have_except = TRUE;
	have_result = FALSE;
	req->env()->exception (ex);
    } else {
	have_result = TRUE;
	have_except = FALSE;
    }
    return TRUE;
}

CORBA::Boolean
MICO::LocalRequest::copy_in_args (CORBA::ORBRequest *_req)
{
    assert (0);
    return FALSE;
}

const char *
MICO::LocalRequest::type ()
{
    return "local";
}

CORBA::Request_ptr
MICO::LocalRequest::request ()
{
    return req;
}

MICO::LocalRequest *
MICO::LocalRequest::_duplicate (LocalRequest *r)
{
    if (!CORBA::is_nil (r))
	r->_ref();
    return r;
}

MICO::LocalRequest *
MICO::LocalRequest::_nil ()
{
    return NULL;
}
