/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#ifndef _WINDOWS
#include <string.h>
#endif
#include <iostream.h>
#include <mico/intercept.h>
#include <mico/template_impl.h>
#include <mico/util.h>


CORBA::ImplementationBase::ImplementationBase ()
{
    orb = ORB_instance ("mico-local-orb");
    boa = orb->BOA_instance ("mico-local-boa");
}

CORBA::ImplementationBase::~ImplementationBase ()
{
    CORBA::release (orb);
    CORBA::release (boa);
}

void
CORBA::ImplementationBase::_create_ref (
    const BOA::ReferenceData &id,
    InterfaceDef_ptr interf,
    ImplementationDef_ptr im,
    const char *repoid)
{
    Object_var p = boa->create (id, interf, im, this, repoid);
    assert (!CORBA::is_nil (p));
    Object::operator= (*(Object_ptr)p);
}

void
CORBA::ImplementationBase::_restore_ref (
    Object_ptr orig_obj,
    const BOA::ReferenceData &id,
    InterfaceDef_ptr interf,
    ImplementationDef_ptr im)
{
    Object_var p = boa->restore (orig_obj, id, interf, im, this);
    assert (!CORBA::is_nil (p));
    Object::operator= (*(Object_ptr)p);
}

void
CORBA::ImplementationBase::_dispose_ref ()
{
    boa->dispose (this);
}

const char *
CORBA::ImplementationBase::_impl_name (const char *)
{
    return boa->impl_name();
}

CORBA::ImplementationDef_ptr
CORBA::ImplementationBase::_find_impl (const char *repoid, const char *name)
{
    name = _impl_name (name);

    CORBA::Object_var obj =
	_orbnc()->resolve_initial_references ("ImplementationRepository");
    assert (!CORBA::is_nil (obj));

    CORBA::ImplRepository_var imr = ImplRepository::_narrow (obj);
    assert (!CORBA::is_nil (imr));

    CORBA::ImplRepository::ImplDefSeq_var impls = imr->find_by_name (name);
    CORBA::ImplementationDef_var impl;
    if (impls->length() > 0) {
        MICODebug::instance()->tracer()
            << "found impl repository entry: " << name << endl;

        impl = CORBA::ImplementationDef::_duplicate (impls[(CORBA::ULong)0]);
        CORBA::ImplementationDef::RepoIdList_var repoids = impl->repoids();
        CORBA::ULong i;
        for (i = 0; i < repoids->length(); ++i) {
            if (!strcmp (repoid, repoids[i]))
                break;
        }
        if (i == repoids->length()) {
            MICODebug::instance()->tracer()
                << "adding repo id: " << repoid << endl;

            repoids->length (repoids->length()+1);
            repoids[repoids->length()-1] = repoid;
            impl->repoids (*(CORBA::ImplementationDef::RepoIdList *)repoids);
        }
    } else {
        MICODebug::instance()->tracer()
            << "creating impl repository entry: " << name << endl;

        CORBA::ImplementationDef::RepoIdList repoids;
        repoids.length (1);
        repoids[0] = repoid;

        impl = imr->create (CORBA::ImplementationDef::ActivatePersistent,
                            repoids, name, "");
    }
    return CORBA::ImplementationDef::_duplicate (impl);
}

CORBA::InterfaceDef_ptr
CORBA::ImplementationBase::_find_iface (const char *repoid)
{
    CORBA::Object_var obj =
	_orbnc()->resolve_initial_references ("InterfaceRepository");
    assert (!CORBA::is_nil (obj));

    CORBA::Repository_var ir = Repository::_narrow (obj);
    assert (!CORBA::is_nil (ir));

    CORBA::Contained_var contained = ir->lookup_id ((char *)repoid);
    if (CORBA::is_nil (contained))
	return CORBA::InterfaceDef::_nil();

    CORBA::InterfaceDef_var iface = CORBA::InterfaceDef::_narrow (contained);
    assert (!CORBA::is_nil (iface));
    return CORBA::InterfaceDef::_duplicate (iface);
}

CORBA::Boolean
CORBA::ImplementationBase::_save_object ()
{
    /*
     * FALSE means the object is transient, i.e. it won't be restored
     * later
     */
    return FALSE;
}

CORBA::BOA_ptr
CORBA::ImplementationBase::_boa ()
{
    return CORBA::BOA::_duplicate (boa);
}

CORBA::ORB_ptr
CORBA::ImplementationBase::_orbnc ()
{
    // no _duplicate() returned
    return boa->orb();
}

CORBA::ORB_ptr
CORBA::ImplementationBase::_orb ()
{
    return CORBA::ORB::_duplicate (orb);
}

CORBA::Object_ptr
CORBA::ImplementationBase::_this ()
{
    return CORBA::Object::_duplicate (this);
}


/************************** ServerRequestBase ***************************/


CORBA::ServerRequestBase::~ServerRequestBase ()
{
}

CORBA::ServerRequestBase_ptr
CORBA::ServerRequestBase::_duplicate (ServerRequestBase_ptr r)
{
    if (!CORBA::is_nil (r))
	r->_ref();
    return r;
}

CORBA::ServerRequestBase_ptr
CORBA::ServerRequestBase::_nil ()
{
    return NULL;
}

void
CORBA::release (ServerRequestBase_ptr r)
{
    if (!CORBA::is_nil (r) && r->_deref())
	delete r;
}

CORBA::Boolean
CORBA::is_nil (ServerRequestBase_ptr r)
{
    return r == NULL;
}


/************************ DynamicImplementation *************************/


CORBA::DynamicImplementation::DynamicImplementation ()
{
}

CORBA::DynamicImplementation::~DynamicImplementation ()
{
}

CORBA::ServerRequestBase_ptr
CORBA::DynamicImplementation::make_request (CORBA::ORBRequest *req,
					    CORBA::Object_ptr obj,
					    CORBA::ULong msgid,
					    CORBA::ObjectAdapter *oa,
					    CORBA::Principal_ptr pr)
{
    return new ServerRequest (req, obj, msgid, oa, pr);
}

void
CORBA::DynamicImplementation::doinvoke (ServerRequestBase_ptr req,
					Environment &env)
{
    invoke ((ServerRequest_ptr)req, env);
}


/***************************** ServerRequest ***************************/


CORBA::ServerRequest::ServerRequest (ORBRequest *r, Object_ptr o,
                                     MsgId _id, CORBA::ObjectAdapter *_oa,
				     CORBA::Principal_ptr pr)
    : oa (_oa), msgid (_id),
      obj(Object::_duplicate (o)),
      req(ORBRequest::_duplicate (r)),
      _env (pr)
{
    aborted = FALSE;
    dir_params = NVList::_nil();
    context = Context::_nil();
    iceptreq = Interceptor::ServerInterceptor::_create_request (
	obj, r->op_name(), *r->context(), this);
}

CORBA::ServerRequest::~ServerRequest ()
{
    if (!_canceled) {
        if (!aborted && !Interceptor::ServerInterceptor::
            _exec_before_marshal ((Interceptor::LWServerRequest_ptr)iceptreq,
                                  &_env)) {
            aborted = TRUE;
        }

        // copy back out args and result into ORBRequest
        set_out_args();

        // tell OA we are done
        CORBA::InvokeStatus stat = CORBA::InvokeOk;
        if (exception()) {
            if (CORBA::SystemException::_narrow (exception())) {
                stat = CORBA::InvokeSysEx;
            } else {
                stat = CORBA::InvokeUsrEx;
            }
        }
        oa->answer_invoke (msgid, obj, req, stat);

        if (!aborted && !Interceptor::ServerInterceptor::
            _exec_finish_request ((Interceptor::LWServerRequest_ptr)iceptreq,
                                  &_env)) {
            // XXX how to abort the request here???
        }
    }

    CORBA::release (dir_params);
    CORBA::release (context);
    CORBA::release (iceptreq);
    CORBA::release (obj);
    CORBA::release (req);
}

// begin CORBA 2.2 mapping [20.26.1]
const char *
CORBA::ServerRequest::operation ()
{
  return req->op_name();
}

void
CORBA::ServerRequest::arguments (NVList_ptr & _params)
{
  params (_params);
}

void
CORBA::ServerRequest::set_result (const Any & val)
{
  res = new CORBA::Any (val);
}

void
CORBA::ServerRequest::set_exception (const Any & val)
{
  res = 0;
  _env.exception (Exception::_decode ((Any &) val));
  assert (_env.exception());
}
// end CORBA 2.2 mapping; ctx () as below

const char *
CORBA::ServerRequest::op_name ()
{
    return req->op_name();
}

CORBA::OperationDef_ptr
CORBA::ServerRequest::op_def ()
{
    InterfaceDef_var iface = obj->_get_interface ();
    assert (!CORBA::is_nil (iface));

    Contained_var cont = iface->lookup ((char *)op_name());
    assert (!CORBA::is_nil (cont));

    OperationDef_ptr opdef = OperationDef::_narrow (cont);
    assert (!CORBA::is_nil (opdef));

    return opdef;
}

CORBA::Context_ptr
CORBA::ServerRequest::ctx ()
{
    assert (!CORBA::is_nil (dir_params));
    return context;
}

CORBA::Boolean
CORBA::ServerRequest::params (NVList_ptr params)
{
    assert (CORBA::is_nil (dir_params));
    dir_params = params;

    if (!Interceptor::ServerInterceptor::
	_exec_initialize_request ((Interceptor::LWServerRequest_ptr)iceptreq,
				  &_env)) {
	aborted = TRUE;
	return FALSE;
    }

    if (!req->get_in_args (dir_params, context)) {
        MICODebug::instance()->debugger()
            << "cannot decode args" << endl;
        exception (new CORBA::MARSHAL());
        aborted = TRUE;
        return FALSE;
    }

    if (!Interceptor::ServerInterceptor::
	_exec_after_unmarshal ((Interceptor::LWServerRequest_ptr)iceptreq,
			       &_env)) {
	aborted = TRUE;
	return FALSE;
    }
    return TRUE;
}

void
CORBA::ServerRequest::result (Any *val)
{
    _env.clear ();
    res = val;
}

CORBA::Environment_ptr
CORBA::ServerRequest::environment ()
{
    return &_env;
}

void
CORBA::ServerRequest::exception (Any *val)
{
    res = 0;
    _env.exception (Exception::_decode (*val));
    assert (_env.exception());
    delete val;
}

void
CORBA::ServerRequest::exception (Exception *except)
{
    _env.exception (except);
}

CORBA::Exception *
CORBA::ServerRequest::exception ()
{
    // maybe _nil(), overship maintained by ServerRequest
    return _env.exception();
}

CORBA::Any *
CORBA::ServerRequest::result ()
{
    // maybe 0, overship maintained by ServerRequest
    return res;
}

CORBA::NVList_ptr
CORBA::ServerRequest::arguments ()
{
    // maybe _nil(), overship maintained by ServerRequest
    return dir_params;
}

void
CORBA::ServerRequest::set_out_args ()
{

    if (_env.exception()) {
        req->set_out_args (_env.exception());
    } else if (!CORBA::is_nil (dir_params)) {
	if (!req->set_out_args (res, dir_params)) {
            MICODebug::instance()->debugger()
                << "cannot set out args" << endl;
            CORBA::MARSHAL ex;
            req->set_out_args (&ex);
        }
    }
}

CORBA::ORBRequest *
CORBA::ServerRequest::request ()
{
    return req;
}

CORBA::ServerRequest_ptr
CORBA::ServerRequest::_duplicate (ServerRequest_ptr r)
{
    if (!CORBA::is_nil (r))
	r->_ref();
    return r;
}

CORBA::ServerRequest_ptr
CORBA::ServerRequest::_nil ()
{
    return NULL;
}

void
CORBA::release (ServerRequest_ptr r)
{
    if (!CORBA::is_nil (r) && r->_deref())
	delete r;
}

CORBA::Boolean
CORBA::is_nil (ServerRequest_ptr r)
{
    return r == NULL;
}
