/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#ifndef _WINDOWS
#include <string.h>
#endif
#include <iostream.h>
#include <strstream.h>
#include <mico/impl.h>
#include <mico/template_impl.h>


/*************************** Exception ***************************/

void (*CORBA::Exception::_old_terminate_handler) () = NULL;

char *CORBA::Exception::_last_except_desc = 0;

CORBA::Exception::Exception ()
{
}

CORBA::Exception::Exception (const Exception &e)
{
}

CORBA::Exception::~Exception ()
{
}

CORBA::Exception &
CORBA::Exception::operator= (const Exception &e)
{
    return *this;
}

CORBA::Exception *
CORBA::Exception::_decode (Any &a)
{
    Exception *e;
    if ((e = SystemException::_decode (a)))
	return e;
    if ((e = UserException::_decode (a)))
	return e;
    return NULL;
}

CORBA::Exception *
CORBA::Exception::_decode (DataDecoder &dc)
{
    Exception *e;
    ULong rpos = dc.buffer()->rpos();
    if ((e = SystemException::_decode (dc)))
	return e;
    dc.buffer()->rseek_beg (rpos);
    if ((e = UserException::_decode (dc)))
	return e;
    dc.buffer()->rseek_beg (rpos);
    return NULL;
}

void
CORBA::Exception::_print (ostream &os) const
{
    os << _repoid() << endl;
}

const char *
CORBA::Exception::_repoid () const
{
    return "IDL:omg.org/CORBA/Exception:1.0";
}

CORBA::Exception *
CORBA::Exception::_clone () const
{
    return new Exception (*this);
}

void
CORBA::Exception::_prepare_throw () const
{
    ostrstream ostr;
    _print (ostr);
    ostr << ends;

    if (_last_except_desc)
	CORBA::string_free (_last_except_desc);
    _last_except_desc = CORBA::string_dup (ostr.str());
    ostr.rdbuf()->freeze (0);
}

void
CORBA::Exception::_encode (DataEncoder &ec) const
{
    assert (0);
}

void
CORBA::Exception::_throw_failed (Exception *ex)
{
    cerr << "exception handling not supported," << endl;
    cerr << "the following exception has been thrown:" << endl;
    cerr << "  ";
    ex->_print (cerr);
    abort ();
}

void
CORBA::Exception::_terminate_handler ()
{
    cerr << "uncaught MICO exception:" << endl;
    if (_last_except_desc)
	cerr << "  " << _last_except_desc << flush;
    (*_old_terminate_handler) ();
}

// Exception::_init() is in except2.cc


/************************* SystemException *************************/


CORBA::SystemException::SystemException ()
    : _minor (0), _completed (COMPLETED_NO)
{
}

CORBA::SystemException::SystemException (const SystemException &e)
    : _minor (e._minor), _completed (e._completed)
{
}

CORBA::SystemException::SystemException (ULong minor, CompletionStatus status)
    : _minor (minor), _completed (status)
{
}

CORBA::SystemException::~SystemException ()
{
}

CORBA::SystemException &
CORBA::SystemException::operator= (const SystemException &e)
{
    if (this != &e) {
	Exception::operator= (e);
	_minor = e._minor;
	_completed = e._completed;
    }
    return *this;
}

const char *
CORBA::SystemException::_repoid () const
{
    return "IDL:omg.org/CORBA/SystemException:1.0";
}

CORBA::Exception *
CORBA::SystemException::_clone () const
{
    return new SystemException (*this);
}

CORBA::SystemException *
CORBA::SystemException::_create_sysex (const char *repoid, ULong minor,
				       CompletionStatus completed)
{
    if (!strcmp (repoid, "IDL:omg.org/CORBA/SystemException:1.0"))
	return new SystemException (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/UNKNOWN:1.0"))
	return new UNKNOWN (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/BAD_PARAM:1.0"))
	return new BAD_PARAM (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/NO_MEMORY:1.0"))
	return new NO_MEMORY (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/IMP_LIMIT:1.0"))
	return new IMP_LIMIT (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/COMM_FAILURE:1.0"))
	return new COMM_FAILURE (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/INV_OBJREF:1.0"))
	return new INV_OBJREF (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/NO_PERMISSION:1.0"))
	return new NO_PERMISSION (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/INTERNAL:1.0"))
	return new INTERNAL (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/MARSHAL:1.0"))
	return new MARSHAL (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/INITIALIZE:1.0"))
	return new INITIALIZE (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0"))
	return new NO_IMPLEMENT (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0"))
	return new BAD_TYPECODE (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/BAD_OPERATION:1.0"))
	return new BAD_OPERATION (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/NO_RESOURCES:1.0"))
	return new NO_RESOURCES (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/NO_RESPONSE:1.0"))
	return new NO_RESPONSE (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/PERSIST_STORE:1.0"))
	return new PERSIST_STORE (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0"))
	return new BAD_INV_ORDER (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/TRANSIENT:1.0"))
	return new TRANSIENT (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/FREE_MEM:1.0"))
	return new FREE_MEM (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/INV_IDENT:1.0"))
	return new INV_IDENT (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/INV_FLAG:1.0"))
	return new INV_FLAG (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/INTF_REPOS:1.0"))
	return new INTF_REPOS (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/BAD_CONTEXT:1.0"))
	return new BAD_CONTEXT (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0"))
	return new OBJ_ADAPTER (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/DATA_CONVERSION:1.0"))
	return new DATA_CONVERSION (minor, completed);

    if (!strcmp (repoid, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0"))
	return new OBJECT_NOT_EXIST (minor, completed);

    return NULL;
}

CORBA::Boolean
CORBA::SystemException::_is_sysex (const char *repoid)
{
    return (!strcmp (repoid, "IDL:omg.org/CORBA/SystemException:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/UNKNOWN:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/BAD_PARAM:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/NO_MEMORY:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/IMP_LIMIT:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/COMM_FAILURE:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/INV_OBJREF:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/NO_PERMISSION:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/INTERNAL:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/MARSHAL:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/INITIALIZE:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/BAD_OPERATION:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/NO_RESOURCES:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/NO_RESPONSE:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/PERSIST_STORE:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/TRANSIENT:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/FREE_MEM:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/INV_IDENT:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/INV_FLAG:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/INTF_REPOS:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/BAD_CONTEXT:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/DATA_CONVERSION:1.0") ||
	    !strcmp (repoid, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0"));
}

CORBA::SystemException *
CORBA::SystemException::_narrow (Exception *e)
{
    if (e && _is_sysex (e->_repoid()))
	return (SystemException *)e;
    return NULL;
}

CORBA::SystemException *
CORBA::SystemException::_decode (Any &a)
{
    ULong k, minor = 0;
    CompletionStatus completed;
    String_var repoid;

    if (!(a.except_get_begin (repoid) &&
	  (a >>= minor) &&
	  a.enum_get (k) &&
	  a.except_get_end ())) {
	return NULL;
    }
    completed = (CompletionStatus)k;

    return _create_sysex (repoid, minor, completed);
}

CORBA::SystemException *
CORBA::SystemException::_decode (DataDecoder &dc)
{
    ULong k, minor;
    CompletionStatus completed;
    string repoid;

    if (!(dc.except_begin (repoid) &&
	  dc.get_ulong (minor) &&
	  dc.enumeration (k) &&
	  dc.except_end ())) {
	return NULL;
    }
    completed = (CompletionStatus)k;

    return _create_sysex (repoid.c_str(), minor, completed);
}

void
CORBA::SystemException::_print (ostream &os) const
{
    os << _repoid() << " (" << _minor << ", ";
    switch (_completed) {
    case COMPLETED_YES:
	os << "completed";
	break;
    case COMPLETED_NO:
	os << "not-completed";
	break;
    case COMPLETED_MAYBE:
	os << "maybe-completed";
	break;
    default:
	assert (0);
    }
    os << ")" << endl;
}

void
CORBA::SystemException::_encode (DataEncoder &ec) const
{
    ec.except_begin (_repoid());
    ec.put_ulong (_minor);
    ec.enumeration ((ULong)_completed);
    ec.except_end ();
}

CORBA::ULong
CORBA::SystemException::minor () const
{
    return _minor;
}

void
CORBA::SystemException::minor (ULong minor)
{
    _minor = minor;
}

CORBA::CompletionStatus
CORBA::SystemException::completed () const
{
    return _completed;
}

void
CORBA::SystemException::completed (CompletionStatus status)
{
    _completed = status;
}


/************************* UserException *************************/


CORBA::UserException::~UserException ()
{
}

void
CORBA::UserException::_throwit () const
{
    assert (0);
}

void
CORBA::UserException::_print (ostream &os) const
{
    os << _repoid() << endl;
}

void
CORBA::UserException::_encode (DataEncoder &ec) const
{
    ec.except_begin ("IDL:omg.org/CORBA/UserException:1.0");
    ec.except_end ();
}

const char *
CORBA::UserException::_repoid () const
{
    return "IDL:omg.org/CORBA/UserException:1.0";
}

CORBA::Exception *
CORBA::UserException::_clone () const
{
    return new UserException (*this);
}

CORBA::UserException *
CORBA::UserException::_narrow (Exception *e)
{
    if (!SystemException::_narrow (e))
	return (UserException *)e;
    return NULL;
}

CORBA::UserException *
CORBA::UserException::_decode (Any &a)
{
    return new UnknownUserException (new Any (a));
}

CORBA::UserException *
CORBA::UserException::_decode (DataDecoder &dc)
{
    return new UnknownUserException (dc.clone ());
}


/********************** UnknownUserException **********************/


CORBA::UnknownUserException::UnknownUserException (Any *a)
{
    _excpt = a;
    _static_except = 0;
    _dc = 0;
}

CORBA::UnknownUserException::UnknownUserException (DataDecoder *dc)
{
    _excpt = 0;
    _static_except = 0;
    _dc = dc;
}

CORBA::UnknownUserException::UnknownUserException ()
{
    _excpt = new Any;
    _static_except = 0;
    _dc = 0;
}

CORBA::UnknownUserException::UnknownUserException (
    const UnknownUserException &ex)
{
    _excpt = ex._excpt ? new Any (*ex._excpt) : 0;
    _static_except =
	ex._static_except ? new StaticAny (*ex._static_except) : 0;
    _dc = ex._dc ? ex._dc->clone () : 0;
}

CORBA::UnknownUserException &
CORBA::UnknownUserException::operator= (const UnknownUserException &ex)
{
    if (_excpt)
	delete _excpt;
    if (_static_except)
	delete _static_except;
    if (_dc)
	delete _dc;
    _excpt = ex._excpt ? new Any (*ex._excpt) : 0;
    _static_except =
	ex._static_except ? new StaticAny (*ex._static_except) : 0;
    _dc = ex._dc ? ex._dc->clone () : 0;
    return *this;
}

CORBA::UnknownUserException::~UnknownUserException ()
{
    if (_excpt)
	delete _excpt;
    if (_static_except)
	delete _static_except;
    if (_dc)
	delete _dc;
}

CORBA::Any &
CORBA::UnknownUserException::exception (TypeCode_ptr tc)
{
    if (!_excpt) {
	assert (tc);
	assert (_dc);
	_excpt = new Any;
	CORBA::Boolean r = _excpt->demarshal (tc, *_dc);
	assert (r);
    }
    return *_excpt;
}

CORBA::StaticAny &
CORBA::UnknownUserException::exception (StaticTypeInfo *stc)
{
    if (_static_except)
	return *_static_except;

    _static_except = new CORBA::StaticAny (stc);

    if (_excpt) {
	// XXX slow ...
	MICO::SimpleEncoder ec;
	MICO::SimpleDecoder dc (ec.buffer(), FALSE);

	_excpt->marshal (ec);

	CORBA::Boolean ret = _static_except->demarshal (dc);
	assert (ret);
    } else {
	assert (_dc);
	CORBA::Long rpos = _dc->buffer()->rpos();

	CORBA::Boolean ret = _static_except->demarshal (*_dc);
	assert (ret);

	_dc->buffer()->rseek_beg (rpos);
    }
    return *_static_except;
}

void
CORBA::UnknownUserException::_throwit () const
{
    assert (0);
}

void
CORBA::UnknownUserException::_print (ostream &os) const
{
    os << ((UnknownUserException *)this)->_except_repoid()
       << " within " << _repoid() << endl;
}

void
CORBA::UnknownUserException::_encode (DataEncoder &ec) const
{
    if (_excpt) {
	_excpt->marshal (ec);
    } else {
	assert (_dc);
	// XXX alignment, differing codecs ???
	assert (!strcmp (ec.type(), _dc->type()));
	ec.put_octets (_dc->buffer()->data(), _dc->buffer()->length());
    }
}

const char *
CORBA::UnknownUserException::_except_repoid ()
{
    if (_excpt) {
	String_var repoid;
	CORBA::Boolean r =
	    _excpt->except_get_begin (repoid) &&
	    _excpt->except_get_end ();
	assert (r);
	_ex_repoid = repoid;
    } else {
	assert (_dc);
	ULong rpos = _dc->buffer()->rpos();
	CORBA::Boolean r = _dc->except_begin (_ex_repoid);
	assert (r);
	_dc->buffer()->rseek_beg (rpos);
    }
    return _ex_repoid.c_str();
}

const char *
CORBA::UnknownUserException::_repoid () const
{
    return "IDL:omg.org/CORBA/UnknownUserException:1.0";
}

CORBA::Exception *
CORBA::UnknownUserException::_clone () const
{
    return new UnknownUserException (*this);
}

CORBA::UnknownUserException *
CORBA::UnknownUserException::_narrow (Exception *e)
{
    if (e && !strcmp (e->_repoid(),
		      "IDL:omg.org/CORBA/UnknownUserException:1.0"))
	return (UnknownUserException *)e;
    return NULL;
}


/*********************** SystemException's ***********************/


#define SYSEXC(name)							\
									\
CORBA::name::name ()							\
{									\
}									\
									\
CORBA::name::name (const name &ex)					\
    : CORBA::SystemException (ex)      					\
{									\
}									\
									\
CORBA::name::name (ULong minor, CompletionStatus status)		\
    : CORBA::SystemException (minor, status)			       	\
									\
{									\
}									\
									\
CORBA::name &								\
CORBA::name::operator= (const name &ex)					\
{									\
    MICO_SCOPE(CORBA,SystemException::operator=) (ex);			\
    return *this;							\
}									\
									\
CORBA::name::~name ()							\
{									\
}									\
									\
CORBA::Exception *							\
CORBA::name::_clone () const						\
{									\
    return new name (*this);						\
}									\
									\
const char *								\
CORBA::name::_repoid ()	const						\
{									\
    return "IDL:omg.org/CORBA/" #name ":1.0";			       	\
}									\
									\
CORBA::name *								\
CORBA::name::_narrow (Exception *ex)					\
{									\
    if (ex && !strcmp (ex->_repoid(), "IDL:omg.org/CORBA/" #name ":1.0")) \
	return (name *)ex;						\
    return NULL;							\
}

SYSEXC (UNKNOWN)
SYSEXC (BAD_PARAM)
SYSEXC (NO_MEMORY)
SYSEXC (IMP_LIMIT)
SYSEXC (COMM_FAILURE)
SYSEXC (INV_OBJREF)
SYSEXC (NO_PERMISSION)
SYSEXC (INTERNAL)
SYSEXC (MARSHAL)
SYSEXC (INITIALIZE)
SYSEXC (NO_IMPLEMENT)
SYSEXC (BAD_TYPECODE)
SYSEXC (BAD_OPERATION)
SYSEXC (NO_RESOURCES)
SYSEXC (NO_RESPONSE)
SYSEXC (PERSIST_STORE)
SYSEXC (BAD_INV_ORDER)
SYSEXC (TRANSIENT)
SYSEXC (FREE_MEM)
SYSEXC (INV_IDENT)
SYSEXC (INV_FLAG)
SYSEXC (INTF_REPOS)
SYSEXC (BAD_CONTEXT)
SYSEXC (OBJ_ADAPTER)
SYSEXC (DATA_CONVERSION)
SYSEXC (OBJECT_NOT_EXIST)

#undef SYSEXC

