/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#include <mico/template_impl.h>
#include <iostream.h>
#include <mico/os-math.h>

/*
 * one would usually use pow(10,x) but for some strange
 * reasons this segfaults on AIX using egcs 1.0.2 in
 * from_longdouble(). However, if you write a test program
 * with just what from_longdouble() does in it everything
 * works fine. This is what is called a Heisenbug.
 */
static MICO_LongDouble
pow10 (int x)
{
    if (x == 0)
        return 1;
    int neg = 0;
    if (x < 0) {
	neg = 1;
	x = -x;
    }
    MICO_LongDouble s = 1, t = 10;
    while (x) {
	if (x & 1)
	    s *= t;
	x >>= 1;
	t *= t;
    }
    if (neg)
	s = 1/s;
    return s;
}


FixedBase::FixedBase ()
{
    assert (0);
}

FixedBase::FixedBase (MICO_UShort d, MICO_Short s)
    : _digits(d), _scale(s), _val(0.0)
{
}

FixedBase::FixedBase (const FixedBase &f)
    : _digits(f._digits), _scale(f._scale), _val(f._val)
{
}

FixedBase &
FixedBase::operator= (const FixedBase &f)
{
    _val = f.transform (_scale);
    truncate ();
    return *this;
}

FixedBase::~FixedBase ()
{
}

void
FixedBase::truncate ()
{
    _val = fmodl (_val, pow10 (_digits));
    if (_val < 0) {
	_val = ceill (_val - 0.5);
    } else {
	_val = floorl (_val + 0.5);
    }
}

MICO_LongDouble
FixedBase::transform (MICO_Short nscale) const
{
    if (_scale == nscale)
	return _val;
    return _val * pow10 (nscale - _scale);
}

MICO_LongDouble
FixedBase::to_longdouble () const
{
    return _val / pow10 (_scale);
}

void
FixedBase::from_longdouble (MICO_LongDouble val)
{
    _val = val * pow10 (_scale);
    truncate ();
}

static MICO_LongDouble
epsilon ()
{
    static MICO_LongDouble d = 0.0;
    if (d == 0.0) {
	/*
	 * calculate the error in representing 0.6 in a LongDouble.
	 */
	d = 6;
	d /= 10;
	/*
	 * subtract 0.5 before multiply instead of subtracting 5
	 * afterwards to make sure multiplying by 10 doesn't overflow
	 * mantissa.
	 */
	d -= 0.5;
	d *= 10;
	d -= 1;

	int exp;
	frexpl (d, &exp);

	d = ldexpl (1, exp);
	/*
	 * now d is the smallest power of 2 that is greater than
	 * the error in representing 0.6 as a LongDouble
	 */
    }
    return d;
}

void
FixedBase::compute_params (MICO_LongDouble val,
			   MICO_UShort &digits, MICO_Short &scale)
{
    digits = 0;
    scale = 0;

    if (val == 0.0)
	return;
    if (val < 0)
	val = -val;

    int exp;
    frexpl (val, &exp);
    MICO_LongDouble eps = ldexpl (epsilon(), exp);

    /*
     * if there is no real floorl() then fabs(rem) is still >= 1 afterwards
     * if val doesn't fit into a double. but subtracting floor(rem)
     * again will make fabs(rem) < 1, which is done in the below loop...
     *
     * this way we get the full long double precision even if there are
     * no math functions for long double.
     */
    MICO_LongDouble rem = val - floorl (val);

    /*
     * shift val to the left, until the fractional part or 1 minus
     * the fractional part becomes <= eps.
     * eps takes into account the number of mantissa bits used for the
     * fractional part of the value and grows as we shift to the left.
     * this will only work if the value has at least one significant
     * digit less than the maximum number of significant digits supported
     * by LongDouble. otherwise we cannot decide whether e.g. 0.2999999...
     * is the LongDouble approximation for 0.3 (which has one siginificant
     * digit) or the user specified number 0.2999999... (which has > 1
     * significant digits).
     */
    while (42) {
	rem -= floorl (rem);
	if (1 - rem <= eps || rem <= eps) {
	    val = floorl (val);
	    break;
	}
	val *= 10;
	eps *= 10;
	rem *= 10;
	++scale;
    }

    // shift val to the right, until the 10^0 coeff becomes nonzero
    if (scale == 0) {
	while (fmodl (val, 10) == 0.0) {
	    val /= 10;
	    --scale;
	}
    }

    // now count the nonzero digits to the left of the decimal point
    while (val != 0.0) {
	++digits;
	val = floorl (val / 10);
    }
}

FixedBase::FixedValue*
FixedBase::to_digits () const
{
    FixedValue *digs = new FixedValue;
    digs->length (_digits+1);

    MICO_LongDouble d = fabsl (_val) + 0.5;
    MICO_LongDouble q = pow10 (_digits-1);

    for (int i = 0; i < _digits; ++i) {
	int r = (int)(d/q);
	d -= r*q;
	q /= 10;
	(*digs)[i] = (MICO_Octet)r;
    }
    (*digs)[_digits] = _val < 0 ? 1 : 0;

    return digs;
}

void
FixedBase::from_digits (const FixedValue &digs)
{
    assert (digs.length() == (CORBA::ULong)_digits+1);

    _val = 0.0;

    for (int i = 0; i < _digits; ++i) {
	_val *= 10;
	_val += (CORBA::Octet)digs[i];
    }

    if (digs[_digits])
	_val = -_val;

    // shouldn't be necessary ...
    truncate ();
}

istream &
FixedBase::read (istream &i)
{
    /*
     * operator>>(istream &, long double &) is broken on most systems,
     * so we cheat a bit and use operator>>(istream &, double &)...
     */
    CORBA::Double d;
    i >> d;
    _val = d * pow10 (_scale);
    truncate ();
    return i;
}

ostream &
FixedBase::write (ostream &o) const
{
    FixedValue_var digs = to_digits();
    if (digs[(CORBA::ULong)_digits])
	o << "-";
    if (_scale > _digits) {
	o << ".";
	for (int i = _digits; i < _scale; ++i)
	    o << "0";
    }
    for (CORBA::ULong i0 = 0; i0 < _digits; ++i0) {
	if (i0 == (CORBA::ULong)(_digits-_scale))
	    o << ".";
        o << (int)(CORBA::Octet)digs[i0];
    }
    for (int i1 = _scale; i1 < 0; ++i1)
	o << "0";
    return o;
}

void
FixedBase::add (FixedBase &res, const FixedBase &a1, const FixedBase a2)
{
    MICO_UShort res_scale;

    if (a1._scale == a2._scale) {
	res._val = a1._val + a2._val;
	res_scale = a1._scale;
    } else if (a1._scale > a2._scale) {
	res._val = a1._val + a2.transform (a1._scale);
	res_scale = a1._scale;
    } else {
	res._val = a1.transform (a2._scale) + a2._val;
	res_scale = a2._scale;
    }
    if (res._scale != res_scale)
	res._val *= pow10 (res._scale - res_scale);
    res.truncate ();
}

void
FixedBase::sub (FixedBase &res, const FixedBase &a1, const FixedBase a2)
{
    MICO_UShort res_scale;

    if (a1._scale == a2._scale) {
	res._val = a1._val - a2._val;
	res_scale = a1._scale;
    } else if (a1._scale > a2._scale) {
	res._val = a1._val - a2.transform (a1._scale);
	res_scale = a1._scale;
    } else {
	res._val = a1.transform (a2._scale) - a2._val;
	res_scale = a2._scale;
    }
    if (res._scale != res_scale)
	res._val *= pow10 (res._scale - res_scale);
    res.truncate ();
}

void
FixedBase::mul (FixedBase &res, const FixedBase &a1, const FixedBase a2)
{
    res._val = (a1._val * a2._val) / pow10 (a2._scale);
    if (res._scale != a1._scale)
	res._val *= pow10 (res._scale - a1._scale);
    res.truncate ();
}

void
FixedBase::div (FixedBase &res, const FixedBase &a1, const FixedBase a2)
{
    res._val = (a1._val * pow10 (a2._scale)) / a2._val;
    if (res._scale != a1._scale)
	res._val *= pow10 (res._scale - a1._scale);
    res.truncate ();
}

void
FixedBase::neg (FixedBase &res, const FixedBase &a)
{
    res._val = - a._val;
    if (res._scale != a._scale)
	res._val *= pow10 (res._scale - a._scale);
}

int
FixedBase::eq (const FixedBase &a1, const FixedBase &a2)
{
    if (a1._scale == a2._scale)
	return fabsl (a1._val - a2._val) < 0.1;
    if (a1._scale > a2._scale)
	return fabsl (a1._val - a2.transform (a1._scale)) < 0.1;
    return fabsl (a1.transform (a2._scale) - a2._val) < 0.1;
}

int
FixedBase::lt (const FixedBase &a1, const FixedBase &a2)
{
    if (a1._scale == a2._scale)
	return a1._val < a2._val;
    if (a1._scale > a2._scale)
	return a1._val < a2.transform (a1._scale);
    return a1.transform (a2._scale) < a2._val;
}

int
FixedBase::le (const FixedBase &a1, const FixedBase &a2)
{
    if (a1._scale == a2._scale)
	return a1._val <= a2._val;
    if (a1._scale > a2._scale)
	return a1._val <= a2.transform (a1._scale);
    return a1.transform (a2._scale) <= a2._val;
}

