/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#ifndef _WINDOWS
#include <string.h>
#endif
#include <iostream.h>
#include <mico/impl.h>
#include <mico/intercept.h>
#include <mico/template_impl.h>
#include <mico/util.h>


static MICO::CodesetConvDB the_codeset_db;


/****************************** misc dtors *******************************/


MICO::GIOPConnCallback::~GIOPConnCallback ()
{
}


/**************************** ServiceContext *****************************/


CORBA::IOP::ServiceContext::ServiceContext ()
{
}

CORBA::IOP::ServiceContext::ServiceContext (const ServiceContext &s)
    : context_id (s.context_id),
      context_data (s.context_data)
{
}

CORBA::IOP::ServiceContext::~ServiceContext ()
{
}

CORBA::IOP::ServiceContext &
CORBA::IOP::ServiceContext::operator= (const ServiceContext &s)
{
    context_id = s.context_id;
    context_data = s.context_data;
    return *this;
}


CORBA::IOP::TaggedComponent::TaggedComponent ()
{
}

CORBA::IOP::TaggedComponent::TaggedComponent (const TaggedComponent &s)
    : tag (s.tag), component_data (s.component_data)
{
}

CORBA::IOP::TaggedComponent::~TaggedComponent ()
{
}

CORBA::IOP::TaggedComponent &
CORBA::IOP::TaggedComponent::operator= (const TaggedComponent &s)
{
    tag = s.tag;
    component_data = s.component_data;
    return *this;
}


/***************************** GIOPInContext *****************************/


MICO::GIOPInContext::GIOPInContext (GIOPCodec *codec, CORBA::Buffer *buf,
				    CORBA::CodesetConv *csc,
				    CORBA::CodesetConv *wcsc)
{
#ifdef HAVE_THREADS
    _dc = codec->dc_proto()->clone (buf, TRUE, csc, FALSE, wcsc, FALSE);
    _delete_dc = TRUE;
    _delete_buf = FALSE;
#else
    _dc = codec->dc_proto();
    _dc->buffer (buf, FALSE);
    _dc->converter (csc, FALSE);
    _dc->wconverter (wcsc, FALSE);
    _delete_dc = FALSE;
    _delete_buf = TRUE;
#endif
}

MICO::GIOPInContext::~GIOPInContext ()
{
    if (_delete_buf)
	delete _dc->buffer();
    if (_delete_dc)
	delete _dc;
}

void
MICO::GIOPInContext::buffer (CORBA::Buffer *b)
{
    _delete_buf = FALSE;
    _dc->buffer (b, FALSE);
}

void
MICO::GIOPInContext::converters (CORBA::CodesetConv *csc,
				 CORBA::CodesetConv *wcsc)
{
    _dc->converter (csc, FALSE);
    _dc->wconverter (wcsc, FALSE);
}

CORBA::DataDecoder *
MICO::GIOPInContext::_retn ()
{
#ifdef HAVE_THREADS
    assert (_delete_dc);
    _delete_dc = FALSE;
    return _dc;
#else
    assert (_delete_buf);
    _delete_buf = FALSE;
    return _dc->clone (_dc->buffer(), TRUE,
		       _dc->converter(), FALSE,
		       _dc->wconverter(), FALSE);
#endif
}


/***************************** GIOPOutContext *****************************/


MICO::GIOPOutContext::GIOPOutContext (GIOPCodec *codec,
				      CORBA::CodesetConv *csc,
				      CORBA::CodesetConv *wcsc)
{
#ifdef HAVE_THREADS
    _ec = codec->ec_proto()->clone (new CORBA::Buffer(), FALSE,
				    csc, FALSE, wcsc, FALSE);
#else
    _ec = codec->ec_proto();
    _ec->buffer (new CORBA::Buffer(), FALSE);
    _ec->converter (csc, FALSE);
    _ec->wconverter (wcsc, FALSE);
#endif
    _deleteme = TRUE;
}

MICO::GIOPOutContext::GIOPOutContext (CORBA::DataEncoder *ec)
{
    _ec = ec;
    _deleteme = FALSE;
}

MICO::GIOPOutContext::~GIOPOutContext ()
{
    if (_deleteme) {
	delete _ec->buffer();
#ifdef HAVE_THREADS
	delete _ec;
#endif
    }
}

void
MICO::GIOPOutContext::converters (CORBA::CodesetConv *csc,
				  CORBA::CodesetConv *wcsc)
{
    _ec->converter (csc, FALSE);
    _ec->wconverter (wcsc, FALSE);
}

CORBA::Buffer *
MICO::GIOPOutContext::_retn()
{
    assert (_deleteme);
    _deleteme = FALSE;
    return _ec->buffer();
}


/******************************* GIOPCodec *******************************/


MICO::GIOPCodec::GIOPCodec (CORBA::DataDecoder *_dc, CORBA::DataEncoder *_ec)
    : _dc_proto(_dc), _ec_proto(_ec)
{
    GIOPOutContext ctx (this);
    size_offset = put_header (ctx, CORBA::GIOP::Request);
    headerlen = ctx.ec()->buffer()->length();

    giop_ver_major = 1;
    giop_ver_minor = 0;
}

MICO::GIOPCodec::~GIOPCodec ()
{
    delete _ec_proto;
    delete _dc_proto;
}

CORBA::ULong
MICO::GIOPCodec::put_header (GIOPOutContext &out, CORBA::GIOP::MsgType mt)
{
    CORBA::ULong key;

    CORBA::DataEncoder *ec = out.ec();

    ec->struct_begin();
    {
	ec->arr_begin();
	{
	    // these are really chars, but must be encoded in latin1 ...
	    ec->put_char_raw ('G');
	    ec->put_char_raw ('I');
	    ec->put_char_raw ('O');
	    ec->put_char_raw ('P');
	}
	ec->arr_end();

	ec->struct_begin();
	{
	    ec->put_octet (giop_ver_major);
	    ec->put_octet (giop_ver_minor);
	}
	ec->struct_end();

        if (giop_ver_minor == 0) {
            ec->put_boolean (ec->byteorder() == CORBA::LittleEndian);
        } else {
            ec->put_octet ((ec->byteorder() == CORBA::LittleEndian)
                ? GIOP_BYTEORDER_BIT : 0);
        }
	ec->put_octet (mt);
	key = ec->buffer()->wpos();
	ec->put_ulong (0);
    }
    ec->struct_end();

    return key;
}

void
MICO::GIOPCodec::put_contextlist (GIOPOutContext &out,
				  const CORBA::IOP::ServiceContextList &ctx,
				  CORBA::Boolean codesets)
{
    CORBA::DataEncoder *ec = out.ec();

    if (CORBA::Codeset::disabled())
	codesets = FALSE;

    ec->seq_begin (ctx.length() + !!codesets);
    {
	CORBA::ULong state, key;
	for (CORBA::ULong i = 0; i < ctx.length(); ++i) {
	    ec->struct_begin();
	    {
		ec->put_ulong (ctx[i].context_id);
		key = ec->encaps_begin (state);
		if (ctx[i].context_data.length() > 0) {
		    ec->put_octets (
			(const CORBA::Octet *)&ctx[i].context_data[0],
			ctx[i].context_data.length());
		}
		ec->encaps_end (key, state);
	    }
	    ec->struct_end();
	}

	if (codesets) {
	    // code set service
	    CORBA::Codeset::CodesetId csid = 0, wcsid = 0;

	    if (ec->converter())
		csid = ec->converter()->to()->id();
	    if (ec->wconverter())
		wcsid = ec->wconverter()->to()->id();

	    ec->struct_begin();
	    {
		ec->put_ulong (CORBA::IOP::CodeSets);
		key = ec->encaps_begin (state);
		ec->struct_begin ();
		{
		    ec->put_ulong (csid);
		    ec->put_ulong (wcsid);
		}
		ec->struct_end ();
		ec->encaps_end (key, state);
	    }
	    ec->struct_end();
	}
    }
    ec->seq_end ();
}

void
MICO::GIOPCodec::put_objectkey (GIOPOutContext &out, CORBA::Object_ptr obj)
{
    CORBA::DataEncoder *ec = out.ec();

    // XXX assumes all the profiles of an IOR have the same objectkey ...
    CORBA::IORProfile *prof = obj->_ior()->profile ();
    assert (prof);

    CORBA::Long objkeylen;
    const CORBA::Octet *objkey = prof->objectkey (objkeylen);

    // XXX encapsulated or simple sequence ???
    ec->seq_begin (objkeylen);
    ec->put_octets (objkey, objkeylen);
    ec->seq_end ();
}

void
MICO::GIOPCodec::put_args (GIOPOutContext &out, CORBA::ORBRequest *req,
			   CORBA::Boolean inp)
{
    CORBA::DataEncoder *ec = out.ec();

    ec->struct_begin ();
    {
	if (inp) {
	    CORBA::Boolean ret = req->get_in_args (ec);
            assert (ret);
	} else {
	    CORBA::Boolean is_except;
	    CORBA::Boolean ret = req->get_out_args (ec, is_except);
            assert (ret);
	}
    }
    ec->struct_end ();
}

void
MICO::GIOPCodec::put_size (GIOPOutContext &out, CORBA::ULong key)
{
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ULong end_pos = ec->buffer()->wpos();
    ec->buffer()->wseek_beg (key);
    ec->put_ulong (0);
    ec->struct_end();
    CORBA::ULong start_pos = ec->buffer()->wpos();

    ec->buffer()->wseek_beg (key);
    ec->put_ulong (end_pos - start_pos);

    ec->buffer()->wseek_beg (end_pos);
}

CORBA::Boolean
MICO::GIOPCodec::put_invoke_request (GIOPOutContext &out,
				     CORBA::ULong req_id,
				     CORBA::Boolean response_exp,
				     CORBA::Object_ptr obj,
				     CORBA::ORBRequest *req,
				     CORBA::Principal_ptr pr)
{
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ByteOrder bo = ec->byteorder();
    if (!strcmp (req->type(), "giop"))
        ec->byteorder (((GIOPRequest *)req)->input_byteorder());

    CORBA::ULong key = put_header (out, CORBA::GIOP::Request);

    ec->struct_begin ();
    {
	put_contextlist (out, *req->context(), TRUE);
	ec->put_ulong (req_id);
	ec->put_boolean (response_exp);
        if (giop_ver_minor != 0) {
            ec->put_octet (0);
            ec->put_octet (0);
            ec->put_octet (0);
        }
	put_objectkey (out, obj);
	ec->put_string_raw (req->op_name());
	ec->put_principal (pr ? *pr : CORBA::Principal());
    }
    ec->struct_end ();

    put_args (out, req, TRUE);
    put_size (out, key);

    ec->byteorder (bo);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_invoke_reply_offset (GIOPOutContext &out,
					  CORBA::ORBRequest *req)
{
    CORBA::DataEncoder *ec = out.ec();

    ec->buffer()->wseek_rel (headerlen);

    ec->struct_begin ();
    {
	put_contextlist (out, *req->context());
	ec->put_ulong (0);
	ec->enumeration (0);
    }
    ec->struct_end ();

    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_invoke_reply (GIOPOutContext &out,
				   CORBA::ULong req_id,
				   CORBA::GIOP::ReplyStatusType stat,
				   CORBA::Object_ptr obj,
				   CORBA::ORBRequest *req)
{
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ByteOrder bo = ec->byteorder();
    if (!strcmp (req->type(), "giop"))
        ec->byteorder (((GIOPRequest *)req)->output_byteorder());

    CORBA::ULong key = put_header (out, CORBA::GIOP::Reply);

    ec->struct_begin ();
    {
	put_contextlist (out, *req->context());
	ec->put_ulong (req_id);
	ec->enumeration ((CORBA::ULong)stat);
    }
    ec->struct_end ();

    switch (stat) {
    case CORBA::GIOP::NO_EXCEPTION:
    case CORBA::GIOP::USER_EXCEPTION:
    case CORBA::GIOP::SYSTEM_EXCEPTION:
	put_args (out, req, FALSE);
	break;

    case CORBA::GIOP::LOCATION_FORWARD:
	ec->put_ior (*obj->_ior());
	break;

    default:
	assert (0);
	break;
    }
    put_size (out, key);

    ec->byteorder (bo);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_cancel_request (GIOPOutContext &out, CORBA::ULong req_id)
{
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ULong key = put_header (out, CORBA::GIOP::CancelRequest);

    ec->struct_begin();
    {
	ec->put_ulong (req_id);
    }
    ec->struct_end();

    put_size (out, key);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_locate_request (GIOPOutContext &out,
				     CORBA::ULong req_id,
				     CORBA::Object_ptr obj)
{
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ULong key = put_header (out, CORBA::GIOP::LocateRequest);

    ec->struct_begin();
    {
	ec->put_ulong (req_id);
	put_objectkey (out, obj);
    }
    ec->struct_end();

    put_size (out, key);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_locate_reply (GIOPOutContext &out,
				   CORBA::ULong req_id,
				   CORBA::GIOP::LocateStatusType stat,
				   CORBA::Object_ptr obj)
{
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ULong key = put_header (out, CORBA::GIOP::LocateReply);

    ec->struct_begin();
    {
	ec->put_ulong (req_id);
	ec->enumeration ((CORBA::ULong)stat);
    }
    ec->struct_end();

    if (stat == CORBA::GIOP::OBJECT_FORWARD) {
	ec->put_ior (*obj->_ior());
    }
    put_size (out, key);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_bind_request (GIOPOutContext &out, CORBA::ULong req_id,
				   const char *repoid,
				   const CORBA::ORB::ObjectTag &oid)
{
    /*
     * mapped to an invocation:
     *   GIOP::LocateStatusType _bind (in string repoid,
     *				       in sequence<octet> oid,
     *				       out object obj);
     */
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ULong key = put_header (out, CORBA::GIOP::Request);

    ec->struct_begin ();
    {
	put_contextlist (out, CORBA::IOP::ServiceContextList());
	ec->put_ulong (req_id);
	ec->put_boolean (TRUE);
        if (giop_ver_minor != 0) {
            ec->put_octet (0);
            ec->put_octet (0);
            ec->put_octet (0);
        }
	// object key
	ec->seq_begin (1);
	ec->put_octet (0);
	ec->seq_end ();
	ec->put_string ("_bind");
	ec->put_principal (CORBA::Principal());
    }
    ec->struct_end ();

    ec->struct_begin ();
    {
	ec->put_string (repoid);
	ec->seq_begin (oid.length());
	if (oid.length() > 0)
	    ec->put_octets ((const CORBA::Octet *)&oid[0], oid.length());
	ec->seq_end ();
    }
    ec->struct_end ();

    put_size (out, key);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_bind_reply (GIOPOutContext &out,
				 CORBA::ULong req_id,
				 CORBA::GIOP::LocateStatusType stat,
				 CORBA::Object_ptr obj)
{
    /*
     * mapped to an invocation:
     *   GIOP::LocateStatusType _bind (in string repoid,
     *				       in sequence<octet> oid,
     *				       out object obj);
     */
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ULong key = put_header (out, CORBA::GIOP::Reply);

    ec->struct_begin ();
    {
	put_contextlist (out, CORBA::IOP::ServiceContextList());
	ec->put_ulong (req_id);
	ec->enumeration ((CORBA::ULong)CORBA::GIOP::NO_EXCEPTION);
    }
    ec->struct_end ();

    ec->struct_begin ();
    {
	ec->enumeration ((CORBA::ULong)stat);
	ec->put_ior (stat == CORBA::GIOP::OBJECT_HERE
		     ? *obj->_ior() : CORBA::IOR());
    }
    ec->struct_end ();

    put_size (out, key);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_close_msg (GIOPOutContext &out)
{
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ULong key = put_header (out, CORBA::GIOP::CloseConnection);
    put_size (out, key);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::put_error_msg (GIOPOutContext &out)
{
    CORBA::DataEncoder *ec = out.ec();

    CORBA::ULong key = put_header (out, CORBA::GIOP::MessageError);
    put_size (out, key);
    return TRUE;
}

#define check(exp) if (!(exp)) return FALSE

CORBA::Boolean
MICO::GIOPCodec::get_header (GIOPInContext &in, CORBA::GIOP::MsgType &mt,
			     CORBA::ULong &sz, CORBA::Octet &flags)
{
    CORBA::DataDecoder *dc = in.dc();

    CORBA::Boolean b;
    CORBA::Octet o;
    CORBA::Char c;

    check (dc->struct_begin());
    {
	check (dc->arr_begin());
	{
	    // these are really chars, but must be encoded in latin1 ...
	    check (dc->get_char_raw (c) && c == 'G');
	    check (dc->get_char_raw (c) && c == 'I');
	    check (dc->get_char_raw (c) && c == 'O');
	    check (dc->get_char_raw (c) && c == 'P');
	}
	check (dc->arr_end());

	check (dc->struct_begin());
	{
	    check (dc->get_octet (c) && c == 1);
	    check (dc->get_octet (c) && (c == 0 || c == 1));
            giop_ver_minor = c;
	}
	check (dc->struct_end());

        if (giop_ver_minor == 0) {
            check (dc->get_boolean (b));
            dc->byteorder (b ? CORBA::LittleEndian
                           : CORBA::BigEndian);
            flags = b ? 1 : 0;
        } else {
            check (dc->get_octet (flags));
            dc->byteorder ((flags & GIOP_BYTEORDER_BIT) ? CORBA::LittleEndian
                           : CORBA::BigEndian);
        }
	check (dc->get_octet (o));
	mt = (CORBA::GIOP::MsgType)o;
	check (dc->get_ulong (sz));
    }
    check (dc->struct_end());
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_contextlist (GIOPInContext &in,
				  CORBA::IOP::ServiceContextList &ctx,
				  CORBA::Boolean codesets)
{
    CORBA::DataDecoder *dc = in.dc();

    CORBA::Codeset::CodesetId csid = 0, wcsid = 0;

    CORBA::ULong state, len;
    check (dc->seq_begin (len));
    ctx.length (len);
    {
	CORBA::ULong context_id;
	for (CORBA::ULong i = 0; i < ctx.length(); ) {
	    check (dc->struct_begin());
	    {
		check (dc->get_ulong (context_id));
		check (dc->encaps_begin (state, len));
		if (context_id == CORBA::IOP::CodeSets) {
		    // code set service ...
		    ctx.length (ctx.length()-1);
		    check (dc->struct_begin());
		    {
			check (dc->get_ulong (csid));
			check (dc->get_ulong (wcsid));
		    }
		    check (dc->struct_end());
		} else {
		    // other service ...
		    ctx[i].context_id = context_id;
		    ctx[i].context_data.length (len);
		    if (len > 0) {
			check (dc->get_octets (
			    (CORBA::Octet *)&ctx[i].context_data[0],
			    ctx[i].context_data.length()));
		    }
		    ++i;
		}
		check (dc->encaps_end (state));
	    }
	    check (dc->struct_end());
	}
    }
    check (dc->seq_end ());

    // set up code sets ...
    if (codesets) {
	CORBA::CodesetConv *wconv = 0, *conv = 0;

	if (csid)
	    conv = the_codeset_db.find (csid, CORBA::Codeset::special_cs(
		CORBA::Codeset::NativeCS)->id());

	if (wcsid)
	    wconv = the_codeset_db.find (wcsid, CORBA::Codeset::special_cs(
		CORBA::Codeset::NativeWCS)->id());

	in.converters (conv, wconv);
    }
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_objectkey (GIOPInContext &in, CORBA::Object_ptr obj)
{
    CORBA::DataDecoder *dc = in.dc();

    CORBA::ULong objkeylen;
    CORBA::Octet *objkey;

    check (dc->seq_begin (objkeylen));
    objkey = dc->buffer()->data();
    dc->buffer()->rseek_rel (objkeylen);
    check (dc->seq_end ());

    // XXX should make sure the seeked over data is still valid ...
    obj->_ior()->objectkey (objkey, objkeylen);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::check_header (GIOPInContext &in, CORBA::GIOP::MsgType &mt,
			       CORBA::ULong &sz, CORBA::Octet &flags)
{
    CORBA::ULong pos = in.dc()->buffer()->rpos ();
    CORBA::Boolean ret = get_header (in, mt, sz, flags);
    in.dc()->buffer()->rseek_beg (pos);
    return ret;
}

CORBA::Boolean
MICO::GIOPCodec::get_invoke_request (GIOPInContext &in,
				     CORBA::ULong &req_id,
				     CORBA::Boolean &response_exp,
				     CORBA::Object_ptr obj,
				     CORBA::ORBRequest * &req,
				     CORBA::Principal_ptr pr)
{
    CORBA::DataDecoder *dc = in.dc();

    CORBA::String_var opname;
    CORBA::IOP::ServiceContextList ctx;

    check (dc->struct_begin ());
    {
	check (get_contextlist (in, ctx, TRUE));
	check (dc->get_ulong (req_id));
	check (dc->get_boolean (response_exp));
        if (giop_ver_minor != 0) {
            CORBA::Octet o;
            check (dc->get_octet (o));
            check (dc->get_octet (o));
            check (dc->get_octet (o));
        }
	check (get_objectkey (in, obj));
	check (dc->get_string_raw (opname));
	check (dc->get_principal (*pr));
    }
    check (dc->struct_end ());

    req = new GIOPRequest (opname, in._retn(), this);
    req->context (&ctx);

    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_invoke_reply1 (GIOPInContext &in,
				    CORBA::ULong &req_id,
				    CORBA::GIOP::ReplyStatusType &stat,
				    CORBA::IOP::ServiceContextList &ctx)
{
    CORBA::DataDecoder *dc = in.dc();

    CORBA::ULong k;

    check (dc->struct_begin ());
    {
	check (get_contextlist (in, ctx));
	check (dc->get_ulong (req_id));
	check (dc->enumeration (k));
        stat = (CORBA::GIOP::ReplyStatusType)k;
    }
    check (dc->struct_end ());

    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_invoke_reply2 (GIOPInContext &in,
				    CORBA::ULong req_id,
				    CORBA::GIOP::ReplyStatusType stat,
				    CORBA::Object_ptr &obj,
				    CORBA::ORBRequest *req,
				    CORBA::IOP::ServiceContextList &ctx)
{
    CORBA::DataDecoder *dc = in.dc();

    obj = CORBA::Object::_nil();

    switch (stat) {
    case CORBA::GIOP::NO_EXCEPTION:
	if (req) {
	    // may be NULL for a bind()
	    req->context (&ctx);
	    return req->set_out_args (dc, FALSE);
	}
	break;

    case CORBA::GIOP::USER_EXCEPTION:
    case CORBA::GIOP::SYSTEM_EXCEPTION:
	if (req) {
	    // may be NULL for a bind()
	    req->context (&ctx);
	    return req->set_out_args (dc, TRUE);
	}
	break;

    case CORBA::GIOP::LOCATION_FORWARD: {
        CORBA::IOR *ior = new CORBA::IOR ("", CORBA::IOR::IORProfileVec());
        check (dc->get_ior (*ior));
        obj = new CORBA::Object (ior);
	break;
    }
    default:
	assert (0);
	break;
    }
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_cancel_request (GIOPInContext &in, CORBA::ULong &req_id)
{
    CORBA::DataDecoder *dc = in.dc();

    check (dc->struct_begin());
    {
	check (dc->get_ulong (req_id));
    }
    check (dc->struct_end());
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_locate_request (GIOPInContext &in,
				     CORBA::ULong &req_id,
				     CORBA::Object_ptr obj)
{
    CORBA::DataDecoder *dc = in.dc();

    check (dc->struct_begin());
    {
	check (dc->get_ulong (req_id));
	check (get_objectkey (in, obj));
    }
    check (dc->struct_end());
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_locate_reply (GIOPInContext &in,
				   CORBA::ULong &req_id,
				   CORBA::GIOP::LocateStatusType &stat,
				   CORBA::Object_ptr &obj)
{
    CORBA::DataDecoder *dc = in.dc();

    CORBA::ULong k;

    check (dc->struct_begin());
    {
	check (dc->get_ulong (req_id));
	check (dc->enumeration (k));
	stat = (CORBA::GIOP::LocateStatusType)k;
    }
    check (dc->struct_end());

    if (stat == CORBA::GIOP::OBJECT_FORWARD) {
	CORBA::IOR *ior = new CORBA::IOR;
	check (dc->get_ior (*ior));
	obj = new CORBA::Object (ior);
    } else {
	obj = CORBA::Object::_nil();
    }
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_bind_request (GIOPInContext &in, char * &repoid,
				   CORBA::ORB::ObjectTag &oid)
{
    /*
     * mapped to an invocation:
     *   GIOP::LocateStatusType _bind (in string repoid,
     *				       in sequence<octet> oid,
     *				       out object obj);
     * at this point only the invoke request in args are left in the
     * buffer...
     */
    CORBA::DataDecoder *dc = in.dc();

    check (dc->struct_begin ());
    {
	check (dc->get_string (repoid));
	CORBA::ULong l;
	check (dc->seq_begin (l));
	oid.length (l);
	if (l > 0)
	    check (dc->get_octets ((CORBA::Octet *)&oid[0], oid.length()));
	check (dc->seq_end ());
    }
    check (dc->struct_end ());

    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_bind_reply (GIOPInContext &in,
				 CORBA::GIOP::LocateStatusType &stat,
				 CORBA::Object_ptr &obj)
{
    /*
     * mapped to an invocation:
     *   GIOP::LocateStatusType _bind (in string repoid,
     *				       in sequence<octet> oid,
     *				       out object obj);
     * at this point only the invoke reply out args/result are left in the
     * buffer...
     */
    CORBA::DataDecoder *dc = in.dc();

    CORBA::ULong k;
    CORBA::IOR ior;

    obj = CORBA::Object::_nil();

    check (dc->struct_begin ());
    {
	check (dc->enumeration (k));
	stat = (CORBA::GIOP::LocateStatusType)k;
	check (dc->get_ior (ior));
    }
    check (dc->struct_end ());

    obj = new CORBA::Object (new CORBA::IOR (ior));
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_close_msg (GIOPInContext &in)
{
    CORBA::DataDecoder *dc = in.dc();
    return TRUE;
}

CORBA::Boolean
MICO::GIOPCodec::get_error_msg (GIOPInContext &in)
{
    CORBA::DataDecoder *dc = in.dc();
    return TRUE;
}

#undef check


/****************************** GIOPRequest *****************************/


MICO::GIOPRequest::GIOPRequest (const char *op, CORBA::DataDecoder *indata,
				GIOPCodec *c)
{
    codec = c;

    CORBA::CodesetConv *conv = indata->converter();
    CORBA::CodesetConv *wconv = indata->wconverter();
    CORBA::CodesetConv *rconv = 0;
    CORBA::CodesetConv *rwconv = 0;

    if (conv) {
	rconv = the_codeset_db.reverse (conv);
    }
    if (wconv) {
	rwconv = the_codeset_db.reverse (rwconv);
    }
    opname = op;
    idc = indata;
    istart = idc->buffer()->rpos();

    odc = idc->clone (new CORBA::Buffer(), TRUE, conv, FALSE, wconv, FALSE);
    oec = odc->encoder (odc->buffer(), FALSE, rconv, FALSE, rwconv, FALSE);
    ostart = 0;
    is_except = FALSE;
}

MICO::GIOPRequest::~GIOPRequest ()
{
    delete idc;
    delete odc;
    delete oec;
}

const char *
MICO::GIOPRequest::op_name()
{
    return opname.c_str();
}

CORBA::Boolean
MICO::GIOPRequest::get_in_args (CORBA::NVList_ptr iparams,
				CORBA::Context_ptr &ctx)
{
    idc->buffer()->rseek_beg (istart);

    CORBA::Any *a;
    CORBA::TypeCode_var tc;
    CORBA::Boolean ret;

    for (CORBA::ULong i = 0; i < iparams->count(); ++i) {
	if (iparams->item(i)->flags() & (CORBA::ARG_IN|CORBA::ARG_INOUT)) {
	    a = iparams->item(i)->value();
	    tc = a->type();
	    ret = a->demarshal (tc, *idc);
	    if (!ret)
                return FALSE;
	}
    }
    if (idc->buffer()->length() > 0) {
	ctx = new CORBA::Context ("");
	if (!idc->get_context (*ctx)) {
	    CORBA::release (ctx);
	    ctx = CORBA::Context::_nil();
	}
    }
    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::get_in_args (StaticAnyList *iparams,
				CORBA::Context_ptr &ctx)
{
    idc->buffer()->rseek_beg (istart);

    CORBA::Boolean ret;

    for (mico_vec_size_type i = 0; i < iparams->size(); ++i) {
	if ((*iparams)[i]->flags() & (CORBA::ARG_IN|CORBA::ARG_INOUT)) {
	    ret = (*iparams)[i]->demarshal (*idc);
            if (!ret)
                return FALSE;
	}
    }
    if (idc->buffer()->length() > 0) {
	ctx = new CORBA::Context ("");
	if (!idc->get_context (*ctx)) {
	    CORBA::release (ctx);
	    ctx = CORBA::Context::_nil();
	}
    }
    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::get_in_args (CORBA::DataEncoder *ec)
{
    // XXX alignment, byteorder, differing codecs ???
    assert (!strcmp (ec->type(), idc->type()));
    assert (ec->byteorder() == idc->byteorder());

    idc->buffer()->rseek_beg (istart);
    ec->put_octets (idc->buffer()->data(), idc->buffer()->length());
    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::get_out_args (CORBA::Any *res, CORBA::NVList_ptr oparams,
				 CORBA::Exception *&except)
{
    odc->buffer()->rseek_beg (ostart);
    if (is_except) {
	except = CORBA::Exception::_decode (*odc);
	assert (except);
    } else {
	CORBA::TypeCode_var tc;
	CORBA::Boolean ret;
	CORBA::Any *a;

	except = 0;
	if (res) {
	    tc = res->type();
	    if (!tc->equal (CORBA::_tc_void)) {
		ret = res->demarshal (tc, *odc);
                if (!ret)
                    return FALSE;
	    }
	}
	for (CORBA::ULong i = 0; i < oparams->count(); ++i) {
	    if (oparams->item(i)->flags() & (CORBA::ARG_OUT|CORBA::ARG_INOUT)){
		a = oparams->item(i)->value();
		tc = a->type();
		ret = a->demarshal (tc, *odc);
                if (!ret)
                    return FALSE;
	    }
	}
    }
    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::get_out_args (CORBA::StaticAny *res,
				 StaticAnyList *oparams,
				 CORBA::Exception *&except)
{
    odc->buffer()->rseek_beg (ostart);
    if (is_except) {
	except = CORBA::Exception::_decode (*odc);
	assert (except);
    } else {
	except = 0;

	CORBA::Boolean ret;
	if (res) {
	    ret = res->demarshal (*odc);
            if (!ret)
                return FALSE;
	}
	for (mico_vec_size_type i = 0; i < oparams->size(); ++i) {
	    if ((*oparams)[i]->flags() & (CORBA::ARG_OUT|CORBA::ARG_INOUT)){
		ret = (*oparams)[i]->demarshal (*odc);
		if (!ret)
                    return FALSE;
	    }
	}
    }
    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::get_out_args (CORBA::DataEncoder *ec, CORBA::Boolean &is_ex)
{
    // XXX alignment, byteorder, differing codecs ???
    assert (!strcmp (ec->type(), odc->type()));
    assert (ec->byteorder() == odc->byteorder());

    odc->buffer()->rseek_beg (ostart);
    ec->put_octets (odc->buffer()->data(), odc->buffer()->length());
    is_ex = is_except;

    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::set_out_args (CORBA::Any *res, CORBA::NVList_ptr oparams)
{
    oec->buffer()->reset();

    // this is to get alignment right
    GIOPOutContext out (oec);
    codec->put_invoke_reply_offset (out, this);
    ostart = oec->buffer()->wpos();
    oec->buffer()->rseek_beg (ostart);

    is_except = FALSE;
    if (res)
	res->marshal (*oec);
    for (CORBA::ULong i = 0; i < oparams->count(); ++i) {
	if (oparams->item(i)->flags() & (CORBA::ARG_OUT|CORBA::ARG_INOUT)) {
	    oparams->item(i)->value()->marshal (*oec);
	}
    }
    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::set_out_args (CORBA::StaticAny *res,
				 StaticAnyList *oparams)
{
    oec->buffer()->reset();

    // this is to get alignment right
    GIOPOutContext out (oec);
    codec->put_invoke_reply_offset (out, this);
    ostart = oec->buffer()->wpos();
    oec->buffer()->rseek_beg (ostart);

    is_except = FALSE;
    if (res)
	res->marshal (*oec);
    for (mico_vec_size_type i = 0; i < oparams->size(); ++i) {
	if ((*oparams)[i]->flags() & (CORBA::ARG_OUT|CORBA::ARG_INOUT))
	    (*oparams)[i]->marshal (*oec);
    }
    return TRUE;
}

void
MICO::GIOPRequest::set_out_args (CORBA::Exception *except)
{
    oec->buffer()->reset();

    // this is to get alignment right
    GIOPOutContext out (oec);
    codec->put_invoke_reply_offset (out, this);
    ostart = oec->buffer()->wpos();
    oec->buffer()->rseek_beg (ostart);

    is_except = TRUE;
    except->_encode (*oec);
}

CORBA::Boolean
MICO::GIOPRequest::set_out_args (CORBA::DataDecoder *dc, CORBA::Boolean is_ex)
{
    oec->buffer()->reset();

    // this is to get alignment right
    GIOPOutContext out (oec);
    codec->put_invoke_reply_offset (out, this);
    ostart = oec->buffer()->wpos();
    oec->buffer()->rseek_beg (ostart);

    // XXX alignment, byteorder, differing codecs ???
    assert (!strcmp (dc->type(), oec->type()));
    assert (dc->byteorder() == oec->byteorder());

    oec->put_octets (dc->buffer()->data(), dc->buffer()->length());
    dc->buffer()->rseek_rel (dc->buffer()->length());
    is_except = is_ex;
    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::copy_out_args (CORBA::ORBRequest *req)
{
    if (this == req)
	return TRUE;

    // copy service context list
    copy_svc (req);

    assert (oec);

    oec->buffer()->reset();

    // this is to get alignment right
    GIOPOutContext out (oec);
    codec->put_invoke_reply_offset (out, this);
    ostart = oec->buffer()->wpos();
    oec->buffer()->rseek_beg (ostart);

    req->get_out_args (oec, is_except);
    return TRUE;
}

CORBA::Boolean
MICO::GIOPRequest::copy_in_args (CORBA::ORBRequest *req)
{
    assert (0);
    return FALSE;
}

const char *
MICO::GIOPRequest::type ()
{
    return "giop";
}

CORBA::ByteOrder
MICO::GIOPRequest::input_byteorder ()
{
    return idc->byteorder();
}

CORBA::ByteOrder
MICO::GIOPRequest::output_byteorder ()
{
    return odc->byteorder();
}

CORBA::Codeset::CodesetId
MICO::GIOPRequest::csid()
{
    return idc->converter() ? idc->converter()->from()->id() : 0;
}
    
CORBA::Codeset::CodesetId
MICO::GIOPRequest::wcsid ()
{
    return idc->wconverter() ? idc->wconverter()->from()->id() : 0;
}

MICO::GIOPRequest *
MICO::GIOPRequest::_duplicate (GIOPRequest *r)
{
    if (!CORBA::is_nil (r))
	r->_ref();
    return r;
}

MICO::GIOPRequest *
MICO::GIOPRequest::_nil ()
{
    return NULL;
}


/***************************** GIOPSimpleProf ****************************/


MICO::GIOPSimpleProf::GIOPSimpleProf ()
    : objkey (0), length (0)
{
}

MICO::GIOPSimpleProf::GIOPSimpleProf (const GIOPSimpleProf &p)
    : objkey (0), length (0)
{
    if (p.objkey) {
	objkey = new CORBA::Octet[length = p.length];
	memcpy (objkey, p.objkey, p.length);
    }
}

MICO::GIOPSimpleProf::~GIOPSimpleProf ()
{
    if (objkey)
	delete[] objkey;
}

MICO::GIOPSimpleProf &
MICO::GIOPSimpleProf::operator= (const GIOPSimpleProf &p)
{
    if (this != &p)
	objectkey (p.objkey, p.length);
    return *this;
}

void
MICO::GIOPSimpleProf::encode (CORBA::DataEncoder &) const
{
    assert (0);
}

const CORBA::Address *
MICO::GIOPSimpleProf::addr () const
{
    assert (0);
    return 0;
}

CORBA::IORProfile::ProfileId
MICO::GIOPSimpleProf::id () const
{
    return 10000;
}

CORBA::IORProfile::ProfileId
MICO::GIOPSimpleProf::encode_id () const
{
    return 10000;
}

void
MICO::GIOPSimpleProf::objectkey (CORBA::Octet *o, CORBA::Long l)
{
    if (objkey)
	delete[] objkey;
    objkey = new CORBA::Octet[length = l];
    memcpy (objkey, o, l);
}

const CORBA::Octet *
MICO::GIOPSimpleProf::objectkey (CORBA::Long &l) const
{
    l = length;
    if (!objkey)
	return (const CORBA::Octet *)"";
    return objkey;
}

CORBA::Boolean
MICO::GIOPSimpleProf::reachable ()
{
    return TRUE;
}

void
MICO::GIOPSimpleProf::print (ostream &o) const
{
    o << "giop-simple-profile";
}

CORBA::IORProfile *
MICO::GIOPSimpleProf::clone () const
{
    return new GIOPSimpleProf (*this);
}

CORBA::Long
MICO::GIOPSimpleProf::compare (const CORBA::IORProfile &p) const
{
    if (p.id() != id())
	return (CORBA::Long)id() - (CORBA::Long)p.id();
    const GIOPSimpleProf &gp = (const GIOPSimpleProf &)p;
    if (length != gp.length)
	return (CORBA::Long)length - (CORBA::Long)gp.length;
    return mico_key_compare (objkey, gp.objkey, length);
}

CORBA::Boolean
MICO::GIOPSimpleProf::operator== (const CORBA::IORProfile &p) const
{
    return compare (p) == 0;
}

CORBA::Boolean
MICO::GIOPSimpleProf::operator< (const CORBA::IORProfile &p) const
{
    return compare (p) < 0;
}


/******************************* GIOPConn *******************************/


MICO::GIOPConn::GIOPConn (CORBA::ORB_ptr _orb, CORBA::Transport *_transp,
			  GIOPConnCallback *_cb, GIOPCodec *codec,
			  CORBA::Long tmout)
    : inctx (codec, new CORBA::Buffer())
{
    orb = _orb;
    transp = _transp;
    cb = _cb;
    _codec = codec;

    inbuf = new CORBA::Buffer;
    inlen = _codec->header_length ();
    inflags = 0;
    infrag = 0;

    refcnt = 0;
    idle_tmout = tmout;
    have_tmout = FALSE;
    have_wselect = FALSE;

    transp->block (FALSE);
    transp->rselect (orb->dispatcher(), this);

    check_idle ();
}

MICO::GIOPConn::~GIOPConn ()
{
    assert (refcnt >= 0);

    list<CORBA::Buffer *>::iterator i;
    for (i = outbufs.begin(); i != outbufs.end(); ++i)
	delete *i;
    for (i = inbufs.begin(); i != inbufs.end(); ++i)
	delete *i;

    orb->dispatcher()->remove (this, CORBA::Dispatcher::Timer);
    transp->rselect (orb->dispatcher(), 0);
    transp->wselect (orb->dispatcher(), 0);
    delete transp;
    delete inbuf;
    delete _codec;

    // do not delete: disp
}

void
MICO::GIOPConn::do_read ()
{
    while (42) {
	assert (inlen > 0);
	CORBA::Long r = transp->read (*inbuf, inlen);
	if (r < 0 || (r == 0 && transp->eof())) {
	    // connection broken or EOF
	    transp->rselect (orb->dispatcher(), 0);
	    transp->wselect (orb->dispatcher(), 0);
	    cb->callback (this, GIOPConnCallback::Closed);
	    break;
	} else if (r > 0) {
	    inlen -= r;
	    if (inbuf->length() == _codec->header_length()) {
		// header completely received
		assert (inlen == 0);
		CORBA::GIOP::MsgType mt;
		inctx.buffer (inbuf);
		if (!_codec->check_header (inctx, mt, inlen, inflags)) {
		    // garbled message, send it to input handler ...
		    inbufs.push_back (inbuf);
		    inbuf = new CORBA::Buffer;
		    inlen = _codec->header_length ();
		    cb->callback (this, GIOPConnCallback::InputReady);
		    continue;
		}
		if (mt == CORBA::GIOP::Fragment) {
		    // a fragment, append contents to infrag ...
		    assert (infrag);
		    delete inbuf;
		    inbuf = infrag;
		    infrag = 0;
		}
	    }
	    if (inlen == 0) {
		// message completely received
		if (inflags & GIOP_FRAGMENT_BIT) {
		    // more fragments follow ...
		    assert (!infrag);
		    infrag = inbuf;
		    inbuf = new CORBA::Buffer;
		    inlen = _codec->header_length();
		} else {
		    inbufs.push_back (inbuf);
		    inbuf = new CORBA::Buffer;
		    inlen = _codec->header_length ();
		    cb->callback (this, GIOPConnCallback::InputReady);
		}
	    }
	} else if (r == 0) {
	    break;
	} else {
	    assert (0);
	}
    }
}

void
MICO::GIOPConn::do_write ()
{
    while (42) {
	assert (outbufs.size() > 0);
	CORBA::Buffer *b = outbufs.front();
	CORBA::Long r = transp->write (*b, b->length());
	if (r > 0) {
	    if (b->length() == 0) {
		// message completely sent
		delete b;
		outbufs.pop_front();
		if (outbufs.size() == 0) {
		    check_idle ();
		    break;
		}
	    }
	} else if (r < 0) {
	    // connection broken
	    transp->rselect (orb->dispatcher(), 0);
	    transp->wselect (orb->dispatcher(), 0);
	    cb->callback (this, GIOPConnCallback::Closed);
	    break;
	} else if (r == 0) {
	    break;
	} else {
	    assert (0);
	}
    }
}

void
MICO::GIOPConn::callback (CORBA::Transport *,
			  CORBA::TransportCallback::Event ev)
{
    switch (ev) {
    case Read:
	do_read ();
	break;

    case Write:
	do_write ();
	break;

    default:
	assert (0);
    }
}

void
MICO::GIOPConn::callback (CORBA::Dispatcher *d,
			  CORBA::DispatcherCallback::Event ev)
{
    switch (ev) {
    case CORBA::Dispatcher::Timer:
	cb->callback (this, GIOPConnCallback::Idle);
	break;
    default:
	assert (0);
    }
}

void
MICO::GIOPConn::check_idle ()
{
    if (idle_tmout > 0 && refcnt == 0 && outbufs.size() == 0) {
        if (have_tmout)
            orb->dispatcher()->remove (this, CORBA::Dispatcher::Timer);
	orb->dispatcher()->tm_event (this, idle_tmout);
	have_tmout = TRUE;
    }
    if (have_wselect && outbufs.size() == 0) {
        transp->wselect (orb->dispatcher(), 0);
        have_wselect = FALSE;
    }
}

void
MICO::GIOPConn::check_busy ()
{
    if (have_tmout && (refcnt > 0 || outbufs.size() > 0)) {
	orb->dispatcher()->remove (this, CORBA::Dispatcher::Timer);
	have_tmout = FALSE;
    }
    if (!have_wselect && outbufs.size() > 0) {
        transp->wselect (orb->dispatcher(), this);
        have_wselect = TRUE;
    }
}

void
MICO::GIOPConn::ref ()
{
    ++refcnt;
    check_busy ();
}

CORBA::Boolean
MICO::GIOPConn::deref (CORBA::Boolean all)
{
    if (all)
	refcnt = 0;
    else
	--refcnt;
    check_idle ();
    return (refcnt <= 0);
}

void
MICO::GIOPConn::output (CORBA::Buffer *b)
{
    // try to write as much as possible immediatly
    if (outbufs.size() == 0) {
	transp->write (*b, b->length());
	if (b->length() == 0) {
	    delete b;
	    return;
	}
	outbufs.push_back (b);
    } else {
	outbufs.push_back (b);
	do_write ();
    }
    check_busy ();
}

CORBA::Buffer *
MICO::GIOPConn::input ()
{
    if (inbufs.size() == 0)
	return 0;
    CORBA::Buffer *b = inbufs.front();
    inbufs.pop_front ();
    return b;
}

void
MICO::GIOPConn::cancel (CORBA::ULong reqid)
{
    // XXX kill pending fragment if reqid matches

    // for now disallow CancelRequests inbetween fragments
    assert (!infrag);
}


/******************************* IIOPProxy ******************************/


MICO::IIOPProxy::IIOPProxy (CORBA::ORB_ptr _orb)
{
    /*
     * these are the IOR profile types we can handle. they are sorted
     * by priority, i.e. we will first try to use valid_profiles[0],
     * then valid_profiles[1], ...
     */
    valid_profiles.push_back (CORBA::IORProfile::TAG_UNIX_IOP);
    valid_profiles.push_back (CORBA::IORProfile::TAG_INTERNET_IOP);
    valid_profiles.push_back (CORBA::IORProfile::TAG_SSL_UNIX_IOP);
    valid_profiles.push_back (CORBA::IORProfile::TAG_SSL_INTERNET_IOP);

    orb = _orb;
    orb->register_oa (this);
}

MICO::IIOPProxy::~IIOPProxy ()
{
    orb->unregister_oa (this);
    /*
     * the address keys point to GIOPConn::Transport->addr(), so do not
     * delete them.
     * the GIOPConn entries in the 'ids' map just pointers to the entries
     * in the 'conns' map, so do not delete them
     */
    for (MapAddrConn::iterator i = conns.begin(); i != conns.end(); ++i)
	delete (*i).second;
}

MICO::IIOPProxyInvokeRec *
MICO::IIOPProxy::get_invoke (MsgId id)
{
    MapIdConn::iterator i = ids.find (id);
    return i == ids.end() ? 0 : &(*i).second;
}

MICO::GIOPConn *
MICO::IIOPProxy::make_conn (const CORBA::Address *addr)
{
    MapAddrConn::iterator i = conns.find (addr);
    if (i != conns.end()) {
	GIOPConn *conn = (*i).second;
	return conn;
    }
    MICODebug::instance()->tracer()
	<< "making new connection to " << addr->stringify() << endl;
    CORBA::Transport *t = addr->make_transport();
    if (!t->connect (addr)) {
        MICODebug::instance()->tracer()
            << "connect: " << t->errormsg() << endl;
        delete t;
        return 0;
    }
    GIOPConn *conn =
	new GIOPConn (orb, t, this,
		      new GIOPCodec (new CDRDecoder, new CDREncoder),
		      0L /* no tmout */);
    conns[t->peer()] = conn;
    return conn;
}

MICO::GIOPConn *
MICO::IIOPProxy::make_conn (CORBA::Object_ptr obj)
{
    CORBA::IORProfile *prof;
    const CORBA::Address *addr;
    MICO::GIOPConn *conn;

    for (mico_vec_size_type i = 0; i < valid_profiles.size(); ++i) {
	prof = obj->_ior()->profile (valid_profiles[i]);
	if (prof) {
	    addr = prof->addr ();
	    assert (addr);
            conn = make_conn (addr);
            if (conn)
                return conn;
	}
    }
    return 0;
}

void
MICO::IIOPProxy::kill_conn (GIOPConn *conn)
{
    CORBA::Boolean again;

    do {
	again = FALSE;
	for (MapAddrConn::iterator i = conns.begin(); i != conns.end(); ++i) {
	    if ((*i).second == conn) {
		conns.erase (i);
		again = TRUE;
		break;
	    }
	}
    } while (again);
    do {
	again = FALSE;
	IIOPProxyInvokeRec *rec;
	for (MapIdConn::iterator i = ids.begin(); i != ids.end(); ++i) {
	    rec = &(*i).second;
	    if (rec->conn() == conn) {
		// XXX notify orb about failed invocation ...
		MICODebug::instance()->printer()
		    << "invocation(" << (*i).first << ") aborted" << endl;

		switch (orb->request_type ((*i).first)) {
		case CORBA::RequestInvoke: {
		    CORBA::Object_var obj = new CORBA::Object;
		    CORBA::Request_var req =
			new CORBA::Request (obj, "someop");
		    LocalRequest orbreq (req);
                    orbreq.set_out_args (new CORBA::COMM_FAILURE);
		    orb->answer_invoke ((*i).first, CORBA::InvokeSysEx,
					CORBA::Object::_nil(), &orbreq);
		    break;
		}
		case CORBA::RequestLocate:
		    orb->answer_locate ((*i).first, CORBA::LocateUnknown,
					CORBA::Object::_nil());
		    break;

		case CORBA::RequestBind:
		    orb->answer_bind ((*i).first, CORBA::LocateUnknown,
				      CORBA::Object::_nil());
		    break;

		default:
		    assert (0);
		}

		ids.erase (i);
		again = TRUE;
		break;
	    }
	}
    } while (again);
    delete conn;
}

void
MICO::IIOPProxy::conn_error (GIOPConn *conn, CORBA::Boolean send_error)
{
    if (!send_error) {
	kill_conn (conn);
	return;
    }

    GIOPOutContext out (conn->codec());
    conn->codec()->put_error_msg (out);
    conn->output (out._retn());
    // prepare shutdown (i.e. wait until MessageError has been sent)
    conn->deref (TRUE);
}

CORBA::Boolean
MICO::IIOPProxy::get_codeset_ids (CORBA::Object_ptr obj,
				  CORBA::Codeset::CodesetId &csid,
				  CORBA::Codeset::CodesetId &wcsid)
{
    CORBA::IORProfile *prof;
    MICO::CodesetComponent *csc = 0;

    prof = obj->_ior()->profile (CORBA::IORProfile::TAG_MULTIPLE_COMPONENTS,
				 TRUE);
    if (prof) {
	CORBA::MultiComponent *mc =
	    ((MICO::MultiCompProfile *)prof)->components();
	CORBA::Component *comp =
	    mc->component (CORBA::Component::TAG_CODE_SETS);
	if (comp)
	    csc = (MICO::CodesetComponent *)comp;
    }

    if (csc) {
	csid = csc->selected_cs();
	wcsid = csc->selected_wcs();
    } else {
	// no code set info, use defaults ...
	csid = CORBA::Codeset::special_cs(CORBA::Codeset::DefaultCS)->id();
	wcsid = CORBA::Codeset::special_cs(CORBA::Codeset::DefaultWCS)->id();
    }
    return TRUE;
}

const char *
MICO::IIOPProxy::get_oaid () const
{
    return "mico-iiop-proxy";
}

CORBA::Boolean
MICO::IIOPProxy::has_object (CORBA::Object_ptr obj)
{
    /*
     * we have every object whose IOR has a profile tag that is listed
     * in valid_profiles. we dont need to check the concrete address,
     * because the ORB will ask local object adapters first...
     */
    for (mico_vec_size_type i = 0; i < valid_profiles.size(); ++i) {
	if (obj->_ior()->profile (valid_profiles[i]))
	    return TRUE;
    }
    return FALSE;
}

CORBA::Boolean
MICO::IIOPProxy::is_local () const
{
    return FALSE;
}

CORBA::Boolean
MICO::IIOPProxy::invoke (MsgId msgid, CORBA::Object_ptr obj,
			 CORBA::ORBRequest *req,
			 CORBA::Principal_ptr pr, CORBA::Boolean response_exp)
{
    GIOPConn *conn = make_conn (obj);
    if (!conn) {
        CORBA::COMM_FAILURE ex;
	req->set_out_args (&ex);
	orb->answer_invoke (msgid, CORBA::InvokeSysEx, CORBA::Object::_nil(),
			    req);
	return FALSE;
    }

    // set up code set converters 
    CORBA::Codeset::CodesetId csid = 0, wcsid = 0;
    CORBA::CodesetConv *conv = 0, *wconv = 0;

    if (!strcmp (req->type(), "giop")) {
	// its a GIOP request, use code sets that the client selected
	GIOPRequest *greq = (GIOPRequest *)req;
	csid = greq->csid();
	wcsid = greq->wcsid();
	if (csid) {
	    conv = the_codeset_db.find (csid, csid);
	}
	if (wcsid) {
	    wconv = the_codeset_db.find (wcsid, wcsid);
	}
    } else {
	// run code set selection algorithm
	CORBA::Boolean r = get_codeset_ids (obj, csid, wcsid);
	assert (r);

	if (csid) {
	    conv = the_codeset_db.find (
		CORBA::Codeset::special_cs(CORBA::Codeset::NativeCS)->id(),
		csid);
	}
	if (wcsid) {
	    wconv = the_codeset_db.find (
		CORBA::Codeset::special_cs(CORBA::Codeset::NativeWCS)->id(),
		wcsid);
	}
    }

    GIOPOutContext out (conn->codec(), conv, wconv);
    conn->codec()->put_invoke_request (out, msgid, response_exp, obj, req, pr);

    if (response_exp) {
	conn->ref();
	add_invoke (msgid, IIOPProxyInvokeRec (conn, req, csid, wcsid));
    }
    conn->output (out._retn());
    return TRUE;
}

CORBA::Boolean
MICO::IIOPProxy::bind (MsgId msgid, const char *repoid,
		       const CORBA::ORB::ObjectTag &oid,
		       CORBA::Address *addr)
{
    if (!addr || addr->is_local())
	return FALSE;

    GIOPConn *conn = make_conn (addr);
    if (!conn) {
        orb->answer_bind (msgid, CORBA::LocateUnknown, CORBA::Object::_nil());
        return TRUE;
    }

    GIOPOutContext out (conn->codec());
    conn->codec()->put_bind_request (out, msgid, repoid, oid);

    conn->ref();
    assert (ids.count (msgid) == 0);
    add_invoke (msgid, IIOPProxyInvokeRec (conn));

    conn->output (out._retn());
    return TRUE;
}

CORBA::Boolean
MICO::IIOPProxy::locate (MsgId msgid, CORBA::Object_ptr obj)
{
    GIOPConn *conn = make_conn (obj);
    if (!conn) {
	orb->answer_locate (msgid, CORBA::LocateUnknown,
			    CORBA::Object::_nil());
	return FALSE;
    }

    GIOPOutContext out (conn->codec());
    conn->codec()->put_locate_request (out, msgid, obj);

    conn->ref();
    assert (ids.count (msgid) == 0);
    add_invoke (msgid, IIOPProxyInvokeRec (conn));

    conn->output (out._retn());
    return TRUE;
}

CORBA::Object_ptr
MICO::IIOPProxy::skeleton (CORBA::Object_ptr)
{
    assert (0);
    return CORBA::Object::_nil ();
}

void
MICO::IIOPProxy::cancel (MsgId msgid)
{
    IIOPProxyInvokeRec *rec = get_invoke (msgid);
    if (!rec)
	return;

    GIOPOutContext out (rec->conn()->codec());
    rec->conn()->codec()->put_cancel_request (out, msgid);

    rec->conn()->output (out._retn());
    rec->conn()->deref ();

    del_invoke (msgid);
}

void
MICO::IIOPProxy::shutdown (CORBA::Boolean wait_for_completion)
{
    // XXX make sure all invocations have completed
    orb->answer_shutdown (this);
}

void
MICO::IIOPProxy::answer_invoke (CORBA::ULong, CORBA::Object_ptr,
				CORBA::ORBRequest *, CORBA::InvokeStatus)
{
    assert (0);
}

void
MICO::IIOPProxy::exec_invoke_reply (GIOPInContext &in, CORBA::ULong req_id,
				    CORBA::GIOP::ReplyStatusType stat,
				    CORBA::Object_ptr obj,
				    CORBA::ORBRequest *req,
				    GIOPConn *conn)
{
    switch (orb->request_type (req_id)) {
    case CORBA::RequestBind: {
	// it was a bind
	CORBA::GIOP::LocateStatusType bind_stat;
	CORBA::Object_ptr bind_obj;

	if (stat != CORBA::GIOP::NO_EXCEPTION ||
	    !conn->codec()->get_bind_reply (in, bind_stat, bind_obj) ||
	    bind_stat != CORBA::GIOP::OBJECT_HERE) {
	    orb->answer_bind (req_id, CORBA::LocateUnknown,
			      CORBA::Object::_nil());
	} else {
	    orb->answer_bind (req_id, CORBA::LocateHere, bind_obj);
	    CORBA::release (bind_obj);
	}
	break;
    }
    case CORBA::RequestInvoke: {
	// it was a normal invocation
	CORBA::InvokeStatus orb_stat = CORBA::InvokeOk;

	switch (stat) {
	case CORBA::GIOP::NO_EXCEPTION:
	    orb_stat = CORBA::InvokeOk;
	    break;

	case CORBA::GIOP::USER_EXCEPTION:
	    orb_stat = CORBA::InvokeUsrEx;
	    break;

	case CORBA::GIOP::SYSTEM_EXCEPTION:
	    orb_stat = CORBA::InvokeSysEx;
	    break;

	case CORBA::GIOP::LOCATION_FORWARD:
	    orb_stat = CORBA::InvokeForward;
	    break;

	default:
	    assert (0);
	    break;
	}
	orb->answer_invoke (req_id, orb_stat, obj, req);
	break;
    }
    case CORBA::RequestUnknown:
	// request was canceled or reply contains invalid id
	break;
    default:
	assert (0);
    }
}

void
MICO::IIOPProxy::handle_invoke_reply (GIOPConn *conn, GIOPInContext &in)
{
    CORBA::ULong req_id;
    CORBA::GIOP::ReplyStatusType stat;
    CORBA::Object_ptr obj = CORBA::Object::_nil();
    CORBA::IOP::ServiceContextList ctx;

    // XXX take care, get_invoke_reply() does a in._retn()
    if (!conn->codec()->get_invoke_reply1 (in, req_id, stat, ctx)) {
	MICODebug::instance()->printer()
	    << "cannot decode Reply1" << endl;
	conn_error (conn);
	return;
    }

    IIOPProxyInvokeRec *rec = get_invoke (req_id);

    /*
     * set up code set converters. this must be done before
     * get_invoke_reply2(), because it creates a GIOPRequest
     * which needs these converters.
     */
    CORBA::CodesetConv *conv = 0, *wconv = 0;

    if (rec->csid())
	conv = the_codeset_db.find (
	    rec->csid(),
	    CORBA::Codeset::special_cs(CORBA::Codeset::NativeCS)->id());

    if (rec->wcsid())
	wconv = the_codeset_db.find (
	    rec->wcsid(),
	    CORBA::Codeset::special_cs(CORBA::Codeset::NativeWCS)->id());

    in.converters (conv, wconv);
	
    if (!conn->codec()->get_invoke_reply2 (in, req_id, stat, obj,
					   rec->request(), ctx)) {
	MICODebug::instance()->printer()
	    << "reply marshalling error" << endl;
        if (rec->request()) {
            CORBA::MARSHAL ex;
            rec->request()->set_out_args (&ex);
            stat = CORBA::GIOP::SYSTEM_EXCEPTION;
        }
    }
    exec_invoke_reply (in, req_id, stat, obj, rec->request(), conn);
    CORBA::release (obj);

    conn->deref();
    del_invoke (req_id);
}

void
MICO::IIOPProxy::handle_locate_reply (GIOPConn *conn, GIOPInContext &in)
{
    CORBA::ULong req_id;
    CORBA::GIOP::LocateStatusType stat = CORBA::GIOP::OBJECT_HERE;
    CORBA::LocateStatus orb_stat = CORBA::LocateHere;
    CORBA::Object_ptr obj = CORBA::Object::_nil();

    if (!conn->codec()->get_locate_reply (in, req_id, stat, obj)) {
	MICODebug::instance()->printer()
	    << "cannot decode LocateReply" << endl;
	conn_error (conn);
	return;
    }
    switch (stat) {
    case CORBA::GIOP::UNKNOWN_OBJECT:
	orb_stat = CORBA::LocateUnknown;
	break;

    case CORBA::GIOP::OBJECT_HERE:
	orb_stat = CORBA::LocateHere;
	break;

    case CORBA::GIOP::OBJECT_FORWARD:
	orb_stat = CORBA::LocateForward;
	break;

    default:
	assert (0);
	break;
    }
    orb->answer_locate (req_id, orb_stat, obj);
    CORBA::release (obj);

    conn->deref();
    del_invoke (req_id);
}

void
MICO::IIOPProxy::handle_input (GIOPConn *conn)
{
    GIOPInContext in (conn->codec(), conn->input());

    CORBA::GIOP::MsgType mt;
    CORBA::ULong size;
    CORBA::Octet flags;

    // XXX size is wrong for fragmented messages ...
    if (!conn->codec()->get_header (in, mt, size, flags)) {
	MICODebug::instance()->printer()
	    << "cannot decode header" << endl;
	conn_error (conn);
	return;
    }
    switch (mt) {
    case CORBA::GIOP::Reply:
	handle_invoke_reply (conn, in);
	break;

    case CORBA::GIOP::LocateReply:
	handle_locate_reply (conn, in);
	break;

    case CORBA::GIOP::CloseConnection:
	if (!conn->codec()->get_close_msg (in)) {
	    MICODebug::instance()->printer()
		<< "cannot decode CloseConnection" << endl;
	    conn_error (conn);
	} else {
	    MICODebug::instance()->printer()
		<< "got CloseConnection" << endl;
	    kill_conn (conn);
	}
	break;

    case CORBA::GIOP::MessageError:
	if (!conn->codec()->get_error_msg (in)) {
	    MICODebug::instance()->printer()
		<< "cannot decode MessageError" << endl;
	    conn_error (conn, FALSE);
	    break;
	}
	MICODebug::instance()->printer()
	    << "got MessageError" << endl;
	kill_conn (conn);
	break;

    default:
	MICODebug::instance()->printer()
	    << "bad message type: " << mt << endl;
	break;
    }
}

void
MICO::IIOPProxy::callback (GIOPConn *conn, GIOPConnCallback::Event ev)
{
    switch (ev) {
    case GIOPConnCallback::InputReady:
	handle_input (conn);
	break;

    case GIOPConnCallback::Idle:
	MICODebug::instance()->tracer()
	    << "shutting down idle conn" << endl;
	kill_conn (conn);
	break;

    case GIOPConnCallback::Closed:
	MICODebug::instance()->tracer()
	    << "conn closed or broken" << endl;
	kill_conn (conn);
	break;

    default:
	assert (0);
    }
}


/************************* IIOPServerInvokeRec ************************/


MICO::IIOPServerInvokeRec::IIOPServerInvokeRec (
    GIOPConn *conn, MsgId reqid,
    MsgId orbid, CORBA::ORBRequest *req,
    CORBA::Object_ptr obj,
    CORBA::Principal_ptr pr)
{
    _conn = conn;
    _orbid = orbid;
    _reqid = reqid;
    _req = req;
    _obj = obj;
    _pr = pr;
}

MICO::IIOPServerInvokeRec::IIOPServerInvokeRec (
    GIOPConn *conn, MsgId reqid,
    MsgId orbid, CORBA::Object_ptr obj)
{
    _conn = conn;
    _orbid = orbid;
    _reqid = reqid;
    _req = CORBA::ORBRequest::_nil();
    _obj = obj;
    _pr = CORBA::Principal::_nil();
}

MICO::IIOPServerInvokeRec::~IIOPServerInvokeRec ()
{
    CORBA::release (_req);
    CORBA::release (_obj);
    CORBA::release (_pr);

    // XXX do not free connection
}


/******************************* IIOPServer *****************************/


MICO::IIOPServer::IIOPServer (CORBA::ORB_ptr _orb)
{
    orb = _orb;
}

MICO::IIOPServer::~IIOPServer ()
{
    /*
     * the GIOPConn entries in the 'orbids' and 'reqids' maps are just
     * pointers to the entries in the 'conns' list, so do not delete them
     */
    for (ListConn::iterator i0 = conns.begin(); i0 != conns.end(); ++i0)
	delete *i0;
    for (MapIdConn::iterator i1 = orbids.begin(); i1 != orbids.end(); ++i1) {
	orb->cancel ((*i1).first);
	delete (*i1).second;
    }
    for (mico_vec_size_type i2 = 0; i2 < tservers.size(); ++i2) {
	tservers[i2]->aselect (orb->dispatcher(), 0);
	delete tservers[i2];
    }
}

CORBA::Boolean
MICO::IIOPServer::listen (CORBA::Address *addr)
{
    CORBA::TransportServer *tserv = addr->make_transport_server ();
    if (!tserv->bind (addr)) {
        MICODebug::instance()->printer()
	    << "bind: " << tserv->errormsg() << endl;
        assert (0);
    }
    tserv->block (FALSE);
    tserv->aselect (orb->dispatcher(), this);

    CORBA::IORProfile *prof
	= tserv->addr()->make_ior_profile ((CORBA::Octet *)"", 1,
					   CORBA::MultiComponent());

    /*
     * install an IIOP profile tag in the ORB's object template.
     * object adapters will use this template to generate new object
     * references...
     */
    MICODebug::instance()->tracer()
	<< "binding to " << prof->addr()->stringify() << endl;

    orb->ior_template()->add_profile (prof);

    tservers.push_back (tserv);
    return TRUE;
}

CORBA::Boolean
MICO::IIOPServer::listen ()
{
    InetAddress addr ("0.0.0.0", 0);
    return listen (&addr);
}

MICO::IIOPServerInvokeRec *
MICO::IIOPServer::get_invoke_reqid (MsgId msgid, GIOPConn *conn)
{
    // XXX slow, but only needed for cancel

    IIOPServerInvokeRec *rec;
    for (MapIdConn::iterator i = orbids.begin(); i != orbids.end(); ++i) {
	rec = (*i).second;
	if (rec->reqid() == msgid && rec->conn() == conn)
	    return rec;
    }
    return 0;
}

MICO::IIOPServerInvokeRec *
MICO::IIOPServer::get_invoke_orbid (MsgId msgid)
{
    MapIdConn::iterator i = orbids.find (msgid);
    return (i != orbids.end() ? (*i).second : 0);
}

void
MICO::IIOPServer::add_invoke (IIOPServerInvokeRec *rec)
{
    assert (orbids.count (rec->orbid()) == 0);
    orbids[rec->orbid()] = rec;
}

void
MICO::IIOPServer::del_invoke_orbid (MsgId msgid)
{
    MapIdConn::iterator i = orbids.find (msgid);
    if (i != orbids.end()) {
	delete (*i).second;
	orbids.erase (i);
    }
}

void
MICO::IIOPServer::del_invoke_reqid (MsgId msgid, GIOPConn *conn)
{
    // XXX slow, but rarely used
    IIOPServerInvokeRec *rec;
    for (MapIdConn::iterator i = orbids.begin(); i != orbids.end(); ++i) {
	rec = (*i).second;
	if (rec->reqid() == msgid && rec->conn() == conn) {
	    delete rec;
	    orbids.erase (i);
	}
    }
}

void
MICO::IIOPServer::kill_conn (GIOPConn *conn)
{
    CORBA::Boolean again;

    do {
	again = FALSE;
	for (ListConn::iterator i = conns.begin(); i != conns.end(); ++i) {
	    if (*i == conn) {
		conns.erase (i);
		again = TRUE;
		break;
	    }
	}
    } while (again);
    do {
	again = FALSE;
	IIOPServerInvokeRec *rec;
	for (MapIdConn::iterator i = orbids.begin(); i != orbids.end(); ++i) {
	    rec = (*i).second;
	    if (rec->conn() == conn) {
		orb->cancel (rec->orbid());
		orbids.erase (rec->orbid());
		delete rec;
		again = TRUE;
		break;
	    }
	}
    } while (again);
    delete conn;
}

void
MICO::IIOPServer::conn_error (GIOPConn *conn, CORBA::Boolean send_error)
{
    if (!send_error) {
	kill_conn (conn);
	return;
    }

    GIOPOutContext out (conn->codec());
    conn->codec()->put_error_msg (out);
    conn->output (out._retn());
    // prepare shutdown (i.e. wait until MessageError has been sent)
    conn->deref (TRUE);
}

void
MICO::IIOPServer::handle_input (GIOPConn *conn)
{
    GIOPInContext in (conn->codec(), conn->input());

    CORBA::GIOP::MsgType mt;
    CORBA::ULong size;
    CORBA::Octet flags;

    // XXX size is wrong for fragmented messages ...
    if (!conn->codec()->get_header (in, mt, size, flags)) {
	MICODebug::instance()->printer()
	    << "cannot decode header" << endl;
	conn_error (conn);
	return;
    }
    switch (mt) {
    case CORBA::GIOP::Request:
	handle_invoke_request (conn, in);
	break;

    case CORBA::GIOP::LocateRequest:
	handle_locate_request (conn, in);
	break;

    case CORBA::GIOP::MessageError:
	if (!conn->codec()->get_error_msg (in)) {
	    MICODebug::instance()->printer()
		<< "cannot decode MessageError" << endl;
	    conn_error (conn, FALSE);
	} else {
	    MICODebug::instance()->printer()
		<< "got MessageError" << endl;
	    kill_conn (conn);
	}
	break;

    case CORBA::GIOP::CancelRequest:
	handle_cancel_request (conn, in);
	break;

    default:
	MICODebug::instance()->printer()
	    << "bad message type: " << mt << endl;
	break;
    }
}

MICO::IIOPServer::MsgId
MICO::IIOPServer::exec_invoke_request (GIOPInContext &in,
				       CORBA::Object_ptr obj, 
				       CORBA::ORBRequest *req,
				       CORBA::Principal_ptr pr,
				       CORBA::Boolean resp_exp,
				       GIOPConn *conn,
				       MsgId orbid)
{
    if (!strcmp (req->op_name(), "_bind")) {
	// its a bind
	CORBA::String_var repoid;
	CORBA::ORB::ObjectTag oid;
	CORBA::Boolean r = conn->codec()->get_bind_request (in, repoid, oid);
	assert (r);
	/*
	 * orb makes copies of repoid and oid so we can delete them
	 * after the call to bind_async()
	 */
	return orb->bind_async (repoid, oid, 0, this, orbid);
    } else {
	// its a normal invocation
	return orb->invoke_async (obj, req, pr, resp_exp, this, orbid);
    }
}

void
MICO::IIOPServer::handle_invoke_request (GIOPConn *conn, GIOPInContext &in)
{
    CORBA::ULong req_id;
    CORBA::Boolean resp;
    CORBA::ORBRequest *req;
    CORBA::Principal_ptr pr = conn->transport()->get_principal();
    CORBA::Object_ptr obj = new CORBA::Object (new CORBA::IOR);
    obj->_ior()->add_profile (new GIOPSimpleProf);

    // XXX take care, get_invoke_request() does a in._retn()
    if (!conn->codec()->get_invoke_request (in, req_id, resp, obj, req, pr)) {
	CORBA::release (obj);
	CORBA::release (pr);
	MICODebug::instance()->printer()
	    << "cannot decode Request" << endl;
	conn_error (conn);
	return;
    }
    // XXX obj is incomplete, see IOR::operator== ...

    /*
     * code sets are set up in get_contextlist(). GIOPRequest will
     * set up converters for out args, so we dont need to do anything
     * special.
     */

    /*
     * must install the invocation record before we call the ORB, because
     * may invoke callback before returning from invoke_async ...
     */
    MsgId orbid = orb->new_msgid();
    conn->ref ();
    add_invoke (new IIOPServerInvokeRec (conn, req_id, orbid, req, obj, pr));

    MsgId orbid2 = exec_invoke_request (in, obj, req, pr, resp, conn, orbid);
    assert (orbid == orbid2 || (orbid2 == 0 && !resp));

    if (!resp) {
	conn->deref ();
	del_invoke_orbid (orbid);
    }
}

void
MICO::IIOPServer::handle_locate_request (GIOPConn *conn, GIOPInContext &in)
{
    CORBA::ULong req_id;
    CORBA::Object_ptr obj
	= new CORBA::Object (new CORBA::IOR (*orb->ior_template()));

    if (!conn->codec()->get_locate_request (in, req_id, obj)) {
	CORBA::release (obj);
	MICODebug::instance()->printer()
	    << "cannot decode LocateRequest" << endl;
	conn_error (conn);
	return;
    }
    // XXX obj is incomplete, see IOR::operator== ...

    /*
     * must install the invocation record before we call the ORB, because
     * may invoke callback before returning from invoke_async ...
     */
    MsgId orbid = orb->new_msgid();
    conn->ref ();
    add_invoke (new IIOPServerInvokeRec (conn, req_id, orbid, obj));

    MsgId orbid2 = orb->locate_async (obj, this, orbid);
    assert (orbid == orbid2);
}

void
MICO::IIOPServer::handle_cancel_request (GIOPConn *conn, GIOPInContext &in)
{
    CORBA::ULong req_id;

    if (!conn->codec()->get_cancel_request (in, req_id)) {
	MICODebug::instance()->printer()
	    << "cannot decode CancelRequest" << endl;
	conn_error (conn);
	return;
    }
    conn->cancel (req_id);

    IIOPServerInvokeRec *rec = get_invoke_reqid (req_id, conn);
    if (!rec) {
	// request already finished or no such id
	return;
    }
    orb->cancel (rec->orbid());

    // maybe rec->conn() != conn ...
    rec->conn()->deref ();
    del_invoke_orbid (rec->orbid());
}

void
MICO::IIOPServer::handle_invoke_reply (MsgId msgid)
{
    CORBA::ORBRequest *req;
    CORBA::Object_ptr obj;
    CORBA::InvokeStatus stat = orb->get_invoke_reply (msgid, obj, req);

    CORBA::GIOP::ReplyStatusType giop_stat = CORBA::GIOP::NO_EXCEPTION;
    switch (stat) {
    case CORBA::InvokeSysEx:
	giop_stat = CORBA::GIOP::SYSTEM_EXCEPTION;
	break;

    case CORBA::InvokeUsrEx:
	giop_stat = CORBA::GIOP::USER_EXCEPTION;
	break;

    case CORBA::InvokeOk:
	giop_stat = CORBA::GIOP::NO_EXCEPTION;
	break;

    case CORBA::InvokeForward:
	giop_stat = CORBA::GIOP::LOCATION_FORWARD;
	break;
    }

    IIOPServerInvokeRec *rec = get_invoke_orbid (msgid);
    assert (rec);

    GIOPOutContext out (rec->conn()->codec());
    rec->conn()->codec()->put_invoke_reply (out, rec->reqid(), giop_stat,
					    obj, req);
    if (stat == CORBA::InvokeForward)
	CORBA::release (obj);

    rec->conn()->output (out._retn());
    rec->conn()->deref ();
    del_invoke_orbid (rec->orbid());
}

void
MICO::IIOPServer::handle_locate_reply (MsgId msgid)
{
    CORBA::Object_ptr obj = CORBA::Object::_nil();
    CORBA::LocateStatus stat = orb->get_locate_reply (msgid, obj);

    CORBA::GIOP::LocateStatusType giop_stat = CORBA::GIOP::OBJECT_HERE;
    switch (stat) {
    case CORBA::LocateHere:
	giop_stat = CORBA::GIOP::OBJECT_HERE;
	break;

    case CORBA::LocateUnknown:
	giop_stat = CORBA::GIOP::UNKNOWN_OBJECT;
	break;

    case CORBA::LocateForward:
	giop_stat = CORBA::GIOP::OBJECT_FORWARD;
	break;
    }

    IIOPServerInvokeRec *rec = get_invoke_orbid (msgid);
    assert (rec);

    GIOPOutContext out (rec->conn()->codec());
    rec->conn()->codec()->put_locate_reply (out, rec->reqid(), giop_stat, obj);
    CORBA::release (obj);

    rec->conn()->output (out._retn());
    rec->conn()->deref ();
    del_invoke_orbid (rec->orbid());
}

void
MICO::IIOPServer::handle_bind_reply (MsgId msgid)
{
    CORBA::Object_ptr obj = CORBA::Object::_nil();
    CORBA::LocateStatus stat = orb->get_bind_reply (msgid, obj);

    CORBA::GIOP::LocateStatusType giop_stat = CORBA::GIOP::OBJECT_HERE;
    switch (stat) {
    case CORBA::LocateHere:
	giop_stat = CORBA::GIOP::OBJECT_HERE;
	break;

    case CORBA::LocateUnknown:
	giop_stat = CORBA::GIOP::UNKNOWN_OBJECT;
	break;

    case CORBA::LocateForward:
	giop_stat = CORBA::GIOP::OBJECT_FORWARD;
	break;
    }

    IIOPServerInvokeRec *rec = get_invoke_orbid (msgid);
    assert (rec);

    GIOPOutContext out (rec->conn()->codec());
    rec->conn()->codec()->put_bind_reply (out, rec->reqid(), giop_stat, obj);
    CORBA::release (obj);

    rec->conn()->output (out._retn());
    rec->conn()->deref ();
    del_invoke_orbid (rec->orbid());
}

void
MICO::IIOPServer::callback (CORBA::ORB_ptr, MsgId msgid,
			    CORBA::ORBCallback::Event ev)
{
    switch (ev) {
    case CORBA::ORBCallback::Invoke:
	handle_invoke_reply (msgid);
	break;

    case CORBA::ORBCallback::Locate:
	handle_locate_reply (msgid);
	break;

    case CORBA::ORBCallback::Bind:
	handle_bind_reply (msgid);
	break;

    default:
	assert (0);
    }
}

void
MICO::IIOPServer::callback (GIOPConn *conn, GIOPConnCallback::Event ev)
{
    switch (ev) {
    case GIOPConnCallback::InputReady:
	handle_input (conn);
	break;

    case GIOPConnCallback::Idle:
	MICODebug::instance()->tracer()
	    << "shutting down idle conn" << endl;
	// XXX better send CloseConnection ...
	kill_conn (conn);
	break;

    case GIOPConnCallback::Closed: {
	MICODebug::instance()->tracer()
	    << "conn closed or broken" << endl;
	const CORBA::Address *addr = conn->transport()->peer();
	assert (addr);
	Interceptor::ConnInterceptor::
	    _exec_client_disconnect (addr->stringify().c_str());
	kill_conn (conn);
	break;
    }
    default:
	assert (0);
    }
}

void
MICO::IIOPServer::callback (CORBA::TransportServer *tserv,
			    CORBA::TransportServerCallback::Event ev)
{
    switch (ev) {
    case CORBA::TransportServerCallback::Accept: {
	CORBA::Transport *t = tserv->accept();
	if (t) {
	    MICODebug::instance()->tracer()
		<< "new connection from " << t->peer()->stringify() << endl;
            if (t->bad()) {
                MICODebug::instance()->tracer()
                    << "bad conn: " << t->errormsg() << endl;
                delete t;
                break;
            }
	    const CORBA::Address *addr = t->peer();
	    assert (addr);
	    CORBA::Boolean r = Interceptor::ConnInterceptor::
		_exec_client_connect (addr->stringify().c_str());
	    if (!r) {
		delete t;
		break;
	    }
	    GIOPConn *conn =
		new GIOPConn (orb, t, this,
			      new GIOPCodec (new CDRDecoder, new CDREncoder),
			      0L /* no tmout */);
	    conns.push_back (conn);
	}
	break;
    }
    default:
	assert (0);
    }
}
