/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#ifndef _WINDOWS
#include <string.h>
#endif
#include <iostream.h>
#include <mico/template_impl.h>
#include <mico/util.h>


/************************* ServerlessObject **************************/


CORBA::ServerlessObject::~ServerlessObject ()
{
    magic = 0;
}

void
CORBA::ServerlessObject::_check() const
{
    assert (magic == 0x31415927);
}

void
CORBA::ServerlessObject::_ref()
{
    _check();
    ++refs;
}

CORBA::Boolean
CORBA::ServerlessObject::_deref()
{
    _check();
    return --refs <= 0;
}

CORBA::Long
CORBA::ServerlessObject::_refcnt() const
{
    _check();
    return refs;
}

CORBA::ServerlessObject_ptr
CORBA::ServerlessObject::_duplicate (ServerlessObject_ptr o)
{
    if (!CORBA::is_nil (o))
	o->_ref();
    return o;
}

CORBA::ServerlessObject_ptr
CORBA::ServerlessObject::_nil ()
{
    return NULL;
}


/****************************** Object *******************************/


// consumes IOR* argument
CORBA::Object::Object (IOR *i)
{
    ior = i;
    orb = CORBA::ORB_instance ("mico-local-orb");
    assert (orb);
}

CORBA::Object::Object (const Object &o)
{
    ior = o.ior ? new IOR (*o.ior) : 0;
    orb = CORBA::ORB::_duplicate (o.orb);
    _managers = o._managers;
}

CORBA::Object &
CORBA::Object::operator= (const Object &o)
{
    if (this != &o) {
	if (ior)
	    delete ior;
	ior = o.ior ? new IOR (*o.ior) : 0;
	CORBA::release (orb);
	orb = CORBA::ORB::_duplicate (o.orb);
	_managers = o._managers;
    }
    return *this;
}

CORBA::Object::~Object ()
{
    if (ior)
        delete ior;
    CORBA::release (orb);
}

void
CORBA::Object::_setup_domains (CORBA::Object_ptr parent)
{
    if (CORBA::is_nil (parent)) {
	CORBA::DomainManager_var def_manager;
	orb->get_default_domain_manager (def_manager);

	_managers.length (1);

	CORBA::Policy_var p =
	    def_manager->_get_policy (CORBA::SecConstruction);
	assert (!CORBA::is_nil (p));

	CORBA::ConstructionPolicy_var cp =
	    CORBA::ConstructionPolicy::_narrow (p);

	if (cp->constr_policy (this)) {
	    _managers[0] = def_manager._retn();
	} else {
	    _managers[0] = def_manager->copy();
	}
    } else {
	CORBA::DomainManagerList_var dml = parent->_get_domain_managers();
	_managers.length (dml->length());

	for (CORBA::ULong i = 0; i < dml->length(); ++i) {
	    CORBA::Policy_var p =
		dml[i]->_get_policy (CORBA::SecConstruction);
	    assert (!CORBA::is_nil (p));

	    CORBA::ConstructionPolicy_var cp =
		CORBA::ConstructionPolicy::_narrow (p);

	    if (cp->constr_policy (this)) {
		_managers[i] = CORBA::DomainManager::_duplicate (dml[i]);
	    } else {
		_managers[i] = dml[i]->copy();
	    }
	}
    }
}

const char *
CORBA::Object::_ident ()
{
    IORProfile *prof = _ior()->profile (IORProfile::TAG_ANY);
    assert (prof);
    Long len;
    const Octet *key = prof->objectkey (len);
    ident = "";
    for (int i = 0; i < len; ++i) {
	ident += mico_to_xdigit ((key[i] >> 4) & 0xf);
	ident += mico_to_xdigit (key[i] & 0xf);
    }
    return ident.c_str();
}

void *
CORBA::Object::_narrow_helper (const char *repoid)
{
    return NULL;
}

CORBA::ImplementationDef_ptr
CORBA::Object::_get_implementation ()
{
    return _orbnc()->get_impl (this);
}

CORBA::InterfaceDef_ptr
CORBA::Object::_get_interface ()
{
    return _orbnc()->get_iface (this);
}

CORBA::Status
CORBA::Object::_create_request (Context_ptr ctx,
				const char *op,
				NVList_ptr args,
				NamedValue_ptr result,
				Request_ptr &request,
				Flags flags)
{
    request = new Request (this, ctx, op, args, result, flags);
    return 0;
}

CORBA::Status
CORBA::Object::_create_request (Context_ptr ctx,
				const char *op,
				NVList_ptr args,
				NamedValue_ptr result,
				ExceptionList_ptr elist,
				ContextList_ptr clist,
				Request_ptr &request,
				Flags flags)
{
    request = new Request (this, ctx, op, args, result, elist, clist, flags);
    return 0;
}

CORBA::Request_ptr
CORBA::Object::_request (const char *op)
{
    return new Request (this, op);
}

CORBA::Policy_ptr
CORBA::Object::_get_policy (PolicyType policy_type)
{
    for (CORBA::ULong i = 0; i < _managers.length(); ++i) {
	CORBA::Policy_ptr p = _managers[i]->get_domain_policy (policy_type);
	if (!CORBA::is_nil (p))
	    return p;
    }
    // XXX should throw BAD_PARAM ...
    return CORBA::Policy::_nil();
}

CORBA::DomainManagerList *
CORBA::Object::_get_domain_managers ()
{
    return new DomainManagerList (_managers);
}

CORBA::Boolean
CORBA::Object::_is_a (const char *repoid)
{
    // check some trivial cases first
    if (!strcmp (repoid, "IDL:omg.org/CORBA/Object:1.0"))
	return TRUE;
    if (!strcmp (repoid, _repoid()))
	return TRUE;

    // try IDL compiler generated narrowing support
    if (_narrow_helper (repoid))
	return TRUE;

    return _is_a_remote (repoid);
}

CORBA::Boolean
CORBA::Object::_is_a_remote (const char *repoid)
{
    // only ask remote if we are a stub
    if (_orbnc()->is_impl (this))
	return FALSE;
    return _orbnc()->is_a (this, repoid);
}

CORBA::Boolean
CORBA::Object::_non_existent ()
{
    return _orbnc()->non_existent (this);
}

CORBA::Boolean
CORBA::Object::_is_equivalent (Object_ptr o)
{
    if (this == o)
	return TRUE;
    return *_ior() == *o->_ior();
}

CORBA::ULong
CORBA::Object::_hash (ULong max)
{
    string s = _ior()->stringify();

    ULong g, v = 0;
    for (mico_vec_size_type i = 0; i < s.length(); ++i) {
	v = (v << 4) + s[i];
	if ((g = v & 0xf0000000)) {
	    v = v ^ (g >> 24);
	    v = v ^ g;
	}
    }
    return v % max;
}

CORBA::Object_ptr
CORBA::Object::_duplicate (Object_ptr o)
{
    if (!CORBA::is_nil (o))
	o->_ref();
    return o;
}

CORBA::Object_ptr
CORBA::Object::_nil ()
{
    return NULL;
}

void
CORBA::release (ServerlessObject *o)
{
    if (!CORBA::is_nil (o) && o->_deref ())
	delete o;
}

CORBA::Boolean
CORBA::is_nil (ServerlessObject *o)
{
    return !o;
}

