/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#include <CORBA.h>
#ifndef _WINDOWS
#include <string.h>
#endif
#include <iostream.h>
#include <mico/imr_impl.h>
#include <mico/ir_impl.h>
#include <mico/impl.h>
#include <mico/intercept.h>
#include <mico/template_impl.h>
#include <mico/util.h>
#include <algorithm>


/************************** misc dtors ****************************/


CORBA::ObjectAdapter::~ObjectAdapter ()
{
}

CORBA::ORBCallback::~ORBCallback ()
{
}


/************************** ORBRequest ****************************/


CORBA::ORBRequest::~ORBRequest ()
{
}

CORBA::ORBRequest *
CORBA::ORBRequest::_duplicate (ORBRequest *r)
{
    if (!CORBA::is_nil (r))
	r->_ref();
    return r;
}

CORBA::ORBRequest *
CORBA::ORBRequest::_nil ()
{
    return NULL;
}

void
CORBA::release (ORBRequest *r)
{
    if (!CORBA::is_nil (r) && r->_deref())
	delete r;
}

CORBA::Boolean
CORBA::is_nil (ORBRequest *r)
{
    return r == NULL;
}


/************************* InvokeRecord ***************************/


CORBA::InvokeRec::InvokeRec (ORB_ptr _orb, MsgId id, ORBRequest *r,
			     ORBCallback *callback, ObjectAdapter *oa)
{
    have_result = FALSE;
    type = RequestInvoke;

    orb = _orb;
    myid = id;
    obj = Object::_nil();
    req = CORBA::ORBRequest::_duplicate (r);
    adapter = oa;
    cb = callback;
}

CORBA::InvokeRec::InvokeRec (ORB_ptr _orb, MsgId id, const char *,
			     const ORB::ObjectTag &tag,
			     ORBCallback *callback, ObjectAdapter *oa)
{
    have_result = FALSE;
    type = RequestBind;
    
    orb = _orb;
    myid = id;
    obj = Object::_nil();
    req = CORBA::ORBRequest::_nil();
    adapter = oa;
    cb = callback;
    objtag = tag;
}

CORBA::InvokeRec::InvokeRec (ORB_ptr _orb, MsgId id, Object_ptr,
			     ORBCallback *callback, ObjectAdapter *oa)
{
    have_result = FALSE;
    type = RequestLocate;
    
    orb = _orb;
    myid = id;
    obj = Object::_nil();
    req = CORBA::ORBRequest::_nil();
    adapter = oa;
    cb = callback;
}

CORBA::InvokeRec::~InvokeRec ()
{
    // if caller needs the object then he must _duplicate() it
    CORBA::release (obj);
    CORBA::release (req);
}

void
CORBA::InvokeRec::set_answer_invoke (InvokeStatus state,
				     Object_ptr o,
				     ORBRequest *r)
{
    assert (type == RequestInvoke);
    assert (!have_result);

    have_result = TRUE;
    invoke_stat = state;

    switch (state) {
    case InvokeOk:
    case InvokeUsrEx:
    case InvokeSysEx:
        if (!req->copy_out_args (r)) {
            CORBA::MARSHAL ex;
            req->set_out_args (&ex);
            state = InvokeSysEx;
        }
	break;

    case InvokeForward:
	obj = Object::_duplicate (o);
	break;

    default:
	assert (0);
    }
}

void
CORBA::InvokeRec::set_answer_bind (LocateStatus state, Object_ptr o)
{
    assert (type == RequestBind);
    assert (!have_result);

    have_result = TRUE;
    locate_stat = state;

    switch (state) {
    case LocateForward:
    case LocateUnknown:
	break;

    case LocateHere:
	obj = Object::_duplicate (o);
	break;

    default:
	assert (0);
    }
}

void
CORBA::InvokeRec::set_answer_locate (LocateStatus state, Object_ptr o)
{
    assert (type == RequestLocate);
    assert (!have_result);

    have_result = TRUE;
    locate_stat = state;

    switch (state) {
    case LocateHere:
    case LocateUnknown:
	break;

    case LocateForward:
	obj = Object::_duplicate (o);
	break;

    default:
	assert (0);
    }
}

CORBA::Boolean
CORBA::InvokeRec::get_answer_invoke (InvokeStatus &state, Object_ptr &o,
				     ORBRequest *&r)
{
    assert (type == RequestInvoke);
    if (!have_result)
        return FALSE;

    o = obj;
    state = invoke_stat;
    r = req;
    return TRUE;
}

CORBA::Boolean
CORBA::InvokeRec::get_answer_bind (LocateStatus &state, Object_ptr &o)
{
    assert (type == RequestBind);
    if (!have_result)
        return FALSE;

    o = obj;
    state = locate_stat;
    return TRUE;
}

CORBA::Boolean
CORBA::InvokeRec::get_answer_locate (LocateStatus &state, Object_ptr &o)
{
    assert (type == RequestLocate);
    if (!have_result)
        return FALSE;

    o = obj;
    state = locate_stat;
    return TRUE;
}


/**************************** ORB *********************************/


CORBA::ORB::ORB (int &argc, char **argv, const char *rcfile)
{
    disp = new MICO::SelectDispatcher ();
    theid = 1;
    stopped = TRUE;
    wait_for_completion = FALSE;
    _rcfile = rcfile;

    tmpl = new IOR;
}

CORBA::ORB::~ORB ()
{
    delete disp;
    delete tmpl;

    map<MsgId, InvokeRec *, less<MsgId> >::iterator i;
    for (i = invokes.begin(); i != invokes.end(); ++i)
	delete (*i).second;
    // XXX deallocate adapters entries ...
}

char *
CORBA::ORB::object_to_string (Object_ptr obj)
{
    return CORBA::string_dup (obj->_ior()->stringify().c_str());
}

CORBA::Object_ptr
CORBA::ORB::string_to_object (const char *str)
{
    IOR *ior = new IOR;
    if (!ior->from_string (str)) {
	delete ior;
	return Object::_nil();
    }
    return ior_to_object (ior);
}

CORBA::Object_ptr
CORBA::ORB::ior_to_object (IOR *ior)
{
    Object_var obj = new Object (ior);

    ObjectAdapter *oa = get_oa (obj);
    if (oa && oa->is_local ()) {
        CORBA::Object_ptr skel = oa->skeleton (obj);
        if (!CORBA::is_nil (skel))
            return Object::_duplicate (skel);
    }
    return Object::_duplicate (obj);
}

CORBA::Boolean
CORBA::ORB::is_impl (Object_ptr obj)
{
    ObjectAdapter *oa = get_oa (obj);
    return oa && oa->is_local () && obj == oa->skeleton (obj);
}

const char *
CORBA::ORB::rcfile()
{
    return _rcfile.c_str();
}

char *
CORBA::ORB::tag_to_string (const ObjectTag &oid)
{
    string s;

    for (CORBA::ULong i = 0; i < oid.length(); ++i) {
	assert (oid[i]);
	s += (char)(CORBA::Octet)oid[i];
    }
    return CORBA::string_dup (s.c_str());
}

CORBA::ORB::ObjectTag *
CORBA::ORB::string_to_tag (const char *s)
{
    CORBA::ULong len = strlen (s);

    CORBA::ORB::ObjectTag *oid = new CORBA::ORB::ObjectTag;
    oid->length (len);

    for (CORBA::ULong i = 0; i < len; ++i) {
        (*oid)[i] = (CORBA::Octet) s[i];
    }
    return oid;
}

CORBA::Status
CORBA::ORB::create_list (Long len, NVList_ptr &res)
{
    res = new NVList (len);
    return 0;
}

CORBA::Status
CORBA::ORB::create_operation_list (OperationDef_ptr op, NVList_ptr &res)
{
    res = new NVList (0);
    ParDescriptionSeq_var params = op->params();
    for (CORBA::ULong i = 0; i < params->length(); ++i) {
	switch (params[i].mode) {
	case PARAM_IN:
	    res->add(ARG_IN)->value()->type (params[i].type);
	    break;
	case PARAM_OUT:
	    res->add(ARG_OUT)->value()->type (params[i].type);
	    break;
	case PARAM_INOUT:
	    res->add(ARG_INOUT)->value()->type (params[i].type);
	    break;
	default:
	    assert (0);
	}
    }
    return 0;
}

CORBA::Status
CORBA::ORB::create_named_value (NamedValue_ptr &res)
{
    res = new NamedValue;
    return 0;
}

CORBA::Status
CORBA::ORB::create_exception_list (ExceptionList_ptr &res)
{
    res = new ExceptionList;
    return 0;
}

CORBA::Status
CORBA::ORB::create_context_list (ContextList_ptr &res)
{
    res = new ContextList;
    return 0;
}

CORBA::Status
CORBA::ORB::create_environment (Environment_ptr &res)
{
    res = new Environment;
    return 0;
}

CORBA::Status
CORBA::ORB::get_default_context (Context_ptr &res)
{
    res = new Context ("");
    return 0;
}

CORBA::Status
CORBA::ORB::get_default_domain_manager (DomainManager_ptr &dm)
{
    if (CORBA::is_nil (_def_manager))
	_def_manager = new MICO::DomainManager_impl;
    dm = CORBA::DomainManager::_duplicate (_def_manager);
    return 0;
}

CORBA::Boolean
CORBA::ORB::get_service_information (
    CORBA::ServiceType service_type,
    CORBA::ServiceInformation *& service_information)
{
    return FALSE;
}

CORBA::Status
CORBA::ORB::send_multiple_requests_oneway (const RequestSeq &req)
{
    CORBA::Status r;
    for (CORBA::ULong i = 0; i < req.length(); ++i) {
	r = req[i]->send_oneway ();
	assert (r == 0);
    }
    return 0;
}

CORBA::Status
CORBA::ORB::send_multiple_requests_deferred (const RequestSeq &req)
{
    CORBA::Status r;
    for (CORBA::ULong i = 0; i < req.length(); ++i) {
	r = req[i]->send_deferred ();
	assert (r == 0);
    }
    return 0;
}

CORBA::Boolean
CORBA::ORB::poll_next_response ()
{
    InvokeRec *rec;
    ORBRequest *orbreq;

    for (InvokeMap::iterator i = invokes.begin(); i != invokes.end(); ++i) {
	rec = (*i).second;
	orbreq = rec->request ();
	if (rec->request_type() == RequestInvoke &&
	    rec->completed() &&
	    !strcmp (orbreq->type(), "local")) {
	    // found a local request that has completed
	    return TRUE;
	}
    }
    return FALSE;
}

CORBA::Status
CORBA::ORB::get_next_response (Request_ptr &req)
{
    InvokeRec *rec;
    ORBRequest *orbreq;

    for (InvokeMap::iterator i = invokes.begin(); i != invokes.end(); ++i) {
	rec = (*i).second;
	orbreq = rec->request ();
	if (rec->request_type() == RequestInvoke &&
	    rec->completed() &&
	    !strcmp (orbreq->type(), "local")) {
	    // found a local request that has completed
	    req = Request::_duplicate (
		((MICO::LocalRequest *)orbreq)->request());
	    return TRUE;
	}
    }
    req = Request::_nil();
    return 1;
}

CORBA::Boolean
CORBA::ORB::work_pending ()
{
    // if !stopped run() has been called and cares for events ...
    return stopped && !disp->idle();
}

void
CORBA::ORB::perform_work ()
{
    if (stopped)
	disp->run (FALSE);
}

void
CORBA::ORB::run ()
{
    stopped = FALSE;
    while (!stopped)
	disp->run (FALSE);

    if (wait_for_completion)
	shutting_down_adapters = adapters;

    // work on copy in case OA's unregister ifself during shutdown ...
    OAVec adapters_copy = adapters;
    for (ULong i = 0; i < adapters_copy.size(); ++i)
	adapters_copy[i]->shutdown (wait_for_completion);

    if (wait_for_completion) {
	while (shutting_down_adapters.size() > 0)
	    disp->run (FALSE);
    }
}

void
CORBA::ORB::answer_shutdown (ObjectAdapter *oa)
{
    for (ULong i = 0; i < shutting_down_adapters.size(); ++i) {
	if (shutting_down_adapters[i] == oa) {
	    shutting_down_adapters.erase (shutting_down_adapters.begin()+i);
	    break;
	}
    }
}

void
CORBA::ORB::shutdown (Boolean wait)
{
    wait_for_completion = wait;
    stopped = TRUE;
}

CORBA::BOA_ptr
CORBA::ORB::BOA_init (int &argc, char **argv, const char *id)
{
    MICOGetOpt::OptMap opts;
    opts["-OAId"] = "arg-expected";

    MICOGetOpt opt_parser (opts);
    CORBA::Boolean r = opt_parser.parse (rcfile(), TRUE);
    assert (r);
    r = opt_parser.parse (argc, argv, TRUE);
    assert (r);
    const MICOGetOpt::OptVec &o = opt_parser.opts();

    for (MICOGetOpt::OptVec::const_iterator i0 = o.begin(); i0 != o.end(); ++i0) {
	string arg = (*i0).first;
	string val = (*i0).second;
	if (arg == "-OAId") {
	    id = val.c_str();
	}
    }

    assert (id);

    for (ULong i1 = 0; i1 < adapters.size(); ++i1) {
	if (!strcmp (id, adapters[i1]->get_oaid()))
	    return CORBA::BOA::_duplicate ((MICO::BOAImpl *)adapters[i1]);
    }
    if (strcmp ("mico-local-boa", id)) {
	MICODebug::instance()->printer()
	    << "ORB::BOA_init(): unknown OA id: " << id << endl;
	assert (0);
    }
    CORBA::BOA_ptr boa = new MICO::BOAImpl (this, argc, argv);

    // make sure iface and impl repository are active ...
    Object_var obj;
#if 0
    obj = resolve_initial_references ("InterfaceRepository");
#endif
    obj = resolve_initial_references ("ImplementationRepository");

    return CORBA::BOA::_duplicate (boa);
}

CORBA::BOA_ptr
CORBA::ORB::BOA_instance (const char *id, Boolean create)
{
    for (ULong i = 0; i < adapters.size(); ++i) {
	if (!strcmp (id, adapters[i]->get_oaid()))
	    return CORBA::BOA::_duplicate ((MICO::BOAImpl *)adapters[i]);
    }
    if (!create)
	return BOA::_nil();
    int argc = 0;
    char *argv = 0;
    return BOA_init (argc, &argv, id);
}

CORBA::ORB::ObjectIdList *
CORBA::ORB::list_initial_services ()
{
    ObjectIdList_ptr idlist = new ObjectIdList;
    idlist->length (_init_refs.size());

    int j;
    InitialRefMap::iterator i;
    for (j = 0, i = _init_refs.begin(); i != _init_refs.end(); ++i, ++j)
	(*idlist)[j] = (*i).first.c_str();

    return idlist;
}

CORBA::Object_ptr
CORBA::ORB::resolve_initial_references (const char *id)
{
    InitialRefMap::iterator i = _init_refs.find (id);

    if (i == _init_refs.end()) {
	if (!strcmp (id, "ImplementationRepository")) {
	    _init_refs[id] = new ImplRepository_impl;
	} else if (!strcmp (id, "InterfaceRepository")) {
	    _init_refs[id] = new Repository_impl;
	} else {
	    MICODebug::instance()->printer()
		<< "ORB::resolve_initial_references(): bad id: " << id << endl;
	    assert (0);
	}
	i = _init_refs.find (id);
    }
    return Object::_duplicate ((*i).second);
}

CORBA::Boolean
CORBA::ORB::set_initial_reference (const char *id, Object_ptr obj)
{
    _init_refs[id] = CORBA::Object::_duplicate (obj);
    return TRUE;
}

CORBA::TypeCode_ptr
CORBA::ORB::create_struct_tc (const char *rep_id,
			      const char *name,
			      const StructMemberSeq &members)
{
    return TypeCode::create_struct_tc (rep_id, name, members);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_union_tc (const char *rep_id,
			     const char *name,
			     TypeCode_ptr disc_type,
			     const UnionMemberSeq &members)
{
    return TypeCode::create_union_tc (rep_id, name, disc_type, members);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_enum_tc (const char *rep_id,
			    const char *name,
			    const EnumMemberSeq &members)
{
    return TypeCode::create_enum_tc (rep_id, name, members);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_exception_tc (const char *rep_id,
				 const char *name,
				 const StructMemberSeq &mems)
{
    return TypeCode::create_exception_tc (rep_id, name, mems);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_alias_tc (const char *rep_id,
			     const char *name,
			     TypeCode_ptr orig_type)
{
    return TypeCode::create_alias_tc (rep_id, name, orig_type);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_interface_tc (const char *rep_id, const char *name)
{
    return TypeCode::create_interface_tc (rep_id, name);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_string_tc (ULong bound)
{
    return TypeCode::create_string_tc (bound);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_wstring_tc (ULong bound)
{
    return TypeCode::create_wstring_tc (bound);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_fixed_tc (UShort digits, Short scale)
{
    return TypeCode::create_fixed_tc (digits, scale);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_sequence_tc (ULong bound, TypeCode_ptr el_type)
{
    return TypeCode::create_sequence_tc (bound, el_type);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_recursive_sequence_tc (ULong bound, ULong offset)
{
    return TypeCode::create_recursive_sequence_tc (bound, offset);
}

CORBA::TypeCode_ptr
CORBA::ORB::create_array_tc (ULong length, TypeCode_ptr el_type)
{
    return TypeCode::create_array_tc (length, el_type);
}

CORBA::ObjectAdapter *
CORBA::ORB::get_oa (Object_ptr o)
{
    // first try local adapters
    for (ULong i0 = 0; i0 < adapters.size(); ++i0) {
	if (adapters[i0]->is_local() && adapters[i0]->has_object (o))
	    return adapters[i0];
    }
    // non local adaptors
    for (ULong i1 = 0; i1 < adapters.size(); ++i1) {
	if (!adapters[i1]->is_local() && adapters[i1]->has_object (o))
	    return adapters[i1];
    }
    return NULL;
}

CORBA::ORB::MsgId
CORBA::ORB::new_msgid ()
{
    if (theid == 0)
        theid = 1;
    return theid++;
}

void
CORBA::ORB::add_invoke (InvokeRec *rec)
{
    invokes[rec->id()] = rec;
}

CORBA::InvokeRec *
CORBA::ORB::get_invoke (MsgId id)
{
    map<MsgId, InvokeRec *, less<MsgId> >::iterator i;
    i = invokes.find (id);
    if (i == invokes.end())
	return NULL;
    return (*i).second;
}

void
CORBA::ORB::del_invoke (MsgId id)
{
    map<MsgId, InvokeRec *, less<MsgId> >::iterator i;
    i = invokes.find (id);
    if (i != invokes.end()) {
	delete (*i).second;
	invokes.erase (i);
    }
}

const vector<string> &
CORBA::ORB::bindaddrs () const
{
    return _bindaddrs;
}

void
CORBA::ORB::bindaddrs (const vector<string> &av)
{
    _bindaddrs = av;
}

CORBA::Dispatcher *
CORBA::ORB::dispatcher ()
{
    return disp;
}

void
CORBA::ORB::dispatcher (Dispatcher *_disp)
{
    if (_disp) {
	// move the pending events from disp to _disp
	disp->move (_disp);
    }
    if (disp) {
	delete disp;
    }
    disp = _disp;
}

CORBA::IOR *
CORBA::ORB::ior_template ()
{
    return tmpl;
}

void
CORBA::ORB::register_oa (ObjectAdapter *oa)
{
    adapters.push_back (oa);
}

void
CORBA::ORB::unregister_oa (ObjectAdapter *oa)
{
    for (ULong i = 0; i < adapters.size(); ) {
	if (adapters[i] == oa)
	    adapters.erase (adapters.begin() + i);
	else
	    ++i;
    }
}

CORBA::Boolean
CORBA::ORB::builtin_invoke (MsgId msgid, Object_ptr obj,
                            ORBRequest *req, Principal_ptr pr)
{
    /*
     * _not_existent is the only special one, _implementation, _interface,
     * and _is_a are just passed through to the OA.
     */
    if (!strcmp (req->op_name(), "_not_existent")) {
        if (!get_oa (obj)) {
            Any res;
            res <<= Any::from_boolean (TRUE);
	    NVList nv;
            if (!req->set_out_args (&res, &nv)) {
                CORBA::MARSHAL ex;
                req->set_out_args (&ex);
                answer_invoke (msgid, InvokeSysEx, Object::_nil(), req);
            } else {
                answer_invoke (msgid, InvokeOk, Object::_nil(), req);
            }
            return TRUE;
        }
    }
    return FALSE;
}

CORBA::ORB::MsgId
CORBA::ORB::invoke_async (Object_ptr obj,
			  ORBRequest *req,
			  Principal_ptr pr,
			  Boolean response_exp,
			  ORBCallback *cb,
			  MsgId msgid)
{
    if (msgid == 0)
	msgid = new_msgid();
    InvokeRec *rec = 0;
    if (response_exp) {
        rec = new InvokeRec (this, msgid, req, cb);
        add_invoke (rec);
    }

    if (!builtin_invoke (msgid, obj, req, pr)) {
        ObjectAdapter *oa = get_oa (obj);
	if (!oa) {
            CORBA::OBJECT_NOT_EXIST ex;
	    req->set_out_args (&ex);
	    answer_invoke (msgid, InvokeSysEx, Object::_nil(), req);
	    return msgid;
	}
        if (response_exp)
            rec->oa (oa);
        oa->invoke (msgid, obj, req, pr, response_exp);
    }
    return response_exp ? msgid : 0;
}

CORBA::ORB::MsgId
CORBA::ORB::locate_async (Object_ptr obj, ORBCallback *cb, MsgId msgid)
{
    ObjectAdapter *oa = get_oa (obj);
    if (msgid == 0)
	msgid = new_msgid();
    add_invoke (new InvokeRec (this, msgid, obj, cb, oa));

    if (!oa) {
        answer_locate (msgid, LocateUnknown, Object::_nil());
    } else {
        oa->locate (msgid, obj);
    }
    return msgid;
}

CORBA::ORB::MsgId
CORBA::ORB::bind_async (const char *repoid,
			const ObjectTag &oid,
			Address *addr, ORBCallback *cb,
			MsgId msgid)
{
    if (msgid == 0)
	msgid = new_msgid();
    InvokeRec *rec = new InvokeRec (this, msgid, repoid, oid, cb);
    add_invoke (rec);
    for (ULong i = 0; i < adapters.size(); ++i) {
	rec->oa (adapters[i]);
        if (adapters[i]->bind (msgid, repoid, rec->tag(), addr))
	    return msgid;
    }
    answer_bind (msgid, LocateUnknown, Object::_nil());
    return msgid;
}

CORBA::InvokeStatus
CORBA::ORB::invoke (Object_ptr &obj, ORBRequest *req,
		    Principal_ptr pr, Boolean reply_exp)
{
    MsgId id = invoke_async (obj, req, pr, reply_exp);
    if (!reply_exp)
	return InvokeOk;
    assert (id != 0);
    CORBA::Boolean r = wait (id);
    assert (r);
    ORBRequest *dummy;
    return get_invoke_reply (id, obj, dummy);
}

CORBA::LocateStatus
CORBA::ORB::locate (Object_ptr &obj)
{
    MsgId id = locate_async (obj);
    CORBA::Boolean r = wait (id);
    assert (r);
    return get_locate_reply (id, obj);
}

CORBA::LocateStatus
CORBA::ORB::bind (const char *repoid, const ObjectTag &oid,
		  Address *addr, Object_ptr &obj)
{
    MsgId id = bind_async (repoid, oid, addr);
    if (id == 0)
	return LocateUnknown;
    CORBA::Boolean r = wait (id);
    assert (r);
    return get_bind_reply (id, obj);
}

CORBA::Object_ptr
CORBA::ORB::bind (const char *repoid, const ObjectTag &oid, const char *addr)
{
    vector<string> *addrs, _addrs;
    if (addr) {
	// use given address
	_addrs.push_back (addr);
	addrs = &_addrs;
    } else {
	// use default addresses
	addrs = &_bindaddrs;
    }

    Address *a;
    for (mico_vec_size_type i = 0; i < addrs->size(); ++i) {
	a = Address::parse ((*addrs)[i].c_str());
	if (a) {
	    Object_ptr obj;
	    LocateStatus s = bind (repoid, oid, a, obj);
	    delete a;
	    if (s == LocateHere)
		return obj;
	} else {
	    MICODebug::instance()->printer()
		<< "ORB::bind(): bad address: " << (*addrs)[i] << endl;
	}
    }
    return Object::_nil();
}

CORBA::Object_ptr
CORBA::ORB::bind (const char *repoid, const char *addr)
{
    return bind (repoid, ObjectTag(), addr);
}

void
CORBA::ORB::cancel (MsgId id)
{
    InvokeRec *rec = get_invoke (id);
    if (rec) {
        if (rec->oa())
            rec->oa()->cancel (id);
	del_invoke (id);
    }
}

CORBA::Long
CORBA::ORB::wait (vector<MsgId> &ids, Long tmout)
{
    Timeout t (disp, tmout);
    InvokeRec *rec;

    while (42) {
	vector<MsgId> ready;
	for (mico_vec_size_type i = 0; i < ids.size(); ++i) {
	    rec = get_invoke (ids[i]);
	    assert (rec);
	    if (rec->completed())
		ready.push_back (ids[i]);
	}
	if (ready.size() > 0) {
	    ids = ready;
	    return ids.size();
	}
	if (t.done())
	    return 0;
	disp->run (FALSE);
    }
}

CORBA::Boolean
CORBA::ORB::wait (MsgId id, Long tmout)
{
    Timeout t (disp, tmout);
    InvokeRec *rec = get_invoke (id);

    while (42) {
        if (rec->completed())
	    return TRUE;
	if (t.done())
	    return FALSE;
	disp->run (FALSE);
    }
}

CORBA::InvokeStatus
CORBA::ORB::get_invoke_reply (MsgId id, Object_ptr &obj, ORBRequest *&r)
{
    InvokeRec *rec = get_invoke (id);
    assert (rec);

    InvokeStatus state;
    Object_ptr o;
    CORBA::Boolean ret = rec->get_answer_invoke (state, o, r);
    assert (ret);
    obj = Object::_duplicate (o);
    del_invoke (id);
    return state;
}

CORBA::LocateStatus
CORBA::ORB::get_locate_reply (MsgId id, Object_ptr &obj)
{
    InvokeRec *rec = get_invoke (id);
    assert (rec);

    LocateStatus state;
    Object_ptr o;
    CORBA::Boolean r = rec->get_answer_locate (state, o);
    assert (r);
    obj = Object::_duplicate (o);
    del_invoke (id);
    return state;
}

CORBA::LocateStatus
CORBA::ORB::get_bind_reply (MsgId id, Object_ptr &obj)
{
    InvokeRec *rec = get_invoke (id);
    assert (rec);

    LocateStatus state;
    Object_ptr o;
    CORBA::Boolean r = rec->get_answer_bind (state, o);
    assert (r);
    obj = Object::_duplicate (o);
    del_invoke (id);
    return state;
}

void
CORBA::ORB::answer_invoke (MsgId id, InvokeStatus stat, Object_ptr obj,
			   ORBRequest *req)
{
    InvokeRec *rec = get_invoke (id);
    if (rec) {
	rec->set_answer_invoke (stat, obj, req);
	if (rec->callback())
	    rec->callback()->callback (this, rec->id(), ORBCallback::Invoke);
    }
}

void
CORBA::ORB::answer_bind (MsgId id, LocateStatus stat, Object_ptr obj)
{
    InvokeRec *rec = get_invoke (id);
    if (rec) {
	rec->set_answer_bind (stat, obj);
	if (rec->callback())
	    rec->callback()->callback (this, rec->id(), ORBCallback::Bind);
    }
}

void
CORBA::ORB::answer_locate (MsgId id, LocateStatus stat, Object_ptr obj)
{
    InvokeRec *rec = get_invoke (id);
    if (rec) {
	rec->set_answer_locate (stat, obj);
	if (rec->callback())
	    rec->callback()->callback (this, rec->id(), ORBCallback::Locate);
    }
}

CORBA::RequestType
CORBA::ORB::request_type (MsgId id)
{
    InvokeRec *rec = get_invoke (id);
    if (!rec)
	return RequestUnknown;
    return rec->request_type ();
}

CORBA::ImplementationDef_ptr
CORBA::ORB::get_impl (Object_ptr obj)
{
    // [12-17]
    Request_ptr req = obj->_request ("_implementation");
    req->result()->value()->type (CORBA::_tc_ImplementationDef);
    req->invoke ();
    assert (!req->env()->exception());
    ImplementationDef_ptr impl;
    CORBA::Boolean r = (*req->result()->value() >>= impl);
    assert (r);
    CORBA::release (req);
    return impl;
}

CORBA::InterfaceDef_ptr
CORBA::ORB::get_iface (Object_ptr obj)
{
    // [12-17]
    Request_ptr req = obj->_request ("_interface");
    req->result()->value()->type (CORBA::_tc_InterfaceDef);
    req->invoke ();
    assert (!req->env()->exception());
    InterfaceDef_ptr iface;
    CORBA::Boolean r = (*req->result()->value() >>= iface);
    assert (r);
    CORBA::release (req);
    return iface;
}

CORBA::Boolean
CORBA::ORB::is_a (Object_ptr obj, const char *repo_id)
{
    // XXX this assumes RepoIds are globally unique
    string key = string(obj->_repoid()) + "$" + repo_id;
    list<string>::iterator i;
    for (i = isa_cache.begin(); i != isa_cache.end(); ++i) {
        if (*i == key)
	    break;
    }
    if (i != isa_cache.end()) {
        isa_cache.erase (i);
        isa_cache.push_front (key);
        return TRUE;
    }

    // [12-17]
    Request_ptr req = obj->_request ("_is_a");
    req->add_in_arg ("logical_type_id") <<= repo_id;
    req->result()->value()->type (CORBA::_tc_boolean);
    req->invoke ();
    assert (!req->env()->exception());
    Boolean res;
    CORBA::Boolean r = (*req->result()->value() >>= Any::to_boolean (res));
    assert (r);
    CORBA::release (req);

    if (res) {
        isa_cache.push_front (key);
        // XXX CACHE size = 50
        if (isa_cache.size() > 50)
            isa_cache.pop_back();
    }
    return res;
}

CORBA::Boolean
CORBA::ORB::non_existent (Object_ptr obj)
{
    // [12-17]
    // XXX is this really called "_not_existent" (not "_non_existent") ?
    Request_ptr req = obj->_request ("_not_existent");
    req->result()->value()->type (CORBA::_tc_boolean);
    req->invoke ();
    Boolean res;
    if (req->env()->exception()) {
        // an exceptions means we cannot contact the object ...
        res = TRUE;
    } else {
        CORBA::Boolean r = (*req->result()->value() >>= Any::to_boolean (res));
	assert (r);
    }
    CORBA::release (req);
    return res;
}

CORBA::ORB_ptr
CORBA::ORB::_duplicate (ORB_ptr r)
{
    if (!CORBA::is_nil (r))
	r->_ref();
    return r;
}

CORBA::ORB_ptr
CORBA::ORB::_nil ()
{
    return NULL;
}

void
CORBA::release (ORB_ptr r)
{
    if (!CORBA::is_nil (r) && r->_deref())
	delete r;
}

CORBA::Boolean
CORBA::is_nil (ORB_ptr r)
{
    return r == NULL;
}


/*********************** ORB initialization **************************/


static CORBA::ORB_ptr orb_instance = CORBA::ORB::_nil();

CORBA::ORB_ptr
CORBA::ORB_instance (const char *id, Boolean create)
{
    if (strcmp ("mico-local-orb", id)) {
	MICODebug::instance()->printer()
	    << "ORB_instance(): unknown ORB id: " << id << endl;
	assert (0);
    }
    if (create && CORBA::is_nil (orb_instance)) {
	int argc = 0;
	char *argv = 0;
	return ORB_init (argc, &argv, id);
    }
    return ORB::_duplicate (orb_instance);
}

CORBA::ORB_ptr
CORBA::ORB_init (int &argc, char **argv, const char *id)
{
    Exception::_init ();

    Boolean run_iiop_server = TRUE;
    Boolean run_iiop_proxy = TRUE;
    string imr_ior, imr_addr;
    string ir_ior, ir_addr;
    string naming_ior, naming_addr;
    ULong debug_level = 0;
    vector<string> bindaddrs, iiopaddrs;
    string rcfile = "~/.micorc";

    // look for config file name ...
    {
	MICOGetOpt::OptMap opts;
	opts["-ORBConfFile"] = "arg-expected";

	MICOGetOpt opt_parser (opts);

	CORBA::Boolean r = opt_parser.parse (argc, argv, TRUE);
	assert (r);
	const MICOGetOpt::OptVec &o = opt_parser.opts();

	MICOGetOpt::OptVec::const_iterator i;
	for (i = o.begin(); i != o.end(); ++i) {
	    string arg = (*i).first;
	    string val = (*i).second;
	    if (arg == "-ORBConfFile") {
		rcfile = val;
	    }
	}
    }

    MICOGetOpt::OptMap opts;
    opts["-ORBNoIIOPServer"]  = "";
    opts["-ORBNoIIOPProxy"]   = "";
    opts["-ORBIIOPAddr"]      = "arg-expected";
    opts["-ORBId"]            = "arg-expected";
    opts["-ORBImplRepoIOR"]   = "arg-expected";
    opts["-ORBImplRepoAddr"]  = "arg-expected";
    opts["-ORBIfaceRepoIOR"]  = "arg-expected";
    opts["-ORBIfaceRepoAddr"] = "arg-expected";
    opts["-ORBNamingIOR"]     = "arg-expected";
    opts["-ORBNamingAddr"]    = "arg-expected";
    opts["-ORBDebugLevel"]    = "arg-expected";
    opts["-ORBBindAddr"]      = "arg-expected";

    MICOGetOpt opt_parser (opts);
    CORBA::Boolean r = opt_parser.parse (rcfile.c_str(), TRUE);
    assert (r);
    r = opt_parser.parse (argc, argv, TRUE);
    assert (r);
    const MICOGetOpt::OptVec &o = opt_parser.opts ();

    for (MICOGetOpt::OptVec::const_iterator i = o.begin(); i != o.end(); ++i) {
	string arg = (*i).first;
	string val = (*i).second;
	if (arg == "-ORBNoIIOPServer") {
	    run_iiop_server = FALSE;
	} else if (arg == "-ORBNoIIOPProxy") {
	    run_iiop_proxy = FALSE;
	} else if (arg == "-ORBIIOPAddr") {
	    iiopaddrs.push_back (val);
	} else if (arg == "-ORBId") {
	    id = val.c_str();
	} else if (arg == "-ORBImplRepoIOR") {
	    imr_ior = val;
	} else if (arg == "-ORBImplRepoAddr") {
	    imr_addr = val;
	} else if (arg == "-ORBIfaceRepoIOR") {
	    ir_ior = val;
	} else if (arg == "-ORBIfaceRepoAddr") {
	    ir_addr = val;
	} else if (arg == "-ORBNamingIOR") {
	    naming_ior = val;
	} else if (arg == "-ORBNamingAddr") {
	    naming_addr = val;
	} else if (arg == "-ORBDebugLevel") {
	    debug_level = atoi (val.c_str());
	} else if (arg == "-ORBBindAddr") {
	    bindaddrs.push_back (val);
	}
    }

    // create Debugger
    if (CORBA::is_nil (orb_instance))
	new MICODebug (debug_level);
    
    if (id && strcmp ("mico-local-orb", id)) {
	MICODebug::instance()->printer()
	    << "ORB_init(): unknown ORB id: " << id << endl;
	assert (0);
    }

    if (!CORBA::is_nil (orb_instance))
	return ORB::_duplicate (orb_instance);

    // create ORB
    orb_instance = new ORB (argc, argv, rcfile.c_str());

    // interceptors
    CORBA::Boolean retval =
	Interceptor::InitInterceptor::_exec_initialize (orb_instance,
							"mico-local-orb",
							argc, argv);
    // XXX what to do on error ??
    assert (retval);

    // set default bind addresses
    orb_instance->bindaddrs (bindaddrs);

    // create IIOP client
    if (run_iiop_proxy) {
	new MICO::IIOPProxy (orb_instance);
    }

    // create IIOP server
    if (run_iiop_server) {
	MICO::IIOPServer *server = new MICO::IIOPServer (orb_instance);
	for (mico_vec_size_type i = 0; i < iiopaddrs.size(); ++i) {
	    Address *addr = Address::parse (iiopaddrs[i].c_str());
	    if (!addr) {
		MICODebug::instance()->printer()
		    << "ORB_init(): bad address: " << iiopaddrs[i] << endl;
		assert (0);
	    }
	    CORBA::Boolean r = server->listen (addr);
	    assert (r);
	    delete addr;
	}
	if (iiopaddrs.size() == 0) {
	    server->listen ();
	}
    } else {
	orb_instance->ior_template()->add_profile (
	    new MICO::LocalProfile ((Octet *)"", 1));
    }

    // connect to / create implementation repository
    Object_var imr;
    if (imr_ior.length() > 0) {
	imr = orb_instance->string_to_object (imr_ior.c_str());
    } else if (imr_addr.length() > 0) {
	imr = orb_instance->bind ("IDL:omg.org/CORBA/ImplRepository:1.0",
				  imr_addr.c_str());
        if (CORBA::is_nil (imr)) {
            MICODebug::instance()->printer()
                << "warning: cannot bind to Implementation Repository at "
                << imr_addr.c_str() << "." << endl
                << "warning: will use a local Implementation Repository"
                << endl;
        }
    }
    if (!CORBA::is_nil (imr)) {
	orb_instance->set_initial_reference ("ImplementationRepository", imr);
    }

    // connect to / create interface repository
    Object_var ir;
    if (ir_ior.length() > 0) {
	ir = orb_instance->string_to_object (ir_ior.c_str());
    } else if (ir_addr.length() > 0) {
	ir = orb_instance->bind ("IDL:omg.org/CORBA/Repository:1.0",
				 ir_addr.c_str());
        if (CORBA::is_nil (ir)) {
            MICODebug::instance()->printer()
                << "warning: cannot bind to Interface Repository at "
                << ir_addr.c_str() << "." << endl
                << "warning: will use a local Interface Repository" << endl;
        }
    }
    if (!CORBA::is_nil (ir)) {
	orb_instance->set_initial_reference ("InterfaceRepository", ir);
    }

    // connect to naming service
    Object_var naming;
    if (naming_ior.length() > 0) {
	naming = orb_instance->string_to_object (naming_ior.c_str());
    } else if (naming_addr.length() > 0) {
	CORBA::ORB::ObjectTag_var tag = CORBA::ORB::string_to_tag ("root");
	naming = orb_instance->bind ("IDL:omg.org/CosNaming/NamingContext:1.0",
				     tag, naming_addr.c_str());
        if (CORBA::is_nil (naming)) {
            MICODebug::instance()->printer()
                << "warning: cannot bind to Naming Service at "
                << naming_addr.c_str() << "." << endl;
        }
    }
    if (!CORBA::is_nil (naming)) {
	orb_instance->set_initial_reference ("NameService", naming);
    }

    return CORBA::ORB::_duplicate (orb_instance);
}
