/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1998 Frank Pilhofer
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

/*
 * Portable Server Module
 */

#include <CORBA.h>
#include <mico/impl.h>
#include <mico/throw.h>
#include <mico/intercept.h>
#include <mico/util.h>
#include <mico/poa_impl.h>
#include <mico/template_impl.h>

/*
 * Static data
 */

MICOPOA::POA_impl::POAMap MICOPOA::POA_impl::AllPOAs;
MICOPOA::POACurrent_impl * MICOPOA::POA_impl::current = NULL;

/*
 * ----------------------------------------------------------------------
 *
 * Initialization via Init Interceptor
 *
 * ----------------------------------------------------------------------
 */

static class POAInit : public Interceptor::InitInterceptor {
private:
  PortableServer::POA_ptr RootPOA;

public:
  POAInit ()
  {
    RootPOA = PortableServer::POA::_nil();
    activate (0);
  }

  ~POAInit ()
  {
    deactivate ();
    CORBA::release (RootPOA);
  }

  Interceptor::Status initialize (CORBA::ORB_ptr orb, const char *orbid,
				  int &argc, char *argv[])
  {
    RootPOA = new MICOPOA::POA_impl (orb, argc, argv);
    orb->set_initial_reference ("RootPOA", RootPOA);
    orb->set_initial_reference ("POACurrent",
				PortableServer::Current::_the_poa_current);
    return Interceptor::INVOKE_CONTINUE;
  }
} InitPOA;

/*
 * ----------------------------------------------------------------------
 *
 * Policy Interfaces
 *
 * ----------------------------------------------------------------------
 */

/*
 * Who defines these? We need 'em!
 */

static const CORBA::PolicyType ThreadPolicy_type = 0;
static const CORBA::PolicyType LifespanPolicy_type = 1;
static const CORBA::PolicyType IdUniquenessPolicy_type = 2;
static const CORBA::PolicyType IdAssignmentPolicy_type = 3;
static const CORBA::PolicyType ImplicitActivationPolicy_type = 4;
static const CORBA::PolicyType ServantRetentionPolicy_type = 5;
static const CORBA::PolicyType RequestProcessingPolicy_type = 6;

/*
 * ----------------------------------------------------------------------
 *
 * POAManager interface
 *
 * ----------------------------------------------------------------------
 */

MICOPOA::POAManager_impl::POAManager_impl ()
{
  _state = HOLDING;
}

MICOPOA::POAManager_impl::~POAManager_impl ()
{
}

PortableServer::POAManager::POAManagerState
MICOPOA::POAManager_impl::state ()
{
  return _state;
}

void
MICOPOA::POAManager_impl::add_managed_poa (PortableServer::POA_ptr poa)
{
  managed.push_back (poa);
}

void
MICOPOA::POAManager_impl::del_managed_poa (PortableServer::POA_ptr poa)
{
  vector<PortableServer::POA_ptr>::iterator it = managed.begin ();

  while (it != managed.end()) {
    if (*it == poa) {
      managed.erase (it);
      return;
    }
    it++;
  }
}

void
MICOPOA::POAManager_impl::change_state (POAManagerState s,
					CORBA::Boolean etherealize_objects,
					CORBA::Boolean wait_for_completion)
{
  if (_state == INACTIVE) {
    mico_throw (PortableServer::POAManager::AdapterInactive());
  }

  if (_state == s) {
    return;
  }

  _state = s;

  vector<PortableServer::POA_ptr>::iterator it = managed.begin();

  while (it != managed.end()) {
    (*it)->poa_manager_callback (_state,
				 etherealize_objects,
				 wait_for_completion);
    it++;
  }
}

void
MICOPOA::POAManager_impl::activate ()
{
  change_state (ACTIVE);
}

void
MICOPOA::POAManager_impl::hold_requests (CORBA::Boolean wait_for_completion)
{
  change_state (HOLDING, FALSE, wait_for_completion);
}

void
MICOPOA::POAManager_impl::discard_requests (CORBA::Boolean wait_for_completion)
{
  change_state (DISCARDING, FALSE, wait_for_completion);
}

void
MICOPOA::POAManager_impl::deactivate (CORBA::Boolean etherealize_objects,
				      CORBA::Boolean wait_for_completion)
{
  change_state (INACTIVE,
		etherealize_objects,
		wait_for_completion);
}

/*
 * ----------------------------------------------------------------------
 *
 * Current interface
 *
 * ----------------------------------------------------------------------
 */

MICOPOA::POACurrent_impl::POACurrent_impl (CORBA::ORB_ptr _orb)
{
  assert (CORBA::is_nil (_the_poa_current));
  _the_poa_current = this;

  orb = _orb;
  cur = FALSE;
}

MICOPOA::POACurrent_impl::~POACurrent_impl ()
{
  _the_poa_current = NULL;
}

PortableServer::POA_ptr
MICOPOA::POACurrent_impl::get_POA ()
{
  if (!cur) {
    mico_throw (PortableServer::Current::NoContext());
  }
  return PortableServer::POA::_duplicate (poa);
}

PortableServer::ObjectId *
MICOPOA::POACurrent_impl::get_object_id ()
{
  if (!cur) {
    mico_throw (PortableServer::Current::NoContext());
  }
  return new PortableServer::ObjectId (*por->get_id());
}

CORBA::Boolean
MICOPOA::POACurrent_impl::iscurrent ()
{
  return cur;
}

CORBA::Object_ptr
MICOPOA::POACurrent_impl::make_ref ()
{
  assert (cur);
  return CORBA::Object::_duplicate (por->get_ref());
}

PortableServer::Servant
MICOPOA::POACurrent_impl::get_serv ()
{
  assert (cur);
  return serv;
}

void
MICOPOA::POACurrent_impl::set (PortableServer::POA_ptr _poa,
			       POAObjectReference * _por,
			       PortableServer::Servant _serv)
{
  assert (!cur);
  poa  = _poa;
  por  = _por;
  serv = _serv;
  cur  = TRUE;
}

void
MICOPOA::POACurrent_impl::unset ()
{
  assert (cur);
  cur = FALSE;
}

/*
 * ----------------------------------------------------------------------
 *
 * Helper Classes
 *
 * ----------------------------------------------------------------------
 */

/*
 * Object References
 */

MICOPOA::POAObjectReference::POAObjectReference (CORBA::ORB_ptr _orb,
						 const char * _poa,
						 const PortableServer::ObjectId &_i,
						 const char * _repoid)
  : orb (_orb), poa (_poa), repoid (_repoid)
{
  assert (_orb);
  assert (_poa);
  assert (_repoid);

  id   = new PortableServer::ObjectId (_i);
  ref  = CORBA::Object::_nil ();
  pid  = OSMisc::getpid ();
  host = MICO::InetAddress::hostid ();

  assert (id);
}

MICOPOA::POAObjectReference::POAObjectReference (CORBA::ORB_ptr _orb,
						 CORBA::Object_ptr _ref)
  : orb (_orb)
{
  id = NULL;
  ref = CORBA::Object::_duplicate (_ref);
}

MICOPOA::POAObjectReference::POAObjectReference (const POAObjectReference &o)
{
  orb = o.orb;

  pid  = o.pid;
  host = o.host;
  poa  = o.poa;

  repoid = o.repoid;
  ref = CORBA::Object::_duplicate (o.ref);

  if (o.id) {
    id = new PortableServer::ObjectId (*o.id);
    assert (id);
  }
  else {
    id = NULL;
  }
}

MICOPOA::POAObjectReference::~POAObjectReference ()
{
  CORBA::release (ref);
  delete id;
}

void
MICOPOA::POAObjectReference::make_ref ()
{
  if (!CORBA::is_nil (ref)) {
    // nothing to do
    return;
  }

  if (id == NULL) {
    // We manage a nil reference
    return;
  }

  /*
   * key = <ip-address> <pid> <poa-name> <nullbyte> <id>
   */

  CORBA::ULong length = host.size() + sizeof(CORBA::Long) + poa.length() + id->length() + 1;
  CORBA::Octet * key = new CORBA::Octet[length];
  CORBA::Long i, j = poa.length();

  for (j=0; j < host.size(); j++) {
    key[j] = (CORBA::Octet)host[j];
  }

  for (i=0; i < sizeof(CORBA::Long); i++, j++) {
    key[j] = ((CORBA::Octet *) &pid)[i];
  }

  memcpy (key+j, poa.c_str(), poa.length());
  j += poa.length();
  key[j++] = 0;

  for (i=0; i<id->length(); i++, j++) {
    key[j] = (CORBA::Octet)((*id)[i]);
  }

  assert (j == length);

  CORBA::IOR * ior = new CORBA::IOR (*orb->ior_template());
  ior->objectkey (key, length);
  ior->objid (repoid.c_str());
  delete [] key;

  ref = new CORBA::Object (ior);
  assert (!CORBA::is_nil (ref));
}

bool
MICOPOA::POAObjectReference::decompose_ref ()
{
  if (id) {
    // nothing to do
    return true;
  }

  if (CORBA::is_nil (ref)) {
    return false;
  }

  CORBA::Long i, j, length;
  const CORBA::Octet * key = ref->_ior()->profile()->objectkey (length);

  if (!key) return false;

  vector<CORBA::Octet> localhost = MICO::InetAddress::hostid ();
  host.clear();

  for (j=0; j<length && j<localhost.size(); j++) {
    host.push_back (key[j]);
  }

  for (i=0; j<length && i<sizeof(CORBA::Long); i++, j++) {
    ((CORBA::Octet *) &pid)[i] = key[j];
  }

  poa = "";
  for (i=0; j<length && key[j]; i++, j++) {
    poa += key[j];
  }

  if (++j > length) return false;

  id = new PortableServer::ObjectId;
  assert (id);

  id->length (length - j);

  for (i=0; j<length; i++, j++) {
    (*id)[i] = key[j];
  }

  repoid = ref->_repoid ();

  return true;
}

bool
MICOPOA::POAObjectReference::is_legal ()
{
  if (id) {
    return true;
  }
  if (!CORBA::is_nil (ref)) {
    return decompose_ref ();
  }
  return false;
}

bool
MICOPOA::POAObjectReference::is_local ()
{
  if (pid != OSMisc::getpid()) {
    return false;
  }

  vector<CORBA::Octet> localhost = MICO::InetAddress::hostid ();

  if (host.size() != localhost.size()) {
    return false;
  }

  for (int i=0; i<host.size(); i++) {
    if (host[i] != localhost[i]) {
      return false;
    }
  }

  return true;
}

const PortableServer::ObjectId *
MICOPOA::POAObjectReference::get_id ()
{
  if (!id) {
    bool r = decompose_ref ();
    assert (r);
  }
  return id;
}

CORBA::Object_ptr
MICOPOA::POAObjectReference::get_ref ()
{
  make_ref ();
  return ref;
}

MICOPOA::POAObjectReference &
MICOPOA::POAObjectReference::operator= (const CORBA::Object_ptr obj)
{
  // assume same ORB
  delete id;
  id = NULL;
  CORBA::release (ref);
  ref = CORBA::Object::_duplicate (obj);
  return *this;
}

MICOPOA::POAObjectReference &
MICOPOA::POAObjectReference::operator= (const POAObjectReference &o)
{
  delete id;
  id = NULL;
  CORBA::release (ref);

  orb = o.orb;
  poa = o.poa;
  repoid = o.repoid;
  ref = CORBA::Object::_duplicate (o.ref);

  if (o.id) {
    id = new PortableServer::ObjectId (*o.id);
    assert (id);
  }
  else {
    id = NULL;
  }
  return *this;
}

const char *
MICOPOA::POAObjectReference::poa_name ()
{
  bool r = decompose_ref ();
  assert (r);
  return poa.c_str();
}

bool
MICOPOA::POAObjectReference::equal (const CORBA::Object_ptr obj)
{
  if (!CORBA::is_nil (ref) && ref == obj) {
    return true;
  }
  if (CORBA::is_nil (ref) && id == NULL && CORBA::is_nil (obj)) {
    return true;
  }

  const CORBA::Octet * k1, * k2;
  CORBA::Long i, j;

  make_ref ();

  k1 = ref->_ior()->profile()->objectkey (i);
  k2 = obj->_ior()->profile()->objectkey (j);

  return (i == j && memcmp (k1, k2, i) == 0);
}

bool
MICOPOA::POAObjectReference::equal (const POAObjectReference &o)
{
  // test if object pointers are equal

  if (!CORBA::is_nil (ref) && !CORBA::is_nil (o.ref) && ref == o.ref) {
    return true;
  }

  // are both references nil?

  if (CORBA::is_nil (ref) && CORBA::is_nil (o.ref) &&
      id == NULL && o.id == NULL) {
    return true;
  }

  // is this reference nil but not the other?

  if (CORBA::is_nil (ref) && id == NULL &&
      (!CORBA::is_nil (o.ref) || o.id != NULL)) {
    return false;
  }

  // is this reference non-nil, but the other?

  if ((!CORBA::is_nil (ref) || id != NULL) &&
      CORBA::is_nil (o.ref) && o.id == NULL) {
    return false;
  }

  // if the other has already computed an object reference,
  // compute our own and compare object keys.

  if (!CORBA::is_nil (o.ref)) {
    const CORBA::Octet * k1, * k2;
    CORBA::Long i, j;

    make_ref ();

    k1 = ref->_ior()->profile()->objectkey (i);
    k2 = o.ref->_ior()->profile()->objectkey (j);

    return (i == j && memcmp (k1, k2, i) == 0);
  }

  // else compare POA identity and object identity

  bool r = decompose_ref ();
  assert (r);

  return (poa == o.poa && *id == *o.id);
}

bool
MICOPOA::POAObjectReference::in_poa (const char * pname)
{
  if (!decompose_ref()) {
    return false;
  }
  return (poa == pname);
}

bool
MICOPOA::POAObjectReference::in_descendant_poa (const char * pname)
{
  if (!decompose_ref()) {
    return false;
  }
  return !strcmp (poa.c_str(), pname);
}

char *
MICOPOA::POAObjectReference::next_descendant_poa (const char * pname)
{
  bool r = decompose_ref ();
  assert (r);
  assert (in_descendant_poa (pname));

  const char * tpn = poa.c_str();
  size_t i, j, k;
  char * res;

  i = strlen (pname);

  assert (tpn[i] != '\0');
  assert (tpn[i] == '.');
  i++;

  for (j=0; tpn[i+j] != '.' && tpn[i+j] != '\0'; j++);

  assert (j);
  res = CORBA::string_alloc (j);
  assert (res);

  for (k=0; k<j; k++) {
    res[k] = tpn[i+k];
  }

  res[j] = '\0';

  return res;
}

/*
 * Unique Id generator
 */

MICOPOA::UniqueIdGenerator::UniqueIdGenerator ()
{
  uid = NULL;
  prefix = NULL;
  pfxlen = 0;
}

MICOPOA::UniqueIdGenerator::UniqueIdGenerator (const char * pfx)
{
  uid = NULL;
  prefix = CORBA::string_dup (pfx);
  pfxlen = strlen (prefix);
}

MICOPOA::UniqueIdGenerator::~UniqueIdGenerator ()
{
  CORBA::string_free (uid);
  CORBA::string_free (prefix);
}

char *
MICOPOA::UniqueIdGenerator::new_id ()
{
  char * id;

  /*
   * Generate a new unique id
   */

  if (uid == NULL) {
    ulen = 1;
    uid  = CORBA::string_alloc (ulen);
    assert (uid);
    uid[0] = '0';
    uid[1] = 0;
  }
  else {
    int i;
    for (i=0; i<ulen; i++) {
      if (uid[i] != '9')
	break;
      uid[i] = '0';
    }
    if (i == ulen) {
      CORBA::string_free (uid);
      uid = CORBA::string_alloc (++ulen);
      assert (uid);
      for (i=0; i<ulen-1; i++) {
	uid[i] = '0';
      }
      uid[ulen-1] = '1';
      uid[ulen] = 0;
    }
    else {
      uid[i] = uid[i] + 1;
    }
  }
  id = CORBA::string_alloc (ulen + pfxlen);
  assert (id);
  strcpy (id, prefix);
  strcpy (id+pfxlen, uid);

  return id;
}

/*
 * ----------------------------------------------------------------------
 *
 * The POA
 *
 * ----------------------------------------------------------------------
 */

/*
 * ObjectRecord and Active Object Map
 */

MICOPOA::POA_impl::ObjectRecord::ObjectRecord (POAObjectReference * _por,
					       PortableServer::Servant _serv)
  : por (_por), serv(_serv)
{
  assert (por);
}

MICOPOA::POA_impl::ObjectRecord::~ObjectRecord ()
{
  delete por;
}

MICOPOA::POA_impl::ORList::~ORList ()
{
  vector<ObjectRecord *>::iterator it = objs.begin ();

  while (it != objs.end()) {
    delete *it;
    it++;
  }
}

void
MICOPOA::POA_impl::ORList::add (ObjectRecord * _por)
{
  objs.push_back (_por);
}

void
MICOPOA::POA_impl::ORList::del (ObjectRecord * _por)
{
  vector<ObjectRecord *>::iterator it = objs.begin ();

  while (it != objs.end()) {
    if (*it == _por) {
      objs.erase (it);
      return;
    }
    it++;
  }
  assert (0);
}

bool
MICOPOA::POA_impl::ORList::empty () const
{
  return objs.empty();
}

MICOPOA::POA_impl::ObjectRecord *
MICOPOA::POA_impl::ORList::first ()
{
  if (empty()) return NULL;
  return objs.front();
}

MICOPOA::POA_impl::ObjectRecord *
MICOPOA::POA_impl::ORList::find (const PortableServer::ObjectId * id)
{
  vector<ObjectRecord *>::iterator it = objs.begin ();

  while (it != objs.end()) {
    if (*(*it)->por->get_id() == *id) {
      return *it;
    }
    it++;
  }

  return NULL;
}

MICOPOA::POA_impl::ObjectRecord *
MICOPOA::POA_impl::ORList::find_obj (CORBA::Object_ptr obj)
{
  vector<ObjectRecord *>::iterator it = objs.begin ();

  while (it != objs.end()) {
    if ((*it)->por->equal (obj)) {
      return *it;
    }
    it++;
  }

  return NULL;
}

MICOPOA::POA_impl::ObjectRecord *
MICOPOA::POA_impl::ORList::find_por (POAObjectReference * por)
{
  vector<ObjectRecord *>::iterator it = objs.begin ();

  while (it != objs.end()) {
    if ((*it)->por->equal (*por)) {
      return *it;
    }
    it++;
  }

  return NULL;
}

MICOPOA::POA_impl::ObjectRecord *
MICOPOA::POA_impl::ORList::find_serv (PortableServer::Servant serv)
{
  vector<ObjectRecord *>::iterator it = objs.begin ();

  while (it != objs.end()) {
    if ((*it)->serv == serv) {
      return *it;
    }
    it++;
  }

  return NULL;
}

MICOPOA::POA_impl::ObjectRecord *
MICOPOA::POA_impl::ORList::operator[] (CORBA::ULong i)
{
  vector<ObjectRecord *>::iterator it = objs.begin ();

  while (it != objs.end()) {
    if (!i) return *it;
    it++;
    i--;
  }

  return NULL;
}

MICOPOA::POA_impl::InvocationRecord::InvocationRecord (CORBA::ORB::MsgId _id,
						       POAObjectReference * _por,
						       CORBA::ORBRequest * _req,
						       CORBA::Principal_ptr _pr)
  : msgid (_id)
{
  por = new POAObjectReference (*_por);
  req = CORBA::ORBRequest::_duplicate (_req);
  pr  = CORBA::Principal::_duplicate (_pr);
  svr = CORBA::ServerRequest::_nil ();
}

MICOPOA::POA_impl::InvocationRecord::~InvocationRecord ()
{
  CORBA::release (req);
  CORBA::release (pr);
  CORBA::release (svr);
  delete por;
}

MICOPOA::POA_impl::InvocationRecord_ptr
MICOPOA::POA_impl::InvocationRecord::_duplicate (InvocationRecord_ptr ir)
{
  if (!is_nil (ir))
    ir->_ref();
  return ir;
}

MICOPOA::POA_impl::InvocationRecord_ptr
MICOPOA::POA_impl::InvocationRecord::_nil ()
{
  return NULL;
}

CORBA::Boolean
MICOPOA::POA_impl::is_nil (InvocationRecord_ptr ir)
{
  return ir == NULL;
}

void
MICOPOA::POA_impl::InvocationRecord::exec (POA_impl * poa)
{
  poa->local_invoke (this);
}

CORBA::ORB::MsgId
MICOPOA::POA_impl::InvocationRecord::id ()
{
  return msgid;
}

CORBA::ORBRequest *
MICOPOA::POA_impl::InvocationRecord::get_or ()
{
  return req;
}

MICOPOA::POAObjectReference *
MICOPOA::POA_impl::InvocationRecord::get_por ()
{
  return por;
}

CORBA::ServerRequest_ptr
MICOPOA::POA_impl::InvocationRecord::make_req (POA_impl * poa)
{
  if (CORBA::is_nil (svr)) {
    svr = new CORBA::ServerRequest (req, por->get_ref(), msgid, poa, pr);
    assert (svr);
  }
  return svr;
}

/*
 * POA attributes
 */

char *
MICOPOA::POA_impl::the_name ()
{
  return CORBA::string_dup (name.c_str());
}

PortableServer::POA_ptr
MICOPOA::POA_impl::the_parent ()
{
  return PortableServer::POA::_duplicate (parent);
}

PortableServer::POAManager_ptr
MICOPOA::POA_impl::the_POAManager ()
{
  return PortableServer::POAManager::_duplicate (manager);
}

PortableServer::AdapterActivator_ptr
MICOPOA::POA_impl::the_activator ()
{
  return PortableServer::AdapterActivator::_duplicate (adapter_activator);
}

void
MICOPOA::POA_impl::the_activator (PortableServer::AdapterActivator_ptr _aa)
{
  adapter_activator = PortableServer::AdapterActivator::_duplicate (_aa);
}

/*
 * Factories for Policy objects
 */

PortableServer::ThreadPolicy_ptr
MICOPOA::POA_impl::create_thread_policy (PortableServer::ThreadPolicyValue _v)
{
  return new ThreadPolicy_impl (ThreadPolicy_type, _v);
}

PortableServer::LifespanPolicy_ptr
MICOPOA::POA_impl::create_lifespan_policy (PortableServer::LifespanPolicyValue _v)
{
  return new LifespanPolicy_impl (LifespanPolicy_type, _v);
}

PortableServer::IdUniquenessPolicy_ptr
MICOPOA::POA_impl::create_id_uniqueness_policy (PortableServer::IdUniquenessPolicyValue _v)
{
  return new IdUniquenessPolicy_impl (IdUniquenessPolicy_type, _v);
}

PortableServer::IdAssignmentPolicy_ptr
MICOPOA::POA_impl::create_id_assignment_policy (PortableServer::IdAssignmentPolicyValue _v)
{
  return new IdAssignmentPolicy_impl (IdAssignmentPolicy_type, _v);
}

PortableServer::ImplicitActivationPolicy_ptr
MICOPOA::POA_impl::create_implicit_activation_policy (PortableServer::ImplicitActivationPolicyValue _v)
{
  return new ImplicitActivationPolicy_impl (ImplicitActivationPolicy_type, _v);
}

PortableServer::ServantRetentionPolicy_ptr
MICOPOA::POA_impl::create_servant_retention_policy (PortableServer::ServantRetentionPolicyValue _v)
{
  return new ServantRetentionPolicy_impl (ServantRetentionPolicy_type, _v);
}

PortableServer::RequestProcessingPolicy_ptr
MICOPOA::POA_impl::create_request_processing_policy (PortableServer::RequestProcessingPolicyValue _v)
{
  return new RequestProcessingPolicy_impl (RequestProcessingPolicy_type, _v);
}

/*
 * ServantManager registration
 */

PortableServer::ServantManager_ptr
MICOPOA::POA_impl::get_servant_manager ()
{
  if (request_processing_policy->value () != PortableServer::USE_SERVANT_MANAGER) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }
  return PortableServer::ServantManager::_duplicate (servant_manager);
}

void
MICOPOA::POA_impl::set_servant_manager (PortableServer::ServantManager_ptr _s)
{
  if (request_processing_policy->value () != PortableServer::USE_SERVANT_MANAGER) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  servant_manager = PortableServer::ServantManager::_duplicate (_s);
}

PortableServer::Servant
MICOPOA::POA_impl::get_servant ()
{
  if (request_processing_policy->value () != PortableServer::USE_DEFAULT_SERVANT) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  if (default_servant == NULL) {
    mico_throw (PortableServer::POA::NoServant());
  }

  return default_servant;
}

void
MICOPOA::POA_impl::set_servant (PortableServer::Servant _s)
{
  if (request_processing_policy->value () != PortableServer::USE_DEFAULT_SERVANT) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  default_servant = _s;
}

/*
 * POA creation and destruction
 */

void
MICOPOA::POA_impl::set_policies (const CORBA::PolicyList * policies)
{
  thread_policy = PortableServer::ThreadPolicy::_nil();
  lifespan_policy = PortableServer::LifespanPolicy::_nil();
  id_uniqueness_policy = PortableServer::IdUniquenessPolicy::_nil();
  id_assignment_policy = PortableServer::IdAssignmentPolicy::_nil();
  implicit_activation_policy =
    PortableServer::ImplicitActivationPolicy::_nil();
  servant_retention_policy = PortableServer::ServantRetentionPolicy::_nil();
  request_processing_policy = PortableServer::RequestProcessingPolicy::_nil();

  if (policies) {
    for (CORBA::ULong i=0; i<policies->length(); i++) {
      CORBA::Policy_var pol = (*policies)[i];
      switch (pol->policy_type()) {
      case ThreadPolicy_type:
	thread_policy = PortableServer::ThreadPolicy::_narrow (pol);
	assert (thread_policy);
	break;
	
      case LifespanPolicy_type:
	lifespan_policy = PortableServer::LifespanPolicy::_narrow (pol);
	assert (lifespan_policy);
	break;
	
      case IdUniquenessPolicy_type:
	id_uniqueness_policy =
	  PortableServer::IdUniquenessPolicy::_narrow (pol);
	assert (id_uniqueness_policy);
	break;
	
      case IdAssignmentPolicy_type:
	id_assignment_policy =
	  PortableServer::IdAssignmentPolicy::_narrow (pol);
	assert (id_assignment_policy);
	break;
	
      case ImplicitActivationPolicy_type:
	implicit_activation_policy =
	  PortableServer::ImplicitActivationPolicy::_narrow (pol);
	assert (implicit_activation_policy);
	break;
	
      case ServantRetentionPolicy_type:
	servant_retention_policy =
	  PortableServer::ServantRetentionPolicy::_narrow (pol);
	assert (servant_retention_policy);
	break;
	
      case RequestProcessingPolicy_type:
	request_processing_policy =
	  PortableServer::RequestProcessingPolicy::_narrow (pol);
	assert (request_processing_policy);
	break;
	
      default:
	{
	  PortableServer::POA::InvalidPolicy ex;
	  ex.index = i;
	  mico_throw (ex);
	}
	break;
      }
    }
  }

  /*
   * set default policies
   */

  if (!thread_policy) {
    thread_policy = create_thread_policy (PortableServer::ORB_CTRL_MODEL);
  }

  if (!lifespan_policy) {
    lifespan_policy = create_lifespan_policy (PortableServer::TRANSIENT);
  }

  if (!id_uniqueness_policy) {
    id_uniqueness_policy =
      create_id_uniqueness_policy (PortableServer::UNIQUE_ID);
  }

  if (!id_assignment_policy) {
    id_assignment_policy =
      create_id_assignment_policy (PortableServer::SYSTEM_ID);
  }

  if (!implicit_activation_policy) {
    implicit_activation_policy =
      create_implicit_activation_policy (PortableServer::NO_IMPLICIT_ACTIVATION);
  }

  if (!servant_retention_policy) {
    servant_retention_policy =
      create_servant_retention_policy (PortableServer::RETAIN);
  }

  if (!request_processing_policy) {
    request_processing_policy =
      create_request_processing_policy (PortableServer::USE_ACTIVE_OBJECT_MAP_ONLY);
  }
}

void
MICOPOA::POA_impl::register_child (const char * name, POA_impl * thechild)
{
  assert (children.find(name) == children.end());
  (void) PortableServer::POA::_duplicate (thechild);
  children[name] = thechild;
}

void
MICOPOA::POA_impl::unregister_child (const char * name)
{
  POAMap::iterator it = children.find (name);
  assert (it != children.end());
  POA_impl * thechild = (*it).second;
  children.erase (it);
  CORBA::release (thechild);
}

void
MICOPOA::POA_impl::register_poa (const char * name, POA_impl * thechild)
{
  assert (AllPOAs.find(name) == AllPOAs.end());
  AllPOAs[name] = thechild;
}

void
MICOPOA::POA_impl::unregister_poa (const char * name)
{
  POAMap::iterator it = AllPOAs.find (name);
  assert (it != AllPOAs.end());
  AllPOAs.erase (it);
}

void
MICOPOA::POA_impl::etherealize ()
{
  /*
   * Can etherealize if RETAIN and USE_SERVANT_MANAGER, and if a
   * ServantActivator has been registered
   */

  if (servant_retention_policy->value() != PortableServer::RETAIN ||
      request_processing_policy->value() != PortableServer::USE_SERVANT_MANAGER) {
    return;
  }

  if (CORBA::is_nil (servant_manager)) {
    return;
  }

  PortableServer::ServantActivator_var sav = 
    PortableServer::ServantActivator::_narrow (servant_manager);

  if (CORBA::is_nil (sav)) {
    return;
  }

  while (!ActiveObjectMap.empty()) {
    ObjectRecord * orec = ActiveObjectMap.first();
    ActiveObjectMap.del (orec);
    ObjectRecord * other = ActiveObjectMap.find_serv (orec->serv);
    sav->etherealize (*orec->por->get_id(), this, orec->serv,
		      TRUE, (other != NULL) ? TRUE : FALSE);
    delete orec;
  }
}

/*
 * Constructor for child POAs. They are not registered with the ORB
 */

MICOPOA::POA_impl::POA_impl (const char * _name,
			     PortableServer::POAManager_ptr _manager,
			     const CORBA::PolicyList & policies,
			     POA_impl * _parent,
			     CORBA::ORB_ptr _orb)
  : name (_name), parent(_parent), orb(_orb),
    idfactory ("_")
{
  destructed = 0;
  unique_id = 0;
  set_policies (&policies);

  /*
   * We keep one reference to ourselves
   */

  (void) PortableServer::POA::_duplicate (this);

  /*
   * Use given POA Manager or create a new one (managers are a
   * dime a dozen, ask Dilbert)
   */

  if (!CORBA::is_nil (_manager)) {
    manager = PortableServer::POAManager::_duplicate (_manager);
  }
  else {
    manager = new POAManager_impl ();
  }

  assert (orb);
  assert (parent);
  assert (manager);

  default_servant   = NULL;
  servant_manager   = PortableServer::ServantManager::_nil ();
  adapter_activator = PortableServer::AdapterActivator::_nil ();

  oaid = parent->oaid + "." + name;
  state = manager->state ();

  /*
   * Register globally, with our manager, and with our parent
   */

  register_poa (oaid.c_str(), this);
  manager->add_managed_poa (this);
  parent->register_child (name.c_str(), this);
}

/*
 * Constructor for root POA
 */

MICOPOA::POA_impl::POA_impl (CORBA::ORB_ptr _orb, int &argc, char **argv)
  : orb (_orb), idfactory ("_")
{
  destructed = 0;
  unique_id = 0;

  /*
   * We keep one reference to ourselves
   */

  (void) PortableServer::POA::_duplicate (this);

  /*
   * The only non-default policy for the RootPOA is the Implicit
   * Activation policy, which is IMPLICIT_ACTIVATION
   */

  CORBA::PolicyList policies (1);
  policies.length (1);
  policies[0] =
    create_implicit_activation_policy (PortableServer::IMPLICIT_ACTIVATION);
  set_policies (&policies);

  parent = NULL;
  manager = new POAManager_impl ();
  current = new POACurrent_impl (orb);

  assert (orb);
  assert (manager);
  assert (current);

  default_servant   = NULL;
  servant_manager   = PortableServer::ServantManager::_nil ();
  adapter_activator = PortableServer::AdapterActivator::_nil ();

  name = (const char *) "RootPOA";
  oaid = name;
  state = PortableServer::POAManager::HOLDING;

  PortableServer::POA::_the_root_poa = this;

  register_poa (oaid.c_str(), this);
  manager->add_managed_poa (this);
  orb->register_oa (this);
}

MICOPOA::POA_impl::~POA_impl ()
{
  destroy (FALSE, FALSE);
  assert (children.empty());
  assert (InvocationQueue.empty());

  if (_the_root_poa == this) {
    _the_root_poa = PortableServer::POA::_nil ();
    CORBA::release (current);
  }

  CORBA::release (manager);
}

PortableServer::POA_ptr
MICOPOA::POA_impl::create_POA (const char * name,
			       PortableServer::POAManager_ptr theManager,
			       const CORBA::PolicyList & policies)
{
  if (children.find (name) != children.end()) {
    mico_throw (PortableServer::POA::AdapterAlreadyExists());
  }

  PortableServer::POA_ptr child = new POA_impl (name, theManager,
						policies, this, orb);
  assert (!CORBA::is_nil (child));
  return child;
}

MICOPOA::POA_impl *
MICOPOA::POA_impl::_find_POA (const char * name,
			      CORBA::Boolean activate_it)
{
  POAMap::iterator it = children.find (name);

  if (it != children.end()) {
    return (*it).second;
  }

  if (!activate_it || CORBA::is_nil (adapter_activator)) {
    return NULL;
  }

  /*
   * May only invoke the adapter activator in the Active state
   */

  if (state != PortableServer::POAManager::ACTIVE) {
    return NULL;
  }

  if (!adapter_activator->unknown_adapter (this, name)) {
    return NULL;
  }

  /*
   * See if the new adapter has registered itself
   */

  if ((it = children.find (name)) != children.end()) {
    return (*it).second;
  }

  return NULL;
}


PortableServer::POA_ptr
MICOPOA::POA_impl::find_POA (const char * name,
			     CORBA::Boolean activate_it)
{
  POA_impl * child = _find_POA (name, activate_it);

  if (!child) {
    mico_throw (PortableServer::POA::AdapterNonExistent());
  }

  return PortableServer::POA::_duplicate (child);
}

void
MICOPOA::POA_impl::destroy (CORBA::Boolean etherealize_objects,
			    CORBA::Boolean wait_for_completion)
{
  if (destructed) {
    return;
  }

  destructed = 1;

  /*
   * Root POA must unregister from the ORB
   */

  if (!parent) {
    orb->unregister_oa (this);
  }

  /*
   * unregister globally, from parent and from manager
   */

  unregister_poa (oaid.c_str());

  if (parent) {
    parent->unregister_child (name.c_str());
  }

  manager->del_managed_poa (this);

  /*
   * If in the holding state, requests may be left in the queue.
   * Set state to DISCARDING and invoke them
   */

  state = PortableServer::POAManager::DISCARDING;

  while (!InvocationQueue.empty()) {
    InvocationRecord_ptr ir = InvocationQueue.front();
    InvocationQueue.erase (InvocationQueue.begin());
    ir->exec (this);
    CORBA::release (ir);
  }

  /*
   * Destroy child POAs
   */

  while (!children.empty()) {
    POAMap::iterator it = children.begin ();
    PortableServer::POA_ptr child = (*it).second;
    child->destroy (etherealize_objects, wait_for_completion);
  }

  /*
   * Etherealize active objects
   */

  etherealize ();

  /*
   * Release the reference that we hold ourselves
   */

  CORBA::release (this);
}

/*
 * Callback from POA Manager
 */

void
MICOPOA::POA_impl::poa_manager_callback (PortableServer::POAManager::POAManagerState s,
					 CORBA::Boolean etherealize_objects,
					 CORBA::Boolean wait_for_completion)
{
  state = s;

  switch (state) {
  case PortableServer::POAManager::HOLDING:
    /*
     * Nothing to do on transition to Holding
     */
    break;

  case PortableServer::POAManager::ACTIVE:
    /*
     * perform queued invocations
     */
    while (!InvocationQueue.empty()) {
      InvocationRecord_ptr ir = InvocationQueue.front();
      InvocationQueue.erase (InvocationQueue.begin());
      ir->exec (this);
      CORBA::release (ir);
    }
    break;

  case PortableServer::POAManager::INACTIVE:
    /*
     * If we were in the holding state before, requests in the queue
     * must be discarded.
     */
    while (!InvocationQueue.empty()) {
      InvocationRecord_ptr ir = InvocationQueue.front();
      InvocationQueue.erase (InvocationQueue.begin());
      ir->exec (this);
      CORBA::release (ir);
    }

    /*
     * Etherealize active objects
     */

    etherealize ();
    break;

  case PortableServer::POAManager::DISCARDING:
    /*
     * local_invoke() will discard queued invocations
     */
    while (!InvocationQueue.empty()) {
      InvocationRecord_ptr ir = InvocationQueue.front();
      InvocationQueue.erase (InvocationQueue.begin());
      ir->exec (this);
      CORBA::release (ir);
    }
    break;

  default:
    assert (0);
  }
}

/*
 * Object activation and deactivation
 */

PortableServer::ObjectId *
MICOPOA::POA_impl::activate_object (PortableServer::Servant servant)
{
  assert (servant);

  /*
   * Requires SYSTEM_ID and RETAIN
   */

  if (id_assignment_policy->value() != PortableServer::SYSTEM_ID ||
      servant_retention_policy->value() != PortableServer::RETAIN) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * See if Servant is already active
   */

  if (id_uniqueness_policy->value() != PortableServer::MULTIPLE_ID) {
    if (ActiveObjectMap.find_serv (servant) != NULL) {
      mico_throw (PortableServer::POA::ServantAlreadyActive());
    }
  }

  /*
   * Generate Object Reference
   */

  CORBA::String_var uid = idfactory.new_id ();
  PortableServer::ObjectId * id = PortableServer::string_to_ObjectId (uid);
  CORBA::String_var iface = servant->_primary_interface (*id, this);
  POAObjectReference * por =
    new POAObjectReference (orb, oaid.c_str(), *id, iface);
  delete id;

  /*
   * Special handling for AdapterActivators and ServantManagers,
   * which are entered into our Active Object Map. But these
   * "pseudo servants" inherit from CORBA::Object themselves
   * and carry their own object references.
   */

  if (servant->_is_a ("IDL:omg.org/PortableServer/AdapterActivator:1.0") ||
      servant->_is_a ("IDL:omg.org/PortableServer/ServantManager:1.0")) {
    /*
     * This is ugly. But we do anything to follow the specs ...
     */
    CORBA::Object_ptr obj;
    if (servant->_is_a ("IDL:omg.org/PortableServer/AdapterActivator:1.0")) {
      obj = POA_PortableServer::AdapterActivator::_narrow (servant);
    }
    else {
      obj = POA_PortableServer::ServantManager::_narrow (servant);
    }
    obj->operator= (*por->get_ref());
    POAObjectReference * npor =
      new POAObjectReference (orb, obj);
    delete por;
    por = npor;
  }

  /*
   * Insert into active object map
   */

  ObjectRecord * orec = new ObjectRecord (por, servant);
  ActiveObjectMap.add (orec);

  return new PortableServer::ObjectId (*orec->por->get_id());
}

void
MICOPOA::POA_impl::activate_object_with_id (const PortableServer::ObjectId & id,
					    PortableServer::Servant servant)
{
  assert (servant);

  /*
   * Requires RETAIN
   */

  if (servant_retention_policy->value() != PortableServer::RETAIN) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * See if Object or Servant is already active
   */

  ObjectRecord * orec;

  if ((orec = ActiveObjectMap.find (&id)) != NULL) {
    mico_throw (PortableServer::POA::ObjectAlreadyActive());
  }

  if (id_uniqueness_policy->value() != PortableServer::MULTIPLE_ID) {
    if ((orec = ActiveObjectMap.find_serv (servant)) != NULL) {
      mico_throw (PortableServer::POA::ServantAlreadyActive());
    }
  }

  /*
   * Generate Object Reference
   */

  CORBA::String_var iface = servant->_primary_interface (id, this);
  POAObjectReference * por = new POAObjectReference (orb, oaid.c_str(),
						     id, iface);

  /*
   * Special handling for AdapterActivators and ServantManagers,
   * which are entered into our Active Object Map. But these
   * "pseudo servants" inherit from CORBA::Object themselves
   * and carry their own object references.
   */

  if (servant->_is_a ("IDL:omg.org/PortableServer/AdapterActivator:1.0") ||
      servant->_is_a ("IDL:omg.org/PortableServer/ServantManager:1.0")) {
    /*
     * This is ugly. But we do anything to follow the specs ...
     */
    CORBA::Object_ptr obj;
    if (servant->_is_a ("IDL:omg.org/PortableServer/AdapterActivator:1.0")) {
      obj = POA_PortableServer::AdapterActivator::_narrow (servant);
    }
    else {
      obj = POA_PortableServer::ServantManager::_narrow (servant);
    }
    assert (!CORBA::is_nil (obj));
    obj->operator= (*por->get_ref());
    POAObjectReference * npor =
      new POAObjectReference (orb, obj);
    delete por;
    por = npor;
  }

  /*
   * Insert into active object map
   */

  orec = new ObjectRecord (por, servant);
  ActiveObjectMap.add (orec);
}

void
MICOPOA::POA_impl::deactivate_object (const PortableServer::ObjectId & id)
{
  /*
   * Requires RETAIN
   */

  if (servant_retention_policy->value() != PortableServer::RETAIN) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  ObjectRecord * orec = ActiveObjectMap.find (&id);

  if (!orec) {
    mico_throw (PortableServer::POA::ObjectNotActive());
  }

  ActiveObjectMap.del (orec);

  ObjectRecord * other = ActiveObjectMap.find_serv (orec->serv);

  /*
   * Etherealize object
   */

  if (!CORBA::is_nil (servant_manager)) {
    PortableServer::ServantActivator_var sav =
      PortableServer::ServantActivator::_narrow (servant_manager);
    assert (sav);
    sav->etherealize (*orec->por->get_id(), this, orec->serv,
		      FALSE, (other != NULL) ? TRUE : FALSE);
  }

  delete orec;
}

CORBA::Object_ptr
MICOPOA::POA_impl::create_reference (const char * repoid)
{
  assert (repoid);

  /*
   * Requires SYSTEM_ID
   */

  if (id_assignment_policy->value() != PortableServer::SYSTEM_ID) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  CORBA::String_var uid = idfactory.new_id ();
  PortableServer::ObjectId * id = PortableServer::string_to_ObjectId (uid);
  POAObjectReference * por = new POAObjectReference (orb, oaid.c_str(),
						     *id, repoid);
  CORBA::Object_ptr obj = CORBA::Object::_duplicate (por->get_ref());
  assert (!CORBA::is_nil (obj));
  delete id;
  delete por;

  return obj;
}

CORBA::Object_ptr
MICOPOA::POA_impl::create_reference_with_id (const PortableServer::ObjectId & id,
					     const char * repoid)
{
  assert (repoid);

  POAObjectReference * por = new POAObjectReference (orb, oaid.c_str(),
						     id, repoid);
  CORBA::Object_ptr obj = CORBA::Object::_duplicate (por->get_ref());
  assert (!CORBA::is_nil (obj));
  delete por;

  return obj;
}

/*
 * Perform activation upon _this()
 *
 * XXX is this correct?
 */

CORBA::Object_ptr
MICOPOA::POA_impl::activate_for_this (PortableServer::Servant servant)
{
  /*
   * Requires RETAIN and UNIQUE_ID
   */

  if (servant_retention_policy->value() != PortableServer::RETAIN ||
      id_uniqueness_policy->value() != PortableServer::UNIQUE_ID) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * See if servant is already active.
   */

  ObjectRecord * orec = ActiveObjectMap.find_serv (servant);

  if (orec) {
    return CORBA::Object::_duplicate (orec->por->get_ref ());
  }

  /*
   * Require IMPLICIT_ACTIVATION
   */

  if (implicit_activation_policy->value() != PortableServer::IMPLICIT_ACTIVATION) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * Activate the object
   */

  return servant_to_reference (servant);
}

/*
 * Identity mapping operations
 */

PortableServer::ObjectId *
MICOPOA::POA_impl::servant_to_id (PortableServer::Servant servant)
{
  assert (servant);

  /*
   * Requires RETAIN
   */

  if (servant_retention_policy->value() != PortableServer::RETAIN) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * If IMPLICIT_ACTIVATION && MULTIPLE_ID, the Servant is activated
   * without checking the active map
   */

  if (implicit_activation_policy->value() == PortableServer::IMPLICIT_ACTIVATION &&
      id_uniqueness_policy->value() == PortableServer::MULTIPLE_ID) {
    return activate_object (servant);
  }

  /*
   * Search in active map
   */

  ObjectRecord * orec = ActiveObjectMap.find_serv (servant);

  if (orec) {
    return new PortableServer::ObjectId (*orec->por->get_id());
  }

  /*
   * if IMPLICIT_ACTIVATION, the Servant is activated
   */

  if (implicit_activation_policy->value() == PortableServer::IMPLICIT_ACTIVATION) {
    return activate_object (servant);
  }

  mico_throw (PortableServer::POA::ServantNotActive());

  /*
   * Never reached
   */

  assert (0);
  return NULL;
}

CORBA::Object_ptr
MICOPOA::POA_impl::servant_to_reference (PortableServer::Servant servant)
{
  assert (servant);

  /*
   * Requires RETAIN
   */

  if (servant_retention_policy->value() != PortableServer::RETAIN) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * If IMPLICIT_ACTIVATION && MULTIPLE_ID, the Servant is activated
   */

  if (implicit_activation_policy->value() == PortableServer::IMPLICIT_ACTIVATION &&
      id_uniqueness_policy->value() == PortableServer::MULTIPLE_ID) {
    PortableServer::ObjectId * id = activate_object (servant);
    ObjectRecord * orec = ActiveObjectMap.find (id);
    assert (orec);
    delete id;
    return CORBA::Object::_duplicate (orec->por->get_ref());
  }

  /*
   * Search in active object map
   */

  ObjectRecord * orec = ActiveObjectMap.find_serv (servant);

  if (orec) {
    return CORBA::Object::_duplicate (orec->por->get_ref());
  }

  /*
   * if IMPLICIT_ACTIVATION, the Servant is activated
   */

  if (implicit_activation_policy->value() == PortableServer::IMPLICIT_ACTIVATION) {
    PortableServer::ObjectId * id = activate_object (servant);
    orec = ActiveObjectMap.find (id);
    delete id;
    return CORBA::Object::_duplicate (orec->por->get_ref());
  }

  mico_throw (PortableServer::POA::ServantNotActive());

  /*
   * never reached
   */

  assert (0);
  return CORBA::Object::_nil();
}

PortableServer::Servant
MICOPOA::POA_impl::reference_to_servant (CORBA::Object_ptr obj)
{
  assert (!CORBA::is_nil (obj));

  /*
   * Requires RETAIN or USE_DEFAULT_SERVANT
   */

  if (servant_retention_policy->value()  != PortableServer::RETAIN &&
      request_processing_policy->value() != PortableServer::USE_DEFAULT_SERVANT) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * Check if the object reference was created by this POA
   */

  POAObjectReference por (orb, obj);

  if (!por.is_legal() || !por.in_poa(oaid.c_str())) {
    mico_throw (PortableServer::POA::WrongAdapter());
  }

  /*
   * If RETAIN, look in active object map
   */

  if (servant_retention_policy->value() == PortableServer::RETAIN) {
    ObjectRecord * orec = ActiveObjectMap.find_por (&por);

    if (orec) {
      return orec->serv;
    }
  }

  /*
   * If USE_DEFAULT_SERVANT, return default servant
   */

  if (request_processing_policy->value() == PortableServer::USE_DEFAULT_SERVANT && default_servant) {
    return default_servant;
  }

  mico_throw (PortableServer::POA::ObjectNotActive());

  /*
   * never reached
   */

  assert (0);
  return NULL;
}

PortableServer::ObjectId *
MICOPOA::POA_impl::reference_to_id (CORBA::Object_ptr obj)
{
  /*
   * Check if the object reference was created by this POA
   */

  POAObjectReference por (orb, obj);

  if (!por.is_legal() || !por.in_poa(oaid.c_str())) {
    mico_throw (PortableServer::POA::WrongAdapter());
  }

  /*
   * Extract Id from Reference
   */

  return new PortableServer::ObjectId (*por.get_id());
}

PortableServer::Servant
MICOPOA::POA_impl::id_to_servant (const PortableServer::ObjectId &id)
{
  /*
   * Requires RETAIN
   */

  if (servant_retention_policy->value() != PortableServer::RETAIN) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * Search in active object map
   */

  ObjectRecord * orec = ActiveObjectMap.find (&id);

  if (orec == NULL || orec->serv == NULL) {
    mico_throw (PortableServer::POA::ObjectNotActive());
  }

  return orec->serv;
}

CORBA::Object_ptr
MICOPOA::POA_impl::id_to_reference (const PortableServer::ObjectId & id)
{
  /*
   * Requires RETAIN
   */

  if (servant_retention_policy->value() != PortableServer::RETAIN) {
    mico_throw (PortableServer::POA::WrongPolicy());
  }

  /*
   * Search in active object map
   */

  ObjectRecord * orec = ActiveObjectMap.find (&id);

  if (orec == NULL || orec->serv == NULL) {
    mico_throw (PortableServer::POA::ObjectNotActive());
  }

  return CORBA::Object::_duplicate (orec->por->get_ref());
}

/*
 * POA -- ObjectAdapter Interface
 */

const char *
MICOPOA::POA_impl::get_oaid () const
{
  return oaid.c_str();
}

/*
 * has_object() is just to test if this is the right OA for the
 * object. Therefore, we can return TRUE if the reference refers
 * to any object in any POA in the local process.
 */

CORBA::Boolean
MICOPOA::POA_impl::has_object (CORBA::Object_ptr obj)
{
  POAObjectReference por (orb, obj);
  return por.is_legal() && por.is_local();
}

CORBA::Boolean
MICOPOA::POA_impl::is_local () const
{
  return TRUE;
}

/*
 * Global invoke. Only called for the RootPOA, which then delegates
 * the invocation to the appropriate descendant POA (including it-
 * self).
 */

CORBA::Boolean
MICOPOA::POA_impl::invoke (CORBA::ORB::MsgId msgid,
			   CORBA::Object_ptr obj,
			   CORBA::ORBRequest * req,
			   CORBA::Principal_ptr pr,
			   CORBA::Boolean response_exp)
{
  assert (this == _the_root_poa);

  /*
   * Decompose reference and see if it is a reference to an object
   * in this or a descendant POA.
   */

  POAObjectReference por (orb, obj);
  assert (por.is_legal());

  POAMap::iterator it = AllPOAs.find (por.poa_name());
  POA_impl * poa;

  if (it != AllPOAs.end()) {
    poa = (*it).second;
  }
  else {
    /*
     * Doesn't exist. Search the most appropriate POA
     */

    POA_impl *poa=this, *iter;
    CORBA::String_var cname;

    while (42) {
      if (por.in_poa (poa->oaid.c_str())) {
	break;
      }
      cname = por.next_descendant_poa (poa->oaid.c_str());
      if ((iter = poa->_find_POA (cname, FALSE)) == NULL) {
	break;
      }
      poa = iter;
    }
  }

  /*
   * Okay, found responsible POA
   */

  InvocationRecord_var ir = new InvocationRecord (msgid, &por, req, pr);

  poa->local_invoke (ir);

  return TRUE;
}

/*
 * Local invoke; only called for objects in this POA.
 *
 * Also called for requests on non-existent descendant POAs that we
 * must recreate via our adapter activator. But we may only do so in
 * the active state.
 */

void
MICOPOA::POA_impl::local_invoke (InvocationRecord_ptr ir)
{
  /*
   * What does our manager want us to do?
   */

  switch (state) {
  case PortableServer::POAManager::HOLDING:
    /*
     * Queue incoming requests
     */
    {
      InvocationQueue.push_back (InvocationRecord::_duplicate (ir));
      return;
    }
    break;

  case PortableServer::POAManager::DISCARDING:
    /*
     * Discard incoming requests
     */
    {
      CORBA::ServerRequest_ptr svreq = ir->make_req (this);
      CORBA::Exception * ex = new CORBA::TRANSIENT (0, CORBA::COMPLETED_NO);
      svreq->exception (ex);
      return;
    }
    break;

  case PortableServer::POAManager::INACTIVE:
    /*
     * raise OBJ_ADAPTER exception for incoming requests
     */
    {
      CORBA::ServerRequest_ptr svreq = ir->make_req (this);
      CORBA::Exception * ex = new CORBA::OBJ_ADAPTER (0, CORBA::COMPLETED_NO);
      svreq->exception (ex);
      return;
    }
    break;

  case PortableServer::POAManager::ACTIVE:
    // exec code below
    break;

  default:
    assert (0);
  }

  /*
   * We are in the active state. If the request is not for us but
   * for a descendant POA, try to call the adapter activator
   */

  if (!ir->get_por()->in_poa (oaid.c_str())) {
    CORBA::String_var cname = ir->get_por()->next_descendant_poa (oaid.c_str());
    POA_impl * poa = _find_POA (cname, TRUE);

    /*
     * Okay, next POA in chain successfully activated
     */

    poa->local_invoke (ir);
    return;
  }

  /*
   * Request for an object in this POA -- Fire away!
   */

  perform_invoke (ir);
}

/*
 * Builtin operations
 */

bool
MICOPOA::POA_impl::builtin_invoke (InvocationRecord_ptr ir)
{
  const char * opname = ir->get_or()->op_name();

  if (strcmp (opname, "_is_a") != 0) {
    return false;
  }

  CORBA::ServerRequest_ptr svr = ir->make_req (this);

  if (strcmp (opname, "_is_a") == 0) {
    CORBA::NVList_ptr args = new CORBA::NVList (1);
    CORBA::String_var repoid;
    CORBA::Boolean isa = FALSE, r;

    args->add (CORBA::ARG_IN);
    args->item(0)->value()->type (CORBA::_tc_string);
    r = svr->params (args);
    assert (r);
    r = (*args->item(0)->value() >>= repoid);
    assert (r);

    /*
     * Search in active map first
     */

    ObjectRecord * orec = ActiveObjectMap.find_por (ir->get_por());

    if (orec) {
      /*
       * Try _is_a; if that doesn't work (maybe because it's a DSI
       * interface), check the primary interface
       */

      if (!(isa = orec->serv->_is_a (repoid))) {
	CORBA::String_var pif =
	  orec->serv->_primary_interface (*orec->por->get_id(), this);
	if (strcmp (repoid, pif) == 0) {
	  isa = TRUE;
	}
      }
    }
    else {
      /*
       * If we have a default servant or a servant manager, we're
       * optimistic that we will eventually be able to serve the
       * repoid
       */
      if (request_processing_policy->value() == PortableServer::USE_DEFAULT_SERVANT ||
	  request_processing_policy->value() == PortableServer::USE_SERVANT_MANAGER) {
	isa = TRUE;
      }
    }

    CORBA::Any * res = new CORBA::Any;
    r = (*res <<= CORBA::Any::from_boolean (isa));
    assert (r);
    svr->result (res);
  }

  return true;
}

/*
 * Actually perform the invocation
 */

void
MICOPOA::POA_impl::perform_invoke (InvocationRecord_ptr ir)
{
  /*
   * Handle Builtin operations
   */

  if (builtin_invoke (ir)) {
    return;
  }

  /*
   * InvocationRecord maintains ownership of ServerRequest
   */

  CORBA::ServerRequest_ptr svr = ir->make_req (this);

  /*
   * If RETAIN, search the active map first
   */

  if (servant_retention_policy->value() == PortableServer::RETAIN) {
    ObjectRecord * orec = ActiveObjectMap.find_por (ir->get_por());

    if (orec) {
      assert (orec->serv);
      current->set (this, orec->por, orec->serv);
      orec->serv->invoke (svr);
      current->unset ();
      return;
    }
  }

  /*
   * If USE_DEFAULT_SERVANT, invoke it
   */

  if (request_processing_policy->value() == PortableServer::USE_DEFAULT_SERVANT) {
    if (!default_servant) {
      CORBA::Exception * ex =
	new CORBA::OBJ_ADAPTER (0, CORBA::COMPLETED_NO);
      svr->exception (ex);
    }
    else {
      current->set (this, ir->get_por(), default_servant);
      default_servant->invoke (svr);
      current->unset ();
    }

    return;
  }

  /*
   * If USE_SERVANT_MANAGER, ask it to incarnate a servant
   */

  if (request_processing_policy->value() == PortableServer::USE_SERVANT_MANAGER) {
    /*
     * The type of Servant Manager depends on the retention policy.
     * RETAIN:     ServantActivator
     * NON_RETAIN: ServantLocator
     */

    if (servant_retention_policy->value() == PortableServer::RETAIN) {
      PortableServer::ServantActivator_var sav =
	PortableServer::ServantActivator::_narrow (servant_manager);
      
      if (CORBA::is_nil (sav)) {
	CORBA::Exception * ex =
	  new CORBA::OBJ_ADAPTER (0, CORBA::COMPLETED_NO);
	svr->exception (ex);
	return;
      }

      POAObjectReference * npor = new POAObjectReference (*ir->get_por());
      PortableServer::Servant servant;
#ifdef HAVE_EXCEPTIONS
      try {
#endif
	servant = sav->incarnate (*npor->get_id(), this);
#ifdef HAVE_EXCEPTIONS
      } catch (PortableServer::ForwardRequest_var & fwr) {
	svr->cancel ();
	orb->answer_invoke (ir->id(), CORBA::InvokeOk,
			    fwr->forward_reference,
			    ir->get_or());
	return;
      }
#endif

      if (servant == NULL) {
	CORBA::Exception * ex =
	  new CORBA::OBJECT_NOT_EXIST (0, CORBA::COMPLETED_NO);
	svr->exception (ex);
	return;
      }

      /*
       * The new servant is placed in the active map
       */

      ObjectRecord * orec = new ObjectRecord (npor, servant);
      ActiveObjectMap.add (orec);

      /*
       * Now perform the invocation
       */

      current->set (this, orec->por, servant);
      servant->invoke (svr);
      current->unset ();
      return;
    }
    else {
      PortableServer::ServantLocator_var slo =
	PortableServer::ServantLocator::_narrow (servant_manager);

      if (CORBA::is_nil (slo)) {
	CORBA::Exception * ex =
	  new CORBA::OBJ_ADAPTER (0, CORBA::COMPLETED_NO);
	svr->exception (ex);
	return;
      }

      PortableServer::ServantLocator::Cookie cookie;
      PortableServer::Servant servant;
#ifdef HAVE_EXCEPTIONS
      try {
#endif
	servant = slo->preinvoke (*ir->get_por()->get_id(), this,
				  ir->get_or()->op_name(), cookie);
#ifdef HAVE_EXCEPTIONS
      } catch (PortableServer::ForwardRequest_var & fwr) {
	svr->cancel ();
	orb->answer_invoke (ir->id(), CORBA::InvokeOk,
			    fwr->forward_reference,
			    ir->get_or());
	return;
      }
#endif

      if (servant == NULL) {
	CORBA::Exception * ex =
	  new CORBA::OBJECT_NOT_EXIST (0, CORBA::COMPLETED_NO);
	svr->exception (ex);
	return;
      }

      current->set (this, ir->get_por(), servant);
      servant->invoke (svr);
      current->unset ();
      slo->postinvoke (*ir->get_por()->get_id(), this,
		       ir->get_or()->op_name(), cookie,
		       servant);
      return;
    }
  }

  /*
   * Cannot serve the request
   */

  CORBA::Exception * ex =
    new CORBA::OBJECT_NOT_EXIST (0, CORBA::COMPLETED_NO);
  svr->exception (ex);
}

CORBA::Boolean
MICOPOA::POA_impl::bind (CORBA::ORB::MsgId msgid, const char *repoid,
			 const CORBA::ORB::ObjectTag &oid,
			 CORBA::Address *addr)
{
  if (addr && !addr->is_local()) {
    return FALSE;
  }

  /*
   * Check the active map
   */

  if (servant_retention_policy->value() == PortableServer::RETAIN) {
    ObjectRecord * orec;
    for (CORBA::ULong i=0; orec=ActiveObjectMap[i]; i++) {
      /*
       * Try _is_a; if that doesn't work (maybe because it's a DSI
       * interface), check the primary interface
       */

      if (!orec->serv->_is_a (repoid)) {
	CORBA::String_var pif =
	  orec->serv->_primary_interface (*orec->por->get_id(), this);
	if (strcmp (repoid, pif) != 0) {
	  continue;
	}
      }

      /*
       * If the client sent an OID, compare them
       */

      if (oid.length() == 0 || oid == *orec->por->get_id()) {
	orb->answer_bind (msgid, CORBA::LocateHere, orec->por->get_ref());
	return TRUE;
      }
    }
  }

  /*
   * If we have a default servant or a servant manager, generate
   * appropriate object reference
   */

  if (request_processing_policy->value() == PortableServer::USE_DEFAULT_SERVANT ||
      request_processing_policy->value() == PortableServer::USE_SERVANT_MANAGER) {
    CORBA::Object_ptr obj = create_reference_with_id (oid, repoid);
    orb->answer_bind (msgid, CORBA::LocateHere, obj);
    CORBA::release (obj);
    return TRUE;
  }

  /*
   * Ask child POA's
   */

  POAMap::iterator it = children.begin ();

  while (it != children.end()) {
    if ((*it).second->bind (msgid, repoid, oid, addr)) {
      return TRUE;
    }
    it++;
  }

  return FALSE;
}

CORBA::Boolean
MICOPOA::POA_impl::locate (CORBA::ORB::MsgId msgid, CORBA::Object_ptr obj)
{
  orb->answer_locate (msgid, CORBA::LocateHere, CORBA::Object::_nil());
  return TRUE;
}

/*
 * No direct access to the skeleton is allowed in POA-controlled objects.
 * (This means that invocations cannot bypass the POA.)
 */

CORBA::Object_ptr
MICOPOA::POA_impl::skeleton (CORBA::Object_ptr obj)
{
  return CORBA::Object::_nil ();
}

void
MICOPOA::POA_impl::cancel (CORBA::ORB::MsgId msgid)
{
  vector<InvocationRecord *>::iterator it = InvocationQueue.begin ();

  while (it != InvocationQueue.end()) {
    if ((*it)->id() == msgid) {
      delete *it;
      InvocationQueue.erase (it);
      return;
    }
    it++;
  }

  /*
   * If we got here, it's probably meant for one of our children.
   * Unfortunately, we don't know which one.
   */

  POAMap::iterator child = children.begin ();

  while (child != children.end()) {
    (*child).second->cancel (msgid);
    child++;
  }
}

void
MICOPOA::POA_impl::shutdown (CORBA::Boolean wait_for_completion)
{
  destroy (FALSE, wait_for_completion);
  orb->answer_shutdown (this);
}

void
MICOPOA::POA_impl::answer_invoke (CORBA::ORB::MsgId msgid,
				  CORBA::Object_ptr obj,
				  CORBA::ORBRequest * oreq,
				  CORBA::InvokeStatus stat)
{
  orb->answer_invoke (msgid, stat, CORBA::Object::_nil(), oreq);
}
