/*
   GEF.c: The General Exception-Handling Facility Interface.
   Copyright (C) 1998 Bruce W. Bigby, bbigby@alum.MIT.edu

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   MODULE
      General Exception Facility (GEF)

   DESCRIPTION
      The General Exception Facility is a C library that enables a programmer
      to generate program exceptions and to receive and handle operating
      system exceptions, or signals in the case of UNIX, in a synchronous and
      safe manner. This implementation supports primarily UNIX-like operating
      systems, although the code, without the signal-handling support, is
      portable to any platform which supports the C programming language.

   CAVEATS
      (1) This implementation of GEF uses PosixThreads so you must link using
      your platform's pthread library. Typcially, this means adding a
      <-lpthread> option to your link line.

      (2) PosixThreads has not declared the function, <pthread_getspecific>,
      to be async safe. However, this implementation of GEF requires that
      <pthread_gefspecific> be async-safe.  This is because GEF uses the
      function to determine the exception handler to which to transfer control
      in a multithreaded program. Consequently, to use the asynchronous signal
      support, you must enusre that, for your platform, or operating system,
      the implementation of pthread_getspecific is async-safe.  Fortunately,
      for several major, or popular, vendors of operating sytems that support
      pthreads, the function, pthread_getspecific, is async-safe. One
      qualifying operating system is Linux 2.x. Solaris 2.5.1 also appears to
      have an async-safe pthread_getspecific, although the Solaris
      documentation does not declare the function to be async-safe. You need
      to verify whether pthread_getspecific is async-safe for your operating
      system to be comfortable using it. However, if your program does not use
      GEF to handle asynchronous signals, then you have no cause for concern.

      (3) Programmers must always remember to never use the return statement
      from within the scope of an exception block. Doing so would cause the
      GEF checkpoint stack to get out of sync with the regular program stack.
      Consequently, always get in the habit of returning at the end of a
      function.  This constraint encourages good programming practice anyway.

      (4) Likewise, never use the C language "break" command, if doing so
      causes a thread to exit from within the scope of an active exception
      block.

   AUTHOR
      Bruce W. Bigby
      bbigby@alum.MIT.edu

   REFERENCES
      (1) Advanced Programming in the UNIX Environment, by Stevens.
      (2) A Guide to Multithreaded Programming: Threads Primer, by Bil Lewis
          and Daniel J. Berg.
      (3) OSF DCE: Guide to Developing Distributed Applications, by Harold W.
          Lockhart, Jr.
      (4) C++ Reference
      (5) Core Java by Sunsoft
      (6) See references on Eiffel exception-handling.

   COPYRIGHT (C) 1998 Bruce W. Bigby
*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <GEF.h>

/*
  This data structure stores all of the thread synchronization
   objects--the key that we use to access the thread-specific
   exception-handler; the mutex that we must use for signal handler
   initialization; and the list of signal handlers for each signal
   that we override so that we may restore the appropriate signal
   handlers when the last thread exits.
*/
typedef struct {
   pthread_key_t exceptionHandlerKey;
   int signalConfigurationDone;
   sigset_t signalSet;
   pthread_mutex_t signalInitLock;
   unsigned int numberOfThreads;
   void (*signalHandlers[NSIG])(int signo);
   void (*RaiseOSSignal)(int signum);
} GEF_t;

static GEF_t GEF;

static
void
Error(char* format, ...) {
   va_list args;
   va_start(args, format);
   fprintf(stderr, "\n[General Exception Facility]\n");
   fprintf(stderr, "   ");
   vfprintf(stderr, format, args);
   fprintf(stderr, "\n");
   va_end(args);
}

static
void
Abort(int exceptionID) {
#ifndef LINUX
   kill(getpid(), SIGABRT);
   pthread_exit((void*) exceptionID);
#else
   pthread_exit((void*) exceptionID);
#endif
}

static
void
Exit(int code) {
   pthread_exit((void*) code);
}

static
void
OutOfMemory(void) {
   gef_throw((void*) GEF_EXCEPTION_NO_MEMORY);
}

static
void
Delete(void** exceptionID) {
   *exceptionID = NULL;
}

static
void
AssertionViolation(char* fileName, int lineNumber) {
   gef_throw((void*) GEF_EXCEPTION_ASSERTION_VIOLATION);
}

static
void
UnhandledException(void* exceptionID) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   eh->UnhandledException(exceptionID);
}

static
void
signotset(sigset_t* oldset, sigset_t* newset) {
   int sig;

   for (sig = 1; sig < NSIG; sig++) {
      if (!sigismember(oldset, sig)) {
	 sigaddset(newset, sig);
      }
   }
}

void
GEFSetDebug(int value) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   GEFExceptionHandler_SetDebug(eh, value);
}

void
GEFSetAssertions(int value) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   GEFExceptionHandler_SetAssertions(eh, value);
}

void
GEFSetAsyncSupport(int value) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   GEFExceptionHandler_SetAsyncSupport(eh, value);
}

int
GEFGetAsyncSupport(void) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   return(GEFExceptionHandler_GetAsyncSupport(eh));
}

int
GEFGetAssertions(void) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   return(GEFExceptionHandler_GetAssertions(eh));
}

void
GEFExecuteUserCatch(void* userData, void* exceptionID) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   eh->UserCatch(userData, exceptionID);
}

void
GEFBeginCriticalSection(GEFExceptionHandler_t eh) {
   GEFExceptionHandler_BeginCriticalSection(eh);
}

void
GEFEndCriticalSection(GEFExceptionHandler_t eh) {
   GEFExceptionHandler_EndCriticalSection(eh);
}

void
GEFAssert(int expressionValue, char* fileName, int lineNumber) {
   if (!expressionValue) {
      GEFExceptionHandler_t eh = NULL;
      GEFStateID_t current;
      eh = pthread_getspecific(GEF.exceptionHandlerKey);
      if (eh != NULL && eh->assertions) {
         eh->AssertionViolation(fileName, lineNumber);
      }
   }
}

void
GEFRaiseException(void* id) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);

   if (eh != NULL) {
     GEFExceptionHandler_RaiseException(eh, id);
   }

   /* Most likely received a signal before we could initialize the handler */
   /* Just ignore signal, and return. */
}

static
void
GEFRaiseOSSignal(int signo) {
   GEFExceptionHandler_t eh = NULL;
   void* exceptionID = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   exceptionID = eh->OSSignalToException(signo);
   GEFRaiseException(exceptionID);
}

void
GEFCatchSignal(int signum) {
   GEF.RaiseOSSignal(signum);
}

/* static */
GEFSignalHandler_t
GEFSignal(int signum, GEFSignalHandler_t handler) {
  struct sigaction act, oact;

  act.sa_handler = handler;
  sigfillset(&act.sa_mask);

  act.sa_flags = 0;
  if (signum == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT; /* SunOS */
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;   /* SVR4, 4.3+BSD */
#endif
  }
  if (sigaction(signum, &act, &oact) < 0) {
    return(SIG_ERR);
  }
  return(oact.sa_handler);
}

static
GEFSignalHandler_t
GEFAcceptSignal(int signum) {
  GEFSignalHandler_t answer = SIG_DFL;
  GEFExceptionHandler_t eh;

  eh = pthread_getspecific(GEF.exceptionHandlerKey);
  answer = GEFSignal(signum, GEFCatchSignal);
  if (signum == SIGABRT) {
    GEFSignal(signum, answer);
  }

  return(answer);
}

static
void
GEFInitializeSignalHandler(void) {
   int signum;
   int numberOfSignals = NSIG;

   /* Should check return code of lock */
   pthread_mutex_lock(&GEF.signalInitLock);
   if (!GEF.signalConfigurationDone) {
      for (signum = 1; signum < numberOfSignals; signum++) {
         if (sigismember(&GEF.signalSet, signum)) {
            GEF.signalHandlers[signum] = GEFAcceptSignal(signum);
         } /* else do nothing for unsupported signals */
      }
      GEF.signalConfigurationDone = GEFTRUE;
   }
   GEF.numberOfThreads++;
   /* Should check return code of unlock */
   pthread_mutex_unlock(&GEF.signalInitLock);
}

static
void
GEFRestoreSignalHandlers(void) {
   int signum;
   int numberOfSignals = NSIG;

   pthread_mutex_lock(&GEF.signalInitLock);
   if (--GEF.numberOfThreads == 0) {
      for (signum = 1; signum < numberOfSignals; signum++) {
         if (sigismember(&GEF.signalSet, signum)) {
            (void) GEFSignal(signum, GEF.signalHandlers[signum]);
         }
      }
   }
   pthread_mutex_unlock(&GEF.signalInitLock);
}

static
void
GEFDeleteExceptionHandler(void* eh) {
   GEFExceptionHandler_t rep = eh;

   GEFRestoreSignalHandlers();
   GEFExceptionHandler_Delete(&rep);
   pthread_setspecific(GEF.exceptionHandlerKey, (void*) NULL);
}

void
GEFInitializeThread(GEFAttr_t attrs) {
   int current;
   int previous;
   sigset_t allSignals;
   sigset_t previousSignalMask;
   sigset_t threadMask;
   GEFExceptionHandler_t eh = NULL;

   if (attrs.Abort == NULL) attrs.Abort = Abort;
   if (attrs.AssertionViolation == NULL) {
      attrs.AssertionViolation = AssertionViolation;
   }
   if (attrs.UnhandledException == NULL) {
      attrs.UnhandledException = UnhandledException;
   }
   if (attrs.Exit == NULL) attrs.Exit = Exit;
   if (attrs.Error == NULL) attrs.Error = Error;
   if (attrs.OutOfMemory == NULL) attrs.OutOfMemory = OutOfMemory;
   if (attrs.Delete == NULL) attrs.Delete = Delete;

   /* Set signal mask for the thread */
   signotset(&GEF.signalSet, &threadMask);
   pthread_sigmask(SIG_SETMASK, &threadMask, NULL);

   sigfillset(&allSignals);
   if (pthread_sigmask(SIG_BLOCK, &allSignals, &previousSignalMask) < 0) {
      Error("GEFInitialize--pthread_sigmask(SIG_BLOCK, ...) failed!\n");
      Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
   }
   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   if (eh == (GEFExceptionHandler_t) NULL) {
      eh = GEFExceptionHandler_New(attrs, GEF.signalSet);
      if (pthread_setspecific(GEF.exceptionHandlerKey, (void*) eh) != 0) {
         eh->Error("GEFInitializeThread--pthread_setspecific failed!\n");
	 eh->Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
      }
   }
   GEFInitializeSignalHandler();
   if (pthread_sigmask(SIG_SETMASK, &previousSignalMask, NULL) < 0) {
      eh->Error("GEFInitialize--pthread_sigmask(SIG_SETMASK, ...) failed!\n");
      eh->Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
   }
}

void
GEFInitialize(sigset_t signalSet) {
   int current;
   int previous;
   sigset_t allSignals;
   sigset_t previousSignalMask;

   if (sigismember(&signalSet, SIGABRT)) {
     /* Quietly remove SIGABRT, if it is a member of signalSet. */
     /* GEF reserves SIGABRT for its default behavior. */
     sigdelset(&signalSet, SIGABRT);
   }
   sigfillset(&allSignals);
   if (pthread_sigmask(SIG_BLOCK, &allSignals, &previousSignalMask) < 0) {
      fprintf(stderr, "GEFInitialize--pthread_sigmask(SIG_BLOCK, ...) failed!\n");
      pthread_kill(pthread_self(), SIGABRT);
   }
   pthread_key_create(&GEF.exceptionHandlerKey, GEFDeleteExceptionHandler);
   GEF.signalConfigurationDone = GEFFALSE;
   pthread_mutex_init(&GEF.signalInitLock, NULL);
   GEF.signalSet = signalSet;
   GEF.RaiseOSSignal = GEFRaiseOSSignal;
   GEF.numberOfThreads = 0;
   if (pthread_sigmask(SIG_BLOCK, &allSignals, &previousSignalMask) < 0) {
      fprintf(stderr, "GEFInitialize--pthread_sigmask(SIG_BLOCK, ...) failed!\n");
      pthread_kill(pthread_self(), SIGABRT);
   }
}

void
GEFAddSignal(int signum) {
   sigaddset(&GEF.signalSet, signum);
}

void
GEFDeleteSignal(int signum) {
   sigdelset(&GEF.signalSet, signum);
}

void
GEFTerminate(void) {
   pthread_mutex_destroy(&GEF.signalInitLock);
   pthread_key_delete(GEF.exceptionHandlerKey);
}

GEFExceptionHandler_t
GEFGetExceptionHandler(void) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   return(eh);
}

int
GEFStackDepth(void) {
   GEFExceptionHandler_t eh = NULL;

   eh = pthread_getspecific(GEF.exceptionHandlerKey);
   return(GEFExceptionHandler_StackDepth(eh));
}

