/*
  GEF: A General Exception-Handling Facility, which also suppor
  "Programming by Contract."
  Copyright (C) 1998 Bruce W. Bigby
  E-Mail: bbigby@alum.MIT.edu
 */
#ifndef GEF_H
#define GEF_H

#include <GEFAttr.h>
#include <GEFExceptHandler.h>

typedef void (*GEFSignalHandler_t)(int);

extern
GEFExceptionHandler_t
GEFGetExceptionHandler(void);
/*
  Effects: Returns the exception handler for the currently executing thread.
 */

extern
GEFSignalHandler_t
GEFSignal(int signum, GEFSignalHandler_t h);
/*
  Effects: Registers the signal handler, <h>, for signal, <signum>.
 */

extern
void
GEFBeginCriticalSection(GEFExceptionHandler_t eh);
/*
  Effects: Instructs the OS to postpone its delivery of signals that <eh> is
  responsible for handling.
 */

extern
void
GEFEndCriticalSection(GEFExceptionHandler_t eh);
/*
  Effects: Instructs the OS to deliver any pending signals that <eh> is responsible
  for handling.
 */

extern
void
GEFInitialize(sigset_t signalSet);
/*
  Effects: Initialize the General Exception Facility, where <signalSet> is the set of
  OS signals that GEF will handle for the process.
 */

extern void
GEFTerminate(void);
/*
  Requires:The caller must have previously called GEFInitialize, before calling
  this operation.
  Effects: Restores the process wide signal handlers to their
  previous values.
 */

extern
void
GEFInitializeThread(GEFAttr_t attrs);
/*
  Effects: Creates and initializes the exception handler, accodring to the
  configuration, <attrs>,  for the currently running thread, and stores the
  exception handler in the thread's private data area. Each thread must call
  this function prior to using the GEF facilities, even in a single-threaded
  program.
 */

extern
void
GEFAssert(int expressionValue, char* fileName, int lineNumber);
/*
  Effects: If 'expressionValue' is non-zero (true), throws the reserved exception, 
  GEF_EXCEPTION_ASSERTION;  otherwise returns.
 */

extern
void
GEFRaiseException(void*);
/*
  Effects: Transfers control to the most recent checkpoint of an exception handler.
 */

extern
int
GEFStackDepth(void);
/*
  Effects: Returns the depth of the checkpoint stack of the currently executing
  thread.
 */

extern
void
GEFExecuteUserCatch(void* exception, void* data);
/*
  Effects: If the exception handler of the currently executing thread does not
  specify a user-defined call-back function, then the function does nothing;
  otherwise, it calls the user-defined call-back, specifying <data> as the first
  argument and <exception> as the second argument.
 */

extern
void
GEFCatchSignal(int signo);
/*
  Effects: Catches all OS signals and converts them into GEF exceptions.
 */

extern
void
GEFSetDebug(int value);
/*
  Effects: Turns on debugging for the GEF facility. It prints a trace of its actions to
  the standard error stream.
 */

extern
void
GEFSetAsyncSupport(int value);
/*
  Effects: If <value> is non-zero (true), turns on support for asynchronous
  signals; otherwise turns off support for asynchronous signals.
  Caveat: In this implementation, GEF requires that pthread_getspecific is async-
  safe. However, Posix does not require the function to be async safe. 
  Consequently, use async support with caution.  Fortunately, the function,
  <pthread_getspecific>, is safe for several major UNIX operating systems.
  Also note that, in some operating systems, support OS signals can introduce a
  significant amount of overhead, causing the relative performance of handling
  exceptions to decrease by a significant factor.  Results will vary between
  operating systems.
 */

extern
int
GEFGetAsyncSupport(void);
/*
  Effects: Returns non-zero (true), if asynchronous OS signal support is on;
  otherwise returns 0 (false).
 */

#define gef_retry \
GEFExceptionHandler_Retry(gefeh)
/*
  Effects:A keyword macro, <gef_retry>, which transfers the control of the
  currently executing thread to the most recent checkpoint of <gefeh>.
 */

#define gef_break \
GEFExceptionHandler_Break(gefeh)
/*
  Effects: Terminates the exception-handler so that control passes to the first
  statement which appears outside of the scope of the currently executing
  exception block.
 */

#define gef_throw(e) \
GEFRaiseException(e)
/*
  Effects: Transfers control to the most recent checkpoint of the currently
  executing thread and passes <e> as the exception.
 */

#define gef_begin_critical_section \
GEFBeginCriticalSection(gefeh)
/*
  Effects: A keyword macro which prevents the OS from delivering the signals
  that <gefeh> handles.
 */

#define gef_end_critical_section \
GEFEndCriticalSection(gefeh)
/*
  Effects: A keyword macro which allows the OS to deliver the pending signals
  that <gefeh> handles.
 */

#define gef_depth \
GEFStackDepth()
/*
  Effects: Returns the depth of the checkpoint stack in the exception handler of
  the currently executing thread.
 */

#define gef_enable_async_support \
GEFSetAsyncSupport(1)
/*
  Effects: A keyword macro for enabling support for asynchronous OS signals.
 */

#define gef_disable_async_support \
GEFSetAsyncSupport(0)
/*
  Effects: A keyword macro for disabling support for asynchronous OS signals.
 */

#define gef_async_support_is_enabled \
GEFGetAsyncSupport()
/*
  Effects: A keyword macro predicate for determining whether async support is
  enabled.
 */

#define gef_enable_assertions \
 GEFSetAssertions(1)
/*
   Effects: A keyword macro predicate for enabling support for assertion
   statements.
*/

#define gef_disable_assertions \
GEFSetAssertions(0)
/*
   Effects: A keyword macro for disabling support for assertion statements.
*/

#define get_assertions_support_is_enabled \
GEFGetAssertions()
/*
   Effects: A keyword macro predicate for determining whether assertion support
   is enabled.
*/

#define gef_enable_debug \
GEFSetDebug(1)
/*
  Effects: A keyword macro for enabling debugging.
 */

#define gef_disable_debug \
GEFSetDebug(0)
/*
  Effects: A keyword macro for disabling debugging.
 */

#define gef_assert(exp) \
GEFAssert(exp, __FILE__, __LINE__)

#define gef_try \
{  GEFStateID_t gefstate;\
   GEFExceptionHandler_t gefeh = GEFGetExceptionHandler();\
   GEFCheckPoint_t* cp = GEFExceptionHandler_CheckPoint(gefeh, 0);\
   gefstate = setjmp(cp->env);\
   if (!gefeh->assertions && gefstate == 0)  gefstate = GEF_EXCEPT_TRY;\
   do {\
      switch(gefstate) {\
         case GEF_EXCEPT_TRY:\
         case GEF_EXCEPT_RETRY:\
            GEFExceptionHandler_EndCriticalSection(gefeh);\
            {
/*
  Effects: Introduces preamble code that creates a checkpoint and prepares to
  handle an exception.
*/

#define gef_repeat_try(n) \
{  GEFStateID_t gefstate;\
   GEFExceptionHandler_t gefeh = GEFGetExceptionHandler();\
   GEFCheckPoint_t* gefcp = GEFExceptionHandler_CheckPoint(gefeh, n);\
   gefstate = setjmp(gefcp->env);\
   if (!gefeh->assertions && gefstate == 0) gefstate = GEF_EXCEPT_TRY;\
   do {\
      switch(gefstate) {\
         case GEF_EXCEPT_TRY:\
         case GEF_EXCEPT_RETRY:\
            GEFExceptionHandler_EndCriticalSection(gefeh);\
            {
/*
  Effects: Introduces preamble code that creates a checkpoint and prepares to
  handle an exception. However, upon failure of a try block, this code will repeat
  the block automatically until it is successful, or until the number of attempts
  reaches <n>. After <n> attempts, the code transfers control to the catch block,
  if any. If no catch block exists for the current exception block, then GEF
  transfers control to the checkpoint which is on the top of the exception handler
  stack of the currently executing thread.
*/

#define gef_invariants \
               gefstate = GEFExceptionHandler_NextState(gefeh);\
            }\
	    break;\
	 case GEF_EXCEPT_INVARIANTS:\
	    {\
/*
   Effects: Introduces body code into an exception block which handles invariant
   statements. GEF executes all of the statements in this block before entering the
   preconditions block and after exiting the postconditions block.
*/

#define gef_preconditions \
               gefstate = GEFExceptionHandler_NextState(gefeh);\
            }\
	    break;\
	 case GEF_EXCEPT_PRECONDS:\
	    {\
/*
   Effects: Introduces body code into an exception block which handles
   precondition statements.
*/

#define gef_postconditions \
               gefstate = GEFExceptionHandler_NextState(gefeh);\
            }\
	    break;\
	 case GEF_EXCEPT_POSTCONDS:\
	    {\
/*
   Effects: Introduces body code into an exception block which handles
   postconditions statements.
*/

#define gef_finally \
               gefstate = GEFExceptionHandler_NextState(gefeh);\
            }\
            break;\
         case GEF_EXCEPT_FINALLY:\
            {\
/*
  Effects: Introduces body code into an exception block which handles the final
  actions that an exception-handler must perform.
 */

#define gef_catch(e) \
               gefstate = GEFExceptionHandler_NextState(gefeh);\
            }\
            break;\
         case GEF_EXCEPT_CATCH:\
            {\
               void* e = GEFExceptionHandler_GetException(gefeh);
/*
  Effects: Introduces body code into an exception block which catches all
  exceptions.
*/

#define gef_catch_with(e, userData)\
               gefstate = GEFExceptionHandler_NextState(gefeh);\
            }\
            break;\
         case GEF_EXCEPT_CATCH: \
            {\
               void* e = GEFExceptionHandler_GetException(gefeh);\
               gefeh->UserCatch(userData, e);
/*
  Effects: Introduces body code into an exception block which catches all
  exceptions, but transfers control to a user-defined call-back.
*/

#define gef_end \
               gefstate = GEFExceptionHandler_NextState(gefeh);\
            }\
            break;\
         case GEF_EXCEPT_SUCCESS:\
            gefstate = GEFExceptionHandler_NextState(gefeh);\
            break;\
         default:\
            gefstate = GEFExceptionHandler_NextState(gefeh);\
            break;\
      }\
   } while (gefstate != GEF_EXCEPT_DONE);\
   GEFExceptionHandler_EndCriticalSection(gefeh);\
}
/*
  Effects: Introduces code which concludes an exception block.
 */

#define GEFTRUE  1
#define GEFFALSE 0

#endif /* GEF - General Exception Facility */
