/*
  GEFExceptHandler.c:  Implements the exception context data type.
  Copyright (C) 1998 Bruce W. Bigby, bbigby@alum.MIT.edu

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <GEFExceptHandler.h>

static
void
Error(char* format, ...) {
/*
   Description: This routine is the default error reporting routine.  It accepts a
     printf-compatible format string.
*/
   /* Is this routine thread-safe? */
   va_list args;
   va_start(args, format);
   fprintf(stderr, "\n[%s] ", __FILE__);
   vfprintf(stderr, format, args);
   va_end(args);
}

static
void
Exit(int code) {
   exit(code);
}

static
void
Abort(int exceptionID) {
/*
  Description: This routine is the default process or thread termination routine. 
  GEF uses this routine to terminate an errant program or thread.
*/
   kill(getpid(), SIGABRT);
}

static
void
AssertionViolation(char* fileName, int lineNumber) {
  exit(GEF_EXCEPTION_ASSERTION_VIOLATION);
}

static
void
UnhandledException(void* exceptionID) {
   fprintf(stderr, "Unhandled exception occurred, while executing in process, %d.\n\007", (int) getpid());
   exit(GEF_EXCEPTION_UNHANDLED);
}

static
void
OutOfMemory(void) {
   fprintf(stderr, "GEF: Insufficient memory condition occurred, while in process %d!\n", (int) getpid());
   exit(GEF_EXCEPTION_NO_MEMORY);
}

static
void*
OSSignalToException(int signum) {
/*
  Description: Converts a user-defined signal into an operating system signal, if
  possible.  By default, the routine returns the same signal number that it receives.
*/
  return((void*) signum);
}

static
void
UserCatch(void* userData, void* exceptionID) {
}

static
void
GEFExceptionHandler_UnhandledException(GEFExceptionHandler_t eh, void* id) {
/*
   Description: Raises the unhandled-exception exception.
*/
   eh->UnhandledException(id);
#ifdef LINUX
   eh->Exit((int) id);
#else
   eh->Abort((int) id);
#endif
}

static
void
NoActiveCheckPoints(GEFExceptionHandler_t eh) {
   eh->Error("While executing in process, %d, no active check points available.\n\007", eh->ProcessID());
   eh->Abort(GEF_EXCEPTION_FATAL);
}

static
void
CheckUnhandledException(GEFExceptionHandler_t eh, void* id) {
   if (eh == NULL) GEFExceptionHandler_UnhandledException(eh, id);
   if (eh->sp == 0) {
    GEFExceptionHandler_EndCriticalSection(eh);
    GEFExceptionHandler_UnhandledException(eh, id);
  }
}

void
GEFExceptionHandler_SetDebug(GEFExceptionHandler_t eh, int value) {
   eh->debug = (value ? 1 : 0);
}

void
GEFExceptionHandler_SetAssertions(GEFExceptionHandler_t eh, int value) {
   eh->assertions = (value ? 1 : 0);
}

int
GEFExceptionHandler_GetAssertions(GEFExceptionHandler_t eh) {
   return(eh->assertions);
}

void
GEFExceptionHandler_SetAsyncSupport(GEFExceptionHandler_t eh, int value) {
   eh->supportAsyncSignals = (value ? 1 : 0);
}

int
GEFExceptionHandler_GetAsyncSupport(GEFExceptionHandler_t eh) {
   return(eh->supportAsyncSignals);
}

void
GEFExceptionHandler_BeginCriticalSection(GEFExceptionHandler_t eh) {
   sigset_t mask;
   if (eh->supportAsyncSignals) {
      if (!eh->inCriticalSection) {
         if (eh->sp > 0) {
            if (pthread_sigmask(SIG_BLOCK, &eh->signalMask, NULL) < 0) {
               eh->Error("GEFExceptionHandler_BeginCriticalSection--sigmask(SIG_BLOCK, ...):  Failed!\n");
               eh->Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
            }
         } else {
            if (pthread_sigmask(SIG_BLOCK, NULL, &mask) < 0) {
               eh->Error("In GEFExceptionHandler_BeginCriticalSection--sigmask(SIG_BLOCK, ...):  Failed!\n");
               eh->Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
            }
            if (pthread_sigmask(SIG_BLOCK, &eh->signalMask, NULL) < 0) {
               eh->Error("In GEFExceptionHandler_BeginCriticalSection--sigmask(SIG_BLOCK, ...):  Failed!\n");
               eh->Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
            }
            eh->mask = mask;
         }
         eh->inCriticalSection = 1;
      }
   }
}

void
GEFExceptionHandler_EndCriticalSection(GEFExceptionHandler_t eh) {
   int sp;
   GEFCheckPoint_t* env;
   sigset_t* mask;
   if (eh->supportAsyncSignals) {
      if (eh->inCriticalSection) {
         if (eh->sp > 0) {
            sp = eh->sp - 1;
            env = &eh->stack[sp];
            switch(env->current) {
            case GEF_EXCEPT_TRY:
            case GEF_EXCEPT_RETRY:
            case GEF_EXCEPT_DONE:
               eh->inCriticalSection = 0;
               if (pthread_sigmask(SIG_SETMASK, &env->mask, NULL) < 0) {
                  eh->Error("In GEFExceptionHandler_EndCriticalSection--pthread_sigmask(SIG_SETMASK, ...):  Failed!");
                  eh->Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
               }
               break;
            default: break;
            }
         } else {
            if (pthread_sigmask(SIG_SETMASK, &eh->mask, NULL) < 0) {
               eh->Error("In GEFExceptionHandler_EndCriticalSection--pthread_sigmask(SIG_SETMASK, ...):  Failed!");
               eh->Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
            }
         }
      }
   }
}

void
GEFExceptionHandler_RaiseException(GEFExceptionHandler_t eh, void* id) {
   int sp;
   GEFCheckPoint_t* env;
   CheckUnhandledException(eh, id);
   GEFExceptionHandler_BeginCriticalSection(eh);
   sp = eh->sp - 1;
   env = &eh->stack[sp];
   eh->Delete(&env->exceptionID);
   env->exceptionID = id;
   env->previous = env->current;
   env->current = GEF_EXCEPT_CATCH;
   longjmp(env->env, GEF_EXCEPT_CATCH);
   GEFExceptionHandler_EndCriticalSection(eh);
}

static
int
ProcessID() {
   return((int) getpid());
}

static
void
Delete(void** exceptionID) {
   *exceptionID = NULL;
}

static
void
Init(GEFExceptionHandler_t eh) {
   eh->debug = 0;
   eh->assertions = 0;
   eh->jump = 0;
   eh->supportAsyncSignals = 0;
   eh->stack = NULL;
   eh->ssize = 0;
   eh->sp = 0;
   sigemptyset(&eh->signalMask);
   sigemptyset(&eh->mask);
   if (eh->supportAsyncSignals) pthread_sigmask(SIG_BLOCK, NULL, &eh->mask);
   eh->inCriticalSection = 0;
   eh->Abort = Abort;
   eh->AssertionViolation = AssertionViolation;
   eh->UnhandledException = UnhandledException;
   eh->Error = Error;
   eh->Exit = Exit;
   eh->OSSignalToException = OSSignalToException;
   eh->OutOfMemory = OutOfMemory;
   eh->UserCatch = UserCatch;
   eh->ProcessID = ProcessID;
   eh->Delete = Delete;
   eh->states[GEF_EXCEPT_INVARIANTS].id = GEF_EXCEPT_INVARIANTS;
   eh->states[GEF_EXCEPT_INVARIANTS].name = "Invariants";
   eh->states[GEF_EXCEPT_PRECONDS].id = GEF_EXCEPT_PRECONDS;
   eh->states[GEF_EXCEPT_PRECONDS].name = "Preconditions";
   eh->states[GEF_EXCEPT_TRY].id = GEF_EXCEPT_TRY;
   eh->states[GEF_EXCEPT_TRY].name = "Try";
   eh->states[GEF_EXCEPT_RETRY].id = GEF_EXCEPT_RETRY;
   eh->states[GEF_EXCEPT_RETRY].name = "Try";
   eh->states[GEF_EXCEPT_FINALLY].id = GEF_EXCEPT_FINALLY;
   eh->states[GEF_EXCEPT_FINALLY].name = "Finally";
   eh->states[GEF_EXCEPT_CATCH].id = GEF_EXCEPT_CATCH;
   eh->states[GEF_EXCEPT_CATCH].name = "Catch";
   eh->states[GEF_EXCEPT_POSTCONDS].id = GEF_EXCEPT_POSTCONDS;
   eh->states[GEF_EXCEPT_POSTCONDS].name = "Postconditions";
   eh->states[GEF_EXCEPT_SUCCESS].id = GEF_EXCEPT_SUCCESS;
   eh->states[GEF_EXCEPT_SUCCESS].name = "Success";
   eh->states[GEF_EXCEPT_DONE].id = GEF_EXCEPT_DONE;
   eh->states[GEF_EXCEPT_DONE].name = "End";
   eh->states[GEF_EXCEPT_FAILURE].id = GEF_EXCEPT_FAILURE;
   eh->states[GEF_EXCEPT_FAILURE].name = "Failure";
   eh->states[GEF_EXCEPT_INVALID].id = GEF_EXCEPT_INVALID;
   eh->states[GEF_EXCEPT_INVALID].name = "Invalid";
}

void
GEFExceptionHandler_GrowStack(GEFExceptionHandler_t eh) {
   GEFExceptionHandler_BeginCriticalSection(eh);
   if (eh->stack == (GEFCheckPoint_t*) NULL) {
      eh->stack = (GEFCheckPoint_t*) malloc(sizeof(GEFCheckPoint_t) * 512);
      if (eh->stack == (GEFCheckPoint_t*) NULL) {
         eh->Error("GEF: While executing in process, %d, could not initialize exception handler--aborting!\n", eh->ProcessID());
         eh->Abort(GEF_EXCEPTION_NO_MEMORY);
    } else {
       eh->ssize = 512;
       eh->sp = 0;
    }
  } else {
    size_t newsize = eh->ssize * 2;
    size_t numOfBytes = sizeof(GEFCheckPoint_t) * newsize;
    GEFCheckPoint_t* newstack = (GEFCheckPoint_t*) realloc(eh->stack, numOfBytes);
    if (newstack != (GEFCheckPoint_t*) NULL) {
      eh->stack = newstack;
      eh->ssize = newsize;
    } else {
      eh->OutOfMemory();
    }
  }
  GEFExceptionHandler_EndCriticalSection(eh);
}

static
void
InitInvariants(GEFExceptionHandler_t eh) {
   GEFStateID_t previous;

   for (previous = 0; previous < GEF_EXCEPT_NUMBER; previous++) {
     eh->stateTransitions[GEF_EXCEPT_INVARIANTS][previous] = eh->states[GEF_EXCEPT_PRECONDS];
   }
   eh->stateTransitions[GEF_EXCEPT_INVARIANTS][GEF_EXCEPT_POSTCONDS] = eh->states[GEF_EXCEPT_FINALLY];
}

static
void
InitPreconditions(GEFExceptionHandler_t eh) {
   GEFStateID_t previous;

   for (previous = 0; previous < GEF_EXCEPT_NUMBER; previous++) {
      eh->stateTransitions[GEF_EXCEPT_PRECONDS][previous] = eh->states[GEF_EXCEPT_TRY];
   }
}

static
void
InitTry(GEFExceptionHandler_t eh) {
   GEFStateID_t previous;

   for (previous = 0; previous < GEF_EXCEPT_NUMBER; previous++) {
      eh->stateTransitions[GEF_EXCEPT_TRY][previous] = eh->states[GEF_EXCEPT_POSTCONDS];
   }
}

static
void
InitRetry(GEFExceptionHandler_t eh) {
   eh->stateTransitions[GEF_EXCEPT_RETRY][GEF_EXCEPT_RETRY] = eh->states[GEF_EXCEPT_POSTCONDS];
   eh->stateTransitions[GEF_EXCEPT_RETRY][GEF_EXCEPT_CATCH] = eh->states[GEF_EXCEPT_POSTCONDS];
}

static
void
InitFinally(GEFExceptionHandler_t eh) {
   eh->stateTransitions[GEF_EXCEPT_FINALLY][GEF_EXCEPT_INVARIANTS] = eh->states[GEF_EXCEPT_SUCCESS];
   eh->stateTransitions[GEF_EXCEPT_FINALLY][GEF_EXCEPT_RETRY] = eh->states[GEF_EXCEPT_SUCCESS];
   eh->stateTransitions[GEF_EXCEPT_FINALLY][GEF_EXCEPT_TRY] = eh->states[GEF_EXCEPT_SUCCESS];
   eh->stateTransitions[GEF_EXCEPT_FINALLY][GEF_EXCEPT_CATCH] = eh->states[GEF_EXCEPT_FAILURE];

}

static
void
InitPostconditions(GEFExceptionHandler_t eh) {
   eh->stateTransitions[GEF_EXCEPT_POSTCONDS][GEF_EXCEPT_TRY] = eh->states[GEF_EXCEPT_INVARIANTS];
   eh->stateTransitions[GEF_EXCEPT_POSTCONDS][GEF_EXCEPT_RETRY] = eh->states[GEF_EXCEPT_INVARIANTS];
}

static
void
InitCatch(GEFExceptionHandler_t eh) {
   eh->stateTransitions[GEF_EXCEPT_CATCH][GEF_EXCEPT_PRECONDS] = eh->states[GEF_EXCEPT_FINALLY];
   eh->stateTransitions[GEF_EXCEPT_CATCH][GEF_EXCEPT_TRY] = eh->states[GEF_EXCEPT_FINALLY];
   eh->stateTransitions[GEF_EXCEPT_CATCH][GEF_EXCEPT_RETRY] = eh->states[GEF_EXCEPT_FINALLY];
   eh->stateTransitions[GEF_EXCEPT_CATCH][GEF_EXCEPT_FINALLY] = eh->states[GEF_EXCEPT_FINALLY];
   eh->stateTransitions[GEF_EXCEPT_CATCH][GEF_EXCEPT_CATCH] = eh->states[GEF_EXCEPT_CATCH];
   eh->stateTransitions[GEF_EXCEPT_CATCH][GEF_EXCEPT_POSTCONDS] = eh->states[GEF_EXCEPT_FINALLY];
   eh->stateTransitions[GEF_EXCEPT_CATCH][GEF_EXCEPT_INVARIANTS] = eh->states[GEF_EXCEPT_FINALLY];
}

static
void
InitSuccess(GEFExceptionHandler_t eh) {
   GEFStateID_t previous;
   for (previous = 0; previous < GEF_EXCEPT_NUMBER; previous++) {
      eh->stateTransitions[GEF_EXCEPT_SUCCESS][previous] = eh->states[GEF_EXCEPT_DONE];
   }
}

GEFExceptionHandler_t
GEFExceptionHandler_New(GEFAttr_t attrs, sigset_t signalSet) {
   int current;
   int previous;
   GEFExceptionHandler_t eh = NULL;
   if ((eh = malloc(sizeof(*eh))) != NULL) {
      Init(eh);
      eh->signalMask = signalSet;
      if (attrs.Abort != NULL) eh->Abort = attrs.Abort;
      if (attrs.AssertionViolation != NULL) eh->AssertionViolation = attrs.AssertionViolation;
      if (attrs.UnhandledException != NULL) eh->UnhandledException = attrs.UnhandledException;
      if (attrs.Exit != NULL) eh->Exit = attrs.Exit;
      if (attrs.OSSignalToException != NULL) eh->OSSignalToException = attrs.OSSignalToException;
      if (attrs.Error != NULL) eh->Error = attrs.Error;
      if (attrs.OutOfMemory != NULL) eh->OutOfMemory = attrs.OutOfMemory;
      if (attrs.UserCatch != NULL) eh->UserCatch = attrs.UserCatch;
      if (attrs.Delete != NULL) eh->Delete = attrs.Delete;
      for (current = 0; current < GEF_EXCEPT_NUMBER; current++) {
         for (previous = 0; previous < GEF_EXCEPT_NUMBER; previous ++){
            eh->stateTransitions[current][previous] = eh->states[GEF_EXCEPT_INVALID];
         }
      }
      InitInvariants(eh);
      InitPreconditions(eh);
      InitTry(eh);
      InitRetry(eh);
      InitFinally(eh);
      InitCatch(eh);
      InitPostconditions(eh);
      InitSuccess(eh);
      GEFExceptionHandler_GrowStack(eh);
   }

   return(eh);
}

void
GEFExceptionHandler_Delete(GEFExceptionHandler_t* eh) {
   int debug = (*eh)->debug;
   if (debug) fprintf(stderr, "GEFExceptionHandler_Delete: Deleting stack!\n");
   free((void*) (*eh)->stack);
   if (debug) fprintf(stderr, "GEFExceptionHandler_Delete: Deleting eh!\n");
   free((void*) *eh);
   *eh = NULL;
}

int
GEFExceptionHandler_StackDepth(GEFExceptionHandler_t eh) {
   return(eh->sp);
}

#define CheckNoActiveCheckPoints(eh)\
if (eh->sp == 0) NoActiveCheckPoints(eh)

#define CheckOverflow(eh) \
if (eh->sp == eh->ssize) GEFExceptionHandler_GrowStack(eh)

static
void
InvalidTransition(GEFExceptionHandler_t eh, int current, int previous, int next) {
/*
   Description: This function reports an invalid state transition that the GEF
   facility makes and aborts the currently running thread.
*/
   char* currentName = eh->states[current].name;
   char* previousName = eh->states[previous].name;
   char* nextName = eh->states[next].name;
   eh->Error("While executing in the process, %d, encountered invalid exception state...", eh->ProcessID());
   eh->Error("current = %s, previous = %s, next = %s.\n", currentName, previousName, nextName);
   eh->Abort(GEF_EXCEPTION_FATAL);
}

GEFStateID_t
GEFExceptionHandler_NextState(GEFExceptionHandler_t eh) {
   GEFStateID_t previous;
   GEFStateID_t current;
   GEFStateID_t next;
   int sp;
   int numberOfAttempts;
   GEFCheckPoint_t* env;
   GEFExceptionHandler_BeginCriticalSection(eh);
   CheckNoActiveCheckPoints(eh);
   sp = eh->sp - 1;
   env = &eh->stack[sp];
   current = env->current;
   previous = env->previous;
   next = eh->stateTransitions[current][previous].id;
   if (eh->debug) {
      fprintf(stderr, "Current = %s\n", eh->states[current].name);
      fprintf(stderr, "Previous = %s\n", eh->states[previous].name);
      fprintf(stderr, "Next = %s\n", eh->states[next].name);
   }
   switch(current) {
   case GEF_EXCEPT_TRY:
   case GEF_EXCEPT_RETRY:
      if (next == GEF_EXCEPT_POSTCONDS && !eh->assertions) {
         next = GEF_EXCEPT_FINALLY;
      }
      env->previous = current;
      env->current = next;
      break;
   case GEF_EXCEPT_CATCH:
      numberOfAttempts = env->numberOfAttempts;
      if (numberOfAttempts > 1) {
         eh->Delete(&env->exceptionID);
         env->numberOfAttempts--;
         env->previous = current;
         env->current = GEF_EXCEPT_RETRY;
         return(GEF_EXCEPT_RETRY);
      }
      env->previous = current;
      env->current = next;
      break;
   case GEF_EXCEPT_SUCCESS:
      eh->sp--;
      break;
   case GEF_EXCEPT_FINALLY:
      if (next == GEF_EXCEPT_SUCCESS) {
         eh->sp--;
         next = GEF_EXCEPT_DONE;
      } else {
         env->previous = current;
         env->current = next;
      }
      break;
   case GEF_EXCEPT_FAILURE:
      eh->sp--;
      GEFExceptionHandler_RaiseException(eh, eh->stack[eh->sp].exceptionID);
      break;
   case GEF_EXCEPT_INVALID:
      InvalidTransition(eh, current, previous, next);
      break;
   default:
      env->previous = current;
      env->current = next;
      break;
   }
   if (eh->jump) {
      eh->jump = 0;
      longjmp(env->env, next);
   }
   return(next);
}

void
GEFExceptionHandler_Break(GEFExceptionHandler_t eh) {
   int sp;
   GEFCheckPoint_t* env;
   int inCriticalSection = eh->inCriticalSection;
   GEFExceptionHandler_BeginCriticalSection(eh);
   CheckNoActiveCheckPoints(eh);
   sp = eh->sp - 1;
   env = &eh->stack[sp];
   if (env->current == GEF_EXCEPT_CATCH) {
      eh->Delete(&env->exceptionID);
      env->previous = env->current;
      env->current = GEF_EXCEPT_RETRY;
      eh->jump = 1;
      GEFExceptionHandler_NextState(eh);
   }
   if (!inCriticalSection) GEFExceptionHandler_EndCriticalSection(eh);
}

void
GEFExceptionHandler_Retry(GEFExceptionHandler_t eh) {
   int sp;
   int numberOfAttempts;
   GEFCheckPoint_t* env;
   int inCriticalSection = eh->inCriticalSection;
   GEFExceptionHandler_BeginCriticalSection(eh);
   CheckNoActiveCheckPoints(eh);
   sp = eh->sp - 1;
   env = &eh->stack[sp];
   if (env->current == GEF_EXCEPT_CATCH) {
      eh->Delete(&env->exceptionID);
      numberOfAttempts = env->numberOfAttempts;
      if (numberOfAttempts > 1) {
         env->numberOfAttempts--;
         env->previous = env->current;
         env->current = GEF_EXCEPT_RETRY;
         longjmp(env->env, GEF_EXCEPT_RETRY);
      } else if (numberOfAttempts == 0) {
         env->previous = env->current;
         env->current = GEF_EXCEPT_RETRY;
         longjmp(env->env, GEF_EXCEPT_RETRY);
      } else {
         ; /* Ignore request; satisfied number of attempts */
      }
   }
   if (!inCriticalSection) GEFExceptionHandler_EndCriticalSection(eh);
}

GEFCheckPoint_t*
GEFExceptionHandler_CheckPoint(GEFExceptionHandler_t eh, int max) {
   GEFCheckPoint_t* env;
   sigset_t oldmask;
   max = (max < 0 ? 0 : max);
   CheckOverflow(eh);
   env = &eh->stack[eh->sp];
   if (eh->supportAsyncSignals) {
      if (pthread_sigmask(SIG_BLOCK, NULL, &oldmask) < 0) {
         eh->Error("GEFExceptionHandler_CheckPoint--sigmask(SIG_BLOCK, ...):  Failed!\n");
         eh->Abort(GEF_EXCEPTION_SYSTEM_CALL_FAILURE);
      }
      GEFExceptionHandler_BeginCriticalSection(eh);
      env->mask = oldmask;
   }
   if (eh->assertions) {
      env->current = GEF_EXCEPT_INVARIANTS;
   } else env->current = GEF_EXCEPT_TRY;
   env->previous = GEF_EXCEPT_INVALID;
   env->numberOfAttempts = max;
   env->exceptionID = NULL;
   eh->sp++;
   return(env);
}

void*
GEFExceptionHandler_GetException(GEFExceptionHandler_t eh) {
   CheckNoActiveCheckPoints(eh);
   return(eh->stack[eh->sp - 1].exceptionID);
}

static
int
sigsetisempty(sigset_t* set) {
   int signo;
   for (signo = 1; signo < NSIG; signo++) {
      if (sigismember(set, signo)) {
         return(0);
      }
   }
   return(1);
}
