/*
  GEF: A General Exception-Handling Facility, which also support "Programming by Contract."
  Copyright (C) 1998 Bruce W. Bigby
  E-Mail: bbigby@alum.MIT.edu
 */
#ifndef GEF_EXCEPT_HANDLER_H
#define GEF_EXCEPT_HANDLER_H

#include <signal.h>
#include <stdarg.h>
#include <setjmp.h>
#include <pthread.h>
#include <GEFAttr.h>

#ifndef sig_atomic_t
#define sig_atomic_t int
#define raise(signo) pthread_kill(pthread_self(), signo);
#endif

typedef enum {
  GEF_EXCEPT_INVARIANTS = 0, /* Must be zero - Return from setjmp */
  GEF_EXCEPT_PRECONDS,
  GEF_EXCEPT_TRY,
  GEF_EXCEPT_RETRY,
  GEF_EXCEPT_FINALLY,
  GEF_EXCEPT_CATCH,
  GEF_EXCEPT_POSTCONDS,
  GEF_EXCEPT_SUCCESS,
  GEF_EXCEPT_FAILURE,
  GEF_EXCEPT_INVALID,
  GEF_EXCEPT_DONE,
  GEF_EXCEPT_NUMBER
} GEFStateID_t;

typedef struct {
   GEFStateID_t id;
   char* name;
} GEFState_t;

typedef struct {
   jmp_buf env;
   int numberOfAttempts;
   void* exceptionID;
   GEFStateID_t current;
   GEFStateID_t previous;
   sigset_t mask;
} GEFCheckPoint_t;

/* These exception values are reserved by GEF. Do not mix exception values with these values! */
/* You may redefine these values via the appropriate call-back functions. */
#define GEF_EXCEPTION_NO_MEMORY                      0
#define GEF_EXCEPTION_ASSERTION_VIOLATION 1
#define GEF_EXCEPTION_SYSTEM_CALL_FAILURE 2
#define GEF_EXCEPTION_FATAL                                   3
#define GEF_EXCEPTION_UNHANDLED                      4

typedef struct {
   int debug;
   int assertions;
   int supportAsyncSignals;
   int jump;
   GEFState_t states[GEF_EXCEPT_NUMBER];
   GEFState_t stateTransitions[GEF_EXCEPT_NUMBER][GEF_EXCEPT_NUMBER];
   GEFCheckPoint_t* stack;
   size_t ssize;
   size_t sp;
   sigset_t mask;
   sigset_t signalMask;
   int signalMaskIsEmpty;
   int inCriticalSection;
   void (*Abort)(int);
   void (*AssertionViolation)(char*, int);
   void (*UnhandledException)(void*);
   void (*Exit)(int);
   void* (*OSSignalToException)(int);
   void (*Error)(char* format, ...);
   void (*OutOfMemory)(void);
   void (*UserCatch)(void* userData, void* exceptionID);
   int (*ProcessID)(void);
   void (*Delete)(void** exceptionID);
} *GEFExceptionHandler_t;

extern
void
GEFExceptionHandler_SetDebug(GEFExceptionHandler_t eh, int value);

extern
void
GEFExceptionHandler_SetAssertions(GEFExceptionHandler_t eh, int value);

extern
int
GEFExceptionHandler_GetAssertions(GEFExceptionHandler_t eh);

extern
void
GEFExceptionHandler_SetAsyncSupport(GEFExceptionHandler_t eh, int value);

extern
int
GEFExceptionHandler_GetAsyncSupport(GEFExceptionHandler_t eh);

extern
int
GEFExceptionHandler_StackDepth(GEFExceptionHandler_t eh);

extern
void
GEFExceptionHandler_GrowStack(GEFExceptionHandler_t eh);

extern
GEFExceptionHandler_t
GEFExceptionHandler_New(GEFAttr_t attrs, sigset_t signalSet);

extern
GEFStateID_t
GEFExceptionHandler_NextState(GEFExceptionHandler_t eh);

extern
void
GEFExceptionHandler_Break(GEFExceptionHandler_t eh);

extern
void
GEFExceptionHandler_Delete(GEFExceptionHandler_t* eh);

extern
void
GEFExceptionHandler_Retry(GEFExceptionHandler_t eh);

extern
GEFCheckPoint_t*
GEFExceptionHandler_CheckPoint(GEFExceptionHandler_t eh, int max);

extern
void*
GEFExceptionHandler_GetException(GEFExceptionHandler_t eh);

extern
void
GEFExceptionHandler_RaiseException(GEFExceptionHandler_t eh, void* id);

extern
void
GEFExceptionHandler_BeginCriticalSection(GEFExceptionHandler_t eh);

extern
void
GEFExceptionHandler_EndCriticalSection(GEFExceptionHandler_t eh);

#endif /* GEF_EXCEPT_HANDLER - General Exception Handler */
