/*
  gef_test: A demonstration/test program for the GEF (General Exception-
  Handling Facility) library.
  Copyright (C) 1998 Bruce W. Bigby, bbigby@alum.MIT.edu

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <pthread.h>
#include <GEF.h>
#include <time.h>

#define MAX(a, b) ((a) > (b) ? (a) : (b))

static pthread_t *threads;
static char *threadName;
static int* status;

#ifndef LINUX
static int *signalNotificationList;
static pthread_mutex_t signalNotificationListMutex;
#endif

static int numberOfThreads = 1;

typedef struct {
   int threadNumber;
   int asyncSupport;
   int assertionSupport;
} *ThreadMainArgs_t;

#define EXCEPT_FAILURE  0
#define EXCEPT_OS_VIOLATION  1
#define EXCEPT_INSUFFICIENT_MEMORY 2
#define EXCEPT_ASSERTION_VIOLATION  3
#define EXCEPT_ARITHMETIC_VIOLATION  4
#define EXCEPT_USER1_VIOLATION  5
#define EXCEPT_ALARM_EXPIRED  6

static
void
TestTry(int threadNumber) {
   printf("**** Thread %d: Test Try ****\n", threadNumber);
   gef_try {
      printf("\tTRY...\n");
   } gef_invariants {
     printf("\tINVARIANTS...\n");
   } gef_preconditions {
      printf("\tPRECONDITIONS...\n");
   } gef_postconditions {
      printf("\tPOSTCONDITIONS...\n");
   } gef_finally {
      printf("\tFINALLY...\n");
   } gef_end;
   printf("\tDONE...\n");
}

static
void
TestAssert(int threadNumber) {
   printf("*** Thread %d: Test Assert ***\n", threadNumber);
   gef_try {
      gef_try {
         printf("\tTRY...\n");
      } gef_invariants {
         fflush(stdout);
         gef_assert(1 < 0);
      } gef_postconditions {
         printf("\tPOSTCONDITIONS...\n");
      } gef_end;
   } gef_catch(e) {
      switch((int) e) {
      case EXCEPT_ASSERTION_VIOLATION:
         /*
           Normally, e would be an assertion object with a "Line Number" and
           "File Name" attribute. One would be able to print that information
           here, or reraise the exception.  If no one catches the exception,
           GEF could unparse the exception, using a multinational routine and
           display it to the user.
         */
         printf("\tTestAssert: Caught assertion violation!\n");
         printf("\tLine number and filename information currently unavailable.\n");
         break;
      default: break;
      }
      gef_break;
   } gef_end;
  printf("\tDONE...\n");
}

static
void
TestRepeatTry(int threadNumber, int n) {
   printf("**** Thread %d: Test Repeat Try (%d)  ****\n", threadNumber, n);
   gef_repeat_try(n) {
      printf("\tTRY...\n");
      gef_throw((void*) EXCEPT_FAILURE);
   } gef_preconditions {
      printf("\tPRECONDITIONS...\n");
   } gef_catch(e) {
      if (--n == 0) gef_break;
   } gef_postconditions {
      printf("\tPOSTCONDITIONS...\n");
   } gef_end;
   printf("\tDONE...\n");
}

static
void
TestTryCatch(int threadNumber, int value) {
   int once = value;

   printf("**** Thread %d: Test Try, Catch, and Retry ****\n", threadNumber);
   gef_try {
      printf("\tTRY...\n");
      if (once) gef_throw((void*) EXCEPT_FAILURE);
   } gef_preconditions {
      printf("\tPRECONDITIONS...\n");
   } gef_catch(e) {
      printf("\tCATCH exception, %d...\n", (int) e);
      once = 0;
      gef_retry;
   } gef_postconditions {
      printf("\tPOSTCONDITIONS...\n");
   } gef_end;
   printf("\tDONE...\n");
}

static
void
TestDivideByZero(int threadNumber) {
   volatile int x = 10;

   printf("**** Thread %d: Test Divide By Zero ****\n", threadNumber);
   gef_try {
      printf("\tTRY...\n");
      x = x / 0;
   } gef_preconditions {
      printf("\tPRECONDITIONS...\n");
   } gef_catch(e) {
      if (((int) e) == EXCEPT_ARITHMETIC_VIOLATION) {
         printf("\tCATCH exception ARITHMETIC VIOLATION...\n");
         gef_break;
      } else {
         printf("\tCATCH exception, %d...\n", (int) e);
      }
   } gef_postconditions {
      printf("\tPOSTCONDITIONS...\n");
   } gef_end;
   printf("\tDONE...\n");
}

static
void
TestTryCatchAndFinally(int threadNumber) {
   int once = 1;

   printf("**** Thread %d: Test Try, Catch, Try, and Finally ****\n", threadNumber);
   gef_try {
      printf("\tTRY...\n");
      if (once) gef_throw((void*) EXCEPT_FAILURE);
   } gef_preconditions {
      printf("\tPRECONDITIONS...\n");
   } gef_catch(e) {
      printf("\tCATCH exception, %d...\n", (int) e);
      once = 0;
      gef_retry;
   } gef_postconditions {
      printf("\tPOSTCONDITIONS...\n");
   } gef_finally {
      printf("\tFINALLY...\n");
   } gef_end;
   printf("\tDONE...\n");
}

static
void
OutOfMemory() {
   gef_throw((void*) EXCEPT_INSUFFICIENT_MEMORY);
}

static
void
AssertionViolation(char* fileName, int lineNumber) {
  /*
    Normally, one would throw an object, or a pointer to a structure, that a
    catch block could use to obtain the line number and fileName, instead of
    printing information right here.
    */
  printf("\tAssertionViolation: Occurrence on line %d of file %s.\n", lineNumber, fileName);
   gef_throw((void*) EXCEPT_ASSERTION_VIOLATION);
}

static
void
UnhandledException(void* exceptionID) {
   fprintf(stderr, "\tUnhandled exception, %d, occurred, while excecuting in process, %d.\n\007\n", ((int) exceptionID), (int) getpid());
}

static
void*
OSSignalToException(int signum) {
   /* Map signum into application specific exception id, if necessary */
   switch (signum) {
   case SIGALRM:
      return((void*) EXCEPT_ALARM_EXPIRED);
   case SIGFPE:
      return((void*) EXCEPT_ARITHMETIC_VIOLATION);
   case SIGUSR1:
      return((void*) EXCEPT_USER1_VIOLATION);
   default:
       return((void*) EXCEPT_OS_VIOLATION);
   }
}

static
void*
SynchronousPerformanceMain(void* arg) {
   int exception = 0;
   int stop = 300000;
   int stopold = stop;
   struct rusage usage;
   double elapsed_time;
   GEFAttr_t gefattrs;

   GEFAttr_Init(&gefattrs);
   GEFAttr_SetAssertionViolation(&gefattrs, AssertionViolation);
   GEFAttr_SetUnhandledException(&gefattrs, UnhandledException);
   GEFAttr_SetOutOfMemory(&gefattrs, OutOfMemory);
   GEFAttr_SetOSSignalToException(&gefattrs, OSSignalToException);
   GEFInitializeThread(gefattrs);
   while (stop > 0) {
      gef_try {
         stop--;
      } gef_end;
   }
   if (getrusage(RUSAGE_SELF, &usage) == 0) {
      elapsed_time = usage.ru_utime.tv_sec + usage.ru_stime.tv_sec +
         (usage.ru_utime.tv_usec + usage.ru_stime.tv_usec) / 1000000.0;
   }
   if (elapsed_time > 0.0) {
      printf("\tIterations = %d\n", stopold);
      printf("\tSeconds = %f\n", elapsed_time);
      printf("\tIterations per second = %d\n", (int) (((double) stopold) / elapsed_time));
   } else {
      printf("\tIterations per second = %d\n", (int) ((double) stopold));
   }
   pthread_exit((void*) exception);
}

static
void
TestTryPerformance(int threadNumber) {
   volatile int count = 0;
   int seconds;
   int stop = 350000;
   int stopold = stop;
   clock_t begin, end;
   time_t beginTime, endTime;
   double elapsed_time;

   printf("**** Thread %d: Test to determine Try iterations per second ****\n", threadNumber);
   gef_try {
      printf("\tTRY...\n");
      if (gef_async_support_is_enabled) {
         seconds = 1;
         gef_begin_critical_section;
#ifndef LINUX
         pthread_mutex_lock(&signalNotificationListMutex);
         signalNotificationList[threadNumber * NSIG + SIGUSR1] = 1;
         pthread_mutex_unlock(&signalNotificationListMutex);
#endif
         printf("\tIterating for %d seconds!\n", seconds);
         alarm(seconds);
         gef_end_critical_section;

         while (1) {
            gef_try {
               count++;
            } gef_end;
         }
      } else {
#ifdef CLOCK
         begin = clock();
         beginTime = time((time_t*) NULL);
         while (stop > 0) {
            gef_try {
               stop--;
            } gef_end;
         }
#else
         pthread_t threadID;
         pthread_attr_t attr;
         int status;
         int rc = EINTR;

         pthread_attr_init(&attr);
         pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
         pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
         gef_try {
            gef_begin_critical_section;
            if (pthread_create(&threadID, &attr, SynchronousPerformanceMain, (void*) NULL) != 0)  {
               gef_throw((void*) EXCEPT_FAILURE);
            }
            rc = pthread_join(threadID, (void**) &status);
            gef_end_critical_section;
         } gef_preconditions {
            printf("\tPRECONDITIONS...\n");
         } gef_catch(exception) {
            switch((int) exception) {
            case EXCEPT_FAILURE:
               fprintf(stderr, "Thread %d: Unable to create Synchronous Performance thread!\n", threadNumber);
               gef_break;
            default: break;
            } /* switch */
        } gef_postconditions {
           printf("\tPOSTCONDITIONS...\n");
        } gef_end;
#endif
#ifdef CLOCK
        end = clock();
        endTime = time((time_t*) NULL);
#ifdef LINUX
        if (numberOfThreads == 1) {
           elapsed_time = ((double) (end - begin)) / ((double) CLOCKS_PER_SEC);
        } else elapsed_time = ((double) (endTime - beginTime));
#else
        elapsed_time = ((double) (end - begin)) / ((double) CLOCKS_PER_SEC);
#endif
        if (!gef_async_support_is_enabled) {
           if (elapsed_time > 0.0) {
              printf("\tIterations per second <= %d\n", (int) (((double) stopold) / elapsed_time));
           } else {
              printf("\tIterations per second <= %d\n", (int) ((double) stopold));
           }
        }
#endif
      }
   } gef_preconditions {
      printf("\tPRECONDITIONS...\n");
   } gef_catch(signum) {
      printf("\tCATCH...\n");
#ifndef LINUX
      if (((int) signum) == EXCEPT_USER1_VIOLATION) {
#else
      if (((int) signum) == EXCEPT_ALARM_EXPIRED) {
#endif
         int iterations = ((double) count) / ((double) seconds);
         printf("\tcount = %d, seconds = %d\n", count, seconds);
         printf("\tIterations per second = %d\n", iterations);
#ifndef LINUX
         pthread_mutex_lock(&signalNotificationListMutex);
         signalNotificationList[threadNumber * NSIG + SIGUSR1] = 0;
         pthread_mutex_unlock(&signalNotificationListMutex);
#endif
         gef_break;
      }
   } gef_postconditions {
      printf("\tPOSTCONDITIONS...\n");
   } gef_end;
   printf("\tDONE...\n");
}

static
int
GrowStack(int depth) {
   int maxDepth = 0;

   gef_try {
      if (depth > 1) {
         maxDepth = GrowStack(depth - 1);
      } else maxDepth = gef_depth;
   } gef_end;
   return(maxDepth);
}

static
void
TestStackGrowth(int threadNumber, int depth) {
   int currentDepth = gef_depth;
   int achievedDepth;

   printf("**** Thread %d: Test Try stack growth ****\n", threadNumber);
   achievedDepth = GrowStack(depth);
   printf("\tGEF exception depth is %d\n", achievedDepth - currentDepth);
   printf("\tDONE...\n");
}

void*
ThreadMain(void* arg) {
   int exception = 0;
   ThreadMainArgs_t parms = (ThreadMainArgs_t) arg;
   int threadNumber = parms->threadNumber;
   GEFAttr_t gefattrs;

   GEFAttr_Init(&gefattrs);
   GEFAttr_SetAssertionViolation(&gefattrs, AssertionViolation);
   GEFAttr_SetUnhandledException(&gefattrs, UnhandledException);
   GEFAttr_SetOutOfMemory(&gefattrs, OutOfMemory);
   GEFAttr_SetOSSignalToException(&gefattrs, OSSignalToException);
   GEFInitializeThread(gefattrs);

   if (parms->asyncSupport) gef_enable_async_support;
   if (parms->assertionSupport) gef_enable_assertions;

   gef_try {
      TestTry(threadNumber);
      TestAssert(threadNumber);
      TestRepeatTry(threadNumber, 5);
      TestTryCatch(threadNumber, 1);
      TestTryCatchAndFinally(threadNumber);
      TestStackGrowth(threadNumber, 513);
      TestDivideByZero(threadNumber);
      TestTryPerformance(threadNumber);
   } gef_catch_with(e, (void*) NULL) {
      switch ((int) e) {
      case SIGUSR1:
         fprintf(stderr, "Thread, %d, Interrupted!\007\n", threadNumber);
         gef_break;
      case EXCEPT_ARITHMETIC_VIOLATION:
         fprintf(stderr, "Thread, %d, Caught late arithmetic violation!\007\n", threadNumber);
         gef_break;
      default: break;
      }
   } gef_end;

   if (parms->asyncSupport) gef_disable_async_support;
   if (parms->assertionSupport) gef_disable_assertions;
   if (exception == 0) printf("**** Thread %d: DONE ****\n", threadNumber);
   free((void*) parms);
   pthread_exit((void*) exception);
}

#ifndef LINUX
static
void*
sig_hand(void* arg) {
   pthread_t* parentThreadID = (pthread_t*) arg;
   sigset_t set;
   int sig;
   int rc;
   int i;

   /* The problem with this function is that, if a program registers for a */
   /* signal, such as SIGALRM, and if two alarms arrive before this handler */
   /* responds, the OS does not queue alarm calls.  As a result, we have to */
   /* create a timer service, which keeps track of the times for each thread */
   /* When one expires, it notifies all threads whose timers have expired.  */

   sigfillset(&set); /* Catch all signals */

   while (1) {
#ifndef LINUX
      sig = sigwait(&set);
#else
      rc = sigwait(&set, &sig);
#endif
      if (rc == -1) {
         fprintf(stderr, "sigwait failed!\n");
         exit(1);
      }
      switch(sig) {
      case SIGABRT:
         sigdelset(&set, SIGABRT); /* Remove SIGABRT so that we can core dump when we send it to ourselves */
         pthread_sigmask(SIG_SETMASK, &set, NULL);
         kill(getpid(), sig);
         /* exit(sig); */
         break;
      case SIGTERM:
      case SIGINT:
      case SIGQUIT:
      case SIGALRM:
         /* printf("sig_hand: Caught signal %d!\n", sig); */
         for (i = 0; i < numberOfThreads; i++) {
            pthread_mutex_lock(&signalNotificationListMutex);
            if (signalNotificationList[i * NSIG + SIGUSR1]) {
               (void) pthread_kill(threads[i], SIGUSR1);
            }
            pthread_mutex_unlock(&signalNotificationListMutex);
         }
         break;
      default:
         printf("sig_hand: Interrupted with signal %d...exiting\n", sig);
         exit(0);
      }
   }
}
#endif

int
main(int argc, char** argv) {
   pthread_t sig_hand_thread;
   pthread_attr_t attr;
   int i = 0;
   int j = 0;
   GEFAttr_t gefattrs;
   sigset_t set;
   sigset_t oset;
   int rc = EINTR;
   int asyncSupport = 0;
   int syncSupport = 0;
   int assertionSupport = 0;
   ThreadMainArgs_t parms;
   int sentSignalToChildren = 0;
   sigset_t signalSet;
   pthread_t parentThreadID = pthread_self();

   for (i = 1; i < argc; i++) {
      if (argv[i][0] == '-') {
         if (strcmp(argv[i], "-async") == 0) {
            asyncSupport = 1;
         } else if (strcmp(argv[i], "-sync") == 0) {
            syncSupport = 1;
         } else if (strcmp(argv[i], "-assertions") == 0) {
            assertionSupport = 1;
         } else {
            fprintf(stderr, "Usage: %s [-async] [-sync] [-assertions] number-of-threads\n", argv[0]);
            exit(0);
         }
      } else numberOfThreads = atoi(argv[i]);
   }
   if (numberOfThreads <= 0) numberOfThreads = 1;

#ifndef LINUX
   pthread_mutex_init(&signalNotificationListMutex, NULL);
#endif

   printf("%s: PID = %d\n", argv[0], getpid());
   threads = malloc(numberOfThreads * sizeof(pthread_t));
   threadName = malloc(numberOfThreads * sizeof(char) * 16);
   status = malloc(sizeof(int) * numberOfThreads);

#ifndef LINUX
   signalNotificationList = malloc(sizeof(int) * NSIG * numberOfThreads);
   memset(signalNotificationList, '\0', sizeof(int) * NSIG * numberOfThreads);
#endif

#ifndef LINUX
   sigfillset(&set);
   pthread_sigmask(SIG_SETMASK, &set, NULL);
#endif

   pthread_attr_init(&attr);
   pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
   pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

#ifndef LINUX
   pthread_create(&sig_hand_thread, &attr, sig_hand, (void*) &parentThreadID);
#endif

   pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

   /* Initialize GEF with the signals that you want GEF to handle for the process. */
   sigemptyset(&signalSet);
#ifndef LINUX
   sigaddset(&signalSet, SIGUSR1);
#else
   sigaddset(&signalSet, SIGALRM);
   sigaddset(&signalSet, SIGQUIT);
   sigaddset(&signalSet, SIGINT);
   sigaddset(&signalSet, SIGTERM);
#endif
   sigaddset(&signalSet, SIGFPE);
   GEFInitialize(signalSet);

   /* Initialize the specific call-backs, if any, for this thread. */
   GEFAttr_Init(&gefattrs);
   GEFAttr_SetAssertionViolation(&gefattrs, AssertionViolation);
   GEFAttr_SetUnhandledException(&gefattrs, UnhandledException);
   GEFAttr_SetOutOfMemory(&gefattrs, OutOfMemory);
   GEFAttr_SetOSSignalToException(&gefattrs, OSSignalToException);
   GEFInitializeThread(gefattrs);

#ifndef LINUX
   pthread_sigmask(SIG_SETMASK, &set, &oset);
#endif

   for (i = 0; i < numberOfThreads; i++) {
      sprintf(&threadName[i * 16], "%d", i + 1);
      parms = malloc(sizeof(*parms));
      parms->threadNumber = i;
      if (syncSupport && asyncSupport) {
         if (parms->threadNumber % 2) {
            parms->asyncSupport = 1;
         } else parms->asyncSupport = 0;
      } else if (asyncSupport) {
         parms->asyncSupport = 1;
      } else parms->asyncSupport = 0;
      parms->assertionSupport = assertionSupport;
      if (pthread_create(&threads[i], &attr, ThreadMain, (void*) parms) != 0)  {
         numberOfThreads = i;
         fprintf(stderr, "%s: Could only create %d threads!\n", argv[0], numberOfThreads);
         break;
      }
   }

#ifndef LINUX
   pthread_sigmask(SIG_SETMASK, &oset, NULL);
#endif
   gef_enable_async_support;
   j = 0;
   gef_try {
      for (; j < i; j++) {
         gef_begin_critical_section;
         rc = pthread_join(threads[j], (void**) &status[j]);
         if (rc == 0) {
            printf("\n%s: Thread %d returned with status %d\n\n", argv[0], j, status[j]);
         } else fprintf(stderr, "%s: Join with thread %d failed!\n", argv[0], j);
         gef_end_critical_section;
      }
   } gef_catch(e) {
      fprintf(stderr, "%s: Exception, %d, occurred while joining with thread, %d, ...ignoring.\n", argv[0], e, j);
      if (rc != EINTR) { /* This means that the thread did not enter the critical section. */
         j++;
      }
      rc = EINTR;
      gef_retry;
   } gef_end;

   GEFTerminate();
#ifndef LINUX
   pthread_mutex_destroy(&signalNotificationListMutex);
#endif

   free((void*) threadName);
   free((void*) status);
#ifndef LINUX
   free((void*) signalNotificationList);
#endif
   exit(0);
}
