/*
  square_root: A demonstration/test program for the GEF (General Exception-
  Handling Facility) library.
  Demonstrates the use of exception-handling.
  Copyright (C) 1998 Bruce W. Bigby, bbigby@alum.MIT.edu

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <GEF.h>

/* Define exception values for program */
#define EXCEPT_FAILURE              0 /* This should never occur! */
#define EXCEPT_OS_VIOLATION         1
#define EXCEPT_INSUFFICIENT_MEMORY  2
#define EXCEPT_ASSERTION_VIOLATION  3
#define EXCEPT_ARITHMETIC_VIOLATION 4

static
void
InsufficientMemory() {
   gef_throw((void*) EXCEPT_INSUFFICIENT_MEMORY);
}

static
void
AssertionViolation(char* fileName, int lineNumber) {
   /*
     Can only transfer fileName and lineNumber info via a pointer to a
      structure.  For now, simply throw the integer value.
      */
   fprintf(stderr, "Filename: %s\nLine number: %d\n", fileName, lineNumber);
   gef_throw((void*) EXCEPT_ASSERTION_VIOLATION);
}

static
void
UnhandledException(void* exceptionID) {
   fprintf(stderr, "\tUnhandled exception, %d, occurred, while executing process, %d.\n",((int) exceptionID), (int) getpid());
}

static
void*
OSSignalToException(int signum) {
   switch(signum) {
   case SIGFPE:
      return((void*) EXCEPT_ARITHMETIC_VIOLATION);
   default:
      return((void*) EXCEPT_OS_VIOLATION);
   }
}

#ifndef LINUX
int
matherr(struct exception* e) {
   gef_throw((void*) EXCEPT_ARITHMETIC_VIOLATION);
}
#endif

static
double
SquareRoot(double number) {
#ifdef LINUX
   if (number < 0.0) gef_throw((void*) EXCEPT_ARITHMETIC_VIOLATION);
#endif
   return(sqrt(number));
}

int
main(int argc, char** argv) {
   double answer;
   double number;
   int exitCode = 0;
   sigset_t osSignalSet;
   GEFAttr_t gefAttrs;

   if (argc != 2) {
      fprintf(stderr, "Usage: %s decimal\n", argv[0]);
      exit(1);
   }

   /* Configure GEF for process. */
   sigemptyset(&osSignalSet);
   sigaddset(&osSignalSet, SIGFPE);
   GEFInitialize(osSignalSet);
   /* Configure GEF for thread */
   GEFAttr_Init(&gefAttrs);
   GEFAttr_SetAssertionViolation(&gefAttrs, AssertionViolation);
   GEFAttr_SetUnhandledException(&gefAttrs, UnhandledException);
   GEFAttr_SetOutOfMemory(&gefAttrs, InsufficientMemory);
   GEFAttr_SetOSSignalToException(&gefAttrs, OSSignalToException);
   GEFInitializeThread(gefAttrs);
   gef_try {
      number = atof(argv[1]);
      answer = SquareRoot(number);
      printf("%.2f\n",  answer);
   } gef_preconditions {
      gef_assert(argc == 2);
   } gef_catch(exception) {
      switch((int) exception) {
      case EXCEPT_ARITHMETIC_VIOLATION:
         fprintf(stderr, "%s: Cannot take square root of %f!\n", argv[0], number);
         exitCode = 1;
         gef_break;
      case EXCEPT_ASSERTION_VIOLATION:
         fprintf(stderr, "%s: Assertion violation!\n", argv[0]);
         exitCode = 1;
         gef_break;
      default: break;
      }
   } gef_end;
   GEFTerminate();
   exit(exitCode);
}
